STITLE ERABLE Library User Words

**********************************************************************
*  (c) 01/1997 Bernard Parisse, 
*  Institut Fourier, Universit\'e de Grenoble I, 
*  F-38402 St Martin d'Heres
*  This file is part of the ERABLE package.
*  This file can be redistributed and modified under the terms of
*  the GNU General Public License as published by the Free Software
*  Foundation either version 2.0 or (at your option) any later version*
*
*  This file is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this file; if not, write to the Free Software
*  Foundation, Inc, 675 Mass Ave, Cambridge, MA 02139, USA.
**********************************************************************
	
INCLUDE keydefs.h
INCLUDE erromid.h
INCLUDE erextdec.h
INCLUDE cas2extdec.h

INCLUDE global.s

* ====================================================================
* Basic algebra commands, derivation and integration
* ====================================================================

**********************************************************************
* Desc:		Expand symbolic
* Stack:	( ob --> ob' )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME EXPAND
::
  CK1&Dispatch
  sym FLAGEXPAND
  matrix FLAGEXPAND
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:		Factor symbolic
* Stack:	( ob --> ob' )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME FACTOR
::
  CK1&Dispatch
  sym FLAGFACTOR
  matrix FLAGFACTOR
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:		Execute replacement
* Stack:	( ob repl --> ob' )
* Note:		List replacement returns the list of all the
*		different replacements into ob.
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME SUBST
::
  CK2&Dispatch
  matrix FLAGLISTEXEC
  symb  FLAGSYMBEXEC
  idnt  FLAGIDNTEXEC
  lam   FLAGIDNTEXEC
;

**********************************************************************
* Desc:		Symbolic derivative with respect to current variable.
* Stack:	( ob --> ob' )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME DERVX
::
  CK1&Dispatch
  sym :: DERVX FLAGSYMCOLCT ;
  matrix DERVX
  real DROPZ0
  cmp DROPZ0
;

**********************************************************************
* Desc:		Symbolic integral with respect to current variable.
* Stack:	( ob --> ob' )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME INTVX
::
  CK1&Dispatch
  sym FLAGINTVX
  matrix FLAGINTVX
  real FLAGINTVX
  cmp FLAGINTVX
;

**********************************************************************
* Desc:		Symbolic limit calculation.
* Stack:	( sym x=expr --> sym' )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
tNAME xLIMIT lim
xNAME LIMIT
::
  CK2&Dispatch
  symsym     :: 2QPIext SYMLIMIT ;
  matrixsymb :: 2QPIext FLAGMATRIXLIMIT ;
ASSEMBLE
	IF	f39
RPL
  SYMREAL    SYMLIMIT
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Taylor expansion at 0 order 4
* Stack:	( sym --> sym' )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME TAYLOR0
::
  CK1&Dispatch
  sym TAYLOR0
  matrix TAYLOR0
;

**********************************************************************
* Desc:		Symbolic series expansion
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SERIES
::
  CK3&Dispatch
  #AAA FLAGSERIES
  #AA1 FLAGSERIES
  #A11 :: SWAP QPI SWAP FLAGSERIES ;
;

**********************************************************************
* Desc:		Symbolic solution of equation with respect to
*		current variable.
* Stack:	( sym --> sol )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SOLVEVX
::
ASSEMBLE
	IF	f39
RPL
  CK1&Dispatch
  sym  :: isolflag InCASApplication? ITE ClrSysFlag DROP RCLVX SOLVE1EQ ;
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch
  sym  :: RCLVX SOLVE1EQ ;
  real SOLVEXFLOAT
  cmp  SOLVEXFLOAT
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Launch PLOTTER with stk1 -> EQ
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME PLOT
::
  CK1&Dispatch
  sym PLOTSTK
  real PLOTSTK
ASSEMBLE
	IF	f39
RPL
ASSEMBLE
	ELSE	f39
RPL
  list PLOTSTK
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Add stk1 to EQ and launch PLOTTER
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME PLOTADD
::
  CK1&Dispatch
  sym PLOTADD
  real PLOTADD
ASSEMBLE
	IF	f39
RPL
ASSEMBLE
	ELSE	f39
RPL
  list PLOTADD
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Integration by part
* Stack:	( u'*v u -> u*v -u*v' )
* Note:		if u is a constant return INTVX(u'*v)+u
*		if stack 2 is a list it must be of the form { olduv u'*v }
*		then olduv will be added to u*v at stack level 2
*		This permits multiple IBP in algebraic mode
*		IBP(ASIN(X)^2,X)
*		IBP(ANS(1),sqrt(1-X^2))
*		IBP(ANS(1),C) the last step with an integral containing a cst C
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME IBP
::
  CK2&Dispatch 
  symbsym FLAGIBP
  listsym FLAGIBP
;

**********************************************************************
* Desc:		Evaluation of antiderivative at the bounds
* Stack:	( f inf sup --> f[sup]-f[inf] )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME PREVAL
::
  CK3&Dispatch
  # A00 FLAGPREVAL
;

**********************************************************************
* Desc:		Risch integration
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME RISCH
::
  CK2&Dispatch
  SYMSYM FLAGRISCH
  ARRYSYM MATRIXRISCH
  REALSYM FLAGRISCH
  CMPSYM FLAGRISCH
*  symidnt FLAGRISCH
*  matrixidnt MATRIXRISCH
*  realidnt FLAGRISCH
*  cmpidnt FLAGRISCH
*  symlam FLAGRISCH
*  matrixlam MATRIXRISCH
*  reallam FLAGRISCH
*  cmplam FLAGRISCH
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME DERIV
::
  CK2&Dispatch
  SYMSYM FLAGDERIV
  ARRYSYM FLAGDERIV
  REALSYM 2DROPZ0
  CMPSYM 2DROPZ0
*  symidnt DERIVext
*  matrixidnt DERIVext
*  realidnt 2DROPZ0
*  cmpidnt 2DROPZ0
*  symlam DERIVext
*  matrixlam DERIVext
*  reallam 2DROPZ0
*  cmplam 2DROPZ0
  symmatrix DERIVext
  2matrix DERIVext
ASSEMBLE
	IF	f39
RPL
  symlist :: LIST2MATRIX DERIVext MATRIX2LIST ;
  2LIST :: 2LIST2MATRIX DERIVext MATRIX2LIST ;
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Solve differential equation.
* Stack:	( symb/list symb -> list of solutions )
* Note:		list is used for ode with boundary conditions
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME DESOLVE
::
  CK2&Dispatch
  symb FLAGDESOLVE
;

**********************************************************************
* Desc:		Laplace transform of symbolic.
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME LAP ( Laplace de poly )
::
  CK1&Dispatch
  sym FLAGLAP
  matrix FLAGLAP
  real FLAGLAP
  cmp FLAGLAP
;

**********************************************************************
* Desc:		Inverse Laplace transform of symbolic.
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME ILAP
::
  CK1&Dispatch
  sym FLAGILAP
  matrix FLAGILAP
  real FLAGILAP
  cmp FLAGILAP
;

**********************************************************************
* Desc:		Solve linear diff equation with cst coeff
* Stack:	( sym sym -> symb )
*		( matrix matrix -> matrix )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LDEC
::
  CK2&Dispatch
  symsym   FLAGLDECSOLV
  realsym  FLAGLDECSOLV
  cmpsym   FLAGLDECSOLV
  2matrix  FLAGLDSSOLV
;

* ====================================================================
* Transcendental simplification
* ====================================================================

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TEXPAND
::
  CK1&Dispatch
  sym FLAGTEXPAND
  matrix FLAGTEXPAND
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LIN
::
  CK1&Dispatch
  sym FLAGLIN
  matrix FLAGLIN
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:		Minimize the number of variables in an expression
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TSIMP
::
  CK1&Dispatch
  sym FLAGTSIMP
  matrix FLAGTSIMP
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LNCOLLECT
::
  CK1&Dispatch
  sym FLAGLNCOLLECT
  matrix FLAGLNCOLLECT
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME EXPLN
::
  CK1&Dispatch
  sym FLAGEXPLN
  matrix FLAGEXPLN
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SINCOS
::
  CK1&Dispatch
  sym FLAGSINCOS
  matrix FLAGSINCOS
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TLIN
::
  CK1&Dispatch
  sym FLAGTLIN
  matrix FLAGTLIN
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TCOLLECT
::
  CK1&Dispatch
  sym FLAGTCOLLECT
  matrix FLAGTCOLLECT
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TRIG
::
  CK1&Dispatch
  sym FLAGTRIG
  matrix FLAGTRIG
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TRIGCOS
::
  CK1&Dispatch
  sym FLAGTRIGCOS
  matrix FLAGTRIGCOS
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TRIGSIN
::
  CK1&Dispatch
  sym FLAGTRIGSIN
  matrix FLAGTRIGSIN
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TRIGTAN
::
  CK1&Dispatch
  sym FLAGTRIGTAN
  matrix FLAGTRIGTAN
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TAN2SC
::
  CK1&Dispatch
  sym FLAGTAN2SC
  matrix FLAGTAN2SC
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME HALFTAN
::
  CK1&Dispatch
  sym FLAGHALFTAN
  matrix FLAGHALFTAN
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TAN2SC2
::
  CK1&Dispatch
  sym FLAGTAN2SC2
  matrix FLAGTAN2SC2
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME ATAN2S
::
  CK1&Dispatch
  sym FLAGATAN2S
  matrix FLAGATAN2S
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME ASIN2T
::
  CK1&Dispatch
  sym FLAGASIN2T
  matrix FLAGASIN2T
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME ASIN2C
::
  CK1&Dispatch
  sym FLAGASIN2C
  matrix FLAGASIN2C
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME ACOS2S
::
  CK1&Dispatch
  sym FLAGACOS2S
  matrix FLAGACOS2S
  real NOP
  cmp NOP
;

* ====================================================================
* Unary operators
* ====================================================================

**********************************************************************
* Desc:		Inverse function.
* Stack:	( ob --> inv(ob) )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME inv
*::
*  CK1&Dispatch
*  matrix :: DTYPENDO? NcaseTYPEERR
*            STOMAText MATINV
*            ;
*  any  :: Z1 SWAP xDIV1 ;
*;

**********************************************************************
* Desc:		Abs function
* Stack:	( ob --> abs(ob) )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME abs
*::
*  CK1&Dispatch
*  any ABS
*;

**********************************************************************
* Desc:		Neg fucntion
* Stack:	( ob --> -ob )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME neg
*::
*  CK1&Dispatch
*  any neg
*;

**********************************************************************
* Desc:		Real part function
* Stack:	( ob --> re(ob) )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME re
*::
*  CK1&Dispatch
*  any re
*;

**********************************************************************
* Desc:		Imaginary part function
* Stack:	( ob --> im(ob) )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME im
*::
*  CK1&Dispatch
*  any im
*;

**********************************************************************
* Desc:		Conjugacy function
* Stack:	( ob --> conj(ob) )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME conj
*::
*  CK1&Dispatch
*  any conj
*;

**********************************************************************
* Desc:		Arg function
* Stack:	( ob --> arg(ob) )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME arg
*::
*  CK1&Dispatch
*  any ARG
*;

* ====================================================================
* Binary operators
* ====================================================================

**********************************************************************
* Desc:		Addition
* Stack:	( ob1 ob2 --> ob1+ob2 )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME ADDT
*::
*  CK2&Dispatch
*  any ADDT
*;

**********************************************************************
* Desc:		Substraction
* Stack:	( ob1 ob2 --> ob1-ob2 )
**********************************************************************
*ASSEMBLE
*         CON(3)	000
*RPL
*xNAME SUBT
*::
*  CK2&Dispatch
*  any SUBT
*;

**********************************************************************
* Desc:		Multiplication
* Stack:	( ob1 ob2 --> ob1*ob2 )
**********************************************************************
*ASSEMBLE
*         CON(3)	000
*RPL
*xNAME MULT
*::
*  CK2&Dispatch
*  SIXTYEIGHT	x*
*  TWENTY	x*
*  THIRTYSIX	x*
* ARRYREAL	x*
* FOURTWO	x*
*  FOUR		:: ARRAY2MATRIX xMULT ;
*  any MULTNOCK
*;

**********************************************************************
* Desc:		Power function
* Stack:	( ob1 ob2 --> ob1^ob2 )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME POW
*::
*  CK2&Dispatch
*  # 80     RPext
*  SYMREAL  EvalNoCKx^
*  ARRYREAL :: 'x* MPext ;
*  any      :: RESOROMP xMULT MPext ;
*;

**********************************************************************
* Desc:		Division
* Stack:	( ob1 ob2 --> ob1/ob2 )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME DIV1
*::
*  CK2&Dispatch
*  any CAS/
*;


* ====================================================================
* Arithmetic operators
* ====================================================================

**********************************************************************
* Desc:		Euclidean division
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME DIV2
::
  CK2&Dispatch
  symsym 
  :: FLAGDIV2
ASSEMBLE
	IF 	f39
RPL
  AND39
ASSEMBLE
	ENDIF	f39
RPL
  ;
;

**********************************************************************
* Desc:		Integer Euclidean division
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME IDIV2
::
ASSEMBLE
	IF	f39
RPL
  CK2&Dispatch
  symsym :: STEPIDIV2 AND39 ;
  2REAL :: 2QPI STEPIDIV2 AND39 ;
ASSEMBLE
	ELSE	f39
RPL
  CK2&Dispatch
  symsym   STEPIDIV2
  2REAL :: 2QPI STEPIDIV2 ;
ASSEMBLE
	ENDIF	f39
RPL
;


**********************************************************************
* Desc:		Euclidean division, quotient part
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME QUOT
::
  CK2&Dispatch
  symsym :: FLAGDIV2 DROP ;
;

**********************************************************************
* Desc:		Integer euclidean division, quotient part
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME IQUOT
::
  CK2&Dispatch
  symsym :: STEPIDIV2 DROP ;
  2REAL :: 2QPI STEPIDIV2 DROP ;
;

**********************************************************************
* Desc:		Euclidean division, remainder part
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME REMAINDER
::
  CK2&Dispatch
  symsym :: FLAGDIV2 SWAPDROP ;
;

**********************************************************************
* Desc:		Integer euclidean division, remainder part
* Stack:	
* Note:		Like MOD but works for Gauss integers as well
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME IREMAINDER
::
  CK2&Dispatch
  symsym :: STEPIDIV2 SWAPDROP ;
  2REAL :: 2QPI STEPIDIV2 SWAPDROP ;
;

**********************************************************************
* Desc:		GCD
* Stack:	( ob1 ob2 --> gcd(ob1,ob2) )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME GCD
::
  CK2&Dispatch
  zintzint  GCDSTEP
  symsym FLAGGCD
  2REAL :: 2QPI GCDSTEP ;
;

**********************************************************************
* Desc:
* Stack:	( P1 P2 --> PPMC[P1,P2] )
* Note:		PPMC est obtenu par la formule PPMC*PGCD=P1*P2
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME LCM
::
  CK2&Dispatch
  symsym FLAGLCM
  2REAL :: 2QPI FLAGLCM ;
;

**********************************************************************
* Desc:		Polynomial Extended GCD (au+bv=d)
* Stack:	
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME EGCD
::
  CK2&Dispatch
  symsym PEGCD
;

**********************************************************************
* Desc:		Integer Extended GCD
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME IEGCD
::
  CK2&Dispatch
  symsym IEGCD
  2REAL :: 2QPI IEGCD ;
;

**********************************************************************
* Name:		ABCUV for polynomials
* Desc:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME ABCUV
::
  CK3&Dispatch
  # AAA ABCUV
  # AA1 ABCUV
;

**********************************************************************
* Name:		ABCUV for integers
* Desc:		
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME IABCUV
::
  CK3&Dispatch
  # AAA IABCUV
  # 111 :: ROT OBJINT? NcaseSIZEERR UNROT 2QPI IABCUV ;
;

**********************************************************************
* Desc:		GCD of the elements of a list
* Stack:	( list -> gcd )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LGCD
::
  CK1&Dispatch
  matrix FLAGLGCD
  list FLAGLGCD
;

**********************************************************************
* Desc:		Simplifies stk2 and stk1 keeping the ration stk2/stk1 cst
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SIMP2
::
  CK2&Dispatch
  symsym FLAGSIMP2
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME PARTFRAC
::
  CK1&Dispatch
  symb FLAGPARTFRAC
  matrix FLAGPARTFRAC
  sym NOP
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME PROPFRAC
::
  CK1&Dispatch
  symb FLAGPROPFRAC
  matrix FLAGPROPFRAC
  sym NOP
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME PTAYL
::
  CK2&Dispatch
  symsym FLAGPTAYL
  SYMREAL FLAGPTAYL
  SYMCMP FLAGPTAYL
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME HORNER
::
  CK2&Dispatch
  symsym FLAGHORNER
  SYMREAL FLAGHORNER
  SYMCMP FLAGHORNER
;

**********************************************************************
* Name:		xeuler
* Desc:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME EULER
::
  CK1&Dispatch
  zint EULER
  real EULER
;


**********************************************************************
* Name:		P to a^2 and b^2
* Desc:		returns a+ib such that p=a^2+b^2
*		p must be a prime such that p=1 mod 4
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME PA2B2
::
  CK1&Dispatch
  zint PA2B2
  real PA2B2
;

**********************************************************************
* Name:		Chinese remainder
* Desc:		Given [ a b ] and [ c d ] returns [ e b*d ]
*		such that e=a mod b and e=c mod d
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME CHINREM
::
  CK2&Dispatch
  2matrix FLAGCHINREM
  # 99 FLAGCHINREM
ASSEMBLE
	IF	f39
RPL
  2LIST :: 2LIST2MATRIX FLAGCHINREM MATRIX2LIST ;
ASSEMBLE
	ENDIF	f39
RPL
;
**********************************************************************

**********************************************************************
* Name:		Chinese remainder for integers
* Desc:		Given [ a b ] and [ c d ] returns [ e b*d ]
*		such that e=a mod b and e=c mod d
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME ICHINREM
::
  CK2&Dispatch
  2matrix ICHINREM
  # 99 ICHINREM
;
**********************************************************************

**********************************************************************
* Name:		xISPRIME?
* Desc:		Primality test
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME ISPRIME?
::
  CK1&Dispatch
  zint ISPRIME
  real ISPRIME
;

**********************************************************************
* Name:		xNEXTPRIM
* Desc:		Next prime
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME NEXTPRIME
::
  CK1&Dispatch
  zint Prime+
  real Prime+
;

**********************************************************************
* Name:		xPREVPRIM
* Desc:		Previous prime
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME PREVPRIME
::
  CK1&Dispatch
  zint Prime-
  real Prime-
;

**********************************************************************
* Name:		xPRIMN
* Desc:		Nth prime
**********************************************************************
*	ASSEMBLE
*		CON(1)	8
*	RPL
*	xNAME PRIMN
*	::
*	  CK1 FastLib CK1Z
*	  DUP %0> NOTcase SYNTAXERR
*	  COERCE PrimeN
*	;
*********************************************

* ====================================================================
* Roots, factorization
* ====================================================================

**********************************************************************
* Desc:		Return solution of a (system of) equation
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SOLVE
::
ASSEMBLE
	IF	f39
RPL
  CK2&Dispatch
  symsym  :: isolflag InCASApplication? ITE ClrSysFlag DROP SOLVE1EQ ;
ASSEMBLE
	ELSE	f39
RPL
  CK2&Dispatch
  symsym  SOLVE1EQ
  realsym SOLVE1EQ
  cmpsym  SOLVE1EQ
ASSEMBLE
	ENDIF	f39
RPL
  2matrix SOLVEMANYEQ
;

**********************************************************************
* Desc:		Returns zeros of a (system of) equation
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME ZEROS
::
  CK2&Dispatch
  symsym  ZEROS1EQ
  realsym ZEROS1EQ
  cmpsym ZEROS1EQ
  2matrix ZEROSMANYEQ
;

**********************************************************************
* Desc:		Return polynomial having this roots with multiplicities
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME FCOEF
::
  CK1&Dispatch
  matrix FCOEF
;

**********************************************************************
* Desc:		Inverse of FCOEF, finds roots of a poly (with multiplicities)
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME FROOTS
::
  CK1&Dispatch
  sym FROOTS
;

**********************************************************************
* Desc:		Find factors of a poly with multiplicities
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME FACTORS
::
  CK1&Dispatch
  sym FACTORS
;

**********************************************************************
* Desc:		Find list of divisors
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME DIVIS
::
  CK1&Dispatch
  sym DIVIS
  real :: OBJINT? NcaseSIZEERR DIVIS ;
;


* ====================================================================
* Matrices, vectors
* ====================================================================

**********************************************************************
* Desc:		Transpose matrix without conjugation
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TRAN
::
  CK1&Dispatch
  arry MATTRAN
  list :: LIST2MATRIX MATTRAN MATRIX2LIST ;
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME HADAMARD
::
  CK2&Dispatch
  2matrix STUDMULT
ASSEMBLE
	IF	f39
RPL
  2LIST :: SWAP LIST2MATRIX SWAP LIST2MATRIX STUDMULT ;
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		RREF with particular case list
* Stack:	( matrix -> particular_case_list matrix )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME rref
::
  CK1&Dispatch
  matrix rref
  list :: LIST2MATRIX rref ;
;

**********************************************************************
* Desc:		Sub-diagonal reduction (without particular case list)
*		For numeric matrices, acts like RREF
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME REF
::
  CK1&Dispatch
  matrix MATREF
  list :: LIST2MATRIX MATREF ;
;

**********************************************************************
* Desc:		Convert between symbolic and numeric matrix.
* Stack:	( array -> matrix )
*		( matrix -> array )
*		( matrix -> matrix )
* Note:		if crunched element of matrix are all %|C% return array
*		else return a matrix.
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME AXM
::
  CK1&Dispatch
  arry :: DUPTYPEARRY? case ARRAY2MATRIX MATRIX2ARRAYpatch ;
;

**********************************************************************
* Desc:		Convert between symbolic matrix and list.
* Stack:	( list -> matrix )
*		( matrix -> list )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME AXL
::
  CK1&Dispatch
  matrix MATRIX2LISTpatch
  list LIST2MATRIX
  symb LIST2MATRIX
;

**********************************************************************
* Desc:		Convert symbolic quadratic form to matrix
* Stack:	( symb lidnt -> lvar matrice )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME QXA
::
  CK2&Dispatch
  symmatrix FLAGQXA
ASSEMBLE
	IF	f39
RPL
  symlist :: FLAGQXA MATRIX2LIST ;
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Matrix quad form to symbolic quad form
* Stack:	( matrix lidnt -> symb lidnt )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME AXQ
:: 
  CK2&Dispatch
  2matrix FLAGAXQ
ASSEMBLE
	IF	f39
RPL
  2LIST :: 2LIST2MATRIX FLAGAXQ ;
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Gauss quadratic form reduction
* Stack:	( symb lidnt -> symb' )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME GAUSS
::
ASSEMBLE
	IF	f39
RPL
  CK2&Dispatch
  symmatrix FLAGGAUSS
  symlist :: LIST2MATRIX FLAGGAUSS ;
ASSEMBLE
	ELSE	f39
RPL
  CK2&Dispatch 
  symmatrix FLAGGAUSS
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Gauss quadratic form reduction
* Stack:	( M -> D P )
*		where M= transposed[P]*D*P
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SYLVESTER
::
  CK1&Dispatch
  matrix FLAGSYLVESTER
  list :: LIST2MATRIX FLAGSYLVESTER ;
;



**********************************************************************
* Desc:		Characteristic polynomial of a matrix
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME PCAR
::
  CK1&Dispatch
  matrix PCAR
  list :: LIST2MATRIX PCAR ;
;

**********************************************************************
* Desc:		Complete reduction of endomorphism
* Stack:	arry -> 
* Note:		pmadj pcar entry has been disabled
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME JORDAN
::
  ( endomorphisme  ou   poly-madj, poly carac -> )
  CK1&Dispatch 
  matrix FLAGJORDAN
  list :: LIST2MATRIX FLAGJORDAN ;
;

**********************************************************************
* Desc:		Leverrier/Fadeev algorithm to get char. poly of a matrix
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME MAD
::
  CK1&Dispatch
  matrix MADNOCK
  list :: LIST2MATRIX MADNOCK ;
;

**********************************************************************
* Desc:		Linear system solver
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LINSOLVE
::
  CK2&Dispatch
  2matrix SYSTEM
  # AA   SYSTEM
;

**********************************************************************
* Desc:		Make Vandermonde matrix
* Stack:	vector -> matrix
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME VANDERMONDE
::
  CK1&Dispatch
  matrix VANDERMONDE
  list VANDERMONDE
;

**********************************************************************
* Desc:		Make squared Hilbert matrix
* Stack:	zint/real -> matrix
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME HILBERT
::
  CK1&Dispatch
  real HILBERTNOCK
;

**********************************************************************
* Desc:
* Stack:	( L,C, prg -> matrice )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LCXM
::
  CK3&Dispatch
  # 118 LCPROG2M
  # 11A LCPROG2M 
;



* ====================================================================
* Other commands
* ====================================================================

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME DIV
::
  CK2&Dispatch
  2matrix DIVERGENCE
ASSEMBLE
	IF	f39
RPL
  2LIST :: 2LIST2MATRIX DIVERGENCE ;
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME CURL
::
  CK2&Dispatch
  2matrix CURLsafe
ASSEMBLE
	IF	f39
RPL
  2LIST :: 2LIST2MATRIX CURLsafe MATRIX2LIST ;
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LAPL
::
  CK2&Dispatch
  symmatrix LAPLACIAN
  realmatrix 2DROPZ0
  cmpmatrix 2DROPZ0
ASSEMBLE
	IF	f39
RPL
  symlist LAPLACIAN
  TWENTYONE 2DROPZ0
  THIRTYSEVEN 2DROPZ0
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME HESS
::
  CK2&Dispatch
  symmatrix HESSIAN
  realmatrix HESSIAN
  cmpmatrix HESSIAN
ASSEMBLE
	IF	f39
RPL
  symlist HESSIAN
  TWENTYONE HESSIAN
  THIRTYSEVEN HESSIAN
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:
* Stack:	( % -> polynome de Legendre )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME LEGENDRE
::
  CK1&Dispatch
  zint LEGENDRE
  real LEGENDRE
;

**********************************************************************
* Desc:		Calculate Tchebucheff polynomials
* Stack:	( zint -> list )
* Note:		If zint > 0 : T_n[x]=cos[n*arcos[x]]
*		If zint < 0 : U_n[x]=sin[n*arccos[x]]/sin[arccos[x]]
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME TCHEBYCHEFF
::
  CK1&Dispatch
  zint TCHEBNOCK
  real TCHEBNOCK
;

**********************************************************************
* Desc:		Hermite polynomials
* Stack:	% -> P
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME HERMITE
::
  CK1&Dispatch
  zint HERMITE
  real HERMITE
;

**********************************************************************
* Desc:		Lagrange interpolation
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LAGRANGE
::
  CK1&Dispatch
  matrix LAGRANGE
;

**********************************************************************
* Desc:		Fourier coeff of a 2*pi periodic fnct
* Stack:	( symb n -> fnc )
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME FOURIER
::
  CK2&Dispatch
  symsym FOURIER
  SYMREAL :: QPI FOURIER ;
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SIGNTAB
::
ASSEMBLE
	IF	f39
RPL
  CK0 BINT96 ERABLEERROR
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch
  sym SIGNE
  real SIGNE
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:
* Stack:	( fnct -> fnct, tableau de variations )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TABVAR
::
  CK1&Dispatch
  sym TABVAR
;

**********************************************************************
* Desc:
* Stack:	( symb liste -> liste )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TABVAL
::
  CK2&Dispatch
  symlist 
  :: 
ASSEMBLE
	IF	f39
RPL
  RCLVX SWAP TABVALext SWAPDROP
ASSEMBLE
	ELSE	f39
RPL
  OVER DOSTOE
  RCLVX SWAP TABVALext
ASSEMBLE
	ENDIF	f39
RPL
  ;
;

**********************************************************************
* Desc:		Increasing power euclidean division
* Stack:	( num deno order -> quotient )
**********************************************************************
ASSEMBLE
	CON(3) 000
RPL
xNAME DIVPC
::
  CK3&Dispatch
  symsymreal FLAGDIVPC
;

**********************************************************************
* Desc:		truncation of a series expansion with respect to a rest
* Stack:	( series rest -> series' )
**********************************************************************
ASSEMBLE
	CON(3) 0
RPL
xNAME TRUNC
::
  CK2&Dispatch
  symsym FLAGTRUNC
;

**********************************************************************
* Desc:		Evaluate variable of an expression without further simplif.
* Stack:
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME SEVAL
::
  CK1&Dispatch
  sym FLAGSEVAL
  matrix FLAGSEVAL
  real NOP
  cmp NOP
;

**********************************************************************
* Desc:		Eval object and indicates time of evalutation
* Stack:	( ... -> ... time )
* Note:		time includes time for one OBJ>R and one CLKTICKS instruction
**********************************************************************
ASSEMBLE
	IF	f39
        CON(1)  0*(hasFORM)+0*(hasINTG)+0*(hasWHERE)+0*(hasVUNS)
        CON(1)  0*(hasDER)+0*(hasINV)+0*(hasCOLCT)+0*(hasEXPND)
        CON(1)  0*(hasHELP)+1*(hasPDATA)
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME TEVAL
::
  CK1&Dispatch
  any
  ::
ASSEMBLE
	IF	f39
RPL
  GARBAGE
  CLKTICKS OBJ>R
  xEVAL CLKTICKS R>OBJ
  bit- HXS>% % 8192 %/ %4 RNDXY
  TWO{}N
ASSEMBLE
	ELSE	f39
RPL
  GARBAGE
  CLKTICKS OBJ>R
  EVAL CLKTICKS R>OBJ
  bit- HXS>% % 8192 %/ %4 RNDXY
  "s" >TAG
ASSEMBLE
	ENDIF	f39
RPL
  ;
;

**miscdoc+*********************************************************************
** algebraic function property list elements
**+miscdoc*********************************************************************

** help data (if present)   ( Textbook data )


* parse data (if present)

ASSEMBLE
	IF	f39
RPL
ASSEMBLE
        CON(5)  =DOHSTR
        REL(5)  +
        CON(1)  0
        CON(1)  pdOP		Operator
        CON(1)  pdLP		Open parenthesis
        CON(1)  pdQE		Quoted Expression
        CON(1)  pdRP		Close parenthesis
        CON(1)  0
+
RPL
ASSEMBLE
	ENDIF	f39
RPL
        
* derivative (if present)

     
* inverse (if present)

        
* COLCT data (if present)

        
* EXPND data (if present)

        
* FORM menu (if present)

        
* Integral Data (if present)

        
* WHERE Building Data (if present)

        
* very-unsyminner Data (if present)

        
        
RPL     

**********************************************************************
* Desc:		Eval program recursively on each element of a list
* Stack:	( list/arry/obj prog -> list/arry/obj )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME MAP
::
  CK2&Dispatch
  seco SECOEXEC
  sym SECOEXEC
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME XNUM
::
  CK1&Dispatch 
  any XNUM
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME XQ
::
  CK1&Dispatch
  any 
  :: 
ASSEMBLE
	IF	f39
RPL
ASSEMBLE
	ELSE	f39
RPL
  TWO ClrSysFlag 
ASSEMBLE
	ENDIF	f39
RPL
  SETEXACT QPI 
  ;
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
*ASSEMBLE
*	CON(3)	000
*RPL
*xNAME SYMSTO
*::
*  CK2&Dispatch
*  idnt :: OVERSWAP xSTO ;
*;

**********************************************************************
* Desc:		Reorder expression (using variables at level 1 first)
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME REORDER
::
  CK2&Dispatch
  symsym REORDER
  symmatrix REORDER
  symlist REORDER
;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LVAR
::
ASSEMBLE
	IF	f39
RPL
  CK1&Dispatch 
  sym :: USERLVAR SWAPDROP ;
  matrix :: USERLVAR SWAPDROP ;
  real :: NULL{} SWAPDROP ;
  cmp :: NULL{} SWAPDROP ;
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch 
  sym USERLVAR
  matrix USERLVAR
  real NULL{}
  cmp NULL{}
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Fraction -> numerator and denominator
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME  FXND
::
ASSEMBLE
	IF	f39
RPL
  CK1&Dispatch
  sym :: FXNDext TWO{}N ;
  real :: QPI FXNDext TWO{}N ;
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch
  sym FXNDext
  real %1
  cmp %1
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Equation -> left and right membre
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME EXLR
::
  CK1&Dispatch
ASSEMBLE
	IF	f39
RPL
  sym :: EXLR TWO{}N ;
ASSEMBLE
	ELSE	f39
RPL
  sym EXLR
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME LNAME
::
ASSEMBLE
	IF	f39
RPL
  CK1&Dispatch 
  sym :: USERLIDNT SWAPDROP ;
  matrix :: USERLIDNT SWAPDROP ;
  real :: NULL{} SWAPDROP ;
  cmp :: NULL{} SWAPDROP ;
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch
  sym USERLIDNT
  matrix USERLIDNT
  real NULL{}
  cmp NULL{}
ASSEMBLE
	ENDIF	f39
RPL
;

*----------------------------------------------------------------
* Modular operations
*----------------------------------------------------------------

**********************************************************************
* Name:		xADDTMOD
* Desc:		Modular addition
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME ADDTMOD
::
  CK2&Dispatch
  symsym :: 2QPIext ADDTMODpatch ;
  2matrix :: 2QPIext MADDTMODpatch ;
;

**********************************************************************
* Name:		xSUBTMOD
* Desc:		Modular substraction
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME SUBTMOD
::
  CK2&Dispatch
  symsym :: 2QPIext SUBTMOD ;
  2matrix :: 2QPIext MSUBTMOD ;
;

**********************************************************************
* Name:		xMULTMOD
* Desc:		Modular multiplication
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME MULTMOD
::
  CK2&Dispatch
  symsym :: 2QPIext MULTMOD ;
  SYMARRY :: 2QPIext SC*MATMOD ;
  ARRYSYM :: 2QPIext MAT*SCMOD ;
  2matrix :: 2QPIext MAT*MATMOD ; 
;

**********************************************************************
* Name:		xDIVMOD
* Desc:		Modular multiplication
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME DIVMOD
::
  CK2&Dispatch
  symsym :: 2QPIext DIVMOD ;
  ARRYSYM :: 2QPIext DIVMOD ;
;

**********************************************************************
* Name:		xDIV2MOD
* Desc:		Modular division
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME DIV2MOD
::
  CK2&Dispatch
  symsym :: 2QPIext FLAGDIV2MOD ;
;

**********************************************************************
* Name:		xPOWMOD
* Desc:		Modular power
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME POWMOD
::
  CK2&Dispatch
  # AFF ( symzint ) :: 2QPIext FLAGPOWMOD ;
  SYMREAL :: 2QPIext FLAGPOWMOD ;
  2REAL :: 2QPI FLAGPOWMOD ;
  # 4FF ( arryzint ) FLAGMPOWMOD
;

**********************************************************************
* Name:		xINVMOD
* Desc:		Modular inverse
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME INVMOD
::
  CK1&Dispatch
  sym :: QPI INVMOD ;
  real :: QPI INVMOD ;
  matrix :: QPI MINVMOD ;
;

**********************************************************************
* Name:		xGCDMOD
* Desc:		PGDC modulaire
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME GCDMOD
::
  CK2&Dispatch
  symsym :: 2QPIext GCD1MOD ;
;

**********************************************************************
* Name:		xEXPANDMOD
* Desc:		Modular EXPAND
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME EXPANDMOD
::
  CK1&Dispatch
  sym :: QPI FLAGEXPAMOD ;
  matrix :: QPI EXPAMOD ;
;

**********************************************************************
* Name:		xEXPANDMOD
* Desc:		Modular EXPAND
**********************************************************************
ASSEMBLE
	CON(3)	000
RPL
xNAME FACTORMOD
::
  CK1&Dispatch
  sym :: QPI FLAGFACTORMOD ;
  matrix :: QPI MFACTORMOD ;
;

**********************************************************************
* Name:		xRREFMOD
* Desc:		Modular RREF
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME RREFMOD
::
  CK1&Dispatch
  matrix :: QPI RREFMOD ;
  list :: LIST2MATRIX QPI RREFMOD ;
;

**********************************************************************
* Name:		xMODSTO
* Desc:		Modular MODSTO
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME MODSTO
::
ASSEMBLE
	IF 	f39
RPL
  CK1&Dispatch
  zint :: DUP STOMODULO ;
  real :: DUP QPI STOMODULO ;
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch
  zint STOMODULO
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		affiche en menu les commandes d'erable de numero
*		compris entre les 2 args
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME MENUXY
::
ASSEMBLE
	IF	f39
RPL
  CK0 BINT96 ERABLEERROR
ASSEMBLE
	ELSE	f39
RPL
  CK2&Dispatch
  2REAL :: COERCE2 MENUXYext InitMenu ;
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME KEYEVAL
::
ASSEMBLE
	IF	f39
RPL
  CK0 BINT96 ERABLEERROR
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch
  real KEYEVAL
  zint KEYEVAL
  any  EVAL
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		cree un grob en utilisant AGROB, ->GROB2 ou ->GROB
* Stack:	( obj -> grob )
**********************************************************************
*ASSEMBLE
*	CON(1)	8
*RPL
*xNAME XGROB
*::
*  CK1 XGROBext
*;

**********************************************************************
* Desc:		Add 2 grobs
* Stack:	( grob1 grob2 -> grob )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME GROBADD
::
  CK2&Dispatch
  # CC GROBADDext
;

**********************************************************************
* Desc:		transforme obj en grob met grob dans PICT puis
*		affiche PICT 
* Stack:	( obj -> )
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME SCROLL
::
ASSEMBLE
	IF	f39
RPL
  CK1&Dispatch
  grob :: DUP RunViewObject ;
ASSEMBLE
	ELSE	f39
RPL
  CK1 
  ViewObject 
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		CAS configuration
* Stack:	( -> )
* Note:		waiting for MODE-CAS entry, until then reset cas config
**********************************************************************
ASSEMBLE
         CON(1)  8
RPL
xNAME CASCFG
ASSEMBLE
	IF	f39
RPL
::
  CK0 
  CHANGECASCFG
;
ASSEMBLE
	ELSE	f39
RPL
::
  CK0 FLASHPTR RESETCASCFG
;
ASSEMBLE
	ENDIF	f39
RPL
**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME MAIN
::
  CK0
  chooseflag SysITE FALSE MENUCHOOSE?
ASSEMBLE
	IF 	f39
RPL
  casedrop :: MENUGENE1 "CAS MENU" TOTEMPOB OVER CHOOSE>$ 
              %2 Ck&DoChoose %0= caseDROP
              COERCE 
              DUP#1= case2drop FLASHPTR xMAIN
              #1- NTHCOMPDROP EVAL 
              ;
ASSEMBLE
	ELSE	f39
RPL
  case :: "CAS MENU" TOTEMPOB MENUGENE1 %1 4ROLL EVAL %0= ?SEMI EVAL ;
ASSEMBLE
	ENDIF	f39
RPL
  MENUGENE1
  InitMenu
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME ALGB
::
  CK0
  MENUBASE1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME CMPLX
::
  CK0
  MENUCMPLX1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME TRIGO
::
  CK0
  MENUTRIG1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME MATR
::
  CK0
  MENUMAT1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME DIFF
::
  CK0
  MENUDIFF1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME ARIT
::
  CK0
  MENUARIT1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME SOLVER
::
  CK0
  MENUSOLVE1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
*	IF	f39
*	CON(3) 	000
*	ELSE	f39
	CON(1)	8
*	ENDIF	f39
RPL
xNAME EXP&LN
::
  CK0
  MENUEXPLN1 MENUCHOOSE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
	CON(3) 	000
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME EPSX0
::
  CK1&Dispatch
  any ::  ELMGext TRIMext ;
;


**********************************************************************
**********************************************************************
* Desc:
* Stack:
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME FACTORIAL
*::
*  CK1&Dispatch
*  any	factorial
*;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME ?
::
  CK0 
  SYMB? NcaseSIZEERR
  ?ext
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME \9F
::
  CK0
  EXACT? SYMB? ANDcase PLUSINFext
  ( * create zint that overflow during Z>R conversion * )
  ZINT 500 ONE addtALOG DROP Z>R 
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
	IF	f39
        CON(1)  0*(hasFORM)+0*(hasINTG)+0*(hasWHERE)+0*(hasVUNS)
        CON(1)  0*(hasDER)+0*(hasINV)+0*(hasCOLCT)+0*(hasEXPND)
        CON(1)  0*(hasHELP)+1*(hasPDATA)
	ELSE	f39
	CON(1)	8
	ENDIF	f39
RPL
xNAME PROMPTSTO
::
ASSEMBLE
	IF	f39
RPL
  CK1&Dispatch
  idnt 
  :: 
  DUP ID>$			( id $ )
  "PROMPT FOR " OVER &$         ( id $ Title$ )
  OVER                          ( id $ Title$ Label$ )
  "Enter Value For " 4ROLL &$   ( id Title$ Label$ Help$ )
  4PICK SAFE@_HERE ?SKIP %0	( id Title$ Label$ Help$ value )
  ROMPTR xINPUT			( id value )
  DUPROT InCASApplication? ITE STO UserSto		( -> value )
  ;
  symb :: CARCOMP 'EvalNoCK: ROMPTR xPROMPTSTO ;
ASSEMBLE
	ELSE	f39
RPL
  CK1&Dispatch
  idnt PROMPTSTO1
  lam PROMPTSTO1
ASSEMBLE
	ENDIF	f39
RPL
;


**miscdoc+*********************************************************************
** algebraic function property list elements
**+miscdoc*********************************************************************

** help data (if present)   ( Textbook data )


* parse data (if present)

ASSEMBLE
	IF	f39
RPL

ASSEMBLE
        CON(5)  =DOHSTR
        REL(5)  +
        CON(1)  0
        CON(1)  pdOP		Operator
        CON(1)  pdLP		Open parenthesis
        CON(1)	pdQN		Quoted Name
        CON(1)  pdRP		Close parenthesis
        CON(1)  0
+
RPL
ASSEMBLE
	ENDIF	f39
RPL
        
* derivative (if present)

     
* inverse (if present)

        
* COLCT data (if present)

        
* EXPND data (if present)

        
* FORM menu (if present)

        
* Integral Data (if present)

        
* WHERE Building Data (if present)

        
* very-unsyminner Data (if present)


**********************************************************************
* Desc:		Undefined integration
* Stack:	( f(x), x, y -> F(y) where F'=f )
**********************************************************************
*ASSEMBLE
*         CON(3)  000
*RPL
*xNAME INT
*::
*  CK3&Dispatch
*  # 60 INT3
*;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
ASSEMBLE
         CON(3)  000
RPL
xNAME VER
::
  CK0
ASSEMBLE
	IF	f39
RPL
  "HP40 CAS by Parisse," TOTEMPOB DISPROW1
ASSEMBLE
	ELSE	f39
RPL
  "HP49 CAS by Parisse," TOTEMPOB DISPROW1
ASSEMBLE
	ENDIF	f39
RPL
  "Heiskanen & Fiechter" TOTEMPOB  DISPROW2
  SetDA1Temp
  VERNUMext
	
  ( * reset cas configuration if last error was a try to recover memory * )
  ERROR@ #5= IT FLASHPTR RESETCASCFG
;
**********************************************************************
