INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s
	
	STITLE	Factorization of polynomials

* Look at EVIDENT for deg 1, 2 and 4 factors

**********************************************************************
* EXPORT:
*
* COLCext	( symb -> symb' )
*		Factorization with respect to the current variable of symb
*		and factorization of the integer content of symb
* --------------------------------------------------------------------
* Internal format for factorization is a list or a meta
* with factor (Q) and multiplicity (%)
* For example 45=3^2*5 is 
*		   { Z3 %2 Z5 %1 }
*		   Z3, %2, Z5, %1, #4 
* Many factorization programs are available in the 2 formats
* --------------------------------------------------------------------
*
* MULMULText	( {}, % -> {} )
*
* METAMULMULT	( %, meta -> meta )
*
* METAMM2	( meta, % -> meta )
*		multiplies by % all multiplicities of meta
*
* COMPLIST	( {} -> {} )
*
* METACOMPRIM	( meta -> meta )
*		suppress multiples occurences of the same factor
*		by adding corresponding multiplicities
*
* ADDLISText	( {}, %, obj -> {} )
*		Adds obj with multiplicity % to the list
*
* DIVISext	( obj -> {} )
*		Returns the list of divisors of an object
*
* FACT1ext	( symb -> lvar, Q, {} )
*		{} is the list of root/multiplicity of symb wrt current var
*
* FACTOext	( symb -> lvar, Q, {} )
*		{} is the list of factors/multiplicity of symb
*
* SOLVext	( symb -> {} )
*		Numeric solver for univariate polynomials
*		{} is the list of roots (no multiplicity)
*
* FRND		( obj -> obj )
*		Float Rounding for %%, C%% or list of %%/C%%
*		Used by SOLVext to reconstruct factors
*
* BICARREE?	( P, #5 -> meta, P', TRUE )
*		( P, #5 -> P, #5, FALSE )
*		( P, # -> P, #, FALSE )
*		Search if P is a bisquarred 4th-order equation
*		Returns a meta of factors and the multiplying coeff 
*		in that case
*
* FEVIDENText	( P -> meta-fact, P' )
*		Find all 1st order factors of P and some other factors
*		if it can (e.g. if it remains a 2nd order poly or a
*		4th order bisquarred poly)
*
* m3&m2		( meta3, meta2, obj1 -> meta3&meta2, obj1 )
*		Combines 2 meta objects
*
* EVIDENText	( P -> meta-root, P' )
*		Like FEVIDENText but returns roots instead of factors
*
* DEG2ext	( P -> {} )
*		Returns the roots of a 2nd order polynomial
*
* FDEG2ext	( P -> meta-fact, coef )
*		Returns factors of a 2nd order polynomial and the
*		corresponding multipliying coeff
*
* PIext		( {} -> obj )
*		Returns the product of all elements of the list
*
* RACTOFACext	( {} -> {} )
*		From list of roots to list of factors
*
* FACTORACext	( f -> root, cst_coef )
*		Converts a factor to a root
*
* RFACText	( obj, # -> {}, internal object, meta )
*		{} is the list of variables
*		meta is made of roots or factors of num (N) or deno (D) or 
*		both (N/D) depending of #
*		ZERO for roots N/D
*		ONE for roots N
*		TWO for roots D with numeric solver call
*		THREE for roots D without numeric solver call
*		FOUR for factors N/D
*		FIVE for factors N
*		SIX for factors D with numeric solver call
*		SEVEN for factors D without numeric solver call
*
* RFACT2ext	( obj, {}, # -> {}, internal object, meta )
*		Like RFACText but the list of var is defined
*
* FACTTOOBJext	( { fact mult }, flag, prg*, prg^ -> obj )
*		Rebuild an object from his list of factors (flag=TRUE)
*		or roots (flag=FALSE)
*		using prg* to multiply and prg^ to take multiplicity power
*--------------------------------------------------------------------
* ADDLISText will be tested when improving TSIMP
* x^n+-1 might be tested at the beginning of RFACT2ext (after VXXLext)
* but full factorization would require changing the list of vars on the fly
* which is problematic for e.g. RISCHPF where input and output list
* of variables are not identical. I think the best is to use cyclotomic
* polynomials to get only integer factorization in this case.
**********************************************************************

**********************************************************************
* Desc:		Factorization with respect to the current variable
*		of symb and factorization of the integer content
*		of symb.
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME COLCext
::
  FOUR RFACText			( lvar, obj interne, meta-facteurs mult )
  get1 DROP			( lvar, meta-facteurs mult )
  SYMCOLCTEND
;

NULLNAME SYMCOLCT
( sym -> sym )
::
ASSEMBLE
	IF	f39
RPL
  NOT40case 48SYMCOLCT
ASSEMBLE
	ENDIF	f39
RPL

  CASFLAGEVAL:
  :: VX>LVARext                    ( Q lvar )
     ALG48FCTR0 get1 DROP           ( lvar meta-fact )
     SYMCOLCTEND
     ;
;

LABEL ALG48FCTR0
::                              ( obj lvar -> lvar meta-fact )
  SEVEN THREE TRUE              ( obj, {}, #, rem, ?facteurs )
  5PICK 5PICK
  ' NULLLAM FIVE NDUPN DOBIND   ( obj, {} )

  ( LAM1 = {}, lam2 =obj externe, lam3= flag, lam4= rem, lam5= temporaire )

  VXXLext                               ( {}, intob )
  DUP DUPTYPEZ? case :: MZSQFF ABND ;
             				( {}, intob, intob )
  FXNDext RFACTSTEP3 %-1 METAMM2        ( {}, intob, num, meta-deno )
  get1 RFACTSTEP3 top&                  ( {}, intob, meta-fact )

* need to suppress Z1 coming from METASQFFext

  METACOMPRIM                   ( {}, intob, meta-fact ) 
  ALG48FCTR3			( {}, obint, meta-factors or meta-roots )
  ABND                          ( {}, obint, meta-factors or meta-roots )
;

LABEL ALG48FCTR3
::				( meta-fact )
* Now factorize each square-free factor
  DUP#0= ?SEMI
  ZEROSWAP #2/
  ZERO_DO (DO)                  ( o1, m1, ..., on, mn, meta-res )
     get1 OVER #3+ ROLL         ( ..., on-1, mn-1, meta-res, mn, on )
     ALG48FCTR1 pshtop&		( o1, m1, .., on-1, mn-1, meta-res )
  LOOP				( meta )
;

LABEL ALG48FCTR1
* can be called recursively
::					( mn on -> meta )
  NOTLISTcase RFACTSTEP5
  DUPLENCOMP #1= NOTcase RFACTSTEP5
					( mn on )
  CARCOMP ALG48FCTR2			( meta )
  DUP#0= ?SEMI
  DUP #2/				( meta # )
  ZERO_DO (DO)                  	( meta )
     DUP ROLL OVER #1+ROLL		( meta-2_ob mn on )
     ONE>POLY SWAPROT            	( meta-res )
  LOOP
;

LABEL ALG48FCTR2
::					( mn on )
  RFACTSTEP3				( mn meta )
  METAMULMULT METACOMPRIM		( meta )
  ALG48FCTR3				( meta )
;

LABEL SYMCOLCTEND
( lvar, meta-fact_mult -> symb )
::
  DUP#0= case2drop Z1
  {}N				( lvar, lfact )
  DUPROT ZERO			( lfact, ?, lvar, #0 )
  3LAMBIND			( lfact )
  ZEROZERO ROTDUP >R		( meta-deno, meta-num, lfact )
  LENCOMP #2/
  ZERO_DO (DO)			( meta-deno, meta-num )
     RSWAP 'R'R RSWAP		( meta-deno, meta-num, fact, mult )
     3PUTLAM VAL1M 3GETLAM	( meta-deno, meta-num, meta-fact, mult )
     COLC1
  LOOP
  ABND				( meta-deno, meta-num )
  DUP#0=IT NDROPZ1ONE		( meta-deno, meta-num )
  psh				( meta-num, meta-deno )
  top&addt/			( meta-num/deno )
  CKSYMBN
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME COLC1
::				( meta-deno, meta-num, meta-fact, mult )
  %ABS DUPEQ: %0		( meta-deno, meta-num, meta-fact, mult, ?=0 )
  casedrop NDROP
				( meta-deno, meta-num, mult )
  colc^				( meta-deno, meta-num, meta-fact^mult )
  3GETLAM %0> NOTcase COLC2
  DUP #2+PICK #0=case top& 
  metamultnozck
;

LABEL colc^
( meta-fact mult -> meta-fact^mult )
::
  num#1=case DROP
  R>Z SWAP#1+ addt^
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME COLC2
::
  rot
  DUP#0= ITE_DROP top&addt*
  psh
;

**********************************************************************
* Desc:		Multiplies multiplicities in a factor list by coeff
* Stack:	( {} % --> {}' )
**********************************************************************
NULLNAME MULMULText ( liste, coeff -> liste )
::
  DUPNULL{}? caseDROP
  num#1=case DROP
  SWAPINCOMP

  NULLNAME METAMULMULT		( coeff, meta -> meta )
  :: get1 METAMM2 ;

  {}N
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME METAMM2
::				( o1, %m1, ..., on, %mn, #2n, % )
  SWAPDUP #2/			( o1, %m1, ..., on, %mn, %, #2n, #n )
  DUP#0=csedrp SWAPDROP
  ZERO_DO (DO)			( o1, %m1, ..., on, %mn, %, #2n )
     get1 UNROT			( %m1, ..., on, %mn, o1, %, #2n )
     get1 3PICK RMULText UNROT	( ..., on, %mn, o1, %m1*%, %, #2n )
  LOOP
  SWAPDROP
;

**********************************************************************
* Desc:
* Stack:	( {} --> {}' )
**********************************************************************
NULLNAME COMPLISText
::
  INNERCOMP METACOMPRIM
* on s'assure qu'on n'a pas enleve une racine nulle avec TRIMext juste avant
  {}N
;

**********************************************************************
* Desc:		Suppress multiple occurrances of the same factor
*		by adding corresponding multiplicities.
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME METACOMPRIM
::
  TRUE 1LAMBIND
  METACOMPRIM0 ABND
  ( evite de rendre un meta vide )

  DUP#0= NOT?SEMI
  DROPZ1 %1 TWO
;

LABEL METACOMPRIM0
::
  DUPTWO #/ DROP
  #1= IT :: Z0 OVER#2+UNROL #1+ ;
					( f1 m1...fk mk #2k )
  DUP#0= ?SEMI
  ZEROSWAP #2/				( f1 m1...fk mk meta-res #k )
  ZERO_DO (DO)				( f1 m1...fk mk meta-res )
     get1 rotomo NEWTRIMext	( f1 m1...fk-1 mk-1 meta-res mk fk )

     ( teste si fk est dans meta-res )
     1GETLAM			( true if remove Z1 allowed )
     CODE
		ST=0	sMETACOMPRIM
		GOSBVL	=popflag
		GOC	+
		ST=1	sMETACOMPRIM
+		GOSBVL	=SAVPTR
		C=DAT1	A	
		D0=C		D0=@fk
		A=DAT0	W
		C=A	W
		LCZ1
		?A#C	W
		GOYES	complistlp0
		?ST=1	sMETACOMPRIM
		GOYES	complistlp0
		GOSBVL	=GETPTR
		D1=D1+	10	2DROP
		D=D+1	A
		D=D+1	A
		A=D0		
		LC(5)	(complistlp!)-(complisttst!)
		A=A+C	A	-> LOOP
		D0=A
		GOVLNG	=Loop
complistlp0	D1=D1+	10	D1 pointe sur # du meta-res
complistlp1	C=DAT1	A
		CD1EX
		D1=D1+	5
		A=DAT1	A
		CD1EX
		R0=A.F	A	R0=nombre d'elements a regarder
complistlp	A=R0.F	A
		?A=0	A
		GOYES	complistok
		D1=D1+	10
		C=DAT1	A
		CD1EX
		RSTK=C
		GOSBVL	=EQUAL?	carry clear=EQUAL
		C=RSTK
		CD1EX
		GONC	complistok
		A=R0.F	A
		A=A-1	A
		A=A-1	A
		R0=A.F	A
		GOTO	complistlp
complistok	GOVLNG	=PUSH#LOOP

     ENDCODE
					( f1...fk-1 mk-1 meta-res mk fk # )
     LABEL complisttst!
     DUP#0=ITE METACOMP0 METACOMP1	( f1...fk-1 mk-1 meta-res )
     LABEL complistlp!
  LOOP

;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME METACOMP0
::
  DROP SWAPROT #2+
;

**********************************************************************
* Desc:
* Stack:	( f1...fk-1 mk-1 meta-res mk fk # -> f1...fk-1 mk-1 meta-res )
**********************************************************************
NULLNAME METACOMP1
::
  SWAPDROP ROT				( meta-res' mk # #meta )
  DUPROT #-				( meta-res' mk #meta #pos )
  DUP#2+ #2+ROLL			( meta-res' mk #meta #pos mk' )
  4ROLL RADDext				( meta-res' #meta #pos mk'+mk )

* si ca fait 0, enlever le facteur fk, sinon on le replace

  SWAPOVER				( met-res' #meta mk'+mk #pos mk'+mk )
  DUPTYPEZ? ITE Z0 %0			( met-res' #meta mk'+mk #pos mk'+mk 0 )
  EQUALNOTcase #2+UNROLL
  SWAPDROP #2+ROLL DROP #2-
;

**********************************************************************
* Desc:		Add ob with multiplicity %n to the list. Checks if ob is in {}
* Stack:	( {} %n ob --> {}' )
**********************************************************************
NULLNAME ADDLISText ( liste, multiplicite, objet -> liste )
::	
  DUP 4PICK ListPos			( l m o # )
  DUP#0=csedrp ADDLIST1

  ( * Check if obj was found in the multiplicities * )
  DUPONE #AND #0= casedrop ADDLIST1
					( l m o # )
  SWAPDROP				( l m # )
  #1+ 3PICKOVER NTHCOMPDROP		( l m # m' )
  ROT RADDext SWAPROT PUTLIST
;

**********************************************************************
* Desc:		Add ob with multiplicity %n to the list.
* Stack:	( {} %n ob --> {}' )
**********************************************************************
LABEL ADDLIST1
::
  ROT INNERCOMP get1
  SWAP#1+ get1 SWAP#1+
  {}N
; 


**********************************************************************
* Desc:		Return list of divisors of ob.
* Stack:	( ob --> {divisors} )
* Note:		objet peut etre algebrique ou entier
*		Dans le premier cas, DIVISext renvoie une liste de
*		facteurs premiers. Dans le second cas, la liste des
*		diviseurs si l'entier est < a #FFFFFh
**********************************************************************
NULLNAME DIVISext
::
  DUPTYPEEXT? case ONE{}N
  FACTOext UNROT2DROP			( { factors } )
  DIVISKILL1
					( { facteurs, multiplicite } )

					( * on supprime -1 si necessaire * )
  DUPDUP LENCOMP 
  DIVISKILL-1
  ZERO Z1 ROT DIVIS2 {}N
  DUPNULL{}? casedrop :: Z1 ONE{}N ;
;


LABEL DIVIS2
  ::					( LAM2 metadiv, d, lfact -> metadiv )
    DUPNULL{}? case2DROP
    INNERCOMP #2- {}N			( d1 ... dk #k d f m lfact )
    3PICK j#1=case
    :: UNROT2DROP DIVIS2 ;
    SWAP COERCE #1+
    ZERO_DO (DO)			( d1 ... dk #k d f lfact )
       ROTSWAP DIVIS1			( d1 ... dk #k f d lfact )
       SWAP3PICK RMULText		( d1 ... dk #k f lfact d*f )
       UNROT
    LOOP				( d1 ... dk #k d f lfact )
    3DROP
  ;

LABEL DIVIS1
::	( LAM1: metadiv, facteur, d, lfact -> metadiv, facteur, d,lfact )
   DUPNULL{}? casedrop
   :: ROT#1+UNROT DUP4UNROLL NULL{} ;
					( d1 ... dk #k facteur d lfact )
      THREE psh				( facteur d lfact #3 d1 ... dk #k )
      DUP #4+PICK			( facteur d lfact #3 d1 ... dk #k d )
      OVER #4+PICK			( save_meta d1 ... dk #k d lfact )
      DIVIS2				( save_meta metadiv )
      psh				( metadiv save_meta )
      DROP				( metadiv facteur d lfact )
;



LABEL DIVISKILL1
::					( * on supprime Z1 si necessaire * )
  Z1 OVER ListPos DUP#0=csDROP
					( liste, pos )
  SWAPINCOMP				( pos, o1, ..., on, #n )
  get1 OVER SWAP#-			( pos, o1, ..., on, #n, #n-k )
  DUP #3+ ROLL DROP #1+ ROLL DROP
  #2- {}N
;

LABEL DIVISKILL-1
::
  DUP#0= case2DROP
  #1- NTHCOMPDROP Z-1 EQUAL NOT?SEMI
  ONE OVER LENCOMP #2- SUBCOMP
;



**********************************************************************
* Desc:		{} is the list of root/multiplicity of sym with
*		respect to the current variable.
* Stack:	( symb-poly --> Lvar Q {} )
**********************************************************************
NULLNAME FACT1ext
::					( obj )
  ZERO RFACText 			( lvar Q meta )
  DUP#0=csedrp NULL{}			( -> lvar Q null{} )
					( lvar Q meta )
  {}N					( lvar Q {} )
  DUPLENCOMP SWAP >R			( lvar Q #roots )
  ZEROSWAP #2/				( lvar Q meta #roots )
  ZERO_DO (DO)
	RSWAP 'R'R RSWAP		( lvar Q meta fact mult )
	OVER TYPETAGGED? 		( lvar Q meta fact mult flag )
	ITE 2DROP :: ROT #2+ ;		( lvar Q meta )
	LOOP
  {}N
;

**********************************************************************
* Desc:		{} is the list of factors/multiplicity of symb
* Stack:	( symb --> Lvar Q {} )
**********************************************************************
NULLNAME FACTOext
::
  CK&DISPATCH1
  zint	ZFACTO
  seco	ZFACTO
  any	:: FOUR RFACText {}N ;
;

**********************************************************************
* Desc:
* Stack:	( C --> {} C Lfact )
**********************************************************************
NULLNAME ZFACTO
::
  NULL{} SWAPDUP MSQFF METACOMPRIM {}N
;

**********************************************************************
* Desc:		Numeric solver for univariate polynomials.
*		The list contains the roots without multiplicity.
* Stack:	( symb --> {} )
**********************************************************************
NULLNAME SOLVext ( polynome-liste -> liste des racines )
ASSEMBLE
	IF fSVERSION
RPL
*********************************************************	
* Routine de factorisation pour S/SX
**********************************************************
::
 ( polynome-liste, epsilon, nombre d'iterations -> liste des racines )
	
* Factorise un polynome-liste a coeffcients reels ou complexes
* par la methode de Laguerre ou de Bairstow.

  DUPTYPELIST? NOTcasedrop NULL{}
  DUPLENCOMP FOUR #<case DEG2ext

  numfactorflag TestSysFlag ITE FALSE EXACT?
  case CANTFACTOR 			( unable to factor symbolically )

  ( 6     5      4     3   2      1    )
  "HALT" "RAND" NULL$ DUP "?/10" "?*10"
  MENUext

  ( test et nombre d'iterations apres le test )

  THREE RCLEPS 3PICK LENCOMP UNCOERCE %* ROT

  ( #iterations, epsilon, liste )

  NULL{} SWAP TRIMext 

* conversion des coeff au format long

  INNERDUP #1-SWAP
  ZERO_DO
     get1 FXNDext
     OBJ2FLOAT SWAP
     OBJ2FLOAT SWAP
     RDIVext
  SWAPLOOP
  #1+ {}N
  ROT OBJ2FLOAT UNROT 
		
  ( #iterations, epsilon, root-list, P-liste )

  DUPLENCOMP TWO #< casedrop UNROT2DROP

  ( #iterations, epsilon, root-list, P-liste )

  DUP CARCOMP 5UNROLL		( sauvegarde denominateur )
				( a0, #it, epsilon, root-liste, liste )
  {
   BlankDA12 "SOLV, ?=" 4PICK %%>% DO>STR &$ DISPROW1
   DUP ONESWAP NULL{} VAL2ext DO>STR DISPROW2
   THREE TWO &$ DISPROW3
  }
  Sys1IT

  DUPINCOMP FALSE SWAP
  ZERO_DO (DO)			( Teste s'il y a un complexe dans la liste )
     SWAP XEQTYPE SWAPDROP %22 %= OR
  LOOP
			( a0, #it, epsilon, root-liste, liste , true ou false )
  ITE LAGUERRE BAIRSTOW
  EXACT? CLREXACT
  SWAP SOLV123 
  SWAP ITE SETEXACT CLREXACT
  SWAP &COMP
  UNROT2DROP
  DispMenu.1				( a1, root-l )
  COMPLEX? SWAP EXACT?		( a1, cmp_flag, nroot-l, exact? )
  REBUILDFCTR
;

LABEL OBJ2FLOAT
::
  CK&DISPATCH1
     real   %>%%
     ereal  NOP
     zint   :: Z>R %>%% ;
     cmp    C%>C%%
     ecmp   NOP
     seco   C2C%%
;


LABEL CKC%%
::
  CK&DISPATCH1
  ecmp  NOP
  ereal :: %%0 E%%>C%% ;
;

LABEL RAN2%% ( o1, o2 -> %%, o1, o2, %% )
:: 
  %RAN %.5 %- %>%% UNROT %RAN %.5 %- %>%%
;

LABEL C%%RAN
:: RAN2%% E%%>C%% ; 

LABEL LAGUERREKEY
( liste des touches->actions pour Laguerre )
{ 
    kcMenuKey1 
    :: ( Epsilon * 10 )
       TWELVE ROLL %%10 %%* "?= " OVER %%>% DO>STR
       &$ DISPROW1 TWELVE UNROLL
    ;
    kcMenuKey2
    :: ( Epsilon / 10)
       TWELVE ROLL %%10 %%/ "?= " OVER %%>% DO>STR
       &$ DISPROW1 TWELVE UNROLL
    ;
    kcMenuKey5
    :: ( Random ) TEN ROLL DROP
       C%%RAN TEN UNROLL
    ;
    kcMenuKey6 ( Stop ) DROPFALSE
}

LABEL LAGUERRE
::		
  ( Laguerre: #it, epsilon, root-list, P-liste )
  ( -> #it, epsilon, root-liste, P-liste )
  COMPLEXON
  ( transformation de reels longs en complexes longs )
  ' CKC%% 1LAMBIND LISTSECOext ABND
  ( fin de transformation )
  C%% 0 0 SWAP
  [BEGIN]
    DUPLENCOMP LAGUERREMSG Sys1IT
    THREE #>
  [WHILE]
    LAGUERRELP1
  [REPEAT]
  SWAPDROP
  ( fin de Laguerre )
;

LABEL LAGUERREMSG
{ DUP#1- #>$ EIGHT TWO &$SWAP &$ DISPROW5 SIX SEVEN &$ DISPROW6 } 

LABEL LAGUERRELP1
::
  %%1 UNROT DUPLENCOMP  #1- UNCOERCE%% DUP %%1 %%- 2DUP %%*
  %%2  %%*
  [BEGIN]
    4PICK 6PICK HORNext UNROT HORNext TEN ROLL %%ABS
    DUP %%>% DO>STR DISPROW7 TWELVE PICK %%>
    GETTOUCH IT :: LAGUERREKEY EQLookup NOTcaseDROP EVAL ;
  [WHILE]
    LAGUERRELP2
  [REPEAT]
  4ROLLSWAP C%%C/C C%%C-C 5UNROLL 4DROP
  { DUP C%%>C% DO>STR DISPROW4 } Sys1IT
  4ROLLOVER >TCOMP 4UNROLL HORNext 2DROP
;

LABEL LAGUERRELP2
::
  UNROT HORNext UNROT2DROP 3PICK C%%C*C
  4PICK C%%C*R OVER 6PICK C%%C*R DUP C%%C*C SWAP C%%C-C
  C%%SQRT
  2DUP C%%C+C UNROT C%%C-C DUP C%%ABS 3PICK C%%ABS %%>
  IT SWAP
  DROPDUP C%%ABS %%0=
  :: case2drop C%%RAN
     C%%C/C 4PICK C%%C*R
     ;
  DUP 7UNROLL 6ROLL SWAP C%%C-C 5UNROLL
;


LABEL BAIRSTOWKEY
( LAM4: liste des touches->actions pour Baistow )
( decal it eps [rac] s Po B p %%test -> m^me sch,ma de pile )
{
    kcMenuKey1
    :: ( Epsilon * 10 )
       7ROLL %%10 %%* "?=" OVER
       %%>% DO>STR &$ DISPROW1 7UNROLL
    ;
    kcMenuKey2
    :: ( Epsilon / 10)
       7ROLL %%10 %%/ "?=" OVER
       %%>% DO>STR &$ DISPROW1 7UNROLL
    ;
    kcMenuKey5
    :: ( Random s and p )
       2DROP ROTDROP RAN2%% %%1
    ;
    kcMenuKey6
    :: ( Stop ) DROP %%0 8PICK INDEXSTO ;
  }


LABEL BAIRSTOW
::				( Bairstow )
				( a0, #it, epsilon, root-liste, P-liste )
  DUP CARCOMP ' %%/ LOPAext	( a0, #it, epsilon, root-liste, P-liste )
  %%0 DUPDUP 7UNROLL ROT
  [BEGIN]			( %%decal, #it eps [rac] s p P )
    DUPLENCOMP #1- OVER
    TWO NTHCOMPDROP

    ( conditionne le polynome: somme des racines=0 )

    SWAP UNCOERCE%% %%/ %%CHS
    DUPUNROT PTAYL%%		( %%decal, #it eps {rac} s p %%-newdecal P )
    8ROLL ROT %%+ 7UNROLL	( %%decal #it eps {rac} s p P )
    DUPLENCOMP
    BAIRSTMSG1 Sys1IT
    DUP FIVE #=ITE BAIRST5 :: FOUR #> ;
  [WHILE]
    BAIRSTLP1
  [REPEAT]
				( %%decal #it eps {rac} s p P )
  UNROT2DROP			( %%decal #it eps {rac} P )
  5ROLL %%CHS PTAYL%%		( #it eps {rac} P )
  ( fin de Bairstow )
;

LABEL BAIRSTLP1
::
  6PICK				( nombre d'iterations )
  DUP				( remplace poly B)
  ZERO_DO (DO)			( 2 iterations aprSs le test < EPS )
				( [rac], s, p, poly, deflated_poly )
    DROPDUP 2SWAP		( [rac], poly, poly, s, p )
    BAIRST%%			( [rac], poly, Q, s, p )
    3PICK 4UNROLL BAIRST%%	( [rac] poly Q R s p )
    2SWAP			( [rac] poly s p Q R )
    INNERCOMP SWAPDROP
    #4- 4UNROLL
    THREE psh NDROP DROP	( [rac] poly s p Q c[n-3] c[n-2] c[n-1] )
    4ROLL INNERCOMP #2-
    UNROT TWO psh {}N
    SWAPDROP 8UNROLL	( [rac] poly deflated_poly s p c[n-3] c[n-2] c[n-1] b[n-1] b[n] )

    ( correction de c[n-1]:=c[n-1]-b[n-1] )

    ROT 3PICK %%- UNROTDUP	( S=b[n]*c[n-3]-b[n-1]*c[n-2] )
    6PICK %%*
    3PICK 6PICK %%* %%-	( [rac] poly deflated_poly s p c[n-3] c[n-2] c[n-1] b[n-1] b[n] S )

    ( P=b[n]*c[n-2]-b[n-1]*c[n-1] )

    SWAP 5PICK %%* ROT 4PICK %%* %%-

    ( [rac] poly deflated_poly s p c[n-3] c[n-2] c[n-1] S P )

    ( D=c[n-2]^2-c[n-1]*c[n-3] )

    2SWAP 5ROLL %%* SWAPDUP %%* SWAP %%-	( [rac] Po defl_pol s p S P D )
    DUP %%0=
    ITE :: DROP %RAN %.5 %- %>%% ; %%1/		( [rac] Po B s p S P D^-1 )
    DUPROT %%*UNROT %%*				( [rac] Po B s p Dp Ds )
    DUP %%ABS 3PICK %%ABS %%+ 5UNROLL		( [r] Po B |Dp+Ds| s p Dp Ds )
    UNROT %%+ UNROT %%+				( [rac] Po B |Dp+Ds| p s )
    DUP 6UNROLL %%ABS				( [rac] s Po B |Dp+Ds| p |s| )
    OVER %%ABS %%+ ROTSWAP	( decal it eps [rac] s Po B p |Dp+Ds| |s|+|p| )
    DUP %%0= ITE BAIRSTLP1_0 BAIRSTLP1_1
    UNROT
    LOOP

  SWAPDROP			( [rac] s p B )
  3PICK %%2 %%/ DUPDUP %%*	( [rac] s p B .5*s .25*s^2 )
  4PICK %%- DUP %%0>=		( %%decal #it eps {rac} s p B .5*s delta >=0? )
				( 9        8  7    6    5 4 3  2    1 )
  ITE BAIRSTLP1_2 BAIRSTLP1_3	( [rac] s p B [ r1 r2 true | {poly} false )
  ITE BAIRSTLP1_4 BAIRSTLP1_5
  {}N 4UNROLL
;

LABEL BAIRSTLP1_0
:: 3DROP ROTDROP RAN2%% ZERO INDEXSTO ;

LABEL BAIRSTLP1_1
:: 
   %%/
   BAIRSTMSG4 Sys1IT		( decal it eps [rac] s Po B p %%test )
   GETTOUCH
   IT :: BAIRSTOWKEY EQLookup NOTcaseDROP EVAL ;
   7PICK %%> NOT?SEMI
   ZERO INDEXSTO
;

LABEL BAIRSTLP1_2
:: 
  %%SQRT 2DUP %%+ TEN PICK %%+
  UNROT %%- NINE PICK %%+ TRUE
;

LABEL BAIRSTLP1_3
:: 
  SWAP NINE PICK %%+ SWAP %%CHS %%SQRT
  2DUP %%CHS E%%>C%% UNROT %%>C%% TRUE
;
	
LABEL BAIRSTLP1_4
:: 
  BAIRSTMSG2 Sys1IT
  6ROLL INNERCOMP #2+
;

LABEL BAIRSTLP1_5
:: 
  BAIRSTMSG3 Sys1IT
  5ROLL INNERCOMP #1+
;

LABEL BAIRSTMSG1
{ 
	DUP#1- UNCOERCE DUP %1- DO>STR "," &$SWAP DO>STR &$
	EIGHT TWO &$SWAP &$ DISPROW6
}

LABEL BAIRSTMSG2
{ DUP %%>% DO>STR DISPROW4 OVER %%>% DO>STR DISPROW5 }

LABEL BAIRSTMSG3
{ DUP ONESWAP NULL{} VAL2ext DO>STR DISPROW4 NULL$ DISPROW5 }

LABEL BAIRSTMSG4
{ TWENTYSIX OVER %%>% DO>STR &$ DISPROW7 }
		
LABEL BAIRST5
::					( s, p, P, #5 -> s, p, P )
   DROP UNROT2DROP			( P )
   DUPINCOMP DROP 4ROLLDROP 4ROLLDROP	( P, c,d,e)
   2DUP 5PICK %%2 %%CHS %%/ UNROT	( P, c,d,e,-c/2,d,e )
   %%CHS SWAP				( P, c,d,e,-c/2,-e,d )
   6PICK 5PICK %%*			( P,c,d,e,-c/2,-e,d,ce )
   SWAPDUP %%* %%4 %%/ %%- %%2 %%/
   %%1 4UNROLL				( P,c,d,e,1,-c/2,-e,-d^2/8+ce/2)
   FOUR {}N SOLV123
   DUPLENCOMP NTHCOMPDROP		( P, c,d,e,a)
   DUP %%2 %%* 5ROLL %%- %%ABS %%SQRT	( P, d,e,a,sqrt[2a-c] )
   4ROLL %%0>= ?SKIP %%CHS		( P, e,a, s )
   4UNROLL DUPDUP %%*			( s,P,e,a,a^2 )
   ROT %%- %%ABS %%SQRT %%+ SWAP	( %%decal #it eps {rac}, s,p,P)
   TRUE
;

LABEL BAIRST%%
* Division euclidienne par un polynome quadratique
( P, s, p -> P div X^2-s*X+p & P mod X^2-s*X+p, s, p )
( P, s, p -> Q, s, p )
CODE
		GOSBVL	=SAVPTR
		C=DAT1	A
		CD1EX
		RSTK=C
		GOSUB	getcd1
		GOSBVL  =STCD0		p -> R0:R1
		C=RSTK
		D1=C
		D1=D1+	5
		C=DAT1	A
		CD1EX
		RSTK=C
		GOSUB	getcd1
		GOSBVL  =STCD2		s -> R2:R3
		C=RSTK
		D1=C
		D1=D1+	5
		A=DAT1	A
		R4=A.F	A	R4 -> P
		D0=A		
		AD1EX
		C=DAT1	A
		AD1EX
		LA(5)	=DOLIST
		?C=A	A
		GOYES	bairstlist
		GOVLNG	=GETPTRLOOP
bairstlist	GOSBVL	=SKIPOB	DO -> end P
		A=D0
		C=R4.F	A
		C=A-C	A	C=longueur
		C=C+C	A	
		A=C	A
		C=C+C	A
		C=C+A	A	C=6*long (%%=13 bytes, ROM %%=2.5, 13/2.5<6)
		GOSBVL	=GETTEMP	D0 -> P div X-r
* initialisation
		SETDEC
		A=D0
		AR4EX.F	A	R4 -> P div X-r, A -> P
		D1=A
		D1=D1+	5	( skip DOLIST )
		GOSUB	getab1	( initialisation de X )
		LC(5)	=DOLIST
		DAT0=C	A
		D0=D0+	5
		LC(5)	=DOEREL
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=PUTAB0	( ecrire bn=an )
		C=D0
		RSTK=C
		GOSBVL	=RCCD2	( s )
		GOSBVL	=MULTF	( A:B= s*bn )
		GOSUB	getcd1
		GOSBVL	=RADDF	( A:B=an-1+s*bn )
		C=RSTK
		D0=C
		P=	0
		LC(5)	=DOEREL
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=PUTAB0
* D1 -> ak-1 de P, D0 -> P div X-r, A:B=bk+1, R2:R3=r, 
* dans la boucle: arret si D0 -> SEMI,
* sinon ecrire R0:R1 puis faire R0:R1=R0:R1*R2:R3+ak-1 
bairstloop	C=A	A		sauvegarde A.A
		RSTK=C
		A=DAT1	A		test si D1 -> SEMI
		LC(5)	=SEMI
		?A#C	A
		GOYES	bairstdolp
		GOTO	bairstendlp
bairstdolp	C=RSTK
		A=C	A		restaure A.A -> A:B=bk+1
		GOSBVL	=RCCD2
		GOSBVL	=MULTF		A:B=bk+1*s
		C=D0
		RSTK=C
		GOSUB	getcd1
		GOSBVL	=RADDF		A:B=ak+bk+1*s
		C=RSTK
		D0=C
		GOSBVL	=PUTAB0		sauvegarde temporaire de A:B
		D0=D0-	16
		D0=D0-	16
		D0=D0-	16
		D0=D0-	16
		D0=D0-	4		D0 -> bk+2 (68=21+21+5+21)
		GOSBVL	=GETAB0		A:B=bk+2
		GOSBVL	=RCCD0		C:D=p
		GOSBVL	=MULTF		A:B=bk+2*p
		A=A-1	S
		GOC	bairstchsok
		A=0	S
bairstchsok	D0=D0+	16
		D0=D0+	10		D0 -> sauv. temp. de A:B (26=21+5)
		C=D0
		RSTK=C
		GOSBVL	=GETCD0		C:D=ak+bk+1*s
		GOSBVL	=RADDF
		C=RSTK
		D0=C
		P=	0
		LC(5)	=DOEREL
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=PUTAB0		ecrire bk
		GOTO	bairstloop
bairstendlp	SETHEX
		DAT0=C	A		write SEMI
		D0=D0+	5
		C=RSTK
		C=D0			C -> end P div X-r
		A=R4.F	A		A[A] = -> P div X-r
		C=C-A	A		C[A] = Qsize
		C=C+CON	A,6
		DAT0=C	A		Fix link
		D0=D0+	5		Skip link
		GOSBVL	=recover	Shrink allocated memory
		GOSBVL	=GETPTR		stk1=p, stk2= s, stk3= P
		A=R4.F	A		A[A] = ->Q
		AD1EX
		C=DAT1	10
		AD1EX
		A=C	W
		LA(N)	10
		CON(5)	=DOLIST
		CON(5)	=SEMI
		?A#C	W
		GOYES	bairstendlp1
		LA(5)	=%%0
		GOTO	bairstendlp2
bairstendlp1	A=R4.F	A
bairstendlp2	D1=D1+	10
		DAT1=A	A		stk2= P div X-r
		D1=D1-	10
		GOVLNG	=Loop
ENDCODE

LABEL HORN%%
* Schema de Horner pour des %%
* ( polynome-liste P, r -> P[X] div [X-r], r, P[r] )
* utilise tous les registres sauf R0 et R1
*:: SHALT 
CODE
		GOSBVL	=SAVPTR
		C=DAT1	A
		CD1EX
		RSTK=C
		GOSUB	getcd1
		GOSBVL  =STCD2		r -> R2:R3
		C=RSTK
		D1=C
		D1=D1+	5
		A=DAT1	A
		R4=A.F	A	R4 -> P
		D0=A		
		AD1EX
		C=DAT1	A
		AD1EX
		LA(5)	=DOLIST
		?C=A	A
		GOYES	hornlist
		GOSBVL	=GETPTR
		LC(5)	=%%0
		D1=D1+	5
		DAT1=C	A
		D1=D1-	10
		D=D-1	A
		GONC	hornmemok
		GOSBVL	=GARBAGE1
hornmemok	A=R4.F	A
		GOVLNG	=OverWrALoop
hornlist	GOSBVL	=SKIPOB	DO -> end P
		A=D0
		C=R4.F	A
		C=A-C	A	C=longueur
		C=C+C	A	
		A=C	A
		C=C+C	A
		C=C+A	A	C=6*long (%%=13 bytes, ROM %%=2.5, 13/2.5<6)
		GOSBVL	=GETTEMP	D0 -> P div X-r
* initialisation
* attention a %%0
		A=D0
		AR4EX.F	A	R4 -> P div X-r, A -> P
		D1=A
		D1=D1+	5	( skip DOLIST & DOEREL )
		GOSUB	getab1
		LC(5)	=DOLIST
		DAT0=C	A
		D0=D0+	5
		SETDEC
* D1 -> ak-1 de P, D0 -> P div X-r, A:B=bk, R2:R3=r, 
* dans la boucle: arret si D0 -> SEMI,
* sinon ecrire R0:R1 puis faire R0:R1=R0:R1*R2:R3+ak-1 
hornloop	C=A	A		sauvegarde A.A
		RSTK=C
		A=DAT1	A		test si D1 -> SEMI
		P=	0
		LC(5)	=SEMI
		?A=C	A
		GOYES	hornendlp
		C=RSTK
		A=C	A		restaure A.A
		LC(5)	=DOEREL
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=PUTAB0		ecrire A:B en D0
		CD0EX			sauvegarde D0
		RSTK=C
		GOSBVL	=RCCD2		R2:R3 -> C:D
		GOSBVL	=MULTF		A:B= A:B *C:D
		GOSUB	getcd1		D1 -> C:D
		GOSBVL	=RADDF		A:B= A:B + C:D
		C=RSTK
		CD0EX			restaure D0
		GOTO	hornloop
hornendlp	SETHEX
		DAT0=C	A		write SEMI
		D0=D0+	5
		C=RSTK
		A=C	A		restaure A.A
		GOSBVL	=STAB2		sauvegarde le resultat en R2:R3
		C=D0			C -> end P div X-r
		A=R4.F	A		A[A] = -> P div X-r
		C=C-A	A		C[A] = Qsize
		C=C+CON	A,6
		DAT0=C	A		Fix link
		D0=D0+	5		Skip link
		GOSBVL	=recover	Shrink allocated memory
		GOSBVL	=GETPTR		stk1=r, stk2= P
		A=R4.F	A		A[A] = ->Q
		AD1EX
		C=DAT1	10
		AD1EX
		A=C	W
		LA(N)	10
		CON(5)	=DOLIST
		CON(5)	=SEMI
		?A#C	W
		GOYES	hornendlp1
		LA(5)	=%%0
		GOTO	hornendlp2
hornendlp1	A=R4.F	A
hornendlp2	D1=D1+	5
		DAT1=A	A		stk2= P div X-r
		D1=D1-	5
		GOSBVL	=RCAB2		recupere le resultat
		GOVLNG	=PUSH%%LOOP
getcd1		P=	0
		C=DAT1	A
		D1=D1+	5
		D=C	A
		LC(5)	=DOEREL
		?C#D	A
		GOYES	getcd1_rom
		GOVLNG	=GETCD1
getcd1_rom	CDEX	A	C -> objet en ROM
		CD1EX		D1 -> objet en ROM, C= D1
		RSTK=C		sauvegarde de D1
		D1=D1+	5
		GOSBVL	=GETCD1	C:D
		D1=C		sauvegarde champ A de C
		C=RSTK
		CD1EX		restaure C.A et D1
		RTN
getab1		P=	0
		A=DAT1	A
		D1=D1+	5
		B=A	A
		LA(5)	=DOEREL
		?A#B	A
		GOYES	getab1_rom
		GOVLNG	=GETAB1	( initialisation de X )
getab1_rom	GOSBVL	=XYEX
		GOSUB	getcd1_rom
		GOSBVL	=XYEX
		RTN
  ENDCODE
*;

LABEL PTAYL%%
:: 
  OVER LENCOMP			( P r #n )
  DUP4UNROLL
  ZERO_DO (DO)
        HORN%% 4UNROLL
  LOOP
  2DROP reversym {}N
;

LABEL SOLV123
:: ( LAM9: polynome de degre 1, 2 ou 3 -> racines )
   DUPLENCOMP
   DUP#0=csDROP
   FOUR #=casedrop SOLV3
   #>2case SOLV2
   ( degree 1)
   INCOMPDROP SWAP %%/ %%CHS ONE{}N
;

LABEL SOLV3
:: 
   DUPTWO NTHCOMPDROP %%3 %%/ %%CHS DUPUNROT PTAYL%% DUP
   INCOMPDROP ROTDROP
   3PICK %%2 %%* %%/ %%CHS SWAPROT %%3
   %%* %%/ DUP 2DUP %%* %%* 3PICK DUP %%* %%+
   DUP %%0>= case SOLV3%%+
   DROP %%CHS %%SQRT DUP %%2 %%*UNROT DUPDUP %%* %%* %%/
   DUP %%ABS %%1 %%>
   ITE :: %%0>= ITE %%0 %%2PI %%2 %%/ ; %%ACOSRAD
   THREE ZERO_DO (DO)
           %%2PI %%+ 2DUP %%3 %%/ %%COSRAD %%* 5PICK %%+ 5UNROLL
        LOOP
   4DROP THREE{}N
;

LABEL SOLV3%%+
:: 
   SWAPDROP %%SQRT 2DUP %%+ UNROT %%-
   SOLV3%%XR SWAP SOLV3%%XR %%+
   HORN%% DROP 3PICK %%+ UNROT
   SWAP %%CHS PTAYL%% SOLV123 SWAP >TCOMP
;

LABEL SOLV3%%XR
::
  DUP %%0= ?SEMI DUP %%0>
  DUPUNROT ?SKIP %%CHS %%LN %%3 %%/
  %%EXP SWAP ?SKIP %%CHS
;


LABEL SOLV2
::				( degr\'e 2 { a, b, c } ) 
   DUP CARCOMP
   CK&DISPATCH1
   SIXTYTHREE	SOLV2%%
   any		:: DROP DEG2ext ;
;

LABEL SOLV2%%
::	
  DROPDUP INCOMPDROP
  3PICK %%/ SWAPROT %%2 %%* %%/ %%CHS DUPDUP %%*ROT
  %%- DUP %%0>=
  case SOLV2%%+
  %%CHS %%SQRT 2DUP %%CHS E%%>C%% UNROT
  E%%>C%% TWO{}N SWAPDROP
;

LABEL SOLV2%%+
:: 
  %%SQRT 2DUP %%+ UNROT %%-
  TWO{}N SWAPDROP
;

************************************************************
* Fin de routine de factorisation pour S/SX
*********************************************************
ASSEMBLE
	ELSE
RPL

**********************************************************
* Routine de factorisation pour G/GX
**********************************************************
::
  DUPTYPELIST? NOTcasedrop NULL{}
  DUPLENCOMP DUPTWO #<				( l, #, ? )
  case2drop NULL{}
  FOUR #<case DEG2ext

  numfactorflag TestSysFlag ITE FALSE EXACT?
  case WARN&CANTFACTOR 			( unable to factor symbolically )

  ( * convert back numerical square roots for crunching * )
  ONESWAP NULL{} VAL2ext			( liste )

  ( * crunch and test unit dimensions, bind root unit dimension * )
  ( * set lam2 to an which is non 0, lam1 to the unit dim or %0 for no unit * )
  DUP CARCOMP %0 2LAMBIND 		( list )
  DUP >R 'R CASCRUNCH ONE		( list an #1 )
  ROT LENCOMP				( meta #n+1 )
  ONE_DO 
    RSWAP 'R RSWAP CASCRUNCH		( meta ak )
    SWAP#1+				( meta )
    ( * now check units * )
    CHECKUNIT				( meta )
    ( * remove unit dimension * )
    SWAP DUPTYPEEXT? IT CARCOMP
  SWAPLOOP
  {}N						( liste )

  DUP CARCOMP DUPTYPECMP? ITE C%>C%% %>%%	( liste, a1 )
  SWAPINCOMP					( a1, a1, .., an, #n )
  DUP#1+ NDUP TRUE SWAP
  ZERO_DO (DO)
     SWAP DUPTYPEREAL? ITE DROPTRUE TYPECMP?
     AND
  LOOP
  NOTcase CANTFACTOR			( non-univariate polynomial )
					( unable to factor numerically )
  UNCOERCE ' x\->ARRY EvalNoCK
ASSEMBLE
	IF	f39
RPL
  proot
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR proot 
ASSEMBLE
	ENDIF	f39
RPL
  ARRAY2MATRIX

* conversion des racines en complexes ou reels longs

  COMPLEX? SETCOMPLEX EXACT? ROT	( a1 cmp_flag exact? nrrot-l )
  CLREXACT NULL{} VXXLext		( a1 cmp_flag exact? {} nroot-l )
  SWAPDROPSWAP				( a1 cmp_flag nroot-l exact? )
  ( * rebuild factors with units * )
  REBUILDw/UNIT
;

LABEL CHECKUNIT
* meta -> meta
* check 
::					( meta )
  OVER num#0=case DROP
  2GETLAM EvalNoCKx/			( meta an-k/an )
  'EvalNoCK: xABS INDEX@ UNCOERCE 	( meta |an-k/an| %k )
  SWAP 'EvalNoCK: xXROOT		( meta |an-k/ak|^1/k )
  1GETLAM num#0=case SETUNIT?
					( meta an-k/ak^1/k lam1 )
  CK&DISPATCH1
  2EXT	CHECKSAMEUNIT
  EXTOB SETSIZEERR			( should be inconsistent units )
  EXT   SETSIZEERR			( should be inconsistent units )
  any   2DROP				( -> meta )
;

LABEL SETUNIT?
* unit is not set, obj is not 0, we can set the unit dimension
* + if obj is a unit object, check that we are in real mode
::					( meta an-k/ak^1/k %0 )
  DROP
  CK&DISPATCH1
  EXT SETUNITEXT
  any 1PUTLAM				( -> meta )
;

LABEL SETUNITEXT
::
  COMPLEX? case SETSIZEERR		( unit not allowed in complex mode )
  %1 Unbr>U 1PUTLAM			( -> meta )
;

LABEL CHECKSAMEUNIT
::					( meta an-k/ak^1/k lam1 )
  UM/ 					( meta an-k/ak^[1/k]/lam1 )
  CK&DISPATCH1
  EXT SETSIZEERR			( should be inconsistent units )
  any DROP				( -> meta )
;

LABEL REBUILDw/UNIT
::
  REBUILDFCTR				( list )

  EXACT? case ABND			( -> list: no unit here )

  ( * now sets units for cst coeff in 1st degree factors * )
  ( * and unit^2 for cst coeff, unit^1 for x coeff in 2nd order coeff * )
  ZERO OVER LENCOMP 			( meta&list # )
  #1+_ONE_DO (DO)			( f1..fn list #n )
    OVERINDEX@ NTHCOMPDROP		( f1..fn list #n fk )
    DUPTYPETAG? ?SKIP REBUILDw/ULP	( f1..fn list #n fk' )
    UNROT #1+				( f1..fn+1 list #n+1 )
  LOOP
  SWAPDROP ABND {}N
;

LABEL REBUILDw/ULP
* there are units involved hence approx mode is set
* hence REBUILDFCTR returns complex roots
* therefore we must return the corresponding real factor and suppress
* the conjugate root
::
  CK&DISPATCH1
  ereal LAM1UNIT
  ecmp  REBUILDw/C%%
;

LABEL REBUILDw/C%%
::					( c%% )
  COMPLEX? case LAM1CUNIT
  C%%>%% 				( re im ) 
  DUP%%0= casedrop LAM1UNIT		( -> re )
					( re im )
  OVER %%2 %%* %%CHS 			( re im -2re )
  LAM1UNIT				( re im -2*re_unit )
  UNROTDUP %%* SWAPDUP %%* %%+		( -2re_unit re^2+im^2 )
  %%>% 1GETLAM DUP UM* SWAP Unbr>U	( -2re_unit re^2+im^2_unit )
  %%1 TOTEMPOB UNROT THREE{}N NULL$ >TAG 
  INDEX@ #1+ INDEXSTO
;

LABEL LAM1UNIT
:: 
  1GETLAM DUPTYPEEXT? NOTcaseDROP
  SWAP %%>% Unbr>U 
;

LABEL LAM1CUNIT
::
  1GETLAM DUPTYPEEXT? NOTcaseDROP
  ( should convert C%% to symbolic and add unit )
					( C%% unit )
  SETTYPEERR				
;
	
* Fin de routine de factorisation pour G/GX
	
ASSEMBLE
	ENDIF
RPL

LABEL rndegal!
::					( obj )
  DUP FRND				( obj, obj rnd )
  SWAPOVER RSUBext
  ELMGext TRIMOBJext
  Z0 EQUAL				( obj rnd, obj=obj rnd? )
;



LABEL REBUILDFCTR
::					( a1 cmp_flag nroot-l exact? )
  ROT ?SKIP CLRCOMPLEX			( a1 nroot-l exact? )
  NOTcase SWAPDROP

* reconstruction eventuelle des racines sous forme formelle
* test a1 d'abord
					( a1, nroot-l )
  SWAP rndegal!				( nroot-l, a1, flag )
  NOTcaseDROP
  SWAP SETEXACT
  BINT21 GETERABLEMSG
  displayflag SysITE DISPROW7 DROP	( a1, nroot-l )

  INNERCOMP				( deno, r1, ..., rn, #n )
* convertir les complexes longs dont la prtie imag est nulle en reels long

  DUP #1-SWAP
  ZERO_DO 
     get1 ELMGext
  SWAPLOOP
  #1+ get1 NULL{}			( r1, ..., rn, #n, deno, {} )
  COMPLEX? ?SKIP SUPPRCONJ
  1LAMBIND OVER				( roots, #n, deno, #n )
  BEGIN					( meta-root, deno, #racines_a_tester )
     REBUILDLP
     DUPTWO #<
  UNTIL					( meta-root, deno , #t )
  2DROP {}N 1GETABND			( {roots}, f-roots )
  
					( roots, f-roots )

  DUPNULL{}? 
  ITE :: ( no formal roots found )
	 DROP EXACTOFF
	; 
	&COMP
  COMPLEX? ?SEMI

* pour les racines complexes on renvoie { 1 -2*re norme }

  INNERDUP #1-SWAP
  ZERO_DO 
     get1 REBUILDREAL
  SWAPLOOP
  #1+ {}N
* Fin de routine de factorisation numerique
;


LABEL SUPPRCONJ
::
  ROTDUP #1+SWAP

* teste d'abord par couples les conjugues et n'en laisse qu'un seul
	
				( r1, ..., rn, a1, {}, #n+1, #n )
  ZERO_DO (DO)			( r1, ..., rn, a1, {}, #n+1 )
    get1 SUPPRCONJLP
    LOOP			( r1...rn, a1, {}=f-root-list, #n+1 )
  #1- UNROT
;				( r1, ..., rn, #n, deno, f-root-list )

LABEL SUPPRCONJLP
::
  CK&DISPATCH1
  ecmp	SUPPRCC%%
  any	4UNROLL
;

LABEL SUPPRCC%%
::
  INDEX@ #1+ INDEXSTO
  SWAP get1			( r3...rn a1 {} r1 #n+1 r2=r1barre )
  SWAP#1- UNROT			( r3...rn a1 {} #n r1 r1barre )
  2DUP RMULText
  6PICK RMULText		( r3...rn a1 {} #n r1 r1bar module*a1 )
  rndegal!
  NOTcase2drop 4UNROLL
  SWAP C%%>%% DROPDUP %%+ %%CHS	( r3...rn a1 {} #n r1 module*a1 -2*re[r1] )
  6PICK %%* rndegal!
  NOTcase2drop 4UNROLL
  6PICK SWAPROT THREE{}N
  exactify!			( r3... rn a1 {} #n r1 facteur )
  SWAPDROP NULL$ >TAG
  ROTSWAP >HCOMP SWAP#1-
;


LABEL REBUILDLP
::
  DUPTWO #< ?SEMI
					( meta-root, a1, #t )
* deplace la derniere racine du meta-root en 1ere position 

  4ROLL 4PICK #3+ UNROLL		( meta-root, a1, #t )

* recherche des facteurs d'ordre 2 avec la derniere racine rl de meta-root

  DUP 
  ONE_DO (DO)
    REBUILDLPLP
    LOOP				( meta-root, deno, #t )
  #1-					( meta-root, deno, #t-1 )
;

LABEL REBUILDLPLP
::				( roots, #n, a1, #t )
  5ROLL 5ROLL 2DUP		( roots-2, #n, a1, #t, rk, rl, rk, rl )
  RADDext 5PICK RMULText
  rndegal!
  ITE TRUE REBUILDNEQ		( roots-2 #n a1 #t rk rl [rk+rl]*a1 flag )

* si ca marche pas on remet rl en place, mais rk est place plus en tete

  NOTcasedrop rebuildcase
  RNEGext			( roots #n deno #t rk rl [-rk-rl]*deno )
  UNROT 2DUP RMULText		( roots #n deno #t -rk-rl rk rl rk*rl )
  6PICK RMULText rndegal!	( roots #n deno #t -rk-rl rk rl rk*rl flag )
  NOTcasedrop :: ROTDROP rebuildcase ;
  ExitAtLOOP 
  UNROT2DROP 4PICK UNROT	( roots #n deno #t deno -rk-rl rk*rl )
  THREE{}N			( roots, #n, deno, #t, P-list )
  exactify!
  COMPLEXEVAL DEG2ext		( roots, #n, deno, #t, P-list )
  1GETLAM &COMP			( roots, #n, deno, #t, froot-list )
  1PUTLAM ROT #2- 		( roots, deno, #t, #n-2 )
  UNROT #1-			( roots, #n-2, deno, #t-1 )
;

LABEL rebuildcase
:: 5UNROLL 4PICK #3+ UNROLL ;

LABEL REBUILDNEQ
::
  COMPLEX? caseFALSE
  DROP RCONJext 2DUP RADDext 5PICK RMULText
  rndegal!
;

LABEL exactify!
:: 
  COMPLEX? SETCOMPLEX
  ONE ROT NULL{} VAL2ext NULL{} VXXLext SWAPDROP
  SWAP ?SKIP CLRCOMPLEX
;

LABEL REBUILDREAL
::
  DUPTYPETAG? ?SEMI
  DUP RCONJext 2DUP
  EQUALcase DROP
  2DUP RADDext TRIMOBJext RNEGext
  UNROT RMULText TRIMOBJext Z1 UNROT THREE{}POLY
  NULL$ >TAG
;


**********************************************************************
* Desc:		Float rounding for %%, C%% or list of either type.
*		used by SOLVext to reconstruct factors.
* Stack:	( ob --> ob')
**********************************************************************
NULLNAME FRND
CODE
frnd		GOSBVL	=a=stk1
		LC(5)	=DOEREL
		?A#C	A
		GOYES	frnd1
frnd%%		LA(5)	frnd%%!
		PC=(A)
frnd1		LC(5)	=DOECMP
		?A#C	A
		GOYES	frnd2
		LA(5)	frndc%%!
		PC=(A)
frnd2		LC(5)	=DOLIST
		?A#C	A
		GOYES	frnd3
		LA(5)	frndlist!
		PC=(A)
frnd3		GOVLNG	=Loop
ENDCODE

LABEL frnd%%!
::
  %%>% %0 RNDXY %>%%
;

LABEL frndc%%!
::
  C%%>%% frnd%%!
  SWAP frnd%%!
  SWAP E%%>C%%
;

LABEL frndlist!
::
  INNERDUP #1-SWAP
  ZERO_DO 
     get1 FRND
  SWAPLOOP
  #1+ {}POLY
;

**********************************************************************
* Desc:		Search if P is a bisquared 4-th order equation.
*		returns a meta of factors and the multiplying
*		coeff in that case.
* Stack:	( P #5 --> meta cst_coeff TRUE )
*		( P #5 --> P #5 FALSE )
*		( P #  --> P # FALSE )
**********************************************************************
NULLNAME BICARREE?
::
  DUP #5= NOTcaseFALSE
  OVER INNERCOMP 2DROP			( p, #5, a4, a3, a2, a1 )
  Z0 EQUAL NOTcase2drop DROPFALSE
					( p, #5, a4, a3, a2 )
  DROP Z0 EQUALNOTcase DROPFALSE
					( p, #5, a4 )
  2DROP INCOMPDROP			( a4, a3, a2, a1, a0 )
  SWAPDROP ROTDROP THREE{}POLY FDEG2ext	( meta-fact, coef )
  OVER #1=case REALBICAR		( f1, f2, #2, coef )
  SWAPDROP UNROT			( coef, f1, f2 )
  INCOMPDROP Z0 SWAP THREE{}POLY	( coef, f1, f2' )
  FDEG2ext				( coef, f1, meta-2, coef )
  OBJ>R get1 INCOMPDROP
  Z0 SWAP THREE{}POLY			( coef, meta-2, f1' )
  FDEG2ext OBJ>R top& get1 R>OBJ
  RMULText R>OBJ RMULText TRUE
;

**********************************************************************
* Desc:
* Stack:	( f1, #1, coef -> meta, reste, TRUE )
**********************************************************************
NULLNAME REALBICAR
::
  SWAPDROP RMULText			( f1 )
  SETCOMPLEX FDEG2ext			( f1 f1-conj #2, coef )
  3PICK ONE NTHCOMPDROP			( f1 f1-conj #2, coef f1.1 )
  DUP RCONJext RMULText
  ONE>POLY RMULText
  4UNROLL 2DROP				( coef, f1 )
  INCOMPDROP
  SWAP RDIVext RNEGext SXSQRext		( coef, r1 )

  ( [x-r1]*[x-r1_barre]=x^2-x*[r1+r1_barre]+|r1|^2 )

  DUP RREext
  Z2 RMULText
  SWAPDUP RCONJext RMULText		( coef, 2*r1r, norme )
  FXNDext DUP 4ROLL RMULText		( coef, norme, d, 2*r1*d )
  ROT THREE{}POLY			( coef, f1 )
  DUPINCOMP ROT RNEGext UNROT {}POLY	( coef, fact1, fact1' )
  ROTOVER CARCOMP DUP RMULText RDIVext TWO SWAP
  CLRCOMPLEX TRUE
;

**********************************************************************
* Desc:		Real mode: full factorization over the integer
*		Complex mode: find all 1st order factors of P 
* Stack:	( P --> meta-fact, cst coeff )
**********************************************************************
NULLNAME FEVIDENText
::
  NOTLISTcase ZEROSWAP			( p )

  ALG48FCTR?				( p, FALSE | meta-fact, cst, TRUE )
  ?SEMI

  DUPLENCOMP DUP#0=csedrp INTERNALERR	
					( p # )
  DUP#1= casedrop  ZEROSWAP		( -> #0 p )
					( p # )
  DUPTWO #=case DROPONEZ1		( -> p #1 Z1 )
					( p # )
  DUP #3= casedrop FDEG2ext
					( p # )
  EXACT? NOTcase DROPONEZ1		( -> p #1 Z1 )

* Complex mode, first extract trivial factor (root 0)

  OVERSWAP NTHCOMPDROP DUP Z0 EQUAL	( p, a0, flag )
  ::
    NOTcase ZERO
    DROP INNERCOMP pulldrop		( an, ..., a1, #n-1 )
    2DUP #2+UNROLL {}POLY		( a1, p' )
    INTERNALX
    UNROTSWAP ONE
  ;
				( p a0 #0 | { 1 0 } p' a1 #1 )
  UNROT				( meta p a0 )
  DIVISext			( meta p {divis_a0} )
  OVER CARCOMP DIVISext		( meta p {divis_a0} {divis_an} )
  toLEN_DO (DO)			( meta p Da0 Dan )
    DUPINDEX@ NTHCOMPDROP	( meta p Da0 Dan Dn )
    ROT	toLEN_DO (DO)		( meta p Dan Dn Da0 )
      DUPINDEX@ NTHCOMPDROP	( meta p Dan Dn Da0 D0 )
      3PICKSWAP			( meta p Dan Dn Da0 Dn D0 )
      2DUP TWO{}POLY		( meta p Dan Dn Da0 Dn D0 { Dn D0 } )
      7ROLL SWAP2DUP		( meta Dan Dn Da0 Dn D0 p { Dn D0 } p { Dn D0 } )
      DIVISIBLE?		( meta Dan Dn Da0 Dn D0 p { Dn D0 } [ p' TRUE | obj FALSE ] )
      ::
        NOTcase2DROP
        ROTDROPSWAP 		( meta Dan Dn Da0 Dn D0 p' { Dn D0 } )
        8UNROLL 7ROLL #1+ 7UNROLL
      ;
				( meta Dan Dn Da0 Dn D0 p )
      UNROT RNEGext TWO{}POLY
      2DUP DIVISIBLE?		( meta Dan Dn Da0 p { Dn -D0 } [ p' TRUE | obj FALSE ] )
      ::
        NOTcase2DROP
        ROTDROPSWAP		( meta Dan Dn Da0 p' { Dn -D0 } )
        6UNROLL 5ROLL #1+ 5UNROLL
      ;
				( meta Dan Dn Da0 p )
      DUPLENCOMP FOUR #<
      IT ExitAtLOOP
      4UNROLL			( meta p Dan Dn Da0 )
    LOOP
				( meta p Dan Dn Da0 )
    SWAPDROPSWAP
    3PICK LENCOMP FOUR #<
    IT ExitAtLOOP		( meta p {divis a0} {divis an} )
  LOOP
  2DROP				( meta p )
  SWAP#1+ Z1
;

LABEL DROPONEZ1
:: DROPONE Z1 ;


**********************************************************************
* Desc:		Combine two meta objects
* Stack:	( meta3 meta2 ob1 --> meta3&meta2 ob1 )
**********************************************************************
*NULLNAME m3&m2
*::
*  OBJ>R top& R>OBJ
*;

**********************************************************************
* Desc:		Return the roots of a polynomial P
* Stack:	( P --> meta, cst_coeff )
* Note:		Calls the numeric solver
**********************************************************************
NULLNAME EVIDENText
::
  FEVIDENText			( meta cst_coeff )
  SWAPDUP			( meta' cst_coeff # # )
  DUP#0=csedrp SWAP
  ZERO_DO 			( meta' cst_coef # )
    get1 			( meta'' cst_coef # f1 )
    ROT OBJ>R			( meta'' # f1 )
    EVIDSOLV			( meta'' # meta-root cst )
    R>OBJ RMULText		( meta'' # meta-root cst )
    OBJ>R top& #1- R>OBJ
  SWAPLOOP			( meta' cst_coef # )
  SWAP 				( meta cst_coef )
;

**********************************************************************
* Desc:		Return the roots of a 1st, 2nd order and some other poly
* Stack:	( P --> meta, cst_coeff )
* Note:		Calls the numeric solver if exact solving fails
**********************************************************************
NULLNAME EVIDSOLV
::					( P )
  DUPTYPELIST? NOTcase ZEROSWAP		( -> null-meta, P )
  DUPLENCOMP 				( P, # )
  DUP#1= casedrop ZEROSWAP		( -> null-meta, P )
  BICARREE? case EVIDSOLV4		( -> meta, cst_coeff )
  FOUR #<case EVIDSOLV2			( -> )
  DUP SOLVext INNERCOMP 		( P, meta-rac )
  get1 CARCOMP ONE>POLY			( meta-rac, cst_coeff )
;

LABEL EVIDSOLV2
::					( P )
  METADEG2				( P, meta )
  get1 CARCOMP ONE>POLY			( meta, cst_coeff )
;

LABEL EVIDSOLV4
::				( meta-fact cst_coeff )
  SWAPDUP
  ZERO_DO (DO)			( meta' cst_coef # )
	get1 FACTORACext	( meta'' cst_coef # r cst_coef )
	4ROLL RMULText ROT
	LOOP
  SWAP				( meta-root cst_coeff )
;

**********************************************************************
* Desc:		Return the roots of a 2nd order polynomial
* Stack:	( P --> {} )
**********************************************************************
NULLNAME DEG2ext
::
  DUPTYPELIST? NOTcasedrop NULL{}
  DUPLENCOMP DUPTWO #< case2drop NULL{}
  THREE #>case SOLVext				( list )
  METADEG2					( list, meta-rac )
  {}POLY SWAPDROP				( {} )
;

**********************************************************************
* Desc:		Return the roots of a 2nd order polynomial
* Stack:	( P --> P, meta )
* Note:		P must be of order 1 or 2
**********************************************************************
NULLNAME METADEG2
::
  DUPINCOMP #2= case2drop METADEG1
						( list, a, b, c )
  OVERDUP RMULText				( list, A, B, C, B^2 )
  SWAP4PICK					( list, A, B, B^2, C, A )
  RMULText Z4 RMULText RSUBext			( list, A, B, delta )
  DELTA<0? 					( list, A, B, delta, ?delta<0 )
  case2drop :: DROPDUP NULL$ >TAG ONE ;		( -> list, ::list, #1 )
  SXSQRext					( list, a, b, D )
  SWAPROT Z2 RMULText				( list, D, b, 2a )
  OVER 4PICK RADDext RNEGext			( list, D, b, 2a, -b-D )
  OVER RDIVext 2SWAP				( list, 2a, root1, D, b )
  RSUBext ROT RDIVext TWO			( list, root1, root2, #2 )
;

**********************************************************************
* Desc:		Return the roots of a 1st order polynomial
* Stack:	( P --> P, meta )
* Note:		P must be of order 1
**********************************************************************
NULLNAME METADEG1
:: 
  DUP DEG1 ONE
;

**********************************************************************
* Desc:		Root of a first order factor
* Stack:	( f -->  r )
* Note:		f is one level depth deeper than r
**********************************************************************
NULLNAME DEG1
::
  INCOMPDROP RNEGext 	(  a1 -a0 )
  SWAP RDIVext		(  -a0/a1  )
;

**********************************************************************
* Desc:		Returns factors of a 2nd order polynomial and the
*		corresponding multiplying coefficient.
* Stack:	( P --> meta-fact cst_coef )
* Note:		tests for 1st order polynomial
**********************************************************************
NULLNAME FDEG2ext
::
  DUPINCOMP 				( list, a, b, c, #3 | list, a, b, #2 )
  #2= casedrop DROPONEZ1
  OVERDUP RMULText			( list, A, B, C, B^2 )
  SWAP4PICK				( list, A, B, B^2, C, A )
  RMULText Z4 RMULText RSUBext		( list, A, B, delta )
  DELTA<0? 				( list, A, B, delta, delta<0? )
  case2drop DROPONEZ1

* ax^2+bx+c= (2ax+b+D)(2ax+b-D)/(4a)

  SXSQRext			( list, a, b, D )
  SWAP3PICK Z2 RMULText		( list, a, D, b, 2a )
  OVER 4PICK RADDext		( list, a, D, b, 2a, b+D )
  OVER SIMPext			( list, a, D, b, 2a, b+D', 2a' )
  SWAP3PICK 3PICK RDIVext	( list, a, D, b, 2a, 2a', b+D', 2a/2a' )
  6UNROLL TWO{}POLY		( list, a, 2a/2a', D, b, 2a, fact1 )
  ROT 4ROLL RSUBext		( list, a, 2a/2a', 2a, fact1, b-D )
  3PICK SIMPext SWAP		( list, a, 2a/2a', 2a, fact1, 2a'', b-D'' )
  4ROLL 3PICK RDIVext		( list, a, 2a/2a', fact1, 2a'' b-D'', 2a/2a'' )
  4UNROLL TWO{}POLY		( list, a, 2a/2a', 2a/2a'', fact1, fact2 )
  2SWAP RMULText		( list, a, fact1, fact2, 2a/2a'*2a/2a'' )
  4ROLL Z4 RMULText		( list, fact1, fact2, 2a/2a'*2a/2a'', 4a )
  RDIVext 4ROLLDROP
  TWO SWAP ONE>POLY		( fact1, fact2, #2, coef )
;

LABEL DELTA<0?
* return false in complex mode
:: 				( obj )
  COMPLEX? caseFALSE		( -> obj false )
  DUP ZSIGNE Z-1 EQUAL		( obj obj<0? )
;

**********************************************************************
* Desc:		Returns the product of all elements of the list
* Stack:	( {} ob )
**********************************************************************
NULLNAME PIext
::
  INNERCOMP
  DUP#0=csedrp Z1
  DUP#1= caseDROP
  ONE_DO (DO)
     SWAP RMULText
  LOOP
;

**********************************************************************
* Desc:		Convert root to factor
* Stack:	( r --> n, d )
* Note:		factor is n/d, one level depth deeper than r
**********************************************************************
NULLNAME RACTOFACext
::
  DUP TYPETAGGED? case :: STRIPTAGS Z1 ;

* x - rac.num/rac.deno = (x*rac.deno -rac.num)/rac.deno

  FXNDext			( rac.num, rac.deno )
  DUPROT RNEGext TWO{}POLY	( rac.deno, { rac.deno -rac.num } )
  SWAP ONE>POLY			( { rac.deno -rac.num }, { rac.deno } )
;

**********************************************************************
* Desc:		Convert a factor to a root, solving 1st order factor
* Stack:	( f --> r cst_coef )
* Note:		f and cst_coef are one level depth deeper than r
**********************************************************************
NULLNAME FACTORACext
::	
  DUPLENCOMP #2= 
  NOTcase :: NULL$ >TAG Z1 ;
  DUP DEG1			( P, r )
  SWAP CARCOMP ONE>POLY		( r, cst_coef )
;


**********************************************************************
* Desc:		
* Stack:	( ob # --> {} intob meta )
*
* Note:		{} is the list of variables
*		meta is made of roots or factors of num (N) or deno (D) or 
*		both (N/D) depending of #
*		ZERO  for roots N/D
*		ONE   for roots N
*		TWO   for roots D with numeric solver call
*		THREE for roots D without numeric solver call
*		FOUR  for factors N/D
*		FIVE  for factors N
*		SIX   for factors D with numeric solver call
*		SEVEN for factors D without numeric solver call
**********************************************************************
NULLNAME RFACText
::				( obj, # )
  SWAP VX>LVARext ROT		( obj, lvar, # )
  RFACT2ext
;

**********************************************************************
* Desc:		Like RFACText, but the list of variables is given
* Stack:	( ob {} # --> {} intob meta )
**********************************************************************
NULLNAME RFACT2ext
::				( obj, {}, # )

  DUP FOUR #/ #1=		( obj, {}, #, rem, ?facteurs )
  5PICK 5PICK
  ' NULLLAM FIVE NDUPN DOBIND	( obj, {} )

  ( LAM1 = {}, lam2 =obj externe, lam3= flag, lam4= rem, lam5= temporaire )

  VXXLext			( {}, intob )
  DUP DUPTYPEZ? case
  :: 3GETLAM ABND case MZSQFF DROPZERO ;
				( {}, intob, intob )
  FXNDext

* Partial square-free factorization of denominator only if needed

  4GETLAM #1=
  ITE DROPZERO RFACTSTEP3
  %-1 METAMM2			( {}, intob, num, meta-deno )

* Partial square-free factorization of numerator only if needed

  get1 4GETLAM ONE #>
  ITE DROPZERO RFACTSTEP3
  top&				( {}, intob, meta-fact )

* need to suppress Z1 coming from METASQFFext

  METACOMPRIM			( {}, intob, meta-fact ) 

* Now factorize each square-free factor

  ZEROSWAP #2/
  ZERO_DO (DO)			( {}, intob, o1, m1, ..., on, mn, meta-res )
     get1 rotomo		( ..., on-1, mn-1, meta-res, mn, on )
     RFACTSTEP5 top&		( {}, intob, o1, m1, .., on-1, mn-1, meta-res )
  LOOP
  ABND				( {}, obint, meta-factors or meta-roots )
;

**********************************************************************
* Desc:		Partial square-free factorization w.r.t. the main variable
* Stack:	( obj -> meta-fact )
* Note:		Extract trivial factors 
**********************************************************************
* Etape 3 obj -> meta-fact

NULLNAME RFACTSTEP3
::	
  NOTLISTcase MSQFF
  PPZ %1 ROT			( z %1 P )
  MFactTriv SWAP #2+ SWAP	( meta-fact P' )
  METASQFFext
  2SWAP ROT			( meta-fact S1 %1 .. Se-1 %e-1 Te %e #e Re )
  Z1 EQUALcasedro #2*top&
  %1 ROT#1+ #2*top&
;

LABEL #2*top&
:: #2* top& ;


**********************************************************************
* Desc:		Factorization of a square-free polynomial
* Stack:	( %m on -> add-to-meta-res )
**********************************************************************
NULLNAME RFACTSTEP5
::			
			( %m on )
* just in case:

  OVER %0= case2drop ZERO

* full integer factorization for real coeff polynomials
* find all 1st order factor for complex coeff polynomials

			( %m on )
  ( * added call to MFactTriv for ex. for [x^3-2x]*sqrt[1-x^2] * )
  ( * because trivial factor are not extracted first * )
  MFactTriv		( %m meta-fact on )
  ( * note that meta-fact is f0 %1 f1 %1 ... fk %k, we have to remove %1 * )
  OBJ>R REMOVEMUL R>OBJ
  FEVIDENText		( %m meta-fact f1...fk #k cst_coeff )
  OBJ>R top& R>OBJ	( %m f1...fk #k cst_coeff )

  OVER #0=case RFACT51

  SWAP#1+ get1 5PUTLAM	( f1, ...fk, cst_coeff, #k+1 )

* finish with 2nd order&numeric factorization of the remainder polynomial

  #1-DUP		( f1, .., fk, cst_coeff, #k, #k )
    ZERO_DO (DO)	( f1, ..., fk, cst_coeff, #k )
       get1 ROT		( f2, ..., fk, #k, f1, cst_coeff )
       METASOLV		( f2, ..., fk, #k, fr1 %m ... frj mn #2j cst_coeff  )
       SWAP#1+ top&	( f2, ..., fk, fr1 %m ... frj %m cst_coeff #2j+k+1 )
       #2-		( f2, ..., fk, fr1 %m ... frj %m cst_coeff #2j+k-1 )
    LOOP		( fr1 %m ... frl %m cst_coeff #2l )
  
* we must add the constant coeff for factors, not for roots

  3GETLAM NOTcase SWAPDROP	( r1 %m ... rl %m #2l )

  SWAP				( fr1 %m ... frl %m #2l cst_coef )
  num#1=case DROP		( -> fr1 %m ... frl %m #2l )

* should add a test for -1 factor?

  FXNDext num#1=case		( fr1 %m ... frl %m #2l n d )
  :: DROP 5GETLAM ROT #2+ ;	( -> fr1 %m ... frl %m n %m #2l+2 )
  5GETLAM DUPUNROT %CHS 5ROLL #4+	
				( -> fr1 %m ... frl %m n %m d %-m #2l+4 )
;

LABEL REMOVEMUL
* fk mk ... f1 m1 #2k -> fk^mk...f1^m1 #k
::
  DUP#0= ?SEMI		( -> )
  DUP #2/		( meta #k )
  #1+_ONE_DO		( meta )
    #1- get1		( mk...f1 m1 #2k-1 fk )
    OVER #2+ROLL	( fk-1...f1 m1 #2k-1 fk mk )
    RPext		( fk-1...f1 m1 #2k-1 fk^mk )
  SWAPLOOP		( meta )
;

LABEL REMOVEMULTIPLICITY
* fk mk ... f1 m1 #2k -> fk...f1 #k
::
  DUP#0= ?SEMI		( -> )
  #2/ DUP
  #1+_ONE_DO		( fk mk ... fj mj fj-1 fj-2 ... f1 #k )
    INDEX@ #1+ROLL	( fk mk ... fj fj-1 fj-2 ... f1 #k mj )
  DROPLOOP
;

LABEL RFACT51
:: 				( %m #0, cst_coeff )
  UNROT 			( cst_coeff, %m #0 )
  3GETLAM NOTcase UNROT2DROP 	( -> null-meta )
  #2+				( cst_coeff %m #2 ) 	
;


**********************************************************************
* Desc:		Non-integer factorization (sqrt extensions and numeric)
* Stack:	( pn, cst_coeff -> meta, cst_coeff )
* Note:		multiplicty is in LAM 5,
**********************************************************************
NULLNAME METASOLV
::					( pn, cst_coeff )

  ZERO SWAPROT				( null-meta, cst_coeff, pn )

* look if pn is a constant polynomial

  NOTLISTcase RMULText			( -> null-meta, cst_coeff )
  DUPLENCOMP				( null-meta, cst_coeff, pn, # )
  DUP#1= casedrop RMULText		( -> nul-meta, cst-coeff )

  4ROLLDROP				( cst_coeff, pn, # )

* look if we allow further factorization (square roots and numeric)

  4GETLAM #3= casedrop METASOLVOUT	( -> null-meta, cst_coeff, pn )
					
* Then look for special cases: 1st order, 2nd order (FDEG2)
* and specials 4nd order 

  BICARREE? case METASOLV4		
					( cst_coeff, pn, # )
  FOUR #<case METASOLV2
					( cst_coeff, pn )
* none of above, update constant coefficient and start numeric

  SWAPOVER				( pn cst_coeff pn )
  CARCOMP ONE>POLY RMULText		( pn new_cst_coeff )
  SWAPDUP				( new_cst_coeff pn pn )
  DEPTH 1PUTLAM
  ERRSET SOLVext
  ERRTRAP
  :: DEPTH 1GETLAM #- NDROP		( new_cst_coeff pn %% )
     DROPDUP 				( new_cst_coeff pn pn )
     CARCOMP ONE>POLY			( new_cst_coeff pn an )
     ROTSWAP QDiv			( pn cst_coeff )
     SWAP NULL$ >TAG ONE{}N DUP
  ;
					( cst_coeff [ pn {} | %% { ::pn } ] )
  SWAPDROP				( cst_coeff {} )

  INNERCOMP				( cst_coeff meta-rac )
  DUP#0=case SWAP			( -> null-meta cst_coeff )

  get1					( meta, cst_coeff )

* look if we want roots or factors
  3GETLAM NOTcase ADDMULTIPL		( -> meta cst_coeff )
  SWAPDUP
  ZERO_DO (DO)				( r1 ... rn cst_coeff # )
     get1 RACTOFACext			( r2 ... rn cst-coeff # f1 deno )
     4ROLLSWAP RDIVext 			( r2 ... rn # f1 cst-coeff )
     5GETLAM SWAP4ROLL #1+		( r2 ... rn f1 %m cst-coeff # )
  LOOP					( f1 %m ... fn %m cst-coeff # )
  SWAP
;

NULLNAME METASOLVOUT
::  				( cst_coeff pn )
  ONE Z1 METASOLV4
;

**********************************************************************
* Desc:		Return roots/factors of 1st and 2nd order polynomials
* Stack:	( cst_coeff, p -> fr1 %m [fr2 %m] # cst_coeff )
**********************************************************************
NULLNAME METASOLV2
::
  FDEG2ext 			( cst_coeff, meta, cst_coeff )
  METASOLV4
;

**********************************************************************
* Desc:		Returns factors or convert to roots if needed
* Stack:	( cst1 f1 ... fk #k cst2 -> fr1 %m ...  frn %m #2k cst_coeff )
* Note:		#k=1,2 or 4, fk are of order 1 or 2
**********************************************************************
NULLNAME METASOLV4
::

* looks if we want roots or factors
  rotomo RMULText
  3GETLAM case ADDMULTIPL		( -> meta cst_coeff )
  SWAPDUP
  ZERO_DO (DO)				( f1 ... fn cst_coeff # )
     get1 FACTORACext			( f2 ... fn cst-coeff # r1 coef )
     4ROLLSWAP RMULText 		( f2 ... fn # r1 cst-coeff )
     5GETLAM SWAP4ROLL #1+		( f2 ... fn r1 %m cst-coeff # )
  LOOP					( r1 %m ... rn %m cst-coeff # )
  SWAP

;

**********************************************************************
* Desc:		Add multiplicities to a meta 
* Stack:	( meta cst_coeff -> meta' cst_coeff )
* Note:		multiplicity is in LAM 5
**********************************************************************
NULLNAME ADDMULTIPL
::
  SWAPDUP
  ZERO_DO (DO)			( r1 ... rn cst #n )
    get1 5GETLAM 		( r2 ... rn cst #n r1 %m )
    2SWAP #1+			( r2 ... rn r1 %m cst #n+1 )
    LOOP			( r1 %m ... rn %m cst #2n )
  SWAP
;

**********************************************************************
* Desc:		Rebuild an object from its list of factors (flag=TRUE)
*		or roots (flag=FALSE) using prg* to multiply
*		and prg^to take multiplicity power
* Stack:	( { fact mult }, flag, prg*, prg^ -> obj )
**********************************************************************
NULLNAME FACTOOBJext ( inverse de RFACText )
::
  3LAMBIND
  DUPNULL{}? casedrop :: Z1 ABND ;
  INNERCOMP #2/ Z1 DUPROT
  ZERO_DO (DO)			( fact1, num1, ..., factk, numk, num, deno )
     2SWAP 
     DUPTYPEBINT? IT UNCOERCE 
     DUPTYPEZ? IT Z>R
     dREALNcase SETSIZEERR
     DUP %0< UNROT %ABS 	( .., num, deno, mult<0?, factk, | mult | )
     3GETLAM ?SKIP
     :: SWAP RACTOFACext DROPSWAP ;
     1GETLAM EVAL SWAP		( ..., num, deno, factk, mult<0? )
     ITE
       2GETEVAL
       :: ROTSWAP 2GETEVAL SWAP ;
  LOOP
  NDXFext
  ABND
;
**********************************************************************
