INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

**********************************************************************
* EXPORT:
*
* MADJ		( M -> M^-1, P[M], P[lambda] )
*		computes inverse, matrix polynomial and char. polynomial
*
* MAT+	( M2, M1 -> M2+M1 )
*
* MAT-	( M2, M1 -> M2-M1 )
*
* MAT*	( M2, M1 -> M2*M1 )
*
* MAT*SCL	( M, scalar -> M*scalar )
*
* MAT/SCL	( M, scalar -> 1/scalar*M )
*
* MAT/		( M2 M1 -> M1^-1*M2 )
*		Warning: note the ordering of M1 and M2
*
* MATINV	( M -> M^-1 )
*
* MATCHS, MATCONJ, MATRE, MATIM
*
* XYext		( v2, v1 -> scalar )
*		scalar product
*
* MATTRACE	( M -> trace )
*
* MATTRAN	( M -> M' )
*		transposed
*
* METATRAN	( M -> meta-M transposed )
*
* MATIDN	( M or arry or real or # -> identity matrix )
*
* METAPIVOT	( meta-M, #l, #c -> meta-M, #l, #l', #c, flag )
*		search a pivot in column #c strating from row #l
*		flag is FALSE if pivot is not found,
*		is TRUE if pivot is found, #l' is the row, #c is updated
*
* PIVOTFLOAT	( float -> float, modulus )
*
* STOMAText	( M -> )
*		store M in MATRIX in the current directory
*
* ADDMATOBJext	( array, obj -> array, obj )
*		( obj, array -> obj, array )
*		used for addition of a numeric matrix and a symbolic object
*
* DEPTHEVAL:	( M3, obj2, prg1, #depth -> M )
*		calls << obj2 prg1 >> on each element of M3
*		prg1 must act like a binary operator
*		#depth=2 if M3 is a matrix, 1 for a vector
* VBINARYOP	( v2, v1, binary op -> v )
*		Element by element v[i]=v2[i] op v1[i]
*
* VUNARYOP	( v, unary op -> V )
*		Like above for unary op
*
* MATDET	( M -> determinant )
*		expanding all (not row reduction)
*
* MATREFRREF	( M, #full_red -> M, list, M' )
*		M' is a reduced matrix obtained from M
*		#full_red = #1 for full reduction,
*		otherwise reduction under diagonal
*
* INXREDext	( lvar, #full_red, M  -> lvar, pivot, M )
*
* METAMATRED	( meta-M, lvar, #full_red -> meta-M, lvar, pivot )
*
* METASPARSE	( meta-M', lk, li, coef -> meta-M, lk )
*
* METAELEM	( meta-M, lk -> meta-M, lk )
*
* METAEXL	( meta-M, #l, #l', #c ->  meta-M, #l, #c )
*		exchange lines l and l'
*		change the signe value stored in LAM2 
*
* AFFICHEEXL	( meta-M, #c, #l, #l' ->  meta-M, #c, #l, #l' )
*		displays exchange of lines l and l'
*
* MODULO1MOD	( a, obj -> a mod LAM MODULO )
*
* METAELEMDISP
*		displays elementary line manipulation
*		uses ligne_cur LAM, loop index INDEX@,
*		and variable list liste_var
*
* METAMATDISP	( meta-M, o3, o2, o1, grob -> meta-M, o3, o2, o1 )
*		displays meta-M using liste of var LAM liste_var
**********************************************************************

	STITLE Basic matrix operations


* ====================================================================
*		Basic arithmetic on matrices/vectors
* ====================================================================

**********************************************************************
* Desc:
* Stack:	( M2 M1 --> M2+M1 )
**********************************************************************
NULLNAME MAT+
::
  SAMEMATRIX NOTcase 48MAT+
  CKSAMESIZE NOT?DIMERR
  2DMATRIX? case MADD VADD
;

**********************************************************************
* Desc:
* Stack:	( M2 M1 --> M2+M1 )
**********************************************************************
NULLNAME MADD
::
  ' VADD VBINARYOP
;

**********************************************************************
* Desc:
* Stack:	( M2 M1 --> M2-M1 )
**********************************************************************
NULLNAME MAT-
::
  SAMEMATRIX NOTcase 48MAT-
  CKSAMESIZE NOT?DIMERR
  2DMATRIX? case MSUB VSUB
;

**********************************************************************
* Desc:
* Stack:	( M2 M1 --> M2-M1 )
**********************************************************************
NULLNAME MSUB
::
  ' VSUB VBINARYOP
;

**********************************************************************
* Desc:
* Stack:	( V2 V1 --> V2+V1 )
**********************************************************************
NULLNAME VADD
::
  ' QAdd VBINARYOP
;

**********************************************************************
* Desc:
* Stack:	( V2 V1 --> V2-V1 )
**********************************************************************
NULLNAME VSUB
::
  ' QSub VBINARYOP
;

**********************************************************************
* Desc:		matrix product with size and type checking
* Stack:	( M2 M1 --> M2*M1 )
**********************************************************************
NULLNAME MAT*
::
  OVER MDIMS NOT?DIMERR		( -> vector*matrix is not defined )
				( M2 M1 #l2 #c2 )
  3PICK MDIMS			( M2 M1 #l2 #c2 [ #n1 false | #m1 #n1 true ] )
  NOTcase TRYMAT*VEC
				( M2 M1 #l2 #c2 #l1 #c1 )
  ( * ?c2=l1 * )
  DROP ROTDROP #= NOT?DIMERR	( M2 M1 )
  MMMULT
;

LABEL TRYMAT*VEC
::				( M2 M1 #l2 #c2 #l1 )
  ROTDROP #= NOT?DIMERR
  MVMULT
;

**********************************************************************
* Desc:
* Stack:	( M2 M1 --> M2*M1 )
**********************************************************************
NULLNAME MMMULT
::
  SAMEMATRIX NOTcase 48MAT*

  ( A et B doivent etre des matrices listes valides )
  ( teste compatibilite #colonnes de A=#lignes de B )

  DUPLENCOMP
  3PICK CARCOMP LENCOMP
  #= NOT?DIMERR
						( A, B )
  METATRAN get1

  ( vecteurs colonnes de B b1, ..., bc, #c, A )

  INNERCOMP get1			( b1...bc a1...an #n #c )
  1LAMBIND				( b1...bc a1...an #n )
  DUP #3+
  THREE [DO]
     DUPDUP 1GETLAM #+
     INDEX@ #+SWAP		( b1...bc a1...an ...ck-1 #n #n+#c+k+3 #n)
     #2+PICK			( b1...bc a1...an ...ck-1 #n #n+#c+3+k ak )
     SWAP
     1GETLAM ZERO_DO (DO)	( b1...bc a1...an ...ck-1 #n ak #n+#c+k+3 )
       DUPPICK 3PICK
	XYext
       UNROT
     LOOP
			( b1...bc an...a1 ...ck-1 #n ak.b1...ak.bc ak #2n+3+k )
     2DROP 1GETLAM []N	( apres swap avant loop: b1...bn an...a1...ck #n )
  SWAPLOOP
			( b1...bc an...a1 c1...cn #n )
  DUP 1GETABND #+
  OVER#2+UNROL []N	( b1...bc an...a1 #n+#c C )
  OVER#2+UNROL NDROP
;

**********************************************************************
* Desc:		Product matrix vector
* Stack:	( M V --> V' )
**********************************************************************
NULLNAME MVMULT
::
  SAMEMATRIX NOTcase 48MAT*

  ( matrice symb * vecteur symb -> matrice symb )
  ( teste la compatibilite des longueurs )

				( matr-l, vect-l )
  OVER CARCOMP			( matr-l, vect-l, 1ere ligne de matr )
  LENCOMP OVER LENCOMP
  #= NOT?DIMERR
				( matr-l, vect-l )
  SWAPINCOMP get1 NULL[]
  ROT ZERO_DO (DO)
     OVER 4ROLL XYext >HCOMP
  LOOP
  SWAPDROP
;

**********************************************************************
* Desc:		scalar times matrix
* Stack:	( ob M --> M*ob )
**********************************************************************
NULLNAME SCL*MAT
::
  SWAP MAT*SCL
;

**********************************************************************
* Desc:		matrix times scalar
* Stack:	( M ob --> M*ob )
**********************************************************************
NULLNAME MAT*SCL
::
  SAMEMATSCTYPE NOTcase 48MAT*SCL
					( M ob )
  OVER MATRIXDIM 			( M ob # )
  DEPTHEVAL: QMul			( M*ob )
;

**********************************************************************
* Desc:		Product vector*scalar
* Stack:	( V ob --> V' )
**********************************************************************
NULLNAME VPMULT
::
  SAMEMATSCTYPE NOTcase 48MAT*
  ONE DEPTHEVAL: QMul
;

**********************************************************************
* Desc:		Integral matrix power
* Stack:	( arry zint --> arry )
**********************************************************************
NULLNAME MAT^
::
  OVER MDIMS NcaseTYPEERR
				( arry zint #m #n )
  #= NOT?DIMERR			
				( arry zint )
  DupQIsZero? casedrop MATIDN
  				( arry zint )
  DupZIsNeg? UNROT ZAbs		( flag arry zint )
  OVER TYPEMATRIX? NOTcase 48MAT^ 
  				( flag arry zint )
  SWAP LVARext VXXLext 		( flag zint lvar arry )
  SWAP OBJ>R			( flag zint arry )
  SWAP' MMMULT MPext		( flag arry )
  ONESWAP			( flag #1 arry )
  ROT ITE :: NULL{} SWAP INMATINV DROP ; INNERCOMP 
				( #1 meta-mat )
  R>OBJ BACKCONVERT		( mat )
;

LABEL 48MAT^
:: 
  ' 48MAT* MPext 		( flag arry )
  SWAP NOT?SEMI 48MATINV
;

**********************************************************************
* Desc:		Vector product
* Stack:	( V2 V1 --> V )
**********************************************************************
NULLNAME MATCROSS
::
  SAMEMATRIX NOTcase 48MATCROSS
  SWAP MAT3RRDM					( V1 a b c )
  4ROLL MAT3RRDM				( a b c d e f )
  2DUP 7PICK RMULText
  SWAP 6PICK RMULText RSUBext			( a b c d e f fb-ec )
  7UNROLL 4ROLL 4PICK RMULText
  SWAP 6PICK RMULText RSUBext			( bf-ce a b d e cd-fa )
  5UNROLL 4ROLL RMULText
  UNROT RMULText RSUBext			( bf-ce  cd-fa  ae-bd )
  THREE[]N
;

LABEL MAT3RRDM
::
  INNERCOMP 
  OVER CKMATRIXELEM			( a0..an #n an )
  DROPDUP #3= caseDROP			( -> a b c )
  #2= NOT?DIMERR			( error )
  Z0					( -> a b 0 ) 
;

**********************************************************************
* Desc:		Scalar product with checking
* Stack:	( V2 V1 --> ob )
**********************************************************************
NULLNAME MATDOT
::
  SAMEMATRIX NOTcase 48MATDOT
  OVER 2DMATRIX? caseSIZEERR
  LENCOMP OVER 2DMATRIX? caseSIZEERR
  LENCOMP #= NOT?DIMERR
  XYext
;

**********************************************************************
* Desc:		Round array
* Stack:	( M % --> M )
**********************************************************************
NULLNAME RNDARRY
::
  OVER DTYPEMATRIX? 			( M % M flag )
  NOTcasedrop 48RNDARRY
					( M % M )
  DIMLIMITS LENCOMP 			( M % # )
  DEPTHEVAL: xRND
;

**********************************************************************
* Desc:		Round array
* Stack:	( M % --> M )
**********************************************************************
NULLNAME TRCARRY
::
  OVER DTYPEMATRIX? 			( M % M flag )
  NOTcasedrop 48TRCARRY
					( M % M )
  DIMLIMITS LENCOMP 			( M % # )
  DEPTHEVAL:  xTRNC
;

**********************************************************************
* Desc:		Scalar product of symbolic vectors, no check
* Stack:	( V2 V1 --> ob )
**********************************************************************
NULLNAME XYext
::
  INNERCOMP #1+ROLL INNERCOMP
  DUP#1= casedrop RMULText
  get1				( y1, ..., yn-1, x1, ..., xn, #n, yn )
  ROTSWAP RMULText		( y1, ..., yn-1, x1, ..., xn-1, #n, xn*yn )
  OVER ONE_DO (DO)
     ROT 3PICK #2+PICK
     RMULText RADDext
  LOOP
				( y1, ..., yn-1, #n, X.Y )
  OVER #1+UNROLL #1- NDROP
;

**********************************************************************
* Desc:		matrix divided by scalar
* Stack:	( M ob --> M/ob )
**********************************************************************
NULLNAME MAT/SCL
::
  SAMEMATSCTYPE NOTcase 48MAT/SCL
  OVER MATRIXDIM 			( M ob # )
  DEPTHEVAL: QDiv			( M/ob )
;

**********************************************************************
* Desc:		vector "divided" by matrix
* Stack:	( v M --> M^-1*v )
**********************************************************************
NULLNAME MAT/
::
  SAMEMATRIX NOTcase 
ASSEMBLE
	IF	f39
RPL
  48MAT/
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR 48MAT/
ASSEMBLE
	ENDIF	f39
RPL
  MATINV SWAP MAT*
;


**********************************************************************
* Desc:
* Stack:	( M --> -M )
**********************************************************************
NULLNAME MATCHS
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATCHS
  MUNARYOP QNeg
;

**********************************************************************
* Desc:
* Stack:	( M --> M*M )
**********************************************************************
NULLNAME MATSQUARE
::
  DUP MAT*
;

**********************************************************************
* Desc:
* Stack:	( M --> M_barre )
**********************************************************************
NULLNAME MATCONJ
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATCONJ
  MUNARYOP RCONJext
;

**********************************************************************
* Desc:
* Stack:	( M --> re[M] )
**********************************************************************
NULLNAME MATRE
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATRE
  MUNARYOP RREext
;

**********************************************************************
* Desc:
* Stack:	( M --> im[M] )
**********************************************************************
NULLNAME MATIM
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATIM
  MUNARYOP RIMext
;


**********************************************************************
* Desc:		Matrix trace
* Stack:	( M --> trace )
**********************************************************************
NULLNAME MATTRACE
::
  BESTMATRIXTYPE NOTMATRIXcase
ASSEMBLE
	IF	f39
RPL
  laTRACE
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR laTRACE
ASSEMBLE
	ENDIF
RPL

  INNERCOMP 
  Z0 SWAP
  ZERO_DO (DO)			( l1, ..., ln, res )
     SWAP
     ISTOP-INDEX NTHCOMPDROP
     RADDext
  LOOP
;

**********************************************************************
* Desc:		Matrix transpose and conjugate
* Stack:	( M --> M' )
**********************************************************************
NULLNAME MATTRN
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATTRN
  2DMATRIX? NcaseTYPEERR
  INNERCOMP mattrn []N
;

**********************************************************************
* Desc:		Matrix transpose
* Stack:	( M --> M' )
**********************************************************************
NULLNAME MATTRAN
::
  DUP MDIMS			( M # # true | M # false )
  NOT?DIMERR			( M # # )
  2DROP				( M )
  BESTMATRIXTYPE NOTMATRIXcase 48MATTRAN
  METATRAN []N
;

LABEL 48MATTRAN
:: CKREF !MATTRNnc ;

**********************************************************************
* Desc:		Transpose matrix, return meta-matrix
* Stack:	( M --> Meta-M' )
**********************************************************************
LABEL METATRAN
::
  INNERCOMP mattran
;

**********************************************************************
* Desc:		Transpose meta-matrix
* Stack:	( Meta-M --> Meta-M' )
**********************************************************************
NULLNAME mattran
::				( a1 ... an #n )
  DUP#0= ?SEMI
  OVER TYPEMATRIX? NOT?SEMI
  DUP1LAMBIND			( a1 ... an #n )
  OVER LENCOMP			( a1 ... an #n #m )
  #1+_ONE_DO (DO)		( [lines] a1 ... an #n )
    DUP#0_DO (DO) 		( [lines] a1 ... an #n [] )
      1GETLAM #1+PICK		( [lines] a1 ... an #n [] ak )
      JINDEX@ NTHCOMPDROP	( [lines] a1 ... an #n [] akj )
    LOOP			( [lines] a1 ... an #n a1j ... anj )
    1GETLAM []N			( [lines] a1 ... an #n bj )
    OVER#2+UNROLL		( [lines] a1 ... an #n )
  LOOP
  #1- NDROP 			( [lines] a1 )
  ABND LENCOMP			( meta-matrix )
;

**********************************************************************
* Desc:		Transpose meta-matrix
* Stack:	( Meta-M --> Meta-M' )
**********************************************************************
NULLNAME mattrn
::				( a1 ... an #n )
  DUP1LAMBIND			( a1 ... an #n )
  OVER LENCOMP			( a1 ... an #n #m )
  #1+_ONE_DO (DO)		( [lines] a1 ... an #n )
    DUP#0_DO (DO) 		( [lines] a1 ... an #n [] )
      1GETLAM #1+PICK		( [lines] a1 ... an #n [] ak )
      JINDEX@ NTHCOMPDROP	( [lines] a1 ... an #n [] akj )
      'EvalNoCK: xCONJ		( [lines] a1 ... an #n [] )
    LOOP			( [lines] a1 ... an #n a1j ... anj )
    1GETLAM []N			( [lines] a1 ... an #n bj )
    OVER#2+UNROLL		( [lines] a1 ... an #n )
  LOOP
  #1- NDROP 			( [lines] a1 )
  ABND LENCOMP			( meta-matrix )
;

**********************************************************************
* Desc:		extract submatrix from a matrix
* Stack:	( M rmin nrows cmin ncols { m n } )  --> M' )
**********************************************************************
NULLNAME MATSUB
::
  LENCOMP #1=case VECTSUB
  				( M rmin nrows cmin ncols )
  OBJ>R OBJ>R			( M rmin nrows )
  OVER#+ #1-ROT			( rmin rmax M )
  INNERCOMP get1 		( rmin meta-M rmax )
  metasplit NDROP		( rmin meta-M->#rmax )
  get1 #1- metasplit psh NDROP	( meta-M' )
  mattran			( meta-M' )
  R>OBJ R>OBJ OVER#+ 		( meta-M' cmin cmax )
  #1- submeta mattran []N	( M" )
;

LABEL VECTSUB
::					( M rmin nrows cmin ncols )
  2SWAP 2DROP OVER#+ #1- SUBCOMP		( -> M' )
;

**********************************************************************
* Desc:		extract submeta from a meta
* Stack:	( meta #begin #end  --> meta' )
**********************************************************************
NULLNAME submeta
::
  SWAP#1- OBJ>R 		( meta #end )
  metasplit NDROP		( meta->#end )
  R>OBJ metasplit		( meta-beg meta-keep )
  psh NDROP
;

**********************************************************************
* Desc:		replace part of matrix destination (d) by matrix source (s)
* Stack:	( s d -> d )
* Note: 	LAM1 to 9 must be bound like in Llib/LIMain.s 
*		( 9:r 8:c 7:dmat? 6:f 5:md 4:nd 3:smat? 2:ms 1:ns )
*		Copy begins in matrix d at row r and column c		
**********************************************************************
NULLNAME MATREPL
::					( s d )
  ( * test if we put in a vector or matrix * )
  7GETLAM NOTcase VECTREPL
  ( * we put in a matrix, if s is a vector first -> matrix * )
  ( * then store s in lam6, we do not care about lam 6 * )
  SWAP 3GETLAM ?SKIP ONE[]N 6PUTLAM				( d )
  ( * split d in 3 meta-parts: before, row to modify and after * )
  INNERCOMP 9GETLAM #1- metasplit	( meta-before meta-rest )
  2GETLAM metasplit			( meta-before meta-modify meta-after )
  psh					( ... meta )
  6GETLAM SWAPDUP			( ... meta' s #k #k )
  #1+_ONE_DO (DO)
    get1 				( ... m2..mk s #k m1 )
    3PICK INDEX@ NTHCOMPDROP		( ... m2..mk s #k m1 s1 )
    SWAP PUTVECTINVECT			( ... m2..mk s #k m1' )
    UNROT
  LOOP					( ... meta' s #k )
  SWAPDROP pshtop& top&
  ABND []N
;

LABEL VECTREPL
::
  3GETLAM caseSIZEERR			( can't put a matrix in a vector )
					( s d )
  PUTVECTINVECT ABND			( -> d )
;

**********************************************************************
* Desc:		replace part of vector destination (d) by vector source (s)
* Stack:	( s d -> d )
* Note: 	Copy begins in vector d at position LAM8, length of s
*		is in LAM1
**********************************************************************
LABEL PUTVECTINVECT
::
  SWAP 6PUTLAM				( not used anymore even in MATREPL )
  INNERCOMP 8GETLAM #1- metasplit	( meta-begin meta-rest )
  1GETLAM metasplit psh			( meta-begin meta-after meta-replace )
  NDROP 6GETLAM INNERCOMP		( ... new-meta )
  pshtop& top& []N			( -> d )
;

**********************************************************************
* Desc:		redim a matrix
* Stack:	( M {} --> M' )
**********************************************************************
NULLNAME MATREDIM
::
  OVER TYPEMATRIX? NOTcase 48MATREDIM
  VRRDM
;

**********************************************************************
* Desc:		Vector Right ReDiMension: adds 0 to the right
* Stack:	( []/[[]] {} -> [] )
**********************************************************************
NULLNAME VRRDM
::
  1LAMBIND MATEXPLODE ARSIZE 		( a11..anm #n*m )
  1GETLAM INNERCOMP 			( a11..anm #n*m [#l #c #2 | #l #1] )
  #1=?SKIP #*OVF			( a11..anm #n*m #l )
  VRRDMmeta				( a1..al #l )
  1GETABND SWAP XEQ>ARRAY1		( -> matrix/vector )
;

**********************************************************************
* Desc:		Meta Right ReDiMension: adds 0 to the right
* Stack:	( meta #l -> meta-#l )
**********************************************************************
NULLNAME VRRDMmeta
::
					( a1..ak #k #l )
  2DUP#>				( a1..ak #k #l flag )
  case VRRDMshrink
					( a1..ak #k #l )
  OVER#- 				( a1..ak #k #l-k )
  Z0 SWAP NDUPN top&			( meta )
;

LABEL VRRDMshrink
::				( a1..ak #k #l )
  ( * keep only the beginning * )
  DUPUNROT #-SWAP		( a1 ... an #n-#l #l )
  OVER#2+UNROLL NDROP		( a1 .. a#l #l )
;	

**********************************************************************
* Desc:		Create random matrix
* Stack:	( M -> M' )
**********************************************************************
NULLNAME MATRANM
::
  BESTMATRIXTYPE NOTMATRIXcase
ASSEMBLE
	IF	f39
RPL
  laRANM
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR laRANM
ASSEMBLE
	ENDIF
RPL

  DIMLIMITS 			( {} )
  DIMRANM
;

**********************************************************************
* Desc:		Create symbolic random matrix from dimensions
* Stack:	( {} -> M' )
**********************************************************************
NULLNAME DIMRANM
::
  DUPINCOMP #1=?SKIP #*OVF	( {} #n*m )
  DUP CKMATSIZE			( {} #n*m )
  DUP EXACT? 2LAMBIND
  ZERO_DO (DO)
    ?ATTNQUIT 
ASSEMBLE
	IF	f39
RPL
  laRanInt
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR laRanInt
ASSEMBLE
	ENDIF
RPL
 				( % )
    1GETLAM IT R>Z		( %|zint )
  LOOP
  2GETLAM ABND 			( {} a11 .. anm #n*m )
  #1+ROLL BUILDMATRIX		( -> A )
;


**********************************************************************
* Desc:		Determinant, expanding all (not row reduction)
* Stack:	( M --> det )
**********************************************************************
NULLNAME MATDET
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATDET
  DTYPENDO? NOT?DIMERR
  DUPLENCOMP FOUR #>case MATRDET
  LVARext VXXLext		( lvar matr )
  INMATDET			( lvar det )
  R2SYM				( det )
;

LABEL INMATDET
::		( recursive part of the program : matrix -> determinant )
  INNERCOMP			( meta-mat )
  DUP#1= casedrop CARCOMP	( -> a11 )
  METAMATDET
;

LABEL METAMATDET
::				( meta-mat )
  TWO EQUALcasedro MATDETad-bc
				( meta-mat )
  mattran			( meta-transposed )
  #1-SWAP >R			( c1 ... cn-1 #n-1 )
  mattran			( l1'...ln' #n )
  Z0 1LAMBIND
  DUP#0_DO (DO)			( l1'...ln' #n )
    dup ISTOP-INDEX #1+ROLL	( l1' ..ln' #n l2' ... ln' #n l1' )
    DROP #1- METAMATDET		( l1' ..ln' #n det )
    RSWAP 'R RSWAP QMul		( l1' ..ln' #n det*a1k )
    1GETLAM QSub 1PUTLAM	( l1' ..ln' #n )
  LOOP
  NDROP 1GETABND 		( -> det )
;
	
LABEL MATDETad-bc
( for 2x2 matrix )
::				( { a b }  { c d }  )
  INCOMPDROP ROT INCOMPDROP	( c d a b )
  UNROT QMul
  UNROT QMul
  QSub
;

**********************************************************************
* Desc:		determinant using row reduction
* Stack:	( M -> det )
**********************************************************************
NULLNAME MATRDET
::
  bigdataflag TestSysFlag SWAP
  bigdataflag SetSysFlag
  ONE TOLISText
  ZEROSWAP INXREDext		( flag lvar lpivot matr )
  ::
    INNERCOMP reversym
    SWAP CARCOMP SWAP
    #1+DUP #2= caseDROP
    TWO DO
       SWAPINDEX@ NTHCOMPDROP QMul
    LOOP
  ;
				( flag lvar lpivot numerateur )
  ::
    SWAPINCOMP
    DUP#0=csedrp Z1
    Z1 SWAP #2/
    ZERO_DO (DO)
       UNROT RPext QMul
    LOOP
  ;
  QDiv				( flag lvar determinant )
  ZEROSWAP ROT VAL2ext 		( flag det )
  bigdataflag ROT RestoreFlag	( det )
;


**********************************************************************
* Desc:		Frobenius norm
* Stack:	( M --> obj )
**********************************************************************
NULLNAME MATFNORM
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATFNORM
  MATFNORMSQ xvext
;

LABEL MATFNORMSQ
::
  2DMATRIX? NOTcase SIGMA|xi|^2
  INNERCOMP Z0 SWAP
  ZERO_DO (DO)
    SWAP SIGMA|xi|^2 RADDext
  LOOP
;

LABEL SIGMA|xi|^2
::
  INNERCOMP Z0 SWAP
  ZERO_DO (DO)
    SWAPDUP RCONJext RMULText RADDext
  LOOP
;

**********************************************************************
* Desc:		Row norm
* Stack:	( M --> obj )
**********************************************************************
NULLNAME MATRNORM
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATRNORM
  INNERCOMP SWAP SIGMA|xi|	( a1 .. an-1 #n || )
  SWAP DUP#1= caseDROP		( -> || )
  ONE_DO (DO)			( a1 .. an-1 || )
    SWAP SIGMA|xi| 
    'EvalNoCK: xMAX
  LOOP
;

LABEL SIGMA|xi|
::
  NOTMATRIXcase xABSext
  INNERCOMP Z0 SWAP
  ZERO_DO (DO)		( a1 ... an res )
    SWAP xABSext RADDext
  LOOP
;

**********************************************************************
* Desc:		Column norm
* Stack:	( M --> obj )
**********************************************************************
NULLNAME MATCNORM
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATCNORM
  2DMATRIX? NOTcase SIGMA|xi|
  MATTRAN MATRNORM
;

* ====================================================================
*		Linear algebra and Gauss reduction
* ====================================================================

**********************************************************************
* Desc:		M' is the full reduced matrix obtained from M
* Stack:	( M --> M' )
**********************************************************************
NULLNAME MATRREF
::
  BESTMATRIXTYPE NOTMATRIXcase 
ASSEMBLE
	IF	f39
RPL
  laRREF
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR laRREF
ASSEMBLE
	ENDIF
RPL
  ONE MATREFRREF
  SWAPDROP				( M )
  ' NORMALIZEBY1STNON0 VUNARYOP		( M )
;

LABEL NORMALIZEBY1STNON0
* vector -> vector'
* if vector is non 0, normalize vector by 1st non-0 component
::
  DUP' NONZERO Find1stTrue	( vector [ obj true | false ] )
  NOT?SEMI				( -> vector )
   				( vector elem )
  ONE DEPTHEVAL: QDiv		( vector )
;

LABEL NONZERO
::
  CK&DISPATCH1
  sym Z0EQUALNOT
  real %0<>
  cmp  C%0EQUALNOT
;

LABEL Z0EQUALNOT
::
  Z0 EQUAL NOT
;

LABEL C%0EQUALNOT
:: C%0 EQUAL NOT ;

**********************************************************************
* Desc:		M' is the under-diagonal reduced matrix obtained from M
* Stack:	( M --> M' )
**********************************************************************
NULLNAME MATREF
::
ASSEMBLE
	IF	f39
RPL
  QPI
ASSEMBLE
	ELSE	f39
RPL
  BESTMATRIXTYPE NOTMATRIXcase
  FLASHPTR laRREF
ASSEMBLE
	ENDIF
RPL
  ZERO MATREFRREF
  SWAPDROP
  ' NORMALIZEBY1STNON0 VUNARYOP		( M )
;

**********************************************************************
* Desc:		Rank of a matrix
* Stack:	( M --> Z or % )
**********************************************************************
NULLNAME MATRANK
::
  BESTMATRIXTYPE NOTMATRIXcase
ASSEMBLE
	IF	f39
RPL
  laRANK
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR laRANK
ASSEMBLE
	ENDIF
RPL

  MATREF			( reduced_matrix )
  DUPLENCOMP 			( reduced_matrix # )
  SWAP Z0 ' NULLVECTOR1 POSCOMP	( #n #pos )
  DUP#0=csedrp #>Z		( if #pos=0, no null vector return -> n )
  SWAPDROP #1- #>Z		( else return -> pos-1 )
;

**********************************************************************
* Desc:		
* Stack:	( M --> M^-1 )
**********************************************************************
NULLNAME MATINV
::
  BESTMATRIXTYPE NOTMATRIXcase 48MATINV
  SAFECASEVAL: MATINV2
;

NAMELESS MATINV2
::
  ONESWAP				( #1 matr )

* #1 est la profondeur pour la conversion finale

  LVARext VXXLext			( #1 lvar matr )
  INMATINV				( #1 meta-matrice lvar )

* back-conversion
  BACKCONVERT
;

LABEL BACKCONVERT
::
  ZERO 2LAMBIND 			( #1 meta-matrice )
  METAVAL2
  ABND []N
;

LABEL INMATINV
::
  ( * first check for endomorphism * )
  DUP MDIMS NOT?DIMERR
  #= NOT?DIMERR
				( M )
  INNERCOMP 			( l1..ln #n )
  EXACTMATIDN
  INNERDUP #1+SWAP		( l1...ln i1...in #n+1 #n )
  ZERO_DO (DO)			( l1...lk mk+1...mn i1...ik #n+1 )
    DUP#1+ ROLL ROT &COMP	( l1...lk-1 mk+1...mn i1...ik-1 #n+1 mk )
    OVER UNROLL
  LOOP
  #1-				( lvar, meta-matrice completee  )
  get1 ONE METAMATRED		( meta-matrice completee, lvar, pivot  )
  DROP OBJ>R			( meta-matrice reduite )

  ( reduction de chaque ligne et division par le pivot )

  DUP DUP#1+
  DUPROT #2*
  SWAPDUP #3+ SWAP		( l1...ln, #n, #n+1, #2n , #n+4, #n+1 )
  ONE_DO (DO)			( l1...ln, #n, #n+1, #2n, #n+1 )
     DUPROLL
     DUPINDEX@ NTHCOMPDROP
     SWAP 5PICK 5PICK SUBCOMP
     SWAPONE DEPTHEVAL: RDIVext 5UNROLL
  LOOP
  3DROP				( meta-matrice )
  R>OBJ 			( meta-matrice lvar )
;


**********************************************************************
* Desc:		M' is a reduced matrix obtained from M
*		#full_red = #1 for full reduction
*		otherwise reduction under diagonal
* Stack:	( M #full_ref --> M list M' )
**********************************************************************
NULLNAME MATREFRREF
::
  SWAP					( prg#l1, matr )
  2DMATRIX? NcaseSIZEERR
  ONE TOLISText				( prg#l1, lvar, matr )
  ROTSWAP				( lvar, prg#l1, matr-liste )
  INXREDext
  { ONE TWO }
  FROMLISText				( matr, lpivot )
  SWAP BINT24 GETERABLEMSG >TAG SWAP	( lpivot, matr )
;

**********************************************************************
* Desc:
* Stack:	( Lvar #full_ref M --> Lvar pivot M )
**********************************************************************
NULLNAME INXREDext
::					( lvar, prg#l1, matr )
  UNROT
  OBJ>R OBJ>R
  INNERCOMP
  R>OBJ R>OBJ
  METAMATRED				( meta-matr, lvar , pivot )
  OBJ>R OBJ>R []N R>OBJ R>OBJ		( matr, lvar, pivot )
  ROT					( lvar, pivot, matr )
;

**********************************************************************
* Desc:
* Stack:	( Meta-M Lvar #full_red --> meta-M Lvar pivot )
**********************************************************************
NULLNAME METAMATRED
::				( meta-matr, lvar, prg#l1 )
  ZEROZERO			( matr, lvar, prg#l1, ligne, colonne )
  NULL{} Z1 DUPDUP
  stepflag TestSysFlag
  modularflag TestSysFlag
  bigdataflag TestSysFlag
  ZERO 
  TWELVE ROLL

  ( meta-matr prg#l1 ligne colonne lpivot signe ?display ?mod temp lvar )

* {{  ligne_1 ligne_cur col_cur lpivot signe pivot_cur bareiss disp_flag mod_flag bigdata_flag temp_var liste_var }}
	
DEFINE ligne_1 12GETLAM
DEFINE ligne_cur 11GETLAM
DEFINE col_cur 10GETLAM
DEFINE lpivot 9GETLAM
DEFINE signe 8GETLAM
DEFINE pivot_cur 7GETLAM
DEFINE bareiss 6GETLAM
DEFINE disp_flag 5GETLAM
DEFINE mod_flag 4GETLAM
DEFINE bigdata_flag 3GETLAM
DEFINE temp_var 2GETLAM
DEFINE liste_var 1GETLAM
DEFINE ligne_1! 12PUTLAM
DEFINE ligne_cur! 11PUTLAM
DEFINE col_cur! 10PUTLAM
DEFINE lpivot! 9PUTLAM
DEFINE signe! 8PUTLAM
DEFINE pivot_cur! 7PUTLAM
DEFINE bareiss! 6PUTLAM
DEFINE disp_flag! 5PUTLAM
DEFINE mod_flag! 4PUTLAM
DEFINE bigdata_flag! 3PUTLAM
DEFINE temp_var! 2PUTLAM
DEFINE liste_var! 1PUTLAM
  ' NULLLAM TWELVE NDUPN DOBIND
						( meta-matr )
  DUP#1= case :: liste_var lpivot ABND ;

* ici on multiplie chaque ligne par le LCM des denominateurs
* et on repercute le resultat dans lpivot

  lpivot SWAPDUP
  ZERO_DO (DO)				( l1, .., ln, lpivot, #n )
    get1				( l2, ..., ln, lpivot, #n, l1 )
    METAMATREDPREP			( l2, ..., ln, l1, lpivot, #n )
  LOOP
					( l1, .., ln, lpivot, #n )
  SWAP lpivot!				( meta-matr )
  [BEGIN]
     ligne_cur #1+DUP ligne_cur!
     col_cur #1+DUP col_cur!
     METAPIVOT				( meta-matr, #l, #l', #c, flag )
  [WHILE]
     METAMATREDLP
  [REPEAT]
					( meta-matr, #, #, # )
  BINT82 GETERABLEMSG TOTEMPOB XGROBext ( meta-matr, #, #, #, grob )
  stepflag TestSysFlag 			
  ITE METAMATDISP DROP			( meta-matr, #, #, # )
  3DROP liste_var 			( meta-matr lvar )
  lpivot INNERCOMP signe %1 ROT #2+	( meta-matr lvar pv1 m1...pvk mk s %1 )
  METACOMPRIM {}N			( meta-matr, lvar, pivot )
  ABND
;

LABEL METAMATREDPREP
::					( lpivot, #n, l1 )
  DUP DENOLCMext CKCARCOMP		( lpivot, #n, l1, d1 )
  Z1 EQUALcasedro UNROT			( -> l1, lpivot, #n )

  DUP %1 TWO				( lpivot, #n, l1, d1, d1, %1, #2 )
  7ROLL INNERCOMP top& {}N		( #n, l1, d1, lpivot' )
  4UNROLL VPMULT UNROT		( l1, lpivot, #n )
;


LABEL METAMATREDLP
::
     METAEXL				( meta-matr, #l, #c )
     2DUP col_cur! ligne_cur!		( meta-matr, #l, #c )
     3PICK ROT #- #3+PICK
     DUPROT NTHCOMPDROP			( meta-matr, lk, pivot )
     pivot_cur!				( meta-matr, lk )
     ligne_1 #1= ITE ZERO ligne_cur #1+	( meta-matr, lk, ligne de depart )
     3PICK #1+SWAP 
     2DUP#= casedrop 2DROP
     DO					( meta-matr, lk )
       INDEX@
       ligne_cur #=?SKIP METAELEM
     LOOP				( meta-matr, lk )
     DROP
     bigdata_flag NOT?SEMI
     pivot_cur DUP
     bareiss RDIVext SWAP bareiss!	( meta-matr, pivot/bareiss )
     OVER ligne_1			( meta-matr, pivot, #n, #1 ou 5PICK )
     #1= ITE ONE ligne_cur #-		( meta-matr, pivot, #n, #n-l ou #n-1 )
     UNCOERCE TWO lpivot INNERCOMP 
     pshtop& {}N lpivot!		( meta-matr )
;

**********************************************************************
* Desc:		Exchange lines l and l'.
*		Change the sign value stored in LAM2.
* Stack:	( meta-M #l #l' #c --> meta-M #l #c )
**********************************************************************
LABEL METAEXL
::
  UNROT 2DUP#= case DROPSWAP
  signe RNEGext signe!
  stepflag TestSysFlag IT AFFICHEEXL
					(  meta-matr, #c, #l, #l' )
  META###EXC				( meta-M #l #c )
;

**********************************************************************
* Desc:		Display exchange of lines l and l'
* Stack:	( meta-M #c #l #l' --> meta-M #c #l #l' )
**********************************************************************
LABEL AFFICHEEXL
::				( meta-matr, #c, #l, #l' )
  OVER #>$ CHR_L >H$ $>ID	( meta-matr, #c, #l, #l', 'L#l' )
  OVER #>$ CHR_L >H$ $>ID	( meta-matr, #c, #l, #l', 'L#l', 'L#l'' )
  SYMBOL ID EXCHANGE ;
  TWO ' xFCNAPPLY FIVE SYMBN
  XGROBext			( meta-matr, #c, #l, #l', grob )

* display matrice

  METAMATDISP			( meta-matr, #c, #l, #l' )
;

**********************************************************************
* Desc:
* Stack:	( meta-M lk --> meta-M lk )
**********************************************************************
LABEL METAELEM
::					( meta-matr, lk )
  OVERINDEX@ #-+1 #2+ROLL
  DUP col_cur NTHCOMPDROP		( meta-matr', lk, li, coef )
  bigdata_flag NOTcase METASPARSE

* li' = [lk*coef + li*(-pivot)]/bareiss

  pivot_cur bareiss
  disp_flag ITE METAELEMDISP 2DROP	( meta-matr', lk, li, coef )
  temp_var!				( meta-matr', lk, li )
  METAELEMEXEC METABACK			( meta-matr, lk )
;

**********************************************************************
* Desc:
* Stack:	( meta-M' lk li coef --> meta-M lk )
**********************************************************************
LABEL METASPARSE
::					( meta-matr' lk li coef )
  Z0 EQUALcasedro METABACK
  pivot_cur SIMPext DUP bareiss! %1	( meta-matr' lk li coef' pivot' 1 )
  disp_flag ITE METAELEMDISP 2DROP	( meta-matr' lk li coef' )
  temp_var! OVER INNERCOMP		( meta-matr' lk li ak1 ... akn #n )
  get1 SWAPDUP				( meta-matr' lk ak1 ... akn li #n #n )
  #1+_ONE_DO (DO)			( meta-matr' lk ak1 ... akn li #n )
     get1 temp_var RMULText 
     3PICK INDEX@ NTHCOMPDROP
     bareiss RMULText
     SWAP RSUBext			( ... li #n -coef*ak1+pivot*ai1 )
     mod_flag DUP ITE MODULO1MOD DROP
     UNROT
  LOOP
  SWAPDROP []N			( meta-matr' lk -coef*lk+pivot*li/bareiss )
  METABACK				( meta-matr lk )
  lpivot INNERCOMP
  bareiss %1 TWO top& {}N lpivot!	( meta-matr lk )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
LABEL METABACK
::
  3PICK INDEX@#- #3+ UNROLL
;

**********************************************************************
* Desc:		Do elementary operation
* Stack:
**********************************************************************
LABEL METAELEMEXEC
::				( meta-matr' lk li )
  OVER INNERCOMP		( meta-matr' lk li ak1...akn #n )
  get1 SWAPDUP			( meta-matr' lk ak1...akn li #n #n )
  #1+_ONE_DO (DO)		( meta-matr' lk ak1...akn li #n )
     get1 temp_var RMULText
     3PICK INDEX@ NTHCOMPDROP
     pivot_cur RMULText
     SWAPRSUB			( ... li #n -coef*ak1+pivot*ai1 )
     bareiss
     mod_flag ITE MODULO1MOD RDIVext
     UNROT
  LOOP
  SWAPDROP []N			( meta-matr' lk -coef*lk+pivot*li/bareiss )
;

**********************************************************************
* Desc:
* Stack:	( a ob --> a mod LAM_MODULO )
**********************************************************************
LABEL MODULO1MOD
:: 
  DROP
  LAM MODULO QMODSYMext 
;

**********************************************************************
* Desc:		Display elementary line manipulation. Uses ligne_cur
*		LAM, loop index INDEX@ and variable list liste_var.
* Stack:
**********************************************************************
LABEL METAELEMDISP
::				( meta-matr' lk li coef pivot bareiss )
* cree le grob de manip de lignes

  ZERO 4PICK
  liste_var VAL2ext
  ligne_cur #>$ CHR_L >H$ $>ID
  x*ext				( Mmatr' lk li coef piv bar 'coef*lk' )
  ROT ZEROSWAP
  liste_var VAL2ext
  INDEX@ #>$ CHR_L >H$ $>ID	( Mmatr' lk li coef bar 'coef*lk' piv 'li' )
  DUP4UNROLL x*ext SWAP x-ext	( Mmatr' lk li coef bar 'li' 'piv*li-coef*lk' )
  ZERO 4ROLL
  liste_var VAL2ext NDXFext
  x=ext				( Mmatr' lk li coef manip )
  XGROBext			( Mmatr' lk li coef grob )

* recree la matrice

  3PICK 6PICK
  INDEX@#- #6+ UNROLL		( Mmatr lk li coef grob )
  METAMATDISP			( Mmatr lk li coef )
  4PICK INDEX@#- #5+ ROLL DROP
;

**********************************************************************
* Desc:		Displays meta-M using list of vars LAM liste_var
* Stack:	( meta-M ob3 ob2 ob1 grob --> meta-M ob3 ob2 ob1 )
**********************************************************************
LABEL METAMATDISP
::				( meta-matr o3 o2 o1 grob )
  OBJ>R THREE psh		( o3 o2 o1 #3 meta-matr )
  dup ONE OVER#2+UNROL		( o3 o2 o1 #3 meta-matr #profondeur meta-matr )
  liste_var ZERO 2LAMBIND	( o3 o2 o1 #3 meta-matr #profondeur meta-matr )
  METAVAL2 ABND			( o3 o2 o1 #3 meta-matr meta-matr )
  []N R>OBJ SWAP DISPLAYext	( o3 o2 o1 #3 meta-matr )
  DROP psh DROP
;

LABEL LASTCMETAPIVOT
* calls METAPIVOT searching pivot up to the last column
::
   lastcolflag TestSysFlag NOTcase METAPIVOT
   lastcolflag ClrSysFlag
   METAPIVOT
   lastcolflag SetSysFlag	( meta-v meta-P' meta-l #l #l' #c flag )
;

**********************************************************************
* Desc:		Search a pivot in column #c starting from row #l.
*		Flag is FALSE if pivot is not found. If pivot is
*		found #l' is the row, #c is updated to #c'.
* Stack:	( meta-M #l #c --> meta-M #l #l' #c' flag )
**********************************************************************
NULLNAME METAPIVOT
::					( l1, ..., ln, #n, #l, #c )
  DUP 5PICK LENCOMP			( l1, ..., ln, #n, #l, #c, #c, #realc )
  lastcolflag TestSysFlag IT #1-
  #>case :: DUP FALSE ;
  1LAMBIND				( l1, ..., ln, #n, #l )
  2DUP#< case :: DUP 1GETABND FALSE ;
  BEGIN					( l1, ..., ln, #n, #l )
     DUP %-9 2OVER SWAP#1+SWAP
     DO				( l1, ..., ln, #n, #l, #l', norme )
       ISTOP-INDEX #4+PICK
       1GETLAM NTHCOMPDROP	( l1, ..., ln, #n, #l, #l', norme, obj )

       NULLNAME PIVOTNORM
       CODE
		C=DAT1	A
		CD1EX
		A=DAT1	W
		CD1EX
		LC(5)	=DOINT
		?A#C	A
		GOYES	pivot3
		LCZ0
		P=	9
		?A#C	WP
		GOYES	pivot2
pivot0		P=	0
		LA(5)	=%-9
pivot1		DAT1=A	A
		GOVLNG	=Loop
pivot2		P=	0
		LA(5)	=%9
		GOTO	pivot1
pivot3		LC(5)	=DOLIST
		?A#C	A
		GOYES	pivot5
pivot4		A=PC
		LC(5)	(pivotcomp!)-(*)
pivotbr		A=A+C	A
		PC=(A)
pivot5		LC(5)	=DOSYMB
		?A=C	A
		GOYES	pivot4
		LC(5)	=DOCOL
		?A=C	A
		GOYES	pivot4
		A=PC
		LC(5)	(pivotfloat!)-(*)
		GOTO	pivotbr
       ENDCODE

       2DUP %< ITE
          :: SWAPDROP INDEX@ ROTDROPSWAP ;
          DROP
				( l1, ..., ln, #n, #l, new-#l', new-norme )
     LOOP
				( l1, ..., ln, #n, #l, #l', norme )
     %-9 EQ
  WHILE				( l1, ..., ln, #n, #l, #l )
     1GETLAM DUP#1+ 1PUTLAM
     5PICK LENCOMP
     lastcolflag TestSysFlag IT #1-
     #<case DROP
     1GETABND FALSE 2RDROP
  REPEAT
  1GETABND TRUE
;

LABEL pivotfloat!
::
  PIVOTFLOAT
  DUPEQ: %0 NOT?SEMI
  DROP %-9
;

**********************************************************************
* Desc:
* Stack:	( float --> float_modulus )
**********************************************************************
NULLNAME PIVOTFLOAT
::
  CK&DISPATCH1
  real  %ABS
  TWO   C%ABS
  ereal :: %%ABS %%>% ;
  ecmp  :: C%%>C% C%ABS ;
  sym   :: DROP % -0.001 ;	( * small pivoting value for symbolics * )
;

LABEL pivotcomp!
::
  LENCOMP UNCOERCE %180 %/ %CHS
;

* ====================================================================
*		Linear system solver
* ====================================================================

**********************************************************************
* Desc:
* Stack:	( matrice, linconnues -> )
*		( -> linconns linconnues_rearrangee resultat cas_particulier )
**********************************************************************
NULLNAME SYSText
::
  NULL{} VXXLext SWAPDROP 
  SWAPONE TOLISText		( linconnues, lvar, matrice )
  ONESWAP INXREDext		( linconnues, lvar, lpivot, matrice )
  EXACT? ?SKIP ELMGext
  4ROLL UNROTDUP		( lvar, linconnues, lpivot, matrice, matrice )

    INNERDUP #1-SWAP
    #1+_ONE_DO
       get1 INDEX@ NTHCOMPDROP
    SWAPLOOP
    #1+ {}N
				( lvar linconnues lpivot matrice lcoefdiag )
  ROT &COMP			( lvar linconnues matrice caspart )
  UNROT				( lvar lpivot linconnues matrice_reduite )

  ( supprime les lignes nulles )

  INNERCOMP
  [BEGIN]
     DUP#0= caseSIZEERR
     OVER NULLVECTOR?
  [WHILE]
     pulldrop
  [REPEAT]
  				( lvar lpivot linconnues meta-matrice_reduite )
  DUP #2+PICK INNERCOMP []N	( ... meta-matrice_reduite vect_inconnues )
  SWAP#1+ []N			( lvar lpivot linconnues matrice_reduite )
  DUPLENCOMP
  3PICK LENCOMP SWAPROT		( lpivot linconnues #col #lig mat )

  ( effectue des rotations de colonnes si necessaire )

  ONEONE
  BEGIN			( lvar lpivot lincons #col #lig mat #actcol #actlig )
     OVER 6PICK #<
     OVER 6PICK #< AND
  WHILE			( lvar lpivot lincons #col #lig mat #actcol #actlig )
     3PICKOVER NTHCOMPDROP
     3PICK NTHCOMPDROP
     Z0 EQUAL

     ( teste si le coef #actcol, #actlign de matr est nul )
     ( s'il est nul, echange la colonne avec la derniere )
     ( colonne de la matrice )

     IT
     ::
       OVER 4ROLLSWAP ( EXC )
       SWAPINCOMP get1
       3PICK LENCOMP SWAP#- #2+
       1LAMBIND				( l1, ..., ln, #n )
       DUP#1+ SWAP
       ZERO_DO				( lk, ..., ln, l1, ..., lk-1, #n+1 )
          DUPROLL INNERCOMP
          1GETLAM ROLLSWAP
          []N
       SWAPLOOP
       #1- []N
       ABND ( end of EXC )
       UNROT
     ;
     #1+SWAP #1+SWAP
  REPEAT
  2DROP UNROT2DROP		( caspart lincons matr_reduite_arrangee )
  INNERCOMP NULL{} SWAP
  ONE_DO (DO)			( ...m1 ...mk linconnues { equations }  )
     OVER 4ROLL  ( LXE )	( ..m1..mk-1 linconnues {eq} linconnues mk )
     INNERCOMP trimmeta! 	( ..m1..mk-1 linconnues {eq} linconnues mk-{} )
     INNERCOMP []N DUP CDRCOMP 	( ..m1..mk-1 linconnues {eq} linc. mk mk' )
     DUP NULLCOMP?
     ITE
     :: ROT2DROP CARCOMP Z0 ;
     :: SWAP CARCOMP MAT/SCL	( stk = liste var, coefficients )
        SWAP DUPLENCOMP
        DUP 4PICK LENCOMP
        #-SWAP SUBCOMP			( stk = coeff. ,liste var tronquee )
        DUP CARCOMP
        UNROT CDRCOMP
        MATCHS				( var principale, coeff, var aux )
        XYext
     ;
     x=ext 				( end of LXE )
     >TCOMP
  LOOP
  INNERCOMP reversym []N

  ( lvar caspart lincons linconnues_arrangee tagresultlist )
  4ROLL
  ( lvar lincons lincons_arrangee tegresultlist caspart )
  ONEONE 2DUP FOUR {}N FROMLISText
  "Specific" >TAG
;

**********************************************************************
* Stack:	( matrix2 matrix1 -> matrix2 list )
* Stack:
**********************************************************************
NULLNAME STOSYSText
::
  2DUP TWO{}N UNROT			( {} matrix2 matrix1 )
  MAKESYSText				( matrix_eq matrix lidnt flag )
  NOTcase :: BINT45 ERABLEERROR ;		
					( {} {symb} matrix lidnt )
  OVER LVARext SWAPDROP			( {} {symb} matr lidnt l_other_var )
  NULLCOMP? ?SEMI			( -> {} {symb} res lidnt l_other_var )
  4PICK STOSYSTEM			( {} {symb} res lidnt l_other_var )
;

LABEL STOSYSTEM
* matrix ->
::
  CONTEXT@ CASDIR		( matrix rep )
  SWAP' ID SYSTEM STO		( rep )
  CONTEXT!
;

**********************************************************************
* Desc:		Convert linear equations to a matrix 
*		and check that equation are linear wrt lidnt
* Stack:	( matrix_eq matrix_inc -> matrix_eq matrix lidnt flag )
**********************************************************************
NULLNAME MAKESYSText
::				( {symb} lidnt )
  LIDNTLVAR			( {symb} lidnt lvar )	
  ( * check first that system is polynomial wrt to lidnt * )
  CKPOLYSYST			( {symb} lidnt lvar )
	
  ( * ok! check linearity and compute the matrix * )
  3PICKSWAP VXXLext		( {symb} lidnt lvar { lsym } )
  TRUE 1LAMBIND			( {symb} lidnt lvar { lsym } )
  ZERO 4ROLLROT			( {symb} lvar metares lidnt {lsym} )
  DUPLENCOMP #1+_ONE_DO (DO)	( {symb} lvar metares lidnt {lsym} )
    DUPINDEX@ NTHCOMPDROP	( {symb} lvar metares lidnt {lsym} sym )
    dMATRIXcase SETTYPEERR
    FXNDext DROP		( {symb} lvar mres lidnt {lsym} num )
    ZERO SWAP4PICK LENCOMP	( {symb} lvar mres lidnt {lsym} meta num #id )
    ZERO_DO (DO)		( {symb} lvar mres lidnt {lsym} meta num )
      MAKESYSTLP		( {symb} lvar mres lidnt {lsym} meta num coef )
      SWAPROT #1+SWAP		( {symb} lvar mres lidnt {lsym} meta&coef num )
    LOOP			( {symb} lvar metares lidnt {symb-l} meta num )
    SWAP#1+ []N			( {symb} lvar metares lidnt {symb-l} res )
    4UNROLL ROT #1+ UNROT	( {symb} lvar metares lidnt {symb-l} )
    1GETLAM ?SKIP ExitAtLOOP	( {symb} lvar metares lidnt {symb-l} )
  LOOP				( {symb} lvar metares lidnt {symb-l} )
  DROP OVER#2+UNROL []N		( {symb} lvar lidnt res )
  TWO SWAP4ROLL			( {symb} lidnt #2 res lvar )
  4PICK LENCOMP #1+		( {symb} lidnt #2 res lvar #idnt+1 )
  MINUSONE SUBCOMP		( {symb} lidnt #2 res lvar' )
  VAL2ext SWAP			( {symb} res lidnt )
  Z1 >TCOMP			( {symb} res lidnt&1 )
  1GETABND			( {symb} res lidnt&1 flag )
;

LABEL MAKESYSTLP
( num -> num' coeff )
::				( num )
  ( * no irrq allowed at this level * )
  CKCOEFF
  NOTLISTcase Z0		( -> num 0 )
  DUPLENCOMP DUP#1=		( num # #=1? )
  casedrop :: CARCOMP Z0 ;	( -> num_1 0 )
				( num # )
  #>2case :: ExitAtLOOP FalseFalse 1PUTLAM ;
				( num )
  INCOMPDROP SWAP 		( num_0 num_1 )
  ISTOP-INDEX			( num_0 num_1 # )
  DUP#1= caseDROP		( -> num_0 num_1 )
  ( * look if the coeff num_1 depends non linearly of other vars * )
  ONE_DO (DO)			( num_0 num_1 )
    CKCOEFF DUPTYPELIST?	( num_0 num_1 ? )
    ITE :: DUPLENCOMP		( num_0 num_1 # )
	   #1=case CARCOMP	( -> num_0 num_1' )
           ExitAtLOOP FALSE 1PUTLAM
           ;
    ExitAtLOOP   	      
  LOOP
;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME VARGENext
::
  DUPLENCOMP DUP
  #1+_ONE_DO			( x1, ..., xr, #n )
     "x" INDEX@ UNCOERCE
      DO>STR &$ $>ID
  SWAPLOOP
  {}N
;
**********************************************************************

* ====================================================================
*		Other matrix operations
* ====================================================================
**********************************************************************
* Desc:		Return true if vector is null
* Stack:	( vector --> flag )
**********************************************************************
NULLNAME NULLVECTOR?
::				( vector )
  Z0 EXACT? case NULLVECTOR1	( -> flag )
  DROP SIGMA|xi|^2 ELMGext 	( %%norm )
  %%0 EQUAL 			( flag )
;

LABEL NULLVECTOR1
( vector obj -> flag )
* return true if all components of vect are equal to obj
:: 
   ' EQUALNOT POSCOMP 	( # )
   #0= 		
;

**********************************************************************
* Desc:		Return index # of element {} in array
* Stack:	( {} array --> # TRUE | FALSE )
**********************************************************************
NULLNAME FINDELN
::
  NOTMATRIXcase 48FINDELN
  SWAPINCOMP			( array #n [#m] [ #1 | #2 ] )
  DUP#1= casedrop FINDELNVECT
  TWO EQUALcasedro FINDELNMATR
				( array o1..on #n )
  NDROP DROPFALSE
;

LABEL FINDELNVECT
::				( array #n )
  DUP#0=csedrp DROPFALSE	( -> false )
				( array #n )
  DUPROT 2DMATRIX?		( #n #n array flag )
  case2drop DROPFALSE		( -> false )
				( #n #n array )
  LENCOMP			( #n #n #m )
  #>case DROPFALSE		( -> false )
  				( #n )
  TRUE				( #n true )
;

LABEL FINDELNMATR
::				( array #n #m )
  DUP#0= case2drop DROPFALSE	( -> false )
  SWAP DUP#0= 			( array #m #n ?=0 )
  case2drop DROPFALSE		( -> false )
				( array #m #n )
  ROTDUP LENCOMP SWAP CARCOMP	( #m #n #lin line1 )
  DTYPEMATRIX? NOTcase2drop 2DROPFALSE
				( #m #n #lin line1 )
  LENCOMP UNROT			( #m #col #n #lin )
  OVER #<			( #m #col #n ?out_of_range )
  case2drop DROPFALSE		( -> false )
				( #m #col #n )
  #1- OVER #*			( #m #col #n*#col )
  UNROTOVER 			( #n*#col #m #col #m )
  #< case2drop FALSE		( -> false )
  #+ TRUE			( -> # true )
;

**********************************************************************
* Desc:		Extract element of index # from array
* Stack:	( array # --> array element )
* Note:		array type test is made in assembly for array speed
**********************************************************************
NULLNAME PULLEL[S]
CODE
		D1=D1+	5
		GOSUB	a=d1,c=domat
		D1=D1-	5
		?A=C	A
		GOYES	+
		LA(5)	=48PULLEL[S]
		PC=(A)
+		LA(5)	=MATRIXPULL
		PC=(A)
a=d1,c=domat	C=DAT1	A
		CD1EX
		A=DAT1	A
		CD1EX
		LC(5)	=DOMATRIX
		RTN
ENDCODE

LABEL MATRIXPULL	
::				( array # )
  OVERSWAP 			( array array # )
  OVER CARCOMP TYPEMATRIX?	( array # flag )
  NOTcase NTHCOMPDROP		( -> element )
				( array # )
  #1- 				( array array #-1 )
  OVER CARCOMP LENCOMP		( array array #-1 #m )
  #/				( array array #rem #quot )
  ROTSWAP #1+ NTHCOMPDROP	( array #rem line_#quot )
  SWAP#1+ NTHCOMPDROP		( array element )
;

**********************************************************************
* Desc:		Put element at index # of matrix
* Stack:	( element # matrix --> matrix )
**********************************************************************
NULLNAME BANGARRY
::
  3PICK SAMEMATSCTYPE SWAP		( element # matrix flag element )
  CKOKINMATDROP				( element # matrix flag )
  NOTcase 48BANGARRY
					( element # matrix )
  DUP CARCOMP TYPEMATRIX? 		( element # matrix flag )
  NOTcase PUT[]
  3LAMBIND				( )
  2GETLAM #1- 1GETLAM CARCOMP LENCOMP	( # #col )
  #/ DUPUNROT				( #quot #rem #quot )
  1GETLAM SWAP#1+			( #quot #rem matrix #quot )
  NTHELCOMP NOT?DIMERR			( #quot #rem line )
  SWAP#1+SWAP 3GETLAM UNROT		( #quot element #rem line )
  PUT[]					( #quot new_line )
  SWAP#1+ 1GETABND PUT[]		( -> new_matrix )
;

LABEL CKOKINMATDROP
::
  CK&DISPATCH1
  real DROP
  cmp DROP
  sym DROP
;

**********************************************************************
* Desc:		Replace #i-th vector component by element 
* Stack:	( element #i vector --> vector )
**********************************************************************
NULLNAME PUT[]
::
  INNERCOMP putmeta []N		
;

**********************************************************************
* Desc:		Return max # in an array
* Stack:	( array --> # )
**********************************************************************
NULLNAME ARSIZE
::
  NOTMATRIXcase 48ARSIZE
  MDIMS NOT?SEMI
  #*OVF
;

**********************************************************************
* Desc:		Extract diagonal terms
* Stack:	( A ncols+1 ndiags  --> vector )
* Note:		ncols+1 is there because MATRIX>DIAG is called inside la>DIAG
**********************************************************************
NULLNAME MATRIX>DIAG
::				( A ncols+1 ndiags )
  SWAPDROP			( A ndiags )
  1LAMBIND INNERCOMP		( a_1..a_n )
  1GETLAM #-			( a_1..a_n #n-#ndiags )
  NDROP				( a_1 ... a_ndiags )
  1GETABND			( a_1 ... a_ndiags #n_diags )
  DUPDUP			( a_1 ... a_ndiags #n_diags #n_diags #n_diags )
  #1+_ONE_DO (DO)
    get1			( a_2 ... a_ndiags #n_diags #n_diags a_1 )
    INDEX@ NTHCOMPDROP		( a_2 ... a_ndiags #n_diags #n_diags a_11 ) 
    UNROT
  LOOP				( a_11 a_22 ... a_ndnd #n_diags #n_diags )
  DROP []N
;

**********************************************************************
* Desc:		Construct a matrix from a vector of diagonal terms
* Stack:	( ncol+1 diagV dlen dims{} -> matrix )
**********************************************************************
NULLNAME MATRIXDIAG>
::					( ncol+1 diagV dlen dims{} )
  ROT MATEXPLODE MDIMS IT #* []N UNROT	( ncol+1 diagV dlen dims{} )
  4ROLLDROP INCOMPDROP			( diagV dlen #n #m )

* {{ diag_V diag_dlen diag_n diag_m }}
DEFINE diag_V 4GETLAM
DEFINE diag_dlen 3GETLAM
DEFINE diag_n 2GETLAM
DEFINE diag_m 1GETLAM
DEFINE diag_V! 4PUTLAM
DEFINE diag_dlen! 3PUTLAM
DEFINE diag_n! 2PUTLAM
DEFINE diag_m! 1PUTLAM
	
  4NULLLAM{} BIND			( )
  diag_n
  #1+_ONE_DO (DO)			( [lines] )
    MATRIXDIAG>LP
    []N					( [lines] )
  LOOP
  diag_n ABND []N			( -> matrix )
;

LABEL MATRIXDIAG>LP
::
    ( * if index>diagm then return a line of 0 * )
    INDEX@ diag_m #>case :: Z0 diag_m NDUPN ;
    Z0 INDEX@ #1- NDUPN			( [lines] meta )
    ( * construct line k: if k<=dlen, k-th element is diagV_k else 0 * )
    diag_V INDEX@ NTHELCOMP		( [lines] meta [ vk true | false ] )
    ?SKIP Z0				( [lines] meta [ vk | 0 ] )
    SWAP#1+				( [lines] meta )
    Z0 diag_m INDEX@#- NDUPN top&	( [lines] meta )
;

**********************************************************************
* Desc:		Insert element in symbolic vector at row %i
* Stack:	( vector obj %i --> vector' )
* Note:		Placed here because of the link table of MManp/Manip.s 
**********************************************************************
NULLNAME la+ELEMsym
( V obj i% --> V' )
::
  COERCE 				( V obj #i )
  ROT 2DMATRIX? caseSIZEERR
					( obj #i V )
  ARRAY2MATRIX UNROT			( V obj #i )
  INSERTROW[]
;

**********************************************************************
* Desc:		Insert element/vector in symbolic vector/matrix at row #i
* Stack:	( vector obj #i --> vector )
*		( matrix vector #i -> matrix )
* Note:		check 0 < #i <= #n+1, no check for matrix/vector size
**********************************************************************
NULLNAME INSERTROW[]
::
  ROT INNERCOMP 	( obj #i meta-vect )
  insertrow[]		( meta-vect )
  []N			( vect )
;

**********************************************************************
* Desc:		Insert element/vector in meta-object at position #i
* Stack:	( obj #i meta --> meta )
* Note:		check 0 < #i <= #n+1, no check for matrix/vector size
**********************************************************************
NULLNAME insertrow[]
::
  get1			( obj meta-v #i )
  DUP#0= case SETRANGERR
  SWAP#1+		( obj v1...vn #i #n+1 )
  2DUP#> case SETRANGERR
			( obj v1...vn #i #n+1 )
  get1			( v1...vn #i #n+1 obj )
  OVER 4ROLL #-		( v1...vn #n+1 obj #n+1-i )
  #2+UNROLL		( meta-v )
;

**********************************************************************
* Desc:		Insert vector in symbolic matrix at col #i
* Stack:	( matrix vector #i -> matrix )
* Note:		check 0 < #i <= #n+1, no check for matrix/vector size
**********************************************************************
NULLNAME INSERTCOL[]
::
  ROT INNERCOMP		( vector #i meta )
  mattran		( vector #i meta )
  insertrow[]		( meta )
  mattran []N		( -> matrix )
;

**********************************************************************
* Desc:		Insert matrix2 in matrix3 starting from row #i
* Stack:	( matrix3 matrix2 #i -> matrix )
* Note:		check 0 < #i <= #n+1, no check for matrix sizes
**********************************************************************
NULLNAME INSERT[]ROW[]
::
  FALSE INSERTROWCOL
;

**********************************************************************
* Desc:		Insert matrix2 in matrix3 starting from row #i
* Stack:	( matrix3 matrix2 #i -> matrix )
* Note:		check 0 < #i <= #n+1, no check for matrix sizes
**********************************************************************
NULLNAME INSERT[]COL[]
::
  TRUE INSERTROWCOL
;

LABEL INSERTROWCOL
::				( matrix3 matrix2 #i flag ) 
  3LAMBIND			( matrix3 )
  INNERCOMP			( meta-matrix3 )
  1GETLAM IT mattran		( meta-matrix3 )
  2GETLAM DUP#0=case SETRANGERR
  #1- metasplit			( meta-3.before meta-3.after )
  3GETLAM INNERCOMP 		( meta-3.before meta-3.after meta-matrix2 )
  1GETLAM IT mattran		( meta-3.before meta-3.after meta-matrix2 )
  pshtop& top&			( meta-matrix )
  1GETABND IT mattran		( meta-matrix )
  []N
;

**********************************************************************
* Desc:		Multiplies row #i of symb. matrix M by constant
* Stack:	( ncol i M const -> M' )
* Note:		ncol is not used, it's here because of the stack state
*		at call-time from inside laRCI
**********************************************************************
NULLNAME MATRIXRCI
::
  4ROLLDROP				( #i M const )
  1LAMBIND				( #i M )
  INNERCOMP get1			( meta-M #i )
  2DUP #- #3+ROLL			( meta-M\li #i li )
  1GETABND 				( meta-M\li #i li const )
  V/SCL*SCL				( meta-M\li #i li*const )
  3PICK ROT #- 				( meta-M\li li*const #n-i )
  #2+UNROLL				( meta-M )
  []N
;

LABEL V/SCL*SCL
::
  OVER TYPEMATRIX? ITE VPMULT QMul	
;

**********************************************************************
* Desc:		does Lj <- c*Li+Lj
* Stack:	( ncol #i #j M const -> M' )
* Note:		ncol is not used, it's here because of the stack state
*		at call-time from inside laRCI
**********************************************************************
NULLNAME MATRIXRCIJ
::
  5ROLLDROP				( #i #j M const )
  2SWAP					( M const #i # j )

* {{ rcij_c rcij_i rcij_j  }}
  DEFINE rcij_c 3GETLAM
  DEFINE rcij_i 2GETLAM
  DEFINE rcij_j 1GETLAM
  DEFINE rcij_c! 3PUTLAM
  DEFINE rcij_i! 2PUTLAM
  DEFINE rcij_j! 1PUTLAM
  3LAMBIND				( M )
  INNERCOMP				( meta-mat )
  DUP rcij_i #- #2+PICK			( meta-mat li )
  OVER rcij_j #- #3+ROLL		( meta-mat\lj li lj )
  SWAP rcij_c V/SCL*SCL			( meta-mat\lj lj li*c ) 
  DTYPEMATRIX? ITE VADD QAdd		( meta-mat\lj lj+li*c )
  OVER rcij_j #- #2+UNROLL		( meta-mat )
  ABND []N				( M' )
;

**********************************************************************
* Desc:		Exchange columns c and c' of a symb. matrix
* Stack:	( M #c #c' --> M )
**********************************************************************
NULLNAME MATRIXCSWAP
::
  ROT INNERCOMP 	( #c #c' meta-mat )
  get1			( #c meta-mat #c' )
  rotomo		( meta-mat #c' #c )
  METAMATCSWAP		( meta-mat )
  []N
;

**********************************************************************
* Desc:		Exchange lines r and r' of a symb. matrix
* Stack:	( M #r #r' --> M )
**********************************************************************
NULLNAME MATRIXRSWAP
::
  ROT INNERCOMP 	( #r #r' meta-mat )
  get1			( #r meta-mat #r' )
  rotomo		( meta-mat #r' #r )
  METAMATRSWAP		( meta-mat )
  []N
;

**********************************************************************
* Desc:		Extract row #r from M
* Stack:	( M #r --> M' lr )
* Note:		check boundaries
**********************************************************************
NULLNAME MATRIX-ROW
::
  SWAP MATRIX-ROW1
;

LABEL MATRIX-ROW1
::				( #r M )
  INNERCOMP get1		( meta-mat #r )
  METAMAT-ROW			( meta-mat lr )
  OVER#2+UNROLL []N SWAP	( M' lr )
;

**********************************************************************
* Desc:		Extract row #r from meta-matrix
* Stack:	( meta-mat #r --> meta-mat lr )
* Note:		check boundaries
**********************************************************************
NULLNAME METAMAT-ROW
::
  DUP#0= case SETRANGERR
  OVERSWAP 			( meta-mat #n #r )
  2DUP#< case SETRANGERR
  #-				( meta-mat #n-r )
  SWAP#1-SWAP			( l1..ln #n-1 #n-r )
  #2+ROLL 			( meta-mat lr )
;

**********************************************************************
* Desc:		Extract column #r from matrix
* Stack:	( M #c --> M cc )
* Note:		check boundaries
**********************************************************************
NULLNAME MATRIX-COL
::
  SWAP 2DMATRIX? 		( #c M flag )
  NOTcase MATRIX-ROW1		( -> vector vc )
				( #c M )
  INNERCOMP			( #c meta-mat )
  mattran			( #c meta-mat-tran )
  get1 METAMAT-ROW		( meta-mat-tran cc )
  OVER#2+UNROLL mattran		( cc meta-mat )
  []N SWAP			( M cc )
;

**********************************************************************
* Desc:		Exchange columns c and c' of a meta-matrix
* Stack:	( meta-M #c #c' --> meta-M )
**********************************************************************
NULLNAME METAMATCSWAP
::
  2LAMBIND				( meta-M )
  DUPDUP				( l1 ... ln #n #n #n )
  ZERO_DO (DO)				( l1 ... ln #n #n )
    get1 INNERCOMP			( l2..ln #n #n a11..a1m #m )
    2GETLAM 1GETLAM METAMATRSWAP	( l2..ln #n #n a11..a1m #m )
    []N UNROT				( l2..ln l1' #n #n )
  LOOP
  ABND DROP				( meta-M )
;

**********************************************************************
* Desc:		Exchange lines l and l' of a meta-matrix (or vector)
* Stack:	( meta-M #l #l' --> meta-M  )
**********************************************************************
NULLNAME METAMATRSWAP
::
  DUPUNROT META###EXC 		( meta-M #l #l' )
  2DROP				( meta-M )
;


LABEL META###EXC
( meta-matr # #l #l' -> meta-matr #l # )
::
  4PICK					(  meta-matr, #c, #l, #l',#n )
  DUPROT #-				(  meta-matr, #c, #l, #n, #n-l' )
  SWAP3PICK #-				(  meta-matr, #c, #l, #n-l', #n-l )
  CODE
		GOSBVL	=SAVPTR
		GOSBVL	=POP2#
		B=C	A		B=#n-l
		GOSUB	metaexlml	
		CD0EX			D0=@stk1, 
		B=A	A		B=#n-l'
		GOSUB	metaexlml	
		CD1EX			D1= @stk2
* echange des adresses sur la pile
		C=DAT1	A
		A=DAT0	A
		DAT1=A	A
		DAT0=C	A
		GOVLNG	=GETPTRLOOP
* B=# -> renvoie en C l'adresse de stk #+3, utilise B,C 
metaexlml	C=D1	A
		C=C+B	A
		B=B+B	A
		B=B+B	A
		C=C+B	A
		C=C+CON	A,15
		RTN
  ENDCODE
					(  meta-matr, #c, #l, #n-l', #n-l )
  2DROP SWAP				(  meta-matr, #l, #c )
;

**********************************************************************
* Desc:		Store matrix in 'MATRIX' in current directory.
* Stack:	( M --> )
**********************************************************************
NULLNAME STOMAText
::
  ARRAY2MATRIX LVARext			( matrice, lvar )
  LENCOMP #0=?SEMI
  DUP STOMATRIX
;

LABEL STOMATRIX
* matrix ->
::
  CONTEXT@ CASDIR		( matrix rep )
  SWAP' ID MATRIX STO		( rep )
  CONTEXT!
;

**********************************************************************
* Desc:		Create identity matrix
* Stack:	( matrix or arry or #n --> Mid )
**********************************************************************
NULLNAME MATIDN
::
ASSEMBLE
	IF	f39
RPL
  48MATIDN
ASSEMBLE
	ELSE	f39
RPL
  ::
    CK&DISPATCH1
    real %ABSCOERCE
    arry ENDOSIZE
    bint NOP
  ;
				( # )
  DUPDUP #* CKMATSIZE		( # )
  EXACT? NOTcase :: DUPTWO {}N %0 48MAKEARRY 48MATIDN ;
  EXACTMATIDN
ASSEMBLE
	ENDIF	f39
RPL
;


LABEL EQUALZ1ELSEZ0
:: EQUALcase Z1 Z0 ;

LABEL ENDOSIZE
:: MDIMS NOT?DIMERR OVER #= NOT?DIMERR ;

**********************************************************************
* Desc:		Create constant matrix from matrix
* Stack:	( matrix obj --> matrix )
**********************************************************************
NULLNAME MATCON
::
  SAMEMATSCTYPE NOTcase :: dREALcase 48MATCON SWAP CKCMPARRY SWAP 48MATCON ;
  SWAP DIMLIMITS			( obj { #n [#m] } )
  SWAP MAKEARRY
;

**********************************************************************
* Desc:		Create constant matrix/array from obj type
* Stack:	( {} obj --> matrix )
**********************************************************************
NULLNAME MAKEARRY
::
  FLOAT? case 48MAKEARRY
  SWAP OBJDIMS2MAT
;
	
**********************************************************************
* Desc:		Create constant matrix from dim and obj
* Stack:	( obj {} --> matrix )
**********************************************************************
NULLNAME OBJDIMS2MAT
::					( obj { #n [#m] } )
  INNERCOMP #1=case MAT1DCON
  					( obj #n #m )
  ' 2DROP 4ROLL				( #n #m 2DROP obj )
  2Ob>Seco				( #n #m prg )
  MAKE2DMATRIX				( matrix )
; 


LABEL MAT1DCON
::					( obj #n )
  NDUPN []N
;

**********************************************************************
* Desc:		Fill a matrix of specified size using a program
* Stack:	( n m prg --> matrix )
* Note:		prg must take 2 args and return 1 arg
*		On entry MAKE2DMATRIX provide the indexes as Z integers
**********************************************************************
NULLNAME LCPROG2M
::				( int1 int2 prg )
  DUPTYPECOL? ?SKIP LCXMSYMB
  UNROTSWAP			( prg int2 int1 )
  CKINT>0 NOTcase SETRANGERR
  SWAP				( prg int1 int2 )
  CKINT>0 NOTcase SETRANGERR
				( prg int1 int2 )
  ROT MAKE2DMATRIX		( matrix )
;

LABEL LCXMSYMB
::
ASSEMBLE
	IF	f39
RPL
  { ID S1 ID S2 } 
ASSEMBLE
	ELSE	f39
RPL
  { ID I ID J } 
ASSEMBLE
	ENDIF	f39
RPL
   ' 4ROLL ' 4ROLL ' TWO{}N ' MEVALext 	( symb {} 4ROLL 4ROLL TWO{}N MEVAL )
   SIX ::N 				( prog )
;

**********************************************************************
* Desc:		Create matrix from size and program (with stack checking)
* Stack:	( #n #m prg --> matrix )
* Note:		prg must take 2 args and return 1 arg
*		On entry MAKE2DMATRIX provide the indexes as Z integers
**********************************************************************
NULLNAME MAKE2DMATRIX
::					( #n #m prg )
  make2dmatrix []N
;

**********************************************************************
* Desc:		Create meta-matrix from size and program (with stack checking)
* Stack:	( #n #m prg --> meta-matrix )
* Note:		prg must take 2 args and return 1 arg
*		On entry make2dmatrix provide the indexes as Z integers
**********************************************************************
NULLNAME make2dmatrix
::					( #n #m prg )
  3PICK3PICK #* CKMATSIZE		( #n #m prg )
  2LAMBIND				( #n )
  DUP#0_DO				( #n )
     2GETLAM 				( ... #n #m )
     DUP#0_DO
        1GETLAM DEPTH 1LAMBIND		( ... #n ... #m prg )
        JINDEX@ #1+ #>Z			
        INDEX@ #1+ #>Z			( ... #n ... #m prg indexi indexj )
        ROT ?ATTNQUIT EVAL		( ... #n ... #m result )
        DEPTH 1GETABND 			( ... #n ... #m result # #' )
        #<>case SETSTACKERR
     SWAPLOOP				( ... #n ... #m )
     []N 				( ... #n line_i )
  SWAPLOOP				( ... #n )
  ABND					( meta-matrix )
;

LABEL CKMATSIZE
* ( # -> )
* check that dimension of arry is not too large for 250K of user RAM
* NB: must insure twice the room because of the last []N
::
  # 5000 #< ?SEMI
  BINT75 ERABLEERROR
;

**********************************************************************
* Desc:		Used for addition of numeric matrix and symbolic object.
* Stack:	( arry ob --> arry arry )
*		( ob arry --> arry arry )
**********************************************************************
NULLNAME ADDMATOBJext
::
  DTYPEARRY?
  '
  ::				( obj, array -> array, obj )
    DUP MATIDN ROT		( array, idn, objet )
    CK&DISPATCH1
    real x*
    TWO  x*
    any  ::
           ROT ARRAY2MATRIX UNROT	( array, idn, objet )
           DTYPFMAT? case SWAPDROP
           SWAP ARRAY2MATRIX SWAP
           ONE DEPTHEVAL: x*
         ;
  ;
			( obj, array , TRUE, prg | array, obj, FALSE, prg )
  SWAP case :: EVAL SWAP ;
  ROTSWAP EVAL
;

**********************************************************************
* Desc:		Apply unaryop(v[i]) to get V[i]
* Stack:	( v op --> V )
**********************************************************************
NULLNAME VUNARYOP
::
  1LAMBIND 		( v )
  DUP >R LENCOMP	( # )
  ZEROSWAP		( meta-V # )
  ZERO_DO (DO)		( meta-V )
    RSWAP 'R RSWAP	( meta-V v[i] )
    1GETLAM EVAL	( meta-V V[i] )
    SWAP#1+		( meta-V )
  LOOP			( meta-V )
  ABND []N		( V )
;

**********************************************************************
* Desc:		Apply unaryop(m[i,j]) to get M[i,j]
* Stack:	( m [runstream op] --> M )
* Note:		works for matrices and for vectors
**********************************************************************
LABEL MUNARYOP
::
  'R 1LAMBIND 		( m )
  DUP >R LENCOMP	( # )
  ZEROSWAP		( meta-M # )
  ZERO_DO (DO)		( meta-M )
    RSWAP 'R RSWAP	( meta-M m[i] )
    1GETLAM		( meta-M m[i] unaryop )
    OVER TYPEMATRIX?	( meta-M m[i] unaryop flag )
    ITE VUNARYOP EVAL	( meta-M M[i] )
    SWAP#1+		( meta-M )
  LOOP			( meta-M )
  ABND []N		( M )
;

**********************************************************************
* Desc:		Apply binop(v2[i],v1[i]) to get V[i]
* Stack:	( V2 V1 binop --> V )
* Note:		Works even if V2 and V1 have not the same dimension
**********************************************************************
NULLNAME VBINARYOP
::
  1LAMBIND
  INNERCOMP get1 INNERCOMP get1	( a1...aa b1...bb #a #a )
  ZERO_DO (DO)			( a1...ak ck+1...cb b1...bk #a )
     get1 ROTSWAP
     1GETLAM EVAL
     OVER #1+UNROLL
  LOOP
				( si #a<=#b: c[b-a]+1...cb b1...b[b-a] #a )
  []N ABND
;


**********************************************************************
* Desc:		Horner evaluation
* Stack:	( V r --> P[r] )
* Note:		where coordinates of V represent coefficients of a polynomial
**********************************************************************
NULLNAME PEVAL
::
  SWAP ARRAY2MATRIX 		( r V )
  2DMATRIX? case SETTYPEERR 
  PEvalSlow
;

* ====================================================================
*		Eigenvalue/eigenfunctions computation
* ====================================================================
**********************************************************************
* Desc:		Compute eigenvalues of a matrix like EGVL
* Stack:	( M --> vector of eigenvalues )
**********************************************************************
NULLNAME MATEGVL
::
  DUP MDIMS NcaseSIZEERR
  #= NOT?DIMERR
  BESTMATRIXTYPE NOTMATRIXcase 
ASSEMBLE
	IF	f39
RPL
  laEGVL
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR laEGVL
ASSEMBLE
	ENDIF
RPL

				( matrix )
  ONESWAP			( #1 matrix )
  LVARext VXXLext		( #1 lvar internal_matrix )
  METAMADJ ABND			( #1 lvar mpmadj mpcar )
  EXTRACTCOMMONDENO {}POLY	( #1 lvar mpmadj d pcar )
  SWAPDROP OVER#2+UNROL NDROP	( #1 lvar pcar )
  NULL{} ZERO 			( #1 lvar pcar {} #0 )
  COMPLEXEVAL RFACT2ext []N	( #1 lvar {} internal_pcar [egv] )
  4UNROLL 2DROP			( #1 [egv] lvar )
  VAL2ext			( [egv %mult] )
  ROOTM2ROOT
;

**********************************************************************
* Desc:		Transform list of root/multiplicites to vector of roots
* Stack:	( list/vector --> vector of eigenvalues )
**********************************************************************
NULLNAME ROOTM2ROOT
::				( [egv %mult] )
  DUPDUP NULLCOMP?		( [] [] flag )
  case2drop NULL{}		( -> [] )
  >R LENCOMP #2/		( # )
  ZEROSWAP
  ZERO_DO (DO)			( meta )
    RSWAP 'R'R RSWAP		( meta egv %mult )
    OVER TYPETAGGED?		( meta egv %mult flag )
    ITE DROPZERO COERCE 	( meta egv #mult )
    NDUPN top&			( meta )
  LOOP
  DUP#0=csedrp NULL{} []N	( [ egv ] )
;

**********************************************************************
* Desc:		Compute inverse, matrix polynomial and characteristic
*		polynomial.
* Stack:	( M --> M^-1 P[M] P[lambda] )
**********************************************************************
NULLNAME MADJ
::
  DTYPENDO? NOTcase
  :: TRIMOBJext Z1 OVER RDIVext NULL{} DUP ;
						( A )
  METAMADJ
					( Mpmadj Mpcar )
  {}POLY OBJ>R {}N R>OBJ		( pmadj pcar )
  2DUPSWAP				( pmadj pcar pcar pmadj )
  DUPLENCOMP NTHCOMPDROP		( pmadj pcar pcar [-1]^n*com[A] )
  SWAP DUPLENCOMP NTHCOMPDROP	( pmadj pcar [-1]^n*com[A] [-1]^n*det[A] )
  RNEGext DUP
  1GETABND TWO #/ DROP
  #0= IT RNEGext
  5UNROLL			( det[A] pmadj pcar +/-com[A] +/-det[A] )
  TWO DEPTHEVAL: RDIVext
  UNROT
			( [-1]^n*det[A] inv[A]=1/det[A]*com[A] pmadj pcar )
;

LABEL METAMADJ
( A -> meta-madj meta-pcar )
* binds lam4= I_n, lam3=A, lam2=A_k-1, lam1=#n
::
  DUPLENCOMP EXACTMATIDN DUPUNROT DUPLENCOMP		( I_n, A, I_n, #n )
  4NULLLAM{} BIND

* lam4= I_n, lam3=A, lam2=A_k-1, lam1=#n	( )

  ZERO Z1 ONE				( Meta-pmadj, Meta-pcar )
  1GETLAM #1+_ONE_DO (DO)		( Mpmadj Mpcar )

* complete pmadj

     psh 2GETLAM SWAP#1+ psh		( Mpmadj Mpcar )
     2GETLAM 3GETLAM MMMULT		( Mpmadj Mpcar Ak-1*A )
     DUP MATTRACE			( Mpmadj Mpcar Ak-1*A tr[A_k-1*A] )
     INDEX@ #>Z RNEGext RDIVext		( Mpmadj Mpcar A_k-1*A -tr[A_k-1*A]/k )

* complete pcar

     DUP4UNROLL ROT#1+UNROT		( Mpmadj Mpcar A_k-1*A -tr[A_k-1*A]/k )

* A_k=A_k-1*A-tr[A_k-1*A]/k*I_n

     4GETLAM SWAP MAT*SCL
     MADD 2PUTLAM			( Mpmadj Mpcar ) 
  LOOP
;


**********************************************************************
* Desc:		Compute eigenvalues/eigenvectors of a matrix like EGV
* Stack:	( M --> vector of eigenvalues )
**********************************************************************
NULLNAME MATEGV
::
  BESTMATRIXTYPE NOTMATRIXcase
ASSEMBLE
	IF	f39
RPL
  laEGV
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR laEGV
ASSEMBLE
	ENDIF
RPL

  DTYPENDO? NOT?DIMERR
  COMPLEXEVAL EXECJORDAN 		( pmin pcar list_egvc list_egvl )
  2SWAP 2DROP				( list_egvc list_egvl )
  SWAP' 
  :: TAG> CAR$ CHR_C EQUAL case NOTDIAGERR ; 
  1LAMBIND 				( list_egvl list_egvc )
  LISTSECOext ABND			( list_egvl list_egvc )
  METATRAN DUP#0=ITE {}N []N SWAP	( matrix list_egvl )
  ( * note that the Jordan normal form is not returned * )
;

LABEL NOTDIAGERR
::
  BINT83 ERABLEERROR
;

**********************************************************************
* Desc:		Eigenvalue/eigenfunctions computation
* Stack:	( M -> pmin pcar list_eigenvect list_eigenval )
*		( pmadj pcar -> pmin pcar list_eigenvect list_eigenval )
**********************************************************************
NULLNAME JORDAN
::
  DTYPENDO? case EXECJORDAN
  CK2NOLASTWD 
  OVER TYPELIST? NcaseSIZEERR	( if stk1 is not a matrix, stk2 is be a list )
  FACT1ext			( poly-madj-ext lvar poly_carac lfac )
  2SWAP				( poly_car lfac poly-madj-ext lvar )
  CDRCOMP VXXLext 2SWAP		( lvar poly-madj poly_car lrac )
  COMPUTEEGV 2SWAP		( pmin pcar list_egvc list_egvl )
  ROOTM2ROOT
;

LABEL EXECJORDAN
::
  DUP STOMATRIX 			( matrix )
  LVARext VXXLext MADJ			( lvar +/-det[A] inv[A] pmadj pcar )
  2SWAP 2DROP				( lvar pmadj pcar )
  REWRITEASQ				( lvar pmadj pcar )
  NULL{} ZERO RFACT2ext {}N ROTDROP	( lvar pmadj pcar lrac )
  COMPUTEEGV 2SWAP			( pmin pcar list_egvc list_egvl )
  ROOTM2ROOT
;


**********************************************************************
* Desc:		Eigenvalue/eigenfunctions computation
* Stack:	( lvar pmadj pcar lrac -> pmin pcar list_egvc list_egvl )
**********************************************************************
LABEL COMPUTEEGV
::				( lvar Pmatr Pcar roots )
  stepflag TestSysFlag OBJ>R
  2DUP
  NULL$
  ZEROZERO 2DUP 2DUP ZERO
  Z1 DUPDUP 
  FALSE
  FalseTrue ZERO
  TWENTYONE PICK

* 	{{ lvar_jordan pmadj pcar rootl pmin cur_root root_ext #taillepille #carac #propres #mult ligne_cur col_cur #ligne_exec signe pivot_cur bareiss disp_flag mod_flag bigdata_flag temp_var liste_var }}
DEFINE lvar_jordan 22GETLAM
DEFINE pmadj 21GETLAM
DEFINE pcar 20GETLAM
DEFINE rootl 19GETLAM
DEFINE pmin 18GETLAM
DEFINE cur_root 17GETLAM
DEFINE root_ext 16GETLAM
DEFINE #taillepile 15GETLAM
DEFINE #carac 14GETLAM
DEFINE #propres 13GETLAM
DEFINE #mult 12GETLAM
DEFINE #ligne_exc 9GETLAM
DEFINE lvar_jordan! 22PUTLAM
DEFINE pmadj! 21PUTLAM
DEFINE pcar! 20PUTLAM
DEFINE rootl! 19PUTLAM
DEFINE pmin! 18PUTLAM
DEFINE cur_root! 17PUTLAM
DEFINE root_ext! 16PUTLAM
DEFINE #taillepile! 15PUTLAM
DEFINE #carac! 14PUTLAM
DEFINE #propres! 13PUTLAM
DEFINE #mult! 12PUTLAM
DEFINE #ligne_exc! 9PUTLAM

  ' NULLLAM TWENTYTWO NDUPN DOBIND	( )
  stepflag ClrSysFlag
  ZERO rootl LENCOMP #2/
  #1+_ONE_DO (DO)			( meta-vect )
     rootl INDEX@ #2* 2DUP #1-		( meta-v rootl #2i rootl #2i-1 )
     NTHCOMPDROP DUP cur_root! 
     ZEROSWAP lvar_jordan VAL2ext
     DO>STR root_ext!
     NTHCOMPDROP COERCE			( meta-v #mult )
     COMPUTEEGVLP
  LOOP					( meta-vect )
  {}N 					( M )
  lvar_jordan 2PUTLAM ZERO 1PUTLAM
  rootl ONE INVAL2			( M liste_egvl )
  lvar_jordan RCLVX >HCOMP 2PUTLAM
  pmin VAL1
  pcar VAL1
  					( M liste_egvl pmin pcar )
* abandon des structures locales et restauration des flags display

  ABND stepflag R>OBJ ITE SetSysFlag DROP

;

LABEL COMPUTEEGVLP
::					( meta-v #mult )
  cur_root TYPETAGGED? caseDROP
  DUP#1= casedrop COMPUTEEGV1
  DUP#1- #taillepile! #mult!
  ZEROZERO #carac! #propres!		( meta-v )
  pmadj cur_root #mult			
  ZERO_DO (DO)				( meta-v P ... P[l] pmadj r )
     MHORNext MATTRAN			( P ... P[l] pmadj r P[l+1] )
     UNROT				( P ... P[l+1] pmadj r )
  LOOP
  2DROP					( meta-v P P' ... P[k] )
  #mult reversym
  #1-SWAP INNERCOMP			( meta-v meta-P' l1 ... ln #n )
  CHARREDRESET
  BEGIN					( meta-v meta-P' meta-l )
    #carac #mult #<
  WHILE					( meta-v meta-P' meta-l )
    COMPUTECHAR
  REPEAT				( meta-v meta-P' meta-l )
  NDROP NDROP
;

LABEL CHARREDRESET
::  
   ONEONE ligne_cur! col_cur! 
   Z1 bareiss!	
;

LABEL COMPUTECHAR
( meta-v meta-P' meta-l -> meta-v meta-P' meta-l )
::				( meta-v meta-P' meta-l )
* La reduction se fait en 2 temps:
* d'abord en interdisant les echanges de lignes < #propres
* (passage a la colonne suivante si le pivot est nul)
* ensuite reduction des lignes > #propres

   ligne_cur col_cur
   LASTCMETAPIVOT
   case CHARREDUCE		( -> meta-v meta-P' meta-l )
   3DROP			( meta-v meta-P' meta-l )

* 1er temps de la reduction terminee, maintenant on reduit les
* lignes > #propres

   #propres #1+ ligne_cur! ONE col_cur!
   [BEGIN]
     ligne_cur col_cur 
     LASTCMETAPIVOT		( meta-v meta-P' meta-l #l #l' #c flag )
   [WHILE]
     CHARREDUCE			( meta-v meta-P' meta-l )
   [REPEAT]
				( meta-v meta-P' meta-l #l #l' #c flag )
* reduction terminee, recalcule meta-P', meta-l, meta-v, 
* reinitialise bareiss, l, c, attention #ligne_exl et ligne_cur utilises

   3DROP			( meta-v meta-P' meta-l )
   CHARREDRESET			( meta-v meta-P' meta-l )

* les lignes non nulles de meta-l sont vecteurs propres
* seules celles a partir de #propres sont de nouveaux vecteurs propres

   [BEGIN]			( meta-l )
      SWAPDUP NULLVECTOR?	( l1 ... lk-1 #k lk ? )
      ROTDUP #propres #> 	( l1 ... lk-1 lk ? #k ? )
      ROTAND			( l1 ... lk-1 lk #k ? )
   [WHILE]
      pulldrop
   [REPEAT]
				( meta-v meta-P' meta-l ) 
   DUP#0=csedrp DIVIDEPMIN
					( meta-v meta-P' meta-l )

* on decale et on rajoute les vecteurs propres et carac a meta-v
* sauvegarde du nouveau nombre de vecteurs propres

   DUP #ligne_exc! 		( meta-v meta-P' v1 ..vk..  vl #l )
   []N ONE			( meta-v meta-P' [v1..vk..vl] #1 )
* effectue les decalage dans la pile des matrices
   DECALEMETAP			( meta-v new-meta-P' meta-V )
* puis extraction des cycles
   EXTRACTNEWCYCLES		( meta-v new-meta-P' )
   #ligne_exc #propres #-
   #taillepile #1+ #* 		( meta-v new-meta-P' #num_vect_carac )
   #carac #+ #carac!
   #ligne_exc #propres!
*        #ligne_exc #1+ ligne_cur!
   #taillepile #1- #taillepile!	( meta-v new-meta-P' )
   DUP#0=case DUP
   #1-SWAP INNERCOMP		( meta-v new-meta-P' meta-l )
;

LABEL DECALEMETAP
::				( meta-v meta-P' [v1..vk..vl] #1 )
   #taillepile DUP#0=csDROP
				( meta-v meta-P' meta-V )
   ZERO_DO (DO)			( meta-v meta-P' meta-V )
     DUPINDEX@ #+ #3+ ROLL	( meta-v meta-P' meta-V P' )
     INNERCOMP #ligne_exc #-	( meta-v meta-P' meta-V P'_1 P'_2 )
     #ligne_exc OVER#2+UNROL
     psh []N			( meta-v meta-P' meta-V meta-P'_2 P'_1 )
     OVER #3+ UNROLL		( meta-v meta-P' meta-V meta-P'_2 )
     DUP #4+PICK INNERCOMP
     pshtop& []N		( meta-v meta-P' meta-V new-P' )
     OVERINDEX@ #+ #4+ UNROLL #1+
   LOOP
;

LABEL EXTRACTNEWCYCLES
::				( meta-v new-meta-P' meta-V )
  OVER LENCOMP #propres		( meta-v new-meta-P' meta-V #l #propres )
  #=case NDROP			( -> meta-v new-meta-P' )
  rot get1
  ZERO_DO (DO)			( new-meta-P' v1...vk meta-v )
    get1 			( new-meta-P' v1...vk-1 meta-v vk )
    INNERCOMP
    #propres DUPUNROT #- 	( new-meta-P' v1...vk-1 meta-v vk.1...vk.p #propres #nouveaux )
    DUPUNROT #2+UNROLL
    psh NDROP 			( new-meta-P' v1...vk-1 meta-v vk.propres+1 ... vk.p #nouveaux )
    DUPDUP 
    ZERO_DO (DO)
      get1 ONESWAP
      lvar_jordan VAL2ext
      root_ext 
      JINDEX@ #1+ JSTOP@ #=
      ITE "Eigen: " "Char: " SWAP&$ >TAG
      UNROT
      LOOP
    DROP top&			( new-meta-P' v1 ... vk-1 meta-v )
    LOOP			( new-meta-P' meta-v )
  psh				( meta-v new-meta-P' )
;


LABEL DIVIDEPMIN
::	
					( meta-v meta-P' ) 
  #taillepile #1- #taillepile!		( meta-v meta-P' ) 
  #1-SWAP INNERCOMP			( meta-v new-meta-P' new-meta-l ) 

  pmin cur_root RACTOFACext
  RDIVext RDIVext pmin!	

* remarque: meta-P' ne peut etre vide ici!
;


LABEL CHARREDUCE
::					( meta-v meta-P' meta-l #l #l' #c )
  col_cur!
  DUP #ligne_exc!			( meta-v meta-P' meta-l #l #l' )
  2DUP#= ITE 2DROP CHAREXL		( meta-v meta-P' meta-l )


* reduction: calcul du pivot -> pivot_cur

  DUP ligne_cur #- #2+PICK
  DUP col_cur NTHCOMPDROP 		( meta-v meta-P' meta-l lk pivot )
  DUP Z0 EQUAL case2drop c1+
  pivot_cur!				( meta-v meta-P' meta-l lk )
  DUPLENCOMP #1+ #propres
  #1+ DO				( meta-v meta-P' meta-l lk )
     INDEX@ ligne_cur #=?SKIP CHARELEM	( meta-v meta-P' meta-l lk )
  LOOP
  DROP					( meta-v meta-P' meta-l )

* mise a jour bareiss et suite de la reduction

  pivot_cur bareiss!
  ligne_cur #1+ ligne_cur!
  c1+
;

LABEL c1+
::
  col_cur #1+ col_cur!
;

LABEL CHAREXL
( * exchange line l and l' if #l>#char * )
* if #l<=#propre line exchange is forbidden
::				( meta-v meta-P' meta-l #l #l' )
  OVER #propres #> NOTcase2DROP	( -> meta-v meta-P' meta-l )
				( meta-v meta-P' meta-l #l #l' )
  col_cur METAEXL		( meta-v meta-P' meta-l #l #c )
  2DROP
  #taillepile #0=?SEMI 
  psh				( meta-v meta-l meta-P' )
  DUPDUP 
  ZERO_DO (DO)		( ... P1 ... Pk #k #k )
     get1 INNERCOMP
     ligne_cur #ligne_exc DUP
     METAEXL 2DROP
     []N UNROT
  LOOP
  DROP psh				( meta-v meta-P' meta-l )
;

LABEL CHARELEM
::
  METAELEM			( meta-v meta-P' meta-l lk )

* determine le coef -> temp_var et operation elementaire sur meta-l

  #taillepile #0=?SEMI

* sauvegarde lk 

  OBJ>R psh			( meta-v meta-l meta-P' )
  DUPDUP 
  ZERO_DO (DO)			( ... meta-P[] #n )
    get1 INNERCOMP		( ... P|2| .. P|n| #n #n meta-P|1| #r )
    DUP ligne_cur #- #2+PICK	( ... P|2| .. P|n| #n #n meta-P|1| lk )
    OVER JINDEX@ #-+1 #2+ROLL	( ... P|2| .. P|n| #n #n meta-P|1|' lk li )
    METAELEMEXEC		( ... P|2| .. P|n| #n #n meta-P|1| lk new-li )
    3PICK JINDEX@ #- #3+ UNROLL
    DROP []N UNROT
  LOOP
				( ... meta-P[] #n )
  DROP psh R>OBJ
;


LABEL COMPUTEEGV1
::				( meta-v )
         pmadj cur_root MHORNext
         UNROT2DROP			( meta-v P[r] )
         MATTRAN INNERCOMP
         [BEGIN]
	    DUP#0= caseSIZEERR
            #1-SWAP
            DUP NULLVECTOR?		( *recherche 1 ligne non nulle* )
         [WHILE]
            DROP
         [REPEAT]
					( meta-v meta ligne )
         OVER#2+UNROL NDROP
         LPGCDext
         Z1 SWAP RDIVext VPMULT
         ONESWAP lvar_jordan VAL2ext
         root_ext >TAG 
         SWAP#1+
;

**********************************************************************
* Quadratic forms handling
**********************************************************************
**********************************************************************
* Desc:		Convert symbolic quad form to matrix quad form
* Stack:	( symb lidnt -> matrix lidnt )
**********************************************************************
NULLNAME QXA
::
  ARRAY2MATRIX DUP MDIMS 
  caseSIZEERR DROP			( symb lidnt )
  2LAMBIND				( )
  ZERO 1GETLAM LENCOMP			( meta # )
  #1+_ONE_DO (DO)			( meta )
    2GETLAM 1GETLAM INDEX@ NTHCOMPDROP	( meta symb idntk )
    DERIVext				( meta symb' )
    1GETLAM DERIVext 			( meta 2*line_k )
    SWAP#1+
  LOOP					( meta )
  []N Z2 MAT/SCL SIMP1ext		( matrix )
  ( * now check that the line is constant * )
  LVARext 1GETLAM			( matrix lvar lidnt )
  toLEN_DO (DO)				( matrix lvar lidnt )
    DUPUNROT INDEX@ NTHCOMPDROP 	( matrix lidnt lvar idnt_k )
    Contains? caseSIZEERR		( -> not a quadratic form )
    DROPSWAP				( matrix lvar lidnt )
  LOOP
  SWAPDROP				( matrix lidnt )
;


**********************************************************************
* Desc:		Convert matrix quad form to qymbolic quad form
* Stack:	( matrix lidnt -> symb lidnt )
**********************************************************************
NULLNAME AXQ
::
				( matrix lvar )
  DUP MDIMS caseSIZEERR		( lvar must be a vector! )
				( endo lvar #n )
  ROTDUP MDIMS 			( lvar #n endo [ #m #n true | #m false ] )
  NcaseSIZEERR
				( lvar #n endo #m #n )
  OVER#= NOT?DIMERR
  				( lvar #n endo #m )
  ROT #= NOT?DIMERR		
				( lvar endo )
  ( * symb is lvar.[endo'.lvar] * )
  OVER MVMULT OVER XYext	( lvar symb )
  SIMPLIFY SWAP			( symb lvar )
;

**********************************************************************
* Desc:		gauss reduction of quadr. form (symbolic)
* Stack:	( symb -> D P symb' )
**********************************************************************
NULLNAME GAUSS
::
  QXA SWAP			( lidnt matrix )
  SYLVESTER			( lidnt D P )
  ROT 1LAMBIND 2DUP		( D P D P )
  INNERCOMP 			( D P D P1..Pn #n )
  ZEROSWAP			( D P D P1..Pn meta #n )
  ZERO_DO (DO)			( D P D P1..Pk meta )
     get1 1GETLAM XYext		( D P D P1..Pk-1 meta Pk.lidnt )
     SWAP#1+
  LOOP				( D P D meta-P.idnt )
  reversym get1 INNERCOMP	( D P meta-P.idnt meta-D )
  DUP #4+			( D P meta-P.idnt meta-D # )
  Z0 ROT			( D P px1..pxn #n d1..dn # symb #n )
  ZERO_DO (DO)			( D P px1..pxn #n d1..dn # symb )
     ROT 3PICK PICK 		( D P px1..pxn #n d1..dk # symb dk+1 pxk+1 ) 
     Z2 x^ext x*ext x+ext	( D P px1..pxn #n d1..dk # symb )
  LOOP				( D P px1..pxn #n # symb )
  SWAPDROP OVER#2+UNROLL NDROP	( D P symb )
  1GETABND			( D P symb lidnt )
;

**********************************************************************
* Desc:		gauss reduction of a quadr form (matrix)
* Stack:	( matrix -> D P )
**********************************************************************
NULLNAME SYLVESTER
::
  DUP MDIMS NcaseSIZEERR
  #= NOT?DIMERR				( matrix )
  LVARext VXXLext			( lvar matrix )
  EXACT? NOTcase :: BINT36 ERABLEERROR ;
  SYLVESTER0				( lvar D P )

  { ONE TWO } FROMLISText		( D P )
;

LABEL SYLVESTER0
( matrix -> D P )
::
  ( * if length is 1, return D and P=idn )
  DUPLENCOMP #1=case SYLVESTER1 
					( A )
  ( * if M is 0 return D=0 and P=idn * )
  DUP Z0 MAT*SCL			( A 0 )
  OVER EQUALcase SYLVESTER1
					( A )
  ( * search for a non 0 coeff in the diagonal of A * )
  ( * flag is true if all coeff are 0 * )

  DUPINCOMP reversym			( A ln..l1 #n )
  TRUE OVER				( A ln...l1 #n true #n )
  #1+_ONE_DO (DO)
    ROT INDEX@ NTHCOMPDROP		( A ln..lk+1 #n true lkk )
    Z0 EQUAL AND			( A ln...lk+1 #n boolean )
    ITE TRUE
          ::
            INDEX@#- NDROP
            INDEX@ ExitAtLOOP
            FALSE
          ;
  LOOP
					( A #k false | A #n true )
  case SYLVESTERALL0
					( A #k )
  ( * now reduce the pb to a matrix of order #n-1 * )

  2DUP NTHCOMPDROP			( A #k lk )
  DUP3PICK NTHCOMPDROP 4PICK LENCOMP	( A #k lk akk #n )
					(    4  3  2   1  )
  4NULLLAM{} BIND			( A )
  INNERCOMP reversym			( ln..l1 #n )
  DROP 4GETLAM ROLLDROP ZERO		( ln...lk+1 lk-1... l1 meta )
  1GETLAM 
  ONE_DO (DO)				( ln...lk+1 lk-1...lr meta )
    get1 INNERCOMP reversym		( ...meta arn..ar1 #n )
    4GETLAM #1+ ROLLDROP		( ...meta arn...ark+1 ark-1...ar1 #n )
    3GETLAM INDEX@ DUP#1+		( arn...ark+1 ark-1...ar1 #n lk # # )
    4GETLAM #> IT #1+			( arn...ark+1 ark-1...ar1 #n lk # )
    NTHCOMPDROP				( arn...ark+1 ark-1...ar1 #n akr )
    ONE ROT				( arn...ark+1 ark-1...ar1 meta&akr #n )
    ONE_DO (DO)				( arn...ark+1 ark-1...ar1 meta&akr )
      get1 3PICK 3GETLAM		( ...arj+1 meta&akr arj akr lk )
      INDEX@ DUP#1+ 4GETLAM #> IT #1+	( ...arj+1 meta&akr arj akr lk # )
      NTHCOMPDROP RMULText		( ...arj+1 meta&akr arj akr*akj )
      2GETLAM RDIVext RSUBext		( ...arj+1 meta&akr arj-akr*akj/akk )
      UNROT #1+				( ...arj+1 meta&akr )
    LOOP				( ...meta meta&akr )
    pulldrop []N SWAP#1+		( ln...lk+1 lk-1...lr+1 meta )
  LOOP					( meta )
  []N					( A' )
  4GETLAM 3GETLAM 			( A' #k lk )
  OVER Z1 2GETLAM RDIVext 		( A' #k lk #k 1/akk )
  ABND 5ROLL SYLVESTER0			( #k lk #k 1/akk D P )
  4UNROLL				( #k lk P #k 1/akk D )

  ( * insert 1/akk at proper place in the D vector * )
  ( * note that this is insertion, not replace, we can not use putmeta * )
  INNERCOMP get1 SWAP#1+ get1		( #k lk P d1..dn 1/akk #n+1 #k )
  OVERSWAP #- #2+ ROTSWAP UNROLL []N	( #k lk P D )
  4UNROLL				( D #k lk P )

  ( * insert lk and a column of 0 at proper place in P )
  MATTRAN				( D #k lk tran[P] )
  DUPLENCOMP 4PICK 			( D #k lk tran[P] #tran #k )
  Z0 ROT NDUPN []N			( D #k lk tran[P] #k [0..0] )
  ROT					( D #k lk #k [0..0] tran[P] )
  INNERCOMP get1 SWAP#1+ get1		( D #k lk tran[P]&[0..0] #k )
  OVERSWAP #- 				( D #k lk tran[P]&[0..0] #n-k )
  #2+ ROTSWAP 				( D #k lk tran[P] #+1 [0..0] #n-k+2 )
  UNROLL				( D #k lk tran[P]&[0..0] )
  mattran				( D #k lk meta-P )
  get1 SWAP#1+ get1
  OVERSWAP #- #2+ ROTSWAP UNROLL
  []N					( D P )
;

LABEL SYLVESTER1
( A -> car[A] idn )
:: CARCOMP DUPLENCOMP EXACTMATIDN ;

LABEL SYLVESTERALL0
::					
  ( * all diag coeff are 0 * )
					( A #n )
  ( * make a matrix from identity with * )
  ( * a 1 at row n-1 column n and a -1 at row n column n-1 * )
  DUP' EQUALZ1ELSEZ0 make2dmatrix	( A meta-mat )
  Z1 OVER 5ROLL PUT[] UNROT		( A meta-mat' )
  Z-1 OVER#1- 4ROLL PUT[] SWAP		( A meta-mat" )
  dup mattran []N 			( A meta-mat trans[M] )
  OBJ>R []N SWAP R>OBJ 			( M A trans[M] )

  ( * now call sylvester on M*A*tran[M] * )
  MMMULT MMMULT				( A' )
  SYLVESTER0				( D P )

  ( * finish by multiplying P by trans[M]^-1 * )
  DUPLENCOMP 				( D P #n )
  DUP' EQUALZ1ELSEZ0 make2dmatrix	( D P meta-mat )
  Z1 EXACT? ITE Z2 %%2 QDiv		( D P meta 1/2 )
  DUPDUP 4PICK 7ROLL PUT[]		( D P meta' #n 1/2 1/2 ln-1 )
  4PICK #1-SWAP PUT[]			( D P meta' #n 1/2 ln-1 )
  4UNROLL DUP QNeg			( D P meta 1/2 -1/2 )
  3PICK #1- 5ROLL PUT[]			( D P meta' #n 1/2 ln )
  3PICKSWAP PUT[]			( D P meta' #n ln )
  SWAP []N				( D P M )
  MMMULT
;
