INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s
INCLUDE erext.h

	STITLE	Meta operations

**********************************************************************
*	MetaAdd		( meta, meta -> meta+meta )
*	MetaSub		( meta, meta -> meta-meta )
*	MetaMul		( meta, meta -> meta*meta )
*	MetaDiv		( meta, meta -> meta/meta )
*	MetaPow		( meta, meta -> meta^meta )
*	MetaNeg		( meta -> -meta )
*			These words check for undef, inf and trivial cases
*
*
*	m-1&m+1		( meta -> meta&1&-, meta&1&+ )
*	meta1/meta	( meta -> meta, 1&meta&/ )
*	1&meta		( Meta --> 1&Meta )
*	meta/2		( Meta --> Meta&2&/ )
*	addt2		( Meta --> Meta&2 )
*	addt/		( meta -> meta & x/ )
*	meta2*		( meta -> meta &2&x* )
*	meta1-sq	( meta -> 1&meta&SQ&x- )
*	metasq+1	( meta -> meta&SQ&1&x+ )
*	metasq-1		( meta -> meta&SQ&1&x- )
*	NDROPZERO	( meta -> #0 )
* 	metaadd		( meta2, meta1 -> meta )
*			addition of metas, checks for meta=Z0 ONE
* 	metasub		( meta2, meta1 -> meta )
*			substraction of metas, checks for meta=Z0 ONE
* 	metamult	( meta2, meta1 -> meta )
*			multiplication of metas, checks for meta=Z0/Z1 ONE
* 	meta^		( meta, obj -> meta^obj )
*			integer power of a meta, checks for Z=Z1
*			anc for NEG to an even integral power
* 	top&addt*	( meta2, meta1 -> meta2*meta1 )
*	addti		( meta -> meta&i )
*	metackneg	( meta -> meta&NEG )
*			Checks for NEG and for 0
*	metaEQUAL	( meta2, meta1 -> meta2, meta1, flag )
*			Fast test equality of 2 metas
**********************************************************************

**********************************************************************
* Desc:		Split a meta in 2 metas at position i
* Stack:	( meta #i -> meta-1 meta-2 )
* Note:		meta-1 will contain #i elements
*		meta-2 will contain #n-i elements 
**********************************************************************
NULLNAME metasplit
::
  2DUP#< case SETDIMERR
  DUPUNROT #-			( m1..mn #i #n-i )
  DUPUNROT #2+UNROLL		( m1...mi #i mi+1...mn #n-i )
;

* --------------------------------------------------------------------
* Desc:		Meta utility.
* Stack:	( meta -> meta&1&+, meta&1&- )
* --------------------------------------------------------------------
NULLNAME m-1&m+1
::
  dup Z1 SWAP#1+ addt- psh Z1 SWAP#1+ addt+ 
;

* --------------------------------------------------------------------
* Desc:		Meta utility.
* Stack:	( meta -> meta, 1&meta&/ )
* --------------------------------------------------------------------
NULLNAME meta1/meta
::
  dup 1&meta addt/
;

* --------------------------------------------------------------------
* Desc:		Meta utility.
* Stack:	( Meta --> 1&Meta )
* --------------------------------------------------------------------
NULLNAME 1&meta
::
  Z1 psh1top&
;

* --------------------------------------------------------------------
* Desc:		Meta utility.
* Stack:	( Meta --> Meta&2&/ )
* --------------------------------------------------------------------
NULLNAME meta/2
::
  addt2: Z2 x/ 
;

* --------------------------------------------------------------------
* Desc:		Meta utility.
* Stack:	( Meta --> Meta&2 )
* --------------------------------------------------------------------
NULLNAME addt2
::
  Z2 SWAP#1+
;

* --------------------------------------------------------------------
* Desc:		Append division to Meta
* Stack:	( Meta --> Meta&/ )
* --------------------------------------------------------------------
NULLNAME addt/
::
  addt: x/
;

* --------------------------------------------------------------------
* Desc:		Multiply meta by 2
* Stack:	( Meta --> 2&Meta&* )
* --------------------------------------------------------------------
NULLNAME meta2*
::
  Z2 psh1top& addt*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative utility.
* Stack:	( Meta --> 1&Meta&SQ&- )
* --------------------------------------------------------------------
NULLNAME meta1-sq
::
  addtSQ Z1 psh1top& addt-
;

* --------------------------------------------------------------------
* Desc:		Meta derivative utility.
* Stack:	( Meta --> Meta&SQ&1&+ )
* --------------------------------------------------------------------
NULLNAME metasq+1
::
  addtSQ addt2: Z1 x+
;

* --------------------------------------------------------------------
* Desc:		meta-x -> meta-x^2-1
* Stack:	( Meta --> Meta&SQ&1&- )
* --------------------------------------------------------------------
NULLNAME metasq-1
::
  addtSQ meta-1
;

* --------------------------------------------------------------------
* Desc:		Meta-x -> Meta-x-1
* Stack:	( Meta --> Meta&1&- )
* --------------------------------------------------------------------
NULLNAME meta-1
::
  addt2: Z1 x-
;

* --------------------------------------------------------------------
* Desc:		Replace Meta object with empty Meta object
* Stack:	( Meta --> #0 )
* --------------------------------------------------------------------
NULLNAME NDROPZERO
:: 
  NDROP ZERO 
;

LABEL NDROPONE
:: 
  NDROP ONE
;

LABEL NDROPTWO
:: 
  NDROP TWO
;

NULLNAME 2DROPZ0
::
  2DROP Z0
;

* --------------------------------------------------------------------
* Desc:		Add 2 meta objects with trivial simplifications.
* Stack:	( Meta1 Meta2 --> Meta1+Meta2 )
* NOTE:		metaadd checks for Meta1/2=Z0 ONE
* --------------------------------------------------------------------
NULLNAME metaadd
::
  meta2zint QAdd EvalNoCKx+
  OVER j#0=case 2DROP
  DUP #3+PICK j#0=case swapdrop
  top& OVER EQ: xNEG NOTcase addt+
  pulldrop addt-
;

* --------------------------------------------------------------------
* Desc:		Substract 2 meta objects with trivial simplifications.
* Stack:	( Meta1 Meta2 --> Meta1+Meta2
* NOTE:		metasub checks for Meta1/2=Z0 ONE
* --------------------------------------------------------------------
NULLNAME metasub
::
  meta2zint QSub EvalNoCKx-
  metaEQUAL? case 2NDROPZ0ONE
  OVER j#0=case 2DROP
  DUP #3+PICK 				( meta1 meta2 op1 )
  DUPEQ:
ASSEMBLE
	IF	f39
RPL
  xint
ASSEMBLE
	ELSE	f39
RPL
  xINT
ASSEMBLE
	ENDIF	f39
RPL
  casedrop metasubint
  j#0=case :: swapdrop metackneg ;
  dometasub
;

LABEL dometasub
::
  top& OVER EQ: xNEG NOTcase addt-
  pulldrop addt+
;

LABEL metasubint
::					( meta1 meta2 )
  AEQ1stcase
ASSEMBLE
	IF	f39
RPL
  xint
ASSEMBLE
	ELSE	f39
RPL
  xINT
ASSEMBLE
	ENDIF	f39
RPL

  metasub2int
  dometasub
;

LABEL metasub2int
* fnc var sup1..n xINT #n+3 fnc' var' inf1..k xINT #k+3
* if fnc=fnc' and var=var' return an integral instead
::
  DUP#2+PICK					( meta1 meta2 #n+3 )
  2DUP #+ #3+PICK				( meta1 meta2 #n+3 fnc )
  3PICK #3+PICK EQUAL 				( meta1 meta2 #n+3 fnc=fnc'? )
  NOTcasedrop dometasub				( -> )
						( meta1 meta2 #n+3 )
  OVER#+ #1+PICK				( meta1 meta2 var )
  OVER #1+PICK EQUAL				( meta1 meta2 var var' )
  NOTcase dometasub				( -> )
						( meta1 meta2 )
  SWAPDROP #3-					( meta1 fnc' var' meta-inf )
  get1 DROP get1 DROP				( meta1 meta-inf )
  psh						( meta-inf meta1 )
  SWAPDROP #3- 					( meta-inf fnc var meta-sup )
  get1 SWAP#1+ 					( meta-inf fnc meta-sup&var )
  get1 UNROT #1+				( meta-inf meta-sup&fnc&var )
  top& addt: xINTEGRAL				( meta-integral )
;

* --------------------------------------------------------------------
* Desc:		Multiply 2 meta objects with trivial simplifications.
* Stack:	( Meta1 Meta2 --> Meta1*Meta2
* Note:		Checks for meta1, meta2= Z0 or Z1, checks for xNEG
* --------------------------------------------------------------------
NULLNAME metamult
::
  meta2zint QMul EvalNoCKx*
  metamultnozck
;

LABEL metamultnozck
::
  M1stNEGcase metamultneg2
  OVER j#0=case swapdrop
  OVER j#1=case 2DROP
  DUP #3+PICK num#0=case DROPdrop
  DUPEQ: xNEG casedrop metamultneg
  j#1=case swapdrop
  top& addt*
  ifnegmoveneg
;

LABEL metamultneg2
( meta1, meta2&NEG -> [meta1*meta2]&NEG )
::
  pulldrop metamult metackneg
;

LABEL metamultneg
( meta1&NEG, meta2 -> [meta1*meta2]&NEG )
::
  psh pulldrop psh metamult metackneg
;

LABEL ifnegmoveneg
* tests for xNEG ? # -> ? xNEG #, used for ?=* or /
::
  3PICK EQ: xNEG NOT?SEMI
  ROTSWAP
;

* --------------------------------------------------------------------
* Desc:		Divide 2 meta objects with trivial simplifications.
* Stack:	( Meta2 Meta1 --> Meta2/Meta1
* Note:		Checks for 0, meta2 =1 or Z-1, checks for xNEG
* --------------------------------------------------------------------
NULLNAME metadiv
::
  ( * if meta2=0 simplify meta1 * )
  DUP #3+PICK Z0 EQUAL IT metasimp

  OVER j#0=case metadivby0
  meta2zint SIMPEVALQDiv EvalNoCKx/
  metaEQUAL? case 2NDROPZ1ONE
  OVER num#1=case 3DROP
  j#-1=case 2DROPmetackneg
  DUP #3+PICK num#0=case DROPdrop
  EQ: xNEG case metadivneg
  top& addt/
  ifnegmoveneg
;

LABEL SIMPEVALQDiv
::
  SIMPEVALext QDiv
;

LABEL metadivby0
::             ( meta1 0 #1 )
  DROP Z0 EQUALNOTcase metadivby%0
            ( meta1 )
  ( * added metasimp for meta1 to avoid unrecognized 0/0 form * )
  metasimp OVER     ( meta1 obj )
  j#0=case div0by0
  NDROPunsignedinf
;

LABEL metadivby%0
::
  metasimp OVER
  j#0=case :: NDROP %0 %0 EvalNoCKx/ ;
  NDROP %1 %0 EvalNoCKx/ ONE
;


LABEL div0by0
::						( meta )
  NDROPundef div0by0flag SetSysFlag		( -> meta-undef )
;

LABEL 2DROPmetackneg
:: 
  2DROP metackneg
;

LABEL metadivneg
( meta1&NEG, meta2 -> [meta1/meta2]&NEG )
::
  psh pulldrop psh metadiv metackneg
;

* --------------------------------------------------------------------
* Desc:		power of a meta
* Stack:	( Meta obj --> Meta&obj&^ )
* Note:		If obj=1 returns meta
*		Test for meta=meta'&NEG for integral powers
* --------------------------------------------------------------------
NULLNAME meta^
::
  DUPTYPEBINT? IT #>Z
  OBJINT? NOTcase meta^symb
					( meta Z )
  num#0=case meta^0
  num#1=case DROP
  3PICK DUPTYPEZ? casedrop metaZ^Z
					( meta Z op )
  DUPEQ: xXROOT casedrop metaxroot^
  EQ: x^ case meta^^
					( meta Z )
  SWAP#1+ addt^				( meta )

* tests for even integer power of NEG
* ... xNEG even_integer x^ #
					( meta&Z&^ )
  4PICK EQ: xNEG NOT?SEMI
					( meta' NEG Z ^ # )
  3PICK Z2 ZMod QIsZero? 		( meta' NEG Z ^ # ?Z_even )
  NOTcase 4ROLLSWAP			( -> meta' Z ^ NEG # )
  4ROLLDROP #1-				( -> meta' Z ^ #-1 )
;

LABEL meta^0
::					( meta, 0 )
  DROP metasimp OVER j#0=case NDROPundef
  NDROP Z1ONE
; 

LABEL meta^symb
:: SYMBINCOMP top& addt^ ;

LABEL metaZ^Z
::
  SWAPDROP				( Z Z )
  DupZIsNeg? UNROT			( flag Z Z )
  ZABS Z># RP#				( flag Z )
  SWAP NOTcase ONE			( -> Z one )
  Z1 SWAP' x/ THREE			( -> Z1 Z / three )
;

LABEL z^z
* z1 z2 -> z1^z2
::
  DupZIsNeg? UNROT			( flag Z Z )
  ZABS Z># RP#				( flag Z )
  SWAP					( Z flag )
  NOT?SEMI				( -> Z )
  xINVext				( 1/Z )
;

LABEL meta^^
::					( meta Z )
  OBJ>R pulldrop larg			( meta meta-pow )
  R>OBJ ONE metamult meta^meta
;

* --------------------------------------------------------------------
* Desc:		Append power operator to meta object
* Stac:		( Meta --> Meta&^ )
* --------------------------------------------------------------------
NULLNAME addt^
::
  addt: x^
;

LABEL metaxroot^
::					( meta-xroot z )
  OBJ>R					( meta-xroot )
  dup plDRPpZparg			( meta-xroot meta-obj meta-den )
  CKSYMBN R>OBJ				( meta-xroot meta-obj den power )
  SWAP BIN2EVAL BESTDIV2		( meta-xroot meta-obj quot rem )
  OBJ>R meta^				( meta-xroot meta-obj^quot )
  psh R>OBJ 				( meta-obj^quot meta-xroot rem )
  Z0 EQUALcasedro NDROP			( -> meta-obj^quot )
  Z1 EQUALcasedro metamult		( -> meta-xroot*meta-obj^quot )
  SWAP#1+ addt^ metamult 		( meta-xroot^rem*meta-obj^quot ) 
;

* --------------------------------------------------------------------
* Desc:		meta-power of a meta
* Stack:	( Meta2 Meta1 --> Meta2^Meta1 )
* Note:		If lengthe of Meta1 is ONE calls meta^
* --------------------------------------------------------------------
NULLNAME metapow
::
  meta2zint :: ONESWAP metaZ^Z CKSYMBN ; EvalNoCKx^
  EXACT? NOTcase top&addt^
  DUP#1= casedrop meta^
  M1stNEGcase :: pulldrop metapow Z1ONE psh MetaDiv ;
  metafraction? case meta^fraction
  top&addt^
;
	
LABEL top&addt^
::
  top& addt^
;

LABEL meta^fraction
::
  2DROP				( meta n d )
  Z2 EQUALcasedro meta^n/2
  SWAP OBJ>R 			( meta d )
  ONE psh metaxroot		( xroot[meta,d] )
  R>OBJ meta^			( meta )
;

LABEL meta^n/2
::				( meta n )
  OBJ>R addtSQRT R>OBJ		( meta n )
  meta^
;

* --------------------------------------------------------------------
* Desc:		Test if meta is a fraction of integers
* Stack:	( Meta --> Meta flag )
* --------------------------------------------------------------------
NULLNAME metafraction?
::
  OVER EQ: x/ NOTcaseFALSE
  3PICK TYPEZ? NOTcaseFALSE
  4PICK TYPEZ?
;

* --------------------------------------------------------------------
* Desc:		meta-XROOT of a meta
* Stack:	( Meta2 Meta1 --> Meta2 XROOT Meta1 )
* --------------------------------------------------------------------
NULLNAME metaxroot
::
  meta2zint zxroot EvalNoCKxXROOT
  ( * test if xroot is a square root * )
  DUP #3+PICK Z2 EQUALcase pshNDROPSQRT
  ( * if meta1 is a power do XROOT first and ^ after * )
  M1st^case meta^xroot
  top& addt: xXROOT
;

LABEL meta^xroot
::				( meta2 meta1 )
  pulldrop larg 		( meta2 meta-arg1 meta^ )
  rot metadiv metapow		( meta2 XROOT meta1 )
;

LABEL pshNDROPSQRT
:: psh NDROP addtSQRT ;

LABEL zxroot
* z2 z1 -> z1^[1/z2]=exp[ln[z1]/z2]
::		( z2 z1 )
  num#0=case SWAPDROP
  ONE addtLN get1 ONE MetaDiv metasimp METAEXPEXPA SYMBN
  EXP2POW
;

* --------------------------------------------------------------------
* Desc:		top& addt*
* Stack:	( meta2, meta1 -> meta2*meta1 )
* Note:		No checks
* --------------------------------------------------------------------
NULLNAME top&addt*
::
  DUP#0=csDROP
  top& addt*
;

* --------------------------------------------------------------------
* Desc:		top& addt/
* Stack:	( meta2, meta1 -> meta2/meta1 )
* Note:		No checks
* --------------------------------------------------------------------
NULLNAME top&addt/
:: 
  DUP#0=csDROP
  top& addt/ 
;	



* --------------------------------------------------------------------
* Desc:		addti
* Stack:	( meta -> meta&i )
* --------------------------------------------------------------------
NULLNAME addti
::
  addt: xi
;


**********************************************************************

* --------------------------------------------------------------------
* Desc:		Test equality of 2 metas
* Stack:	( meta2, meta1 -> meta2, meta1, flag )
* --------------------------------------------------------------------
NULLNAME metaEQUAL?
CODE
		GOSBVL	=SAVPTR
		C=DAT1	A
		CD1EX		C -> stack1
		D1=D1+	5
		A=DAT1	A
		R0=A.F	A	R0=#meta1
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A	
		CD1EX		
		D1=D1+	5	D1 -> stack_level of #meta2
		C=DAT1	A
		CD1EX
		D1=D1+	5
		A=DAT1	A
		CD1EX
		C=R0.F	A	C=#meta1, A=#meta2
		?A=C	A
		GOYES	metaequalinit
metanotequal	GOVLNG	=GPPushFLoop
metaequalinit	GOSBVL	=D0=DSKTOP	D0 -> stack level of #meta1
*****************************************************************
*		Compare two metas object
*		A=# stack level to look, 
*		D1 -> stack level of end (#length) of meta1
*		D0 -> stack level of end (#length) of meta2
*		Finish by pushing TRUE or FALSE and RPL Loop
*********************************************************
metaequalloop	C=C-1	A
		GONC	metaequalcmp
		GOVLNG	=GPPushTLoop
metaequalcmp	R2=C.F	A
		D0=D0+	5
		A=DAT0	A
		AD0EX
		R0=A.F	A
		D1=D1+	5
		C=DAT1	A
		CD1EX
		R1=C.F	A
		GOSBVL	=EQUAL?
		GOC	metanotequal
		C=R1.F	A
		CD1EX
		A=R0.F	A
		AD0EX
		C=R2.F	A
		GONC	metaequalloop
ENDCODE

* --------------------------------------------------------------------
* Desc:		Test presence of undef and inf in meta 2 and 1
* Stack:	( meta2, meta1 -> meta2, meta1, # )
*		# is the pos of inf if no undef
* Stack:	( meta2, meta1 -> undef_meta [RDROP] )
*		if undef detected
* --------------------------------------------------------------------
LABEL 2metaundef?
:: 
  2metaundef# 
  #0=case 2metainf#
  RDROP 2NDROPundef
;

* --------------------------------------------------------------------
* Desc:		Test presence of undef in meta2 and 1
* Stack:	( meta2, meta1 -> meta2, meta1, # )
*		# is the pos of undef
* --------------------------------------------------------------------
NULLNAME 2metaundef#
:: ' ROMPTR x? EQUALPOS2META SWAPDROP ;

* --------------------------------------------------------------------
* Desc:		Test presence of undef in meta2 and 1
* Stack:	( meta2, meta1 -> meta2&meta1 )
*		if no undef
* Stack:	( meta2, meta1 -> undef_meta [RDROP] )
*		if undef detected
* --------------------------------------------------------------------
LABEL 2metaundef!
:: 2metaundef# #0=case top& RDROP 2NDROPundef ;

LABEL 2NDROPundef
:: NDROP NDROPundef ;

LABEL NDROPundef
:: NDROP metaundef ;

* --------------------------------------------------------------------
* Desc:		Test presence of undef and infinities in meta
* Stack:	( meta -> meta, # )
*		# is the pos of inf if no undef
* Stack:	( meta -> undef_meta [RDROP] )
*		if undef detected
* --------------------------------------------------------------------
LABEL 1metaundef?
:: 
  1metaundef# #0=case 1metainf#
  RDROP NDROPundef 
;

* --------------------------------------------------------------------
* Desc:		Test presence of undef in meta
* Stack:	( meta -> meta, # )
*		# is the pos of undef
* --------------------------------------------------------------------
NULLNAME 1metaundef#
:: ' ROMPTR x? EQUALPOSMETA SWAPDROP ;

* --------------------------------------------------------------------
* Desc:		Return undef meta
* Stack:	( -> meta )
* --------------------------------------------------------------------
NULLNAME metaundef
:: ' ROMPTR x? TOTEMPOB ONE ;

* --------------------------------------------------------------------
* Desc:		Find position of infinity in meta 2 and 1
* Stack:	( meta2, meta1 -> meta2, meta1, # )
* Note:		meta of length>2 are considered as finite meta
* --------------------------------------------------------------------
NULLNAME 2metainf#
::
  1metainf# 				( meta2 meta1 # )
  DUP#0=csedrp 2metainf#2
  MINUSONE SWAP#- 			( meta2 meta1 # )
;

LABEL 2metainf#2
* meta1 is not inf, check for meta2
::					( meta2 meta1 )
  psh 1metainf#				( meta1 meta2 # )
  DUP#0=csedrp pshZERO
					( meta1 meta2 # )
  OBJ>R psh R>OBJ			( meta2 meta1 # )
;

LABEL pshZERO
::
  psh ZERO
;

* --------------------------------------------------------------------
* Desc:		Find position of infinity in meta
* Stack:	( meta -> meta, # )
* Note:		meta of length>2 are considered as finite meta
* --------------------------------------------------------------------
NULLNAME 1metainf#
:: 
   DUPTWO #>case ZERO 
   ' ROMPTR x\9F EQUALPOSMETA SWAPDROP 
;

* --------------------------------------------------------------------
* Desc:		Return infinty type
* Stack:	( meta -> # )
*		#= 1 for +infinity, 2 for -infinity, 0 for unsigned
* --------------------------------------------------------------------
NULLNAME metainftype
::				( meta )
  DUPTWO #<>case NDROPZERO	( -> #0 )
  DROPSWAPDROP EQ: xNEG 	( flag )
  case TWO ONE			( -> # )
;

* --------------------------------------------------------------------
* Desc:		Return unsigned infinty 
* Stack:	( -> meta )
* --------------------------------------------------------------------
NULLNAME unsignedinf
:: ' ROMPTR x\9F TOTEMPOB ONE ;

LABEL NDROPunsignedinf
:: NDROP unsignedinf ;

LABEL 2NDROPunsignedinf
:: NDROP NDROP unsignedinf ;

* --------------------------------------------------------------------
* Desc:		Return plus infinty 
* Stack:	( -> meta )
* --------------------------------------------------------------------
NULLNAME plusinf
:: ' ROMPTR x\9F TOTEMPOB ' xNEGNEG TWO ;

NULLNAME NDROPplusinf
:: NDROP plusinf ;

* --------------------------------------------------------------------
* Desc:		Return minus infinty 
* Stack:	( -> meta )
* --------------------------------------------------------------------
NULLNAME minusinf
:: ' ROMPTR x\9F TOTEMPOB ' xNEG TWO ;

NULLNAME NDROPminusinf
:: NDROP minusinf ;

* --------------------------------------------------------------------
* Desc:		Add 2 meta objects with trivial simplifications.
* Stack:	( Meta2 Meta1 --> Meta2+Meta1 )
* NOTE:		Checks for infinities then call metaadd
* --------------------------------------------------------------------
NULLNAME MetaAdd
NULLNAME xssSYM+
::
ASSEMBLE
	IF	f39
RPL
  NOT40case 48xssSYM+
ASSEMBLE
	ENDIF	f39
RPL
  2metaundef?
  DUP#0=csedrp metaadd

  ( * one or 2 arg are infinities * )
					( meta2, meta1, # )
  BINT80000 #<case NDROP 		( meta2 is inf, meta1 not )
  
  ( * meta1 is inf * )

  psh 1metainf# #0=case NDROP	( meta2 not inf and meta1 inf )

  ( * both are infinities * )
  metaadd2inf
;

LABEL metaadd2inf
::
  metainftype 			( meta, # )
  DUP#0=csedrp NDROPundef	( -> undef )
  OVER#2+UNROL metainftype	( #, # )
  OVER EQUAL NOTcasedrop metaundef
  #1=case plusinf minusinf
;

* --------------------------------------------------------------------
* Desc:		Substract 2 meta objects with trivial simplifications.
* Stack:	( Meta2 Meta1 --> Meta2-Meta1 )
* NOTE:		Checks for infinities then call metasub
* --------------------------------------------------------------------
NULLNAME MetaSub
NULLNAME xssSYM-
::
ASSEMBLE
	IF	f39
RPL
  NOT40case 48xssSYM-
ASSEMBLE
	ENDIF	f39
RPL
  2metaundef?
  DUP#0=csedrp metasub

  ( * one or 2 arg are infinities * )
					( meta2, meta1, # )
  BINT80000 #<case NDROP 		( meta2 is inf, meta1 not )
  
  ( * meta1 is inf * )

  psh 1metainf# 
  #0=case NDROPneginf		( meta2 not inf and meta1 inf )

  ( * both are infinities * )
  
  psh metaneginf metaadd2inf
;

* --------------------------------------------------------------------
* Desc:		Multiplies 2 meta objects with trivial simplifications.
* Stack:	( Meta2 Meta1 --> Meta2*Meta1 )
* NOTE:		Checks for infinities/0 then call metamult
* --------------------------------------------------------------------
NULLNAME MetaMul
NULLNAME xssSYM*
::
ASSEMBLE
	IF	f39
RPL
  NOT40case 48xssSYM*
ASSEMBLE
	ENDIF	f39
RPL
  2metaundef?
  DUP#0=csedrp metamult
  
  BINT80000 #<case metamultinfobj
  psh 1metainf# #0=case metamultinfobj
  
  ( * Both are infinities -> result is infinity, check sign * )
  
  metainftype 			( meta, # )
  DUP#0=csedrp NDROPunsignedinf
				( meta, # )
  OVER#2+UNROL metainftype 	( #, # )
  DUP#0= case2drop unsignedinf
  #+ #3= case minusinf plusinf
;

* level 2 is infinity, level 1 is not
LABEL metamultinfobj
::
  metasimp OVER j#0=case 2NDROPundef
  metaadjinfsign
;

LABEL metamultinfNEG
::
  pulldrop metaadjinfsign MetaNeg
;

LABEL metaadjinfsign
::					( meta-inf meta-obj )
  zsigne				( meta-inf z )
  num#1=case DROP
  j#0=case NDROPunsignedinf
  metaneginf
;

*::
*  M1stNEGcase metamultinfNEG
*  OVER TYPEZ? casedrop metaadjinf1
*  					( meta-inf meta-obj )
*  { x\pi xe } 3PICK Contains?		( meta-inf meta-obj list last flag )
*  UNROT2DROP				( meta-inf meta-obj flag )
*  case NDROP
*  2NDROPunsignedinf
*;

*LABEL metaadjinf1
*::			( meta, z )
*  ZIsNeg?		( meta, flag )
*  IT metaneginf
*;

* --------------------------------------------------------------------
* Desc:		Divide 2 meta objects with trivial simplifications.
* Stack:	( Meta2 Meta1 --> Meta2/Meta1 )
* NOTE:		Checks for infinities/0 then call metadiv
* --------------------------------------------------------------------
NULLNAME MetaDiv
NULLNAME xssSYM/
::
ASSEMBLE
	IF	f39
RPL
  NOT40case 48xssSYM/
ASSEMBLE
	ENDIF	f39
RPL
  2metaundef?
  DUP#0=csedrp metadiv
  BINT80000 #<case metadivinfobj

  ( * level1 is infinity, check level 2 * )

  psh 1metainf# 			( infinity-meta, meta2, # )
  #0= NOTcase 2NDROPundef		( -> infinity/infinity=undef )
  2NDROPZ0ONE			( -> obj/infinity=0 )
;

LABEL 2NDROPZ0ONE
:: NDROP NDROPZ0ONE ;

LABEL NDROPZ0ONE
:: NDROP Z0ONE	;

LABEL 2NDROPZ1ONE
:: NDROP NDROPZ1ONE ;

LABEL NDROPZ1ONE
:: NDROP Z1ONE ;

LABEL Z1ONE
:: Z1 ONE ;

LABEL 2NDROPZ-1ONE
:: NDROP NDROPZ-1ONE ;

LABEL NDROPZ-1ONE
:: NDROP Z-1ONE ;

LABEL Z-1ONE
:: Z-1 ONE ;

NULLNAME NDROPZ0
:: NDROP Z0 ;

NULLNAME NDROPZ1
:: NDROP Z1 ;

* level 2 is infinity, level 1 not
LABEL metadivinfobj
::
  OVER j#0=case 2NDROPunsignedinf
  metaadjinfsign
;

* --------------------------------------------------------------------
* Desc:		Power
* Stack:	( Meta2 Meta1 --> Meta2^Meta1 )
* NOTE:		Checks for infinities then call metapow
* --------------------------------------------------------------------
NULLNAME MetaPow
NULLNAME xssSYM^
::
ASSEMBLE
	IF	f39
RPL
  NOT40case 48xssSYM^
ASSEMBLE
	ENDIF	f39
RPL
  2metaundef?
  DUP#0=csedrp metapow
  metapowinf
;

LABEL metapowinf
::
  BINT80000 #>case obj^infinity

  ( * level1 is not infinite * )
  ( * infinity^obj=0 if obj<0, =infinity if obj>0, =undef if obj=0 * )

  OVER j#0=case 2NDROPundef
  DUP#1= NOTcase 2NDROPundef
				( meta obj #1 )
  DROP DUPTYPEZ?		( meta obj flag )
  NOTcasedrop NDROPundef	( -> metaundef )
  POSITIFext 			( meta z flag )
  NOTcasedrop NDROPZ0ONE
				( meta z )
  Z2 ZMod 			( meta z_mod_2 )
  j#0=case addtSQinf
				( meta )
;

LABEL obj^infinity
( * obj^infinity= exp[infinity*ln[obj]] * )  
::
  psh addtLN MetaMul addtEXP
;

* --------------------------------------------------------------------
* Desc:		Negate 1 meta object
* Stack:	( Meta --> Meta )
* NOTE:		Checks for infinities then call metaneg
* --------------------------------------------------------------------
NULLNAME MetaNeg
NULLNAME xSYMCHS
::
  1metaundef?
  #0=case metackneg

  ( * change sign of infinity meta * )
  metaneginf
;

LABEL NDROPneginf
::
  NDROP metaneginf
;

LABEL metaneginf
::
  DUPTWO #<> ?SEMI
  #1-SWAP EQ: xNEG 
  NOTcase metaneg
  addt: xNEGNEG
;

* --------------------------------------------------------------------
* Desc:		Checks only for meta finishing by xNEG
* Stack:	( meta -> meta )
* --------------------------------------------------------------------
NULLNAME metaneg
::
  AEQ1stcase xNEG pulldrop
  addtNEGonly
;

LABEL addtNEGonly
:: 
  addt: xNEG
;

* --------------------------------------------------------------------
* Desc:		Like metaneg but checks for meta=obj ONE
* Stack:	( meta -> meta )
* --------------------------------------------------------------------
NULLNAME metackneg
::
  DUP#1= NOTcase metaneg
  OVER IDNTLAM? casedrop metaneg
					( o1 #1 o1 )
  DUPTYPECOL? NOTcase2drop metaQNeg
  					( o1 #1 o1 )
  CARCOMP DUPTYPEZ? case2drop metaQNeg
					( o1 #1 o1.1 )
  EQ: x<< casedrop metaQNeg
  metaneg
;

LABEL metaQNeg
::			( o1 )
  QNeg cassyminner6	( meta )
;

* --------------------------------------------------------------------
* Desc:		Simplify a meta object
* Stack:	( Meta --> Meta )
* NOTE:		Non recursive rational simplification
* --------------------------------------------------------------------
NULLNAME metasimp
::
* set it again when prologs will be fixed
  DUP#1= casedrop metasimp1
  AEQ1stcase x= metasimp=

* we could check for a b x/ the only case that is not simplified
* occurs for b=i (I believe)
  metackundef
  dup sturmflag SysITE LISTOPext LISTOPSQRT LVARDext	( meta lvar )
  DUP 2LAMBIND			( meta )
  VXXL2NR			( Q )
  ZERO 1PUTLAM			( Q )
  VAL1M				( meta )
  ABND
;

LABEL metasimp=
::
  plDRPpZparg metasimp psh metasimp top& addt: x=
;

LABEL metasimp1
::
  CK&DISPATCH1
  zint metasimp1int
  seco metasimp1eval
  romptr metasimp1romptr
  ereal :: %%>% ONE ;
  ecmp :: C%%>C% ONE ;
  any ONE
;

LABEL metasimp1romptr
::
  DUPROMPTR@			( romptr obj TRUE | romptr FALSE )
  NOTcase metasimp1eval		( -> )
				( romptr obj )
  DUPTYPECOL? casedrop metasimp1eval
				( romptr obj )
  SWAPDROP SYMBINCOMP
;

LABEL metasimp1int
::
  EXACT? case valmznock
  Z>RONE
;

LABEL metasimp1eval
::
  DUPTYPEIRRQ? case ONE
  DUPTYPEGAUSSINT? case ONE
  DUP' ROMPTR x\9F EQUALcase ONE
  EVALtometa
;

LABEL EVALtometa
::
  fnConstantsEvalNoCK SYMBINCOMP
;

LABEL metackundef
::
  ' ROMPTR x? EQUALPOSMETA SWAPDROP #0=?SEMI	( -> meta )
  RDROP NDROPundef			( -> [RDROP] meta_undef )
;

LABEL approxmetasimp
:: SYMBN APPROXCOMPEVAL SYMBINCOMP ;
	
* --------------------------------------------------------------------
* Desc:		Test if metas are length 1 
* Stack:	( Meta1 Meta2 --> Meta1 Meta2 [2SKIP] )
* Stack:	( z1 ONE z2 ONE --> SYMBINCOMP[r1[zint1,zint2]] )
*		( f1/z1 ONE f2/z2 ONE -> SYMBINCOMP[r2[zint1,zint2]] )
* --------------------------------------------------------------------
LABEL meta2zint
CODE
		A=DAT1	A
		LC(5)	=ONE
		?A=C	A
		GOYES	+
metanot2zint	LA(5)	=SKIP2
		PC=(A)
+		D1=D1+	10
		A=DAT1	A
		D1=D1-	10
		?A#C	A
		GOYES	metanot2zint
		D1=D1+	15
		GOSBVL	=a=stk1
		D1=D1-	15
		LC(5)	=DOINT
		?A#C	A		* test if stk2 is a zint
		GOYES	stk2notzint
		LA(5)	stk2zint!
		PC=(A)
stk2notzint	D1=D1+	5		* stk2 not zint, test stk1
		RSTK=C
		GOSBVL	=a=stk1
		C=RSTK
		D1=D1-	5
		?A#C	A
		GOYES	meta2float?
stk1zint	LA(5)	stk1zint!
		PC=(A)
meta2float?	LA(5)	meta2float?!
		PC=(A)
ENDCODE

LABEL stk2zint!
* stk2 is a zint, look if stk1 is zint or float
::			( o2 ONE o1 ONE )
  DROPSWAPDROP		( o2 o1 )
  CK&DISPATCH1
  zint meta2zintzint
  real meta2zintfloat
  cmp  meta2zintfloat
  EXT  meta2zintfloat
  any  meta2zintback
;

LABEL stk1zint!
* stk1 is a zint, look if stk2 is float
::
  DROPSWAPDROP
  CK&DISPATCH1
  REALOB meta2floatzint
  CMPOB  meta2floatzint
  EXTOB  meta2floatzint
  any    meta2zintback
;

LABEL meta2zintzint
::
  'REVAL		( ob[z2,z1] )
  RDROP SYMBINCOMP	( meta )
; 

LABEL meta2zintfloat
::				( z float )
  'R DROP			( z float )
  SWAP Z>R SWAP			( z_crunched float )
  'REVAL RDROP SYMBINCOMP	( meta )
;

LABEL meta2floatzint
::				( float z )
  'R DROP Z>R			( float z_crunched )
  'REVAL RDROP SYMBINCOMP	( meta )
;

LABEL meta2zintback
::
  'R'R 2DROP
  ONESWAP ONE			( o1 one o2 one )
;

LABEL meta2float?!
::
  DROPSWAPDROP
  CK&DISPATCH1
  2REAL   meta2floatfloat
  2CMP    meta2floatfloat
  REALCMP meta2floatfloat
  CMPREAL meta2floatfloat
  2EXT    meta2floatfloat
  REALEXT meta2floatfloat
  EXTREAL meta2floatfloat
  any     meta2zintback
;

LABEL meta2floatfloat
::
  'R DROP			( float float )
  'REVAL RDROP ONE		( meta )
;

* --------------------------------------------------------------------
* Desc:		Test presence of pi in a meta
* Stack:	( Meta --> Meta, # )
* NOTE:		# is the last occurence of pi or 0
* --------------------------------------------------------------------
NULLNAME metapi?
:: 'xPI EQUALPOSMETA SWAPDROP ;


LABEL pshNDROP
::
  psh NDROP
;

* --------------------------------------------------------------------
* Desc:		Comparison of 2 meta
* Stack:	( Meta2 Meta1 --> Meta2 Meta1 # )
* Note:		# =0 if undef
*		# =1 if >
*		# =2 if <
*		# =3 if =
*		Assumes generic situation, e.g. X^2 > 0 in real mode
*		Look below STRICTmetaCOMPARE for a more careful comparison
* --------------------------------------------------------------------
NULLNAME metaCOMPARE
::				( m2 m1 )
  metaEQUAL? case THREE
  dup rotdup			( m1 m1 m2 m2 )
  4unroll MetaSub metasimp	( m2 m1 m1-m2 )
  1metainf#			( m2 m1 m1-m2 # )
  #0<> case negmetainftype	( -> m2 m1 # )
  				( m2 m1 m1-m2 )
  OVER j#0=case	:: 2DROP THREE ;		( -> m2 m1 #3 )
				( m2 m1 m1-m2 )
  metaCOMPARE1
;

LABEL negmetainftype
::
  MetaNeg metainftype
;

LABEL metaCOMPARE1
::
  zsigne			( m2 m1 z )
  num#0=case DROPZERO		( -> m2 m1 #0 )
				( m2 m1 z )
  j#1=case TWO
  ONE
;

* --------------------------------------------------------------------
* Desc:		Comparison of 2 meta
* Stack:	( Meta2 Meta1 --> Meta2 Meta1 # )
* Note:		# =0 if undef
*		# =1 if >
*		# =2 if <
*		# =3 if =
*		Unlike metaCOMPARE it does not assume generic situation
* --------------------------------------------------------------------
NULLNAME STRICTmetaCOMPARE
::				( m2 m1 )
  metaEQUAL? case THREE
  dup rotdup			( m1 m1 m2 m2 )
  4unroll MetaSub metasimp	( m2 m1 m1-m2 )
  1metainf#			( m2 m1 m1-m2 # )
  #0<> case negmetainftype		( -> m2 m1 # )
  				( m2 m1 m1-m2 )
  OVER num#0=case 3DROP3	
				( m2 m1 m1-m2 obj )
  DUPTYPEZ? 			( m2 m1 m1-m2 obj ?int )
  SWAP TYPEREAL?		( m2 m1 m1-m2 ?int ?reel )
  ORcase metaCOMPARE1		( -> )
				( m2 m1 m1-m2 )
  COMPLEX? case NDROPZERO
  AEQ1stcase xNEG STRICTCMPNEG
  AEQ1stcase xEXP NDROPTWO
  NDROPZERO
;

LABEL 3DROP3
::
  3DROP THREE
;

LABEL STRICTCMPNEG
::
  pulldrop
  AEQ1stcase xEXP NDROPONE
  NDROPZERO
;

**********************************************************************
* Desc:		Return last occurrence of ob in Meta.
* Stack:	( Meta ob --> Meta ob #pos )
* Note:		if a component of meta is a list/symb then search 
*		if ob is embedded in this component of meta
**********************************************************************
NULLNAME EQUALPOSMETA
CODE
		ST=0	s2META
posmetainit	GOSBVL	=SAVPTR
		C=DAT1	A	
		CD1EX		
		CD0EX			D1 -> obj, D0 -> stk
posmeta2nd	D0=D0+	5		D0 -> # of meta
		C=DAT0	A
		CD0EX
		D0=D0+	5
		A=DAT0	A
		CD0EX
		R4=A.F	A		R4=number of components of meta
posmetalp	A=R4.F	A
		?A=0	A
		GOYES	posnextmeta
		D0=D0+	5
		C=DAT0	A
		CD0EX
		RSTK=C
		GOSBVL	=CONTAINS?	carry clear=EQUAL
		C=RSTK
		CD0EX
		GONC	posmetaend
		A=R4.F	A
		A=A-1	A
		R4=A.F	A
		GOTO	posmetalp
posmetaend	A=R4.F	A
		?ST=0	s2META
		GOYES	posmetaRPL
		A=-A-1	A
posmetaRPL	GOVLNG	=PUSH#ALOOP
posnextmeta	?ST=0	s2META
		GOYES	posmetaRPL
		ST=0	s2META
		GOTO	posmeta2nd
ENDCODE
**********************************************************************
* Desc:		Return last occurrence of ob in Meta1 or in Meta2
* Stack:	( Meta2 Meta1 ob --> Meta2 Meta1 ob #pos )
* Note:		#pos is >0 if in meta2, is <0 if in meta1 (#pos=MINUSONE-#)
**********************************************************************
NULLNAME EQUALPOS2META
CODE
		ST=1	s2META
		GOTO	posmetainit
ENDCODE
**********************************************************************


**********************************************************************
* Desc:		symbolic sum with tests for 2zint
* Stack:	( Meta2 Meta1 --> meta )
* Note:		lam'sumvar bound to 'id/lam' and lam'sumexpr to 'expr'
*************************************************************
NULLNAME vgerxssSYMSUM
::				( meta-llimit meta-ulimit )
  ( * first try to find discrete sum in exact mode * )
  EXACTSUM? ?SEMI
  2metaundef?			( meta-llimit meta-ulimit # )
  #0=case metasigma
  2NDROPundef
;

LABEL EXACTSUM?
* meta-l meta-u -> meta true | meta-l meta-u false
::
* comment next line for exact sigma
*  COLA FALSE
  EXACT? NOTcaseFALSE
				( meta-l meta-u )
  psh ckmetasum psh ckmetasum	( meta-l meta-u )
  lam'sumexpr lam'sumvar 	( meta-l meta-u f x )
  CARCOMP SUM			( meta-l meta-u symb )
  QUOTExSIGMA			( meta-l meta-u symb sigma )
  Contains? SWAPDROP		( meta-l meta-u symb ? )
  casedrop FALSE		( -> meta-l meta-u ? )
				( meta-l meta-u symb )
  OBJ>R Z1ONE MetaAdd CKSYMBN	( meta-l u+1 )
  OVER#2+UNROLL CKSYMBN		( u+1 l )
  R>OBJ SWAPROT			( Sigma l u+1 )
  lam'sumvar CARCOMP 
  NOCAREEVAL PREVALext		( symb )
  SYMBINCOMP TRUE		( meta TRUE )
;

LABEL ckmetasum
* meta -> meta
* if meta is a real/cmplx meta ck that real/cmplx is an integer
::
  OVER CK&DISPATCH1
  real ckmetasum1
  cmp ckmetasum1
  any DROP
;

LABEL ckmetasum1
::			( real/cmp #1 real/cmp )
  OBJINT? 		( real/cmp #1 real/cmp/zint ? )
  NcaseSIZEERR
			( real/cmp #1 zint )
  ROTDROPSWAP		( -> zint #1 )
;

LABEL metasigma
::
  OVER TYPEREALZINT? NOTcase xssSYMSUM
  4PICK TYPEREALZINT? NOTcase xssSYMSUM
  DROPSWAPDROP			( llimit ulimit )
  SWAP OBJ2REAL SWAP OBJ2REAL	( llimit% ulimit% )
  LAM 'sumvar UNROT		( 'id/lam' llimit% ulimit% )
  LAM 'sumexpr 			( 'id/lam' llimit ulimit 'expr' )
  SYMNUMSUM			( obj )
  SYMBINCOMP			( meta )
;

**********************************************************************
* Desc:		If meta contains i switch to complex mode
* Stack:	( meta --> meta )
**********************************************************************
LABEL TESTi
::
  COMPLEX? ?SEMI
  ISMETACOMPLEX? case COMPLEXON
;

*LABEL ISMETACOMPLEX?
* meta -> meta flag
*::
*  OVER TYPECMP? caseTRUE
*  'xi EQUALPOSMETA		( meta i # )
*  SWAPDROP #0<>			( -> meta flag )
*;

**********************************************************************
* Desc:		Distribute /
* Stack:	( meta --> meta' true )
*		( meta --> meta FALSE )
* Note:		Returns FALSE if no distribution done
**********************************************************************
NULLNAME DISTRIB/
::
  plDRPpZparg			( meta-den meta-num )
  DISTRIB/1
;

LABEL DISTRIB/1
* meta-d meta-n -> meta ?
::				( meta-d meta-n )
  M1st+case DISTRIB+-/
  M1st-case DISTRIB+-/
  M1stNEGcase DISTRIBNEG/
  M1st*case DISTRIB*/
  M1st/case DISTRIB//
  pshtop&addt/FALSE
;

LABEL pshtop&addt/FALSE
::
  psh top&addt/ FALSE
;
	
LABEL DISTRIB*/
::
  DISTRIB*			( meta-d meta-n ? )
  NOTcase pshtop&addt/FALSE
				( meta-d meta-n )
  DISTRIB/1			( meta ? )
;

LABEL DISTRIB//
::				( meta-d meta-n )
  pulldrop larg			( meta-d meta-n1 meta-d1 )
  rot MetaMul psh		( meta-d*d1 meta-n1 )
  DISTRIB/1			( meta ? )
;

LABEL DISTRIB+-/
::				( meta-y meta-x )
  #1-SWAP OBJ>R
  pshzerpsharg			( meta-y meta1 meta-2 )
  rotdup			( meta1 meta2 meta-y meta-y )
  rotswap top&addt/ metasimp	( meta1 meta-y meta2/meta-y )
  unrot top&addt/ metasimp	( meta2/meta-y meta1/meta-y )
  top& R>OBJ SWAP#1+		( meta2/meta-y+-meta1/meta-y )
  TRUE
;

LABEL DISTRIBNEG/
::
  pulldrop psh
  top&addt/ addtNEGonly
  TRUE
;

**********************************************************************
* Desc:		M1stNEGcase including negative zint
* Stack:	( meta --> meta )
**********************************************************************
LABEL M1stZNEGcase
::
  CK&DISPATCH1
  #FF0 ZNEGcase
  any M1stNEGcase
;

LABEL ZNEGcase
::
  OVER DUPTYPEZ? NOTcasedrop SKIP
  ZIsNeg? NOTcase SKIP
  DROP QNeg ' xNEG TWO 'REVAL RDROP
;

**********************************************************************
* Desc:		Test if IM[meta]==0
* Stack:	( meta --> meta flag )
**********************************************************************
NULLNAME metareal?
::
  dup addtIM			( meta meta-IM )
  OVER Z0 EQUAL 		( meta meta-IM ? )
  case2drop TRUE
  NDROPFALSE
;

LABEL cassyminner6
::
  DUPTYPESYMB? ITE INNERCOMP ONE
;
