INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE Modular arithmetic
**********************************************************************
* EXPORT:
*
* ModExpa	( Z N/D -> N/D mod Z )
*
* ModAdd	( Qa Qb Zn --> Q )
*
* ModSub	( Qa Qb Zn --> Q )
*
* ModMul	( Qa Qb Zn --> Q )
*
* ModDiv2	( Qa Qb Zn --> Q )
*
* ModInv	(    Zb Zn --> Z )
*
* ModGcd	( Qa Qb Zn --> Q )
*
* ModLGCD	( { Q1 ... Qn } Zn -> Q )
*
* ModLOPD	( { Q1 ... Qn } Q Zn -> { Q1/Q ... Qn/Q } )
*
**********************************************

**********************************************************************
* Desc:
* Stack:	( Zn Fraction -> Fraction modulo Zn )
**********************************************************************
NULLNAME ModExpa
::
  FXNDext				( Zn N D )
  DUPTYPEZ? case ModExpaZ
  SWAP3PICK				( Zn D N Zn )
  QMODSYMext SWAP			( Zn N' D )
  num#1=case DROPSWAPDROP
  3PICK QMODSYMext			( Zn N' D' )
  2DUP 5PICK ModGcd			( Zn N' D' g )
  DUPUNROT 5PICK ModDiv2 DROP		( Zn N' g D'' )
  4UNROLLROT ModDiv2 DROPSWAP		( N'' D'' )
  QDiv
;

LABEL ModExpaZ
::
  num#1=case ModExpa1
  3PICK ModInv				( Zn N D^-1 )
  ROT ModMul				( Zn N*D^-1 )
;

LABEL ModExpa1
::
  DROPSWAP QMODSYMext			( -> N )
;

**********************************************************************
* Desc:		Modular addition.	Z = Q1+Q2 (mod Zn)
* Stack:	( Q1 Q2 Zn --> Z )
**********************************************************************
NULLNAME ModAdd
::
  UNROT RADDext			( Zn Q1+Q2 )
  ModExpa
;

**********************************************************************
* Desc:		Modular substraction.	Z = Q1-Q2 (mod Zn)
* Stack:	( Q1 Q2 Zn --> Z )
**********************************************************************
NULLNAME ModSub
::
  UNROT RSUBext ModExpa
;

**********************************************************************
* Desc:		Modular multiplication. Z = Q1*Q2 (mod Zn)
* Stack:	( Q1 Q2 Zn --> Z )	
**********************************************************************
NULLNAME ModMul
::
  UNROT RMULText ModExpa
;

**********************************************************************
* Desc:		Modular division. Z = Z1/Z2 (mod Zn)
* Stack:	( Z1 Z2 Zn --> Z )
**********************************************************************
NULLNAME ModDiv
::
  DUPUNROT		( Z1 Zn Z2 Zn )
  ModInv		( Z1 Zn Z' )
  SWAP ModMul		( Z'' )
;

**********************************************************************
* Desc:		Modular division.
* Stack:	( Q1 Q2 Zn --> Q1 div Q2, Q1 mod Q2 mod Zn )
* Note:	 if Q1 and Q2 are integers, Q1 mod Q2 mod Zn is always 0
**********************************************************************
NULLNAME ModDiv2
::
  UNROT				( Zn, Q1, Q2 )
  DUPTYPEZ? case
   :: 3PICK ModInv		( Z, Q1, Z2.Inv )
      RMULText
      SWAP QMODSYMext Z0
   ;

  PSEUDOPREP		( Zn, Q1, Q2, a )
  UNROT QDivRem		( Zn, a, a* Q1 div Q2, a* Q1 mod Q2 )
  ROT PPZ		( Zn, a*Q1 div Q2, a*Q1 mod Q2, a.poly, a.z )
  5PICK ModInv		( Zn, a*Q1 div Q2, a*Q1 mod Q2, a.poly, a.z.inv )
  DupQIsZero? case NONINTERR

  ( reduction de a.poly mod Z )

  SWAP 5PICK QMODSYMext
  TRIMOBJext SWAP	( Zn a*Q1_div_Q2 a*Q1_mod_Q2 a.poly a.z.inv )
  4ROLLOVER RMULText	( Zn a*Q1_mod_Q2 a.poly a.z.inv a.z.inv*a*Q1_div_Q2 )
  5PICK QMODSYMext
  3PICK QDiv		( Zn a*Q1_mod_Q2 a.poly a.z.inv a.z.inv*a*Q1_div_Q2/a.poly )
  4ROLLROT RMULText	( Zn a.poly a.z.inv*a*Q1_div_Q2 a.z.inv*a*Q1_mod_Q2 )
  4ROLL QMODSYMext	( a.poly a.z.inv*a*Q1_div_Q2 a.z.inv*a*Q1_mod_Q2 )
  ROT QDiv
;

**********************************************************************
* Desc:		Modular inversion. Z' = INV(Z) (mod Zn)
* Stack:	( Z Zn --> Z' )
* Notes:	NONINTERR if GCD[Z,Zn] <> 1 or if Z = 0
*		(Otherwise the results would be unpredictable.)
**********************************************************************
NULLNAME ModInv
::
					( z, n  )
  Z1 ROT RDIVext FXNDext		( n, 1, z )
  3PICK NOSTEPEVAL IEGCDext		( n, 1, d, u, v / zu+nv=d )
  DROPROT				( n, d, u, 1 )
  RMULText SWAP				( n, u, d )
  RDIVext DUPTYPESYMB?			( n, u, z non inversible )
  case2drop NONINTERR			( --> 0 si non inversible )
					( n, u )
  OVER QMODSYMext
  DUPTYPEZ? NOTcase SWAPDROP
  DupZIsNeg? ITE RADDext SWAPDROP	( u=u+n si necessaire )
;

**********************************************************************
* Name:		ModGcd
* Desc:		modular GCD
* Stack:	( Q1 Q2 Zn --> Q' )
**********************************************************************
NULLNAME ModGcd
::
					( A B Zn )
  UNROTSWAP				( Zn B A )
  NOTLISTcase ModGcdTrivial
  SWAP NOTLISTcase ModGcdTrivial
  [BEGIN]
     DUPUNROT 4PICK			( Zn B A B Zn )
     ModDiv2 FXNDext DROPSWAPDROP	( Zn B C )
     DUPTYPELIST?			( Zn B C flag )
  NOT_UNTIL
					( Zn B C )
  Z0 EQUAL NOTcase2drop Z1
					( Zn G )

* en principe ModGcd est appele pour des poly a 1 var donc simpl. a la fin

  OVER ModLGCD				( Zn G g )
  ROT ModLOPD				( G simplifie )
  NEWTRIMext
;

LABEL ModGcdTrivial
:: 					( Zn B A )
  num#0=case DROPSWAPDROP		( -> B )
  3DROP Z1 				( -> gcd )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME ModLGCD
::						( C, Z )
  OVER TYPELIST? NOTcasedrop Z1
  1LAMBIND DUPINCOMP
  DUP#1= casedrop ABND
  ONE_DO (DO)
     1GETLAM ModGcd
  LOOP
  ABND
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME ModLOPD
::						( A, b, Z )
  3PICK TYPELIST? NOTcasedrop ModDiv2
  1LAMBIND 
  Z1 EQUALcasedro ABND
  SWAPINCOMP get1				( a1, ..., an, #n, b )
  SWAPDUP
  ZERO_DO (DO)					( a1, ..., an, b, #n )
     get1
     3PICK 1GETLAM ModDiv2
     DROP UNROT
  LOOP
  ABND SWAPDROP {}N
* changed back from {}POLY
;

**********************************************************************

**********************************************************************
* Desc:		restes chinois, determine c tel que c=a mod M et c=b mod N
* Stack:	( a, M, b, N -> c, ppcm[M,N] )
* Note:		runstream must contains extended GCD and eucl DIV prog
**********************************************************************
LABEL CHINREMext
::
  ROTSWAP 2DUP 'REVAL			( a, b, M, N, g, m, n )
  DROP 6ROLL 6ROLL OVER RSUBext		( M, N, g, m, a, b-a )
  4PICK 'REVAL
  Z0 EQUALNOTcase NONINTERR
					( M, N, g, m, a, [b-a]/g )
  ROT RMULText				( M, N, g, a, [b-a]/g*m )
  5PICK RMULText RADDext		( M, N, g, c )
  4UNROLL RDIVext RMULText		( c, ppmc[M,N] )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME MODULOMODext
::					( vecteur )
  LAM MODULO
  SWAPINCOMP get1 SWAPDUP
  ZERO_DO (DO)				( v1, ..., vn, modulo, #n )
     get1 3PICK QMODSYMext
     UNROT
  LOOP
  SWAPDROP []N
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME MODULOMAText
::
  INNERDUP #1-SWAP
  ZERO_DO 
     get1 MODULOMODext
  SWAPLOOP
  #1+ []N
;
