        STITLE  Replace, conversions

INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE erext.h
INCLUDE keydefs.h
INCLUDE global.s

**********************************************************************
* EXPORT:
*
* HYP2EXPext    ( symb -> symb )
*               Convert all hyperbolic functions to exp and ln
*               Convert XROOT and ^ to exp and ln
*
* EXPLNext      ( symb -> symb ) *
*               Convert all transcendental functions to exp and ln
*
* TRIGext       ( symb -> symb )
*               Applies sin^2+cos^2=1 to simplify trig. expressions
*               if prefersinflag is set tries to keep only sin, else only cos
*
* TAN2SCext     ( symb -> symb )
*               tan to sin/cos
*
* SIN2TCext     ( symb -> symb )
*               sin to cos*tan
*
* COS2ext       ( symb -> symb' )
*               symb'=sqrt[1-sin^2[symb]]
*
* SIN2ext       ( symb -> symb' )
*               symb'=sqrt[1-cos^2[symb]]
*
* ATAN2Sext     ( symb -> symb )
*               atan[x] -> asin[x/sqrt[1+x^2]]
*
* ATAN2ASIN     ( symb -> symb' )
*               x -> asin[x/sqrt[1+x^2]]
*
* ASIN2Text     ( symb -> symb )
*               asin[x] -> atan[x/sqrt[1-x^2]]
*
* ASIN2ATAN     ( symb -> symb' )
*               x -> atan[x/sqrt[1-x^2]]
*
* ASIN2Cext     ( symb -> symb )
*               asin[x] -> pi/2-acos[x]
*
* ASIN2ACOS     ( symb -> symb' )
*               x -> pi/2-acos[x]
*
* ACOS2ext      ( symb -> symb )
*               acos[x] -> pi/2-asin[x]
*
* ACOS2ASIN     ( symb -> symb' )
*               x -> pi/2-asin[x]
*
* ATAN2LNext    ( symb -> symb ) *
*               * x -> i/2*ln[ [x+i]/[i-x] ]
*
* TAN2SC2ext    ( symb -> symb )
*               tan to sin/cos of the double angle
*
* TAN2SC2       ( symb -> symb' )
*               x -> (1-cos(2x))/sin(2x)
*
* TAN2CS2       ( symb -> symb' )
*               x -> sin(2x)/(1+cos(2x))
*
* SIN2EXPext    ( symb -> symb' ) *
*               x -> (e^ix-1/e^ix)/(2i)
*
* COS2EXPext    ( symb -> symb' ) *
*               x -> (e^ix+1/e^ix)/2
*
* SINH2EXPext   ( symb -> symb' ) *
*               x -> (e^x-1/e^x)/2
*
* COSH2EXPext   ( symb -> symb' ) *
*               x -> (e^x+1/e^x)/2
*
* TANH2EXPext   ( symb -> symb' ) *
*               x -> (e^2x-1)/(e^2x+1)
*
* ASINH2LNext   ( symb -> symb' ) *
*               x -> ln(x+sqrt(x^2+1))
*
* ACOSH2LNext   ( symb -> symb' ) *
*               x -> ln(x+sqrt(x^2-1))
*
* ATANH2LNext   ( symb -> symb' ) *
*               x -> ln((x-1)/(x+1))/2
*
* XROOT2ext     ( symb1, symb2 -> symb' )
*               returns exp[ln[symb1]/symb2]
*
* LN2ext        ( symb -> symb' )
*               If symb contains x, returns -1/inv[neg[ln[x]]] else ln[x]
*               Used by SERIES
*
* SINCOSext     ( symb -> symb )
*               convert exp and ln to exp*sin+cos and ln+i*atan
*
* VAR=LIST      ( idnt, {} -> {} )
*               replace all elements of the initial list by idnt=element
*
* IDNTEXEC      ( symb, idnt -> symb' )
*               Tries to find idnt such that symb=0
*               Return a solution as an equality 'idnt=...' in symb'
*
* LISTEXEC      ( obj, {} -> obj )
*               ( obj, {} -> {} )
*               The list should be of the form 
*               { 'X=1' 'Y=2' ... } in the first case
*               or { 'X=1' 'X=2' } in the second case
*               In the first case X is replaced by 1, Y by 2, etc.
*               In the second case obj is evaluated with X=1, X=2 successively
*
* EQUATION?     ( obj -> obj, flag )
*               Returns TRUE if obj is a symbolic finishing by x=
*
* USERFCN?      ( obj -> obj, flag )
*               Returns TRUE if obj is a symbolic finishing by xFCNAPPLY
*
* SYMBEXEC      ( obj, symb -> obj )
*               If symb is an equation, execute the corresponding change
*               of variables in obj, otherwise tries to find symb s.t.
*               obj is 0
*               Note that change of variable works for change of user fcns
*
* MEVALext      ( obj, list2, list1 -> obj' )
*               Replace all occurences of an element of list2 by the
*               corresponfing element of list1 in obj
*               Looks in obj from outer to inner expressions
*               list2 and list1 may contain secondaries
*
* REPLINText    ( inf, sup, func, symb, int, #4, symb -> symb' )
*               Execute a change of variables in an integral
*               The new variable of integration is set to the first idnt
*               found in symb
*               For example 'int(...,Y)' 'Y=X^2*A' will use X as new
*               integration variable
*
* REPLACE2BY1   ( symb, idnt, a -> symb' )
*               Evaluation of a symbolic replacing an idnt by a value
*               (For example evaluation of F(X) for X=1/2)
*               Does not solve for undetermination (use LIMIT for this)
*               idnt can be a pointer to an operator, in that case a 
*               must be the corresponding action on meta
**********************************************************************


**********************************************************************
* Desc:         Convert trigonometric functions to TAN of the half angle
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME HALFTAN
::
  { xCOS xSIN xTAN }
  { cos2tan/2 sin2tan/2 tan2tan/2 }
  REPLACE2BY1
;

NULLNAME COS2TAN/2
:: SYMBINCOMP cos2tan/2 CKSYMBN ;

NULLNAME cos2tan/2
:: addtTAN/2 1-x^2/1+x^2 ;

NULLNAME 1-x^2/1+x^2
:: dup meta1-sq psh metasq+1 MetaDiv ;

NULLNAME SIN2TAN/2
:: SYMBINCOMP sin2tan/2 CKSYMBN ;

NULLNAME sin2tan/2
:: addtTAN/2 2x/1+x^2 ;

NULLNAME 2x/1+x^2
:: dup meta2* psh metasq+1 MetaDiv ;

NULLNAME TAN2TAN/2
:: SYMBINCOMP tan2tan/2 CKSYMBN ;

NULLNAME tan2tan/2
:: addtTAN/2 dup meta2* psh meta1-sq MetaDiv ;

NULLNAME addtTAN/2
:: meta/2 addtTAN ;

**********************************************************************
* Desc:         Convert sin and cos to tan of the same angle
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME TRIGTAN
::
  ( * step0: replace cos^even power and sin^even power * )
  ' x^ ' TRIGTAN^ REPLACE2BY1           ( symb )

  ( * first replace sin by cos*tan * )
  ' xSIN ' cos*tan REPLACE2BY1          ( symb )

  ( * now extract arg of cos functions * )
  RLVARext                              ( symb lvar )
  DUPNULL{}? casedrop SIMPLIFY          ( -> symb )
                                        ( symb lvar )
  DUP >R LENCOMP ZEROSWAP               ( symb meta-arg-cos )
  ZERO_DO (DO)
    RSWAP 'R RSWAP                      ( symb meta-arg-cos var )
    SYMBINCOMP                          ( symb meta-arg-cos meta-var )
    OVER EQ: xCOS                       ( symb meta-arg-cos meta-var )
    ITE pldrpSYMBNSWP1+ NDROP
  LOOP
  {}N                                   ( symb l_arg-cos )
  DUPNULL{}? casedrop SIMPLIFY
                                        ( symb l_arg-cos )
  SWAP' xCOS ' cos2tan                  ( l_arg-cos symb cos cos2t)
  REPLACE2BY1                           ( l_arg-cos symb )
  SIMPLIFY                              ( l_arg-cos symb )
  ( * replace sqrt[1+tan^2] by 1/cos * )
  OVER QOBJext SQRT1+TAN^2              ( l_arg-cos symb l_sqrt[1+tan^2] )
  ROT QOBJext COS*1+TAN^2
  MEVALext
;

NULLNAME COS2TAN
::
  SYMBINCOMP cos2tan SYMBN
;

NULLNAME cos2tan
::
  addtTAN metasq+1 addtSQRTonly addtINV
;

NULLNAME SIN2TAN
::
  SYMBINCOMP sin2tan SYMBN
;

NULLNAME sin2tan
::
  addtTAN dup metasq+1 addtSQRTonly MetaDiv
;

**********************************************************************
* Desc:         Applies sin^2+cos^2=1 to simplify trigonometric
*               expressions. If prefersinflag is set, tries to keep
*               only sin, else only cos.
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME TRIGext
::
  ' x^ ' TRIGPOW REPLACE2BY1
  RLVARext

  DUPNULL{}? caseDROP
                                        ( symb, lvar )
  DUP >R LENCOMP ZEROSWAP               ( symb meta-arg )
  ZERO_DO (DO)
    RSWAP 'R RSWAP                      ( symb meta-arg var )
    SYMBINCOMP                          ( symb meta-arg meta-var )
    OVER preferSINCOS EQ                ( symb meta-arg meta-var )
    ITE pldrpSYMBNSWP1+ NDROP
  LOOP
  {}N                                   ( symb, l_sin_arg )
  DUPNULL{}? casedrop SIMPLIFY

* on remplace les sin par sqrt[1-cos^] et on simplifie

  DUP
  prefersinflag SysITE
  ::
    QOBJext xCOSext
    SWAP QOBJext COS2ext 
    ROT                                 ( lvarsin, lvarcos, symb )
    ' xCOS ' sqrt1-sin^2
  ;
  ::
    QOBJext xSINext
    SWAP QOBJext SIN2ext 
    ROT                                 ( lvarsin, lvarcos, symb )
    ' xSIN ' sqrt1-cos^2
  ;

  REPLACE2BY1
        
  sqrtflag TestSysFlag OBJ>R sqrtflag ClrSysFlag
  LLVARDext
  ( set sqrtflag: we do not want to mix up square root arguments )
  sqrtflag SetSysFlag 
  VXXLext
  sqrtflag R>OBJ ITE SetSysFlag ClrSysFlag
        
  SWAP VAL2ext 
* on remplace les sqrt[1-cos^] restant par des sin

  UNROT SLVARext SWAP MEVALext          ( symb )
;

LABEL preferSINCOS
:: prefersinflag SysITE 'COS 'SIN ;

LABEL 'SIN
:: ' xSIN ;

LABEL 'COS
:: ' xCOS ;

**********************************************************************
* Desc:         Convert hyperbolic functions to exp and ln.
*               Convert XROOT and ^ to exp and ln.
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME HYP2EXPext
::
  HYP2EXPLST1
  HYP2EXPLST2
  REPLACE2BY1
;

LABEL HYP2EXPLST1
{ 
  x^ xEXP xTAN xSINH xCOSH xTANH 
  xASINH xACOSH xATANH xXROOT 
  xLNP1 xEXPM xLOG xALOG xSIGN
}

LABEL HYP2EXPLST2
{ 
  meta^meta addtEXP 2*1-cos/sin sinh2exp cosh2exp tanh2exp 
  asinh2ln acosh2ln atanh2ln xroot2expln  
  addtLNP1 addtEXPM addtLN/LN10 addt*LN10EXP addtSIGNonly
}

**********************************************************************
* Desc:         Convert all transcendental functions to exp and ln.
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME EXPLNext
::
  EXPLNLST1
  EXPLNLST2             ( obj, liste de secondaries, remplacement )
  REPLACE2BY1
  { xEXP x\v/ xXROOT } { addtEXPonly sqrt2lnexp  xroot2expln } REPLACE2BY1
;

LABEL EXPLNLST1
{ 
  xLN xEXP xSIN xCOS xATAN 
  xTAN xASIN xACOS x^ 
  xSINH xCOSH xTANH xASINH xACOSH xATANH
  xLNP1 xEXPM xLOG xALOG
}

LABEL EXPLNLST2
{ 
  addtLN addtEXPonly sin2exp cos2exp atan2ln
  tan2exp asin2ln acos2ln meta^meta 
  sinh2exp cosh2exp tanh2exp asinh2ln acosh2ln atanh2ln    
  addtLNP1 addtEXPM addtLN/LN10 addt*LN10EXP
}

**********************************************************************
* Desc:         Convert sqrt, ^ and XROOT to EXP/LN
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME SERIESEXPLN
::
  SERIESEXPLN1
  SERIESEXPLN2          ( obj, liste de secondaries, remplacement )
  MEVALext
;

LABEL SERIESEXPLN1
{ 
  x^ x\v/ xXROOT xEXP xLN
}

LABEL SERIESEXPLN2
{ 
  EXPAND^ SERIESSQRT2LNEXP XROOT2ext TSIMPxEXP TSIMPxLN
}

**********************************************************************
* Desc:         Convert LNP1 to LN
* Stack:        symb -> symb
**********************************************************************
NULLNAME LNP12LN
::
  SYMBINCOMP addtLNP1 CKSYMBN
;

**********************************************************************
* Desc:         Convert LOG to LN
* Stack:        symb -> symb
**********************************************************************
NULLNAME LOG2LN
::
  SYMBINCOMP addtLOG CKSYMBN
;

**********************************************************************
* Desc:         Convert ALOG to EXP
* Stack:        symb -> symb
**********************************************************************
NULLNAME ALOG2EXP
::
  SYMBINCOMP addtALOG CKSYMBN
;

**********************************************************************
* Desc:         Convert EXPM to EXP
* Stack:        meta -> meta
**********************************************************************
NULLNAME EXPM2EXP
::
  SYMBINCOMP addtEXPM CKSYMBN
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME SQRT2LNEXP
::
  SYMBINCOMP sqrt2lnexp SYMBN
;

NULLNAME sqrt2lnexp
::
  addtLN meta/2 addtEXP1
;

LABEL SERIESSQRT2LNEXP
::
  SYMBINCOMP seriessqrt SYMBN
;

LABEL seriessqrt
* meta -> meta
* if meta does not contain vx, return sqrtonly[meta] else convert to exp[]
:: 
  RCLVX EQUALPOSMETA SWAPDROP           ( meta # )
  #0=case addtSQRTonly
  sqrt2lnexp
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME TAN2EXP
::
  SYMBINCOMP tan2exp SYMBN
;

NULLNAME tan2exp
::
  meta2* metai* addtEXP         ( exp[2i*obj]-meta )
  m-1&m+1
  metai* top& addt/
;

**********************************************************************
* Desc:
* Stack:        x -> i*ln(x+sqrt(x^2-1))+pi/2
**********************************************************************
NULLNAME ASIN2LN
::      
  SYMBINCOMP asin2ln  SYMBN
;

NULLNAME asin2ln
::
  lnxsqrtsq-1 metai* pisur2 INNERCOMP metaadd
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME ACOS2LN
::
  SYMBINCOMP acos2ln  SYMBN
;

NULLNAME acos2ln
::
  lnxsqrtsq-1 addti addt/
;

LABEL lnxsqrtsq-1
::
  dup metasq-1 addtSQRT metaadd addtLN 
;


**********************************************************************
* Desc:         Convert tan to sin/cos
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME TAN2SCext
::
  ' xTAN ' sin/cos REPLACE2BY1
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME TAN2SC
:: 
  SYMBINCOMP sin/cos  SYMBN 
;

NULLNAME sin/cos
::
  dup addtSIN psh addtCOS top& addt/
;

**********************************************************************
* Desc:         Convert sin to cos*tan
* Stack:        ( symb -_> symb' )
**********************************************************************
NULLNAME SIN2TCext
::
  ' xSIN ' cos*tan REPLACE2BY1
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME SIN2TC
:: 
  SYMBINCOMP cos*tan  SYMBN 
;

NULLNAME cos*tan
::
  dup addtCOS psh addtTAN top&addt*
;

**********************************************************************
* Desc:         Convert symb to sqrt(1-sin^2(symb))
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME COS2ext
::
  SYMBINCOMP sqrt1-sin^2 SYMBN
;

NULLNAME sqrt1-sin^2
::
  addtSIN meta1-sq addtSQRT
;

**********************************************************************
* Desc:         Convert symb to sqrt(1-cos^2(symb))
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME SIN2ext
::
  SYMBINCOMP sqrt1-cos^2 SYMBN
;

NULLNAME sqrt1-cos^2
::
  addtCOS meta1-sq addtSQRT
;

**********************************************************************
* Desc:         Convert atan to asin using asin(x)=atan(x/sqrt(1-x^2))
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME ATAN2Sext
::
  ' xATAN ' atan2asin REPLACE2BY1
;

**********************************************************************
* Desc:         x -> asin[x/sqrt[1+x^2]]
* Stack:        
**********************************************************************
NULLNAME ATAN2ASIN
::
  SYMBINCOMP atan2asin SYMBN
;

NULLNAME atan2asin
::
  dup metasq+1 addtSQRT top& addt/ addtASIN
;

**********************************************************************
* Desc:         Convert asin to atan using asin(x)=atan(x/sqrt(1-x^2))
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME ASIN2Text
::
  ' xASIN ' asin2atan REPLACE2BY1
;

**********************************************************************
* Desc:         x -> atan[x/sqrt[1-x^2]]
* Stack:
**********************************************************************
NULLNAME ASIN2ATAN
::
  SYMBINCOMP asin2atan  SYMBN
;

NULLNAME asin2atan
::
  dup meta1-sq addtSQRT top& addt/ addtATAN
;

**********************************************************************
* Desc:         Convert asin to acos using asin(x)=pi/2-acos(x)
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME ASIN2Cext
::
  ' xASIN  ' pi/2-acos  REPLACE2BY1
;

**********************************************************************
* Desc:         x -> pi/2-acos[x]
* Stack:
**********************************************************************
NULLNAME ASIN2ACOS
::
  SYMBINCOMP pi/2-acos CKSYMBN
;

NULLNAME pi/2-acos
:: addtACOS pi/2-meta ;

NULLNAME pi/2-meta
:: metapi/2 psh MetaSub ;

**********************************************************************
* Desc:         Convert acos to asin using acos(x)=pi/2-asin(x)
* Stack:
**********************************************************************
NULLNAME ACOS2Sext
::
  ' xACOS ' pi/2-asin REPLACE2BY1
;

NULLNAME pi/2-asin
::
  addtASIN pi/2-meta
;

**********************************************************************
* Desc:         x -> pi/2-asin[x]
* Stack:
**********************************************************************
NULLNAME ACOS2ASIN
::
  SYMBINCOMP pi/2-asin  SYMBN
;

**********************************************************************
* Desc:         x -> i/2*ln[ [x+i]/[i-x] ]
* Stack:        (symb --> symb' )
**********************************************************************
NULLNAME ATAN2LNext
::
  SYMBINCOMP atan2ln
  SYMBN
;

NULLNAME atan2ln
::
  ' xi OVER#2+UNROLL 
  #1+ dup addt+ psh addt- metadiv addtLN
  metai meta/2 psh metamult
;

**********************************************************************
* Desc:         Convert tan to sin/cos of the double angle.
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME TAN2SC2ext
::
  ' xTAN
  ' 2*sin/1+cos
  REPLACE2BY1
;

**********************************************************************
* Desc:         x -> (1-cos(2x))/sin(2x)
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME TAN2SC2
::
  SYMBINCOMP 2*1-cos/sin SYMBN
;

NULLNAME 2*1-cos/sin
::
  meta2* metasimp
  dup 
  addtCOS 1&meta addt- 
  psh addtSIN top&
  addt/ 
;

**********************************************************************
* Desc:         x -> sin(2x)/(1+cos(2x))
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME TAN2CS2
::
  SYMBINCOMP 2*sin/1+cos SYMBN
;

NULLNAME 2*sin/1+cos
::
  meta2* metasimp
  dup
  addtSIN 
  psh 
  addtCOS 1&meta addt+ top&
  addt/ 
;

**********************************************************************
* Desc:         x -> (e^ix-1/e^ix)/2
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME SIN2EXPext
::      
  SYMBINCOMP
  sin2exp
  SYMBN
;

NULLNAME sin2exp
::
  tmeta1/meta                           ( meta-exp meta-1/exp )
  metasub/2i
;

**********************************************************************
* Desc:         x -> (e^ix+1/e^ix)/2
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME COS2EXPext
::
  SYMBINCOMP
  cos2exp
  SYMBN
;

NULLNAME cos2exp
::
  tmeta1/meta                           ( meta-exp, meta-1/exp )
  metaadd/2
;

**********************************************************************
* Desc:         x -> (e^x-1/e^x)/2
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME SINH2EXPext
::
  SYMBINCOMP sinh2exp SYMBN
;

NULLNAME sinh2exp
::
  hmeta1/meta metasub meta/2 
;

**********************************************************************
* Desc:         x -> (e^x+1/e^x)/2
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME COSH2EXPext
::
  SYMBINCOMP cosh2exp SYMBN
;

NULLNAME cosh2exp
::
  hmeta1/meta metaadd meta/2
;

* --------------------------------------------------------------------
* Desc:         Trigo utilitie
* Stack:        ( meta -> meta-exp-i*meta, meta-1-exp-i*meta )
* --------------------------------------------------------------------
LABEL tmeta1/meta
::
  metai* addtEXP 
  meta1/meta
;

LABEL tmeta-meta
::
  metai* dup addtEXPNO-CK psh metaneg addtEXPNO-CK
;

LABEL metasub/2i
::
  metasub addt2 addti addt* addt/ 
;

LABEL metaadd/2
:: 
  metaadd meta/2 
;

* --------------------------------------------------------------------
* Desc:         Hyperbolic utilitie
* Stack:        ( obj -> meta-exp-obj, meta-1-exp-obj )
* --------------------------------------------------------------------
LABEL hmeta1/meta
::
  addtEXP meta1/meta 
;

**********************************************************************
* Desc:         x -> (e^2x-1)/(e^2x+1)
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME TANH2EXPext
::
  SYMBINCOMP tanh2exp
  SYMBN
;

NULLNAME tanh2exp
::
  meta2* addtEXP
  m-1&m+1 metadiv
;

**********************************************************************
* Desc:         x -> ln(x+sqrt(x^2+1))
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME ASINH2LNext
::
  SYMBINCOMP asinh2ln SYMBN 
;

NULLNAME asinh2ln
::
  dup metasq+1 asch2ln 
;

**********************************************************************
* Desc:         x -> ln(x+sqrt(x^2-1))
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME ACOSH2LNext
::
  SYMBINCOMP acosh2ln SYMBN
;

NULLNAME acosh2ln
::
  dup metasq-1 asch2ln 
;

LABEL asch2ln
::
  addtSQRT metaadd addtLN
;

**********************************************************************
* Desc:         x -> ln((1+x)/(1-x))/2
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME ATANH2LNext
::
  SYMBINCOMP atanh2ln SYMBN
;

NULLNAME atanh2ln
::
  addtATANH*2 meta/2
;


**********************************************************************
* Desc:         exp(ln(symb1)/symb2)
* Stack:        ( symb1 symb2 --> symb' )
**********************************************************************
NULLNAME XROOT2ext
::
  2SYMBINCOMP xroot2expln SYMBN 
;

NULLNAME xroot2expln
::
  addtLN psh MetaDiv addtEXPonly
;

**********************************************************************
* Desc:         If symb contains x, returns -1/inv(-ln(x)), else ln(x)
* Stack:        ( symb --> symb' )
* Note:         Used by SERIES.
**********************************************************************
NULLNAME LN2ext
:: 
  CKLN symcomp
  DUP RCLVX EQUALPOSCOMP #0=?SEMI
  DUPDUP LENCOMP NTHCOMPDROP EQ: xLN NOT?SEMI
  Z-1 SWAPINCOMP ' xNEG ' xINV ' x/ 4ROLL #4+ SYMBN
;

**********************************************************************
* Desc:         convert exp and ln to exp*sin+cos and ln+i*atan
* Stack:        ( symb --> symb' )
**********************************************************************
NULLNAME SINCOSext
::
  { xEXP xLN } 
  { EXP2SINCOS LN2ATAN } 
  MEVALext
;

LABEL EXP2SINCOS
::
  SYMBINCOMP exp2sincos SYMBN
;

**********************************************************************
* Desc:         return EXP of meta as EXP[RE]*[COS+i*SIN]
* Stack:        ( meta -> meta )
**********************************************************************
NULLNAME exp2sincos
::
  exp2expre_im                          ( exp-re meta-im )
  dup addtCOS psh addtSIN metai*        ( exp-re cos-im i*sin-im )
  metaadd metamult                      ( exp-re*[cos+i*sin] )
;

LABEL exp2expre_im
::
  dup addtRE addtEXP psh addtIM         ( exp-re meta-im )
;

NULLNAME metai*
:: metai psh metamult ;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LN2ATAN
::      
  LVARext VXXLext LNATANext
  ZEROSWAP ROT VAL2ext
;

**********************************************************************
* Desc:         replace all elements of the initial list by idnt=element
* Stack:        ( idnt {} --> {}' )
**********************************************************************
NULLNAME VAR=LIST
::
  NOTLISTcase x=ext
  DUPNULL{}? case SWAPDROP
  INNERCOMP get1 SWAPDUP
  ZERO_DO (DO)                          ( sol1, ..., soln, lvar, #n )
     get1 3PICKSWAP EXPR= UNROT
  LOOP
  SWAPDROP {}N
;

**********************************************************************
* Desc:         Tries to find idnt such that symb=0.
*               Return a solution as an equality 'idnt=..' in symb'
* Stack:        ( symb idnt --> symb' )
**********************************************************************
NULLNAME IDNTEXEC
::
  ISOL1 EXPR=
;

NULLNAME SYMISOL
::
ASSEMBLE
        IF      f39
RPL
  NOTEQWcase 48SYMISOL
ASSEMBLE
        ENDIF   f39
RPL
  SYMQFORM
;

NULLNAME SYMQFORM
:: 
ASSEMBLE
        IF      f39
RPL
  NOTEQWcase 48SYMQFORM
ASSEMBLE
        ENDIF   f39
RPL
  CASFLAGEVAL: 
  :: ISOL2ext 
     DUPNULL{}? case ISOLERR
     VAR=LIST DUPLENCOMP #1= IT CARCOMP
     ;
;

**********************************************************************
* Desc:         The list should be of the form 
*               { 'X=1' 'Y=2' ... } in the first case
*               or { 'X=1' 'X=2' } in the second case
*               In the first case X is replaced by 1, Y by 2, etc.
*               In the second case obj is evaluated with X=1, X=2 successively
* Stack:        ( ob {} --> ob' )
*               ( ob {} --> {}' )
**********************************************************************
NULLNAME LISTEXEC
::
  DUPLENCOMP #1=case :: CARCOMP xSUBST ;
  DUP CARCOMP                           ( obj, liste, obj1 )
  DUPTYPEMATRIX? casedrop LISTEXEC1     ( note: should not occur anymore )
                                        ( obj, liste, obj1 )
  EQUATION?                             ( obj, liste, obj1, ?equation )
  NOTcasedrop SIMP|                     ( *remplacements multiples* )
                                        ( obj, liste, obj1 )
  CARCOMP OVER TWO NTHCOMPDROP          ( obj, liste, obj1.1, obj2 )
  NOTEQUATIONERR
  CARCOMP EQUALcase LISTEXEC1           ( obj, liste )
  INNERCOMP NULL{} DUPROT               ( obj eq1...eqn liste1 liste2 #n )
  ZERO_DO 
     ROT NOTEQUATIONERR                 ( obj eq1...eqn-1 liste1 liste2 eqn )
     EXPR> 2DROP                        ( obj eq1...eqn-1 liste1 liste2 o1 o2 )
     ROTSWAP >HCOMP                     ( obj eq1...eqn-1 liste1 o1 liste2 )
     UNROT >HCOMP
  SWAPLOOP
  MEVALext
;

LABEL NOTEQUATIONERR
::
  EQUATION? ?SEMI
  BINT43 ERABLEERROR
;

**********************************************************************
* Desc:         execute EXEC sur obj de chaque element de liste
* Stack:        ( liste, objet -> liste )
**********************************************************************
NULLNAME LISTEXEC1
::
  INNERCOMP get1 SWAPDUP
  ZERO_DO (DO)                          ( l1, ..., ln, o, #n )
     get1 3PICKSWAP xSUBST UNROT
  LOOP
  SWAPDROP {}N
;

**********************************************************************
* Desc:         exec prog on [each elem] of obj
* Stack:        ( obj prog -> liste )
**********************************************************************
NULLNAME SECOEXEC
::                                      ( obj prg|symb )
  OVER CKSECOEXEC                       ( obj prg|symb )
  MATOPSYMB                             ( obj prg )
  CASFLAGEVAL:
  :: 
     ' ' ' ' ' SWAP ' '                 ( obj prg ' ' SWAP ' )
     5ROLL THREE ' ::N                  ( obj ' ' SWAP ' prg THREE ::N )
     ' ALGEVAL EIGHT ::N        ( obj :: ' ' SWAP ' prg THREE ::N ALGEVAL ; )
     rpnQOBJext
     ;
;

LABEL CKSECOEXEC
* obj ->
::
  CK&DISPATCH1
  seco SETSIZEERR
  list CKSECOEXECCOMP
  any DROP                      ( -> )
;

LABEL CKSECOEXECCOMP
::                              ( composite )
  INNERCOMP                     ( o1..on #n )
  DUP#0=csDROP                  ( -> )
  ZERO_DO (DO)
    CKSECOEXEC                  ( o1..oi )
  LOOP
;

**********************************************************************
* Desc:         Returns TRUE if obj is a symbolic finishing by x=
* Stack:        ( ob --> ob flag )
**********************************************************************
NULLNAME EQUATION?
::
  DUPTYPESYMB? NOTcaseFALSE
  DUPDUP LENCOMP NTHCOMPDROP 
ASSEMBLE
        IF      f39
RPL
  DUPEQ: x= case DROPTRUE
  EQ: x==
ASSEMBLE
        ELSE    f39
RPL
  EQ: x=
ASSEMBLE
        ENDIF   f39
RPL
;

**********************************************************************
* Desc:         Returns TRUE if obj is a symbolic finishing by xFCNAPPLY
* Stack:        ( ob --> ob flag )
**********************************************************************
NULLNAME USERFCN?
::
  DUPTYPESYMB? NOTcaseFALSE
  DUPDUP LENCOMP NTHCOMPDROP EQ: xFCNAPPLY
;

**********************************************************************
* Desc:         If symb is an equation, execute the corresponding
*               change of variables in ob, otherwise tries to find symb
*               so that ob is zero. Note that change of variable works
*               for change of user functions.
* Stack:        ( ob symb --> ob' )
**********************************************************************
NULLNAME SYMBEXEC
:: 
  EQUATION? NOTcase IDNTEXEC
  EXPR> 2DROP                           ( obj, o1, o2 ) 

  ( on devrait aussi pouvoir ecrire obj=liste pour renvoyer une liste )
  ( si o1 est une fnct, on recherche les derivees dans obj )
  ( sinon on isole si possible 1 idnt a gauche et on cree 2 listes )

  OVER
  DUPTYPESYMB? NOTcasedrop MEVALext
                                        ( obj, o1, o2, o1 )
  USERFCN?                              ( obj, o1, o2, o1, ?fnct utilisateur )
  ITE SYMBEXECUFCN SYMBEXECNFCN         ( obj, o1, o2, l1, l2 )
  ROT >HCOMP UNROTSWAP                  ( obj, l2, l1, o1 )
  >HCOMP SWAP MEVALext
;

LABEL SYMBEXECUFCN
::      
  EXPR>                         ( obj o1 o2 arg1...argk %arg idnt )
  SWAP COERCE DUPUNROT
  #2+UNROLL {}N                 ( obj o1 o2 idnt { arg1...argk } )
  5PICK LIDNText                ( obj o1 o2 idnt { arg1...argk } { obids } )

  ( teste si obj2 est une fonction utilisateur )

  4PICK LVARext
  SWAPDROP CARCOMP              ( obj o1 o2 idnt {arg1...argk} { obids } var2 )
  USERFCN? ITE
  ::                    ( cherche la liste des arg de fnct2 )
    EXPR> SWAP COERCE 
    DUPUNROT #2+UNROLL {}N
  ;
  :: DROP 3PICK3PICK ;

                        ( obj o1 o2 idnt {args} {obids} newid {newargs} )
  NULL{} DUP 5ROLL
  [BEGIN]
     DUPNULL{}?
  NOT_WHILE
                        ( ...idnt {args} newid {newargs} l1 l2 {objidnt} )

  ( pour chaque idnt de obj, on teste si "idnt" de obj )
  ( contient "idnt" )

    EXECUFCNLOOP

                        ( ...idnt {args} newid {newargs} l1 l2 {objidnt} )

  [REPEAT]

  DROP 6UNROLL 6UNROLL 4DROP    ( obj, o1, o2, l1, l2 )
; 

*

LABEL EXECUFCNLOOP
::      
  DUP CDRCOMP SWAP CARCOMP
  ( ...idnt {args} newid {newargs} l1 l2 {objidnt} objidnt )
  DUP ID>DERext
  ( ...o1 o2 idnt {args} newidnt {newargs} l1 l2, {objidnt} objidnt liste [stripped_idnt] )
  TEN PICK EQUAL NOTcase2DROP
  ( ...o1 o2 idnt {args} newidnt {newargs} l1 l2 {objidnt} objidnt liste )
  DUPNULL{}? case2DROP

  ( creation de newidnt[newargs] )

* 6PICK INNERDUP #8+ PICK ONE SYMBN SWAP xFCNAPPLY
  TEN PICK

*       fin de modification

  ( ...idnt {args} newidnt {newargs} l1 l2 {objidnt} objidnt liste o2 )

  SWAPDUP LENCOMP
  ZERO_DO 
     ( ...id {args} newid {newargs} l1 l2 {objid} objid result list )
     DUP CDRCOMP SWAP CARCOMP
     ( ...id {args} newid {newargs} l1 l2 {objid} objid result list, nderiv )
     8PICK OVER NTHELCOMP NcaseSIZEERR
     ( ...id {args} newid {newargs} l1 l2 {objid} objid result list nderiv newarg_no )
     4ROLLSWAP 3PICK 
     :: 
       SWAP DUPTYPESYMB? 
       IT :: ' TYPEIDNTLAM? Find1stTrue NOTcase %1 ;
       idntlamcase :: SWAPDROP DERIVext ;
       ( symb, #, obj )
       DROP #>Z DO>STR SWAPINCOMP get1 5ROLL 
       CARCOMP ID>$ &$ CHR_d >H$ $>ID ONE SYMBN 4UNROLL SYMBN
     ;
     ( ...id {args} newid {newargs} l1 l2 {objid} objid list nderiv result )
     TEN PICK ROT NTHELCOMP NcaseSIZEERR
     ( ...id {args} newid {newargs} l1 l2 {objid} objid list result arg_no )
     DUPTYPESYMB? ITE
     :: DUP ' TYPEIDNTLAM? Find1stTrue 
        NOTcaseDROP DERIVext RDIVext
     ;
     DROP
     ( ...id {args} newid {newargs} l1 l2 {objid} objid list result )
  SWAPLOOP

  ( ...id {args} newid {newargs} l1 l2 {objid} objid result list )

  DROP 4ROLLSWAP >TCOMP UNROT

  ( ...id {args} newid {newargs} l1 l2 {objid} objid )

  4ROLL 7PICK INNERDUP #3+ ROLL ONE SYMBN SWAP xFCNAPPLY >TCOMP UNROT

  ( ...id {args} newid {newargs} l1 l2 {objid} )
;

LABEL SYMBEXECNFCN
::      
                                                ( obj, o1, o2, o1 )
  4PICK DUP' x\.S_ALGEBRAIC EQUALPOSCOMP        ( obj, o1, o2, o1, obj, # )
  DUP#0<> IT
  ::  #1- NTHCOMPDROP CARCOMP                   ( obj, o1, o2, o1, idnt )
      2DUP EQUALPOSCOMP
  ;
                                        ( obj o1 o2 o1 [idnt # | obj #0] )
  #0=ITE
     :: DROP' TYPEIDNTLAM? Find1stTrue ;
     SWAPDROPTRUE
                                        ( obj o1 o2 [indt TRUE | FALSE] )
  NOTcase :: RDROP MEVALext ;
                                        ( obj, o1, o2, idnt )
  UNROT 2DUP x-ext                      ( obj, idnt, o1, o2, o1-o2 )
  4ROLL                                 ( obj, o1, o2, o1-o2, idnt )
  PushVStack
  ERRSET 
    ISOL1                               ( obj, o1, o2, idnt, valeur )
  ERRTRAP 
    :: PopVStack 2DROP NULL{} DUP RDROP ( -> obj o1 o2 ) ;
  DropVStack                            ( obj, o1, o2, idnt, valeur )
  ONE{}N SWAP ONE{}N SWAP               ( obj, o1, o2, l1, l2 )
;


**********************************************************************
* Desc:         Replace all occurrances of an element of list2 by
*               the corresponding element of list1 in ob. Looks in
*               ob from outer to inner expressions. list2 and list1
*               may contain secondaries.
*
* Stack:        ( ob list2 list1 --> ob' )
* Note:         if vxxlflag is set SIGN var are leaved unchanged
**********************************************************************
NULLNAME MEVALext
::                                      ( obj X, a )
  DUPNULL{}? case2DROP
  2DUP EQUALcase 2DROP
  MAKE2LIST                             ( obj l1 l2 )
  ( * add a marker for optimization for + - * / ^ * )
  OVER ' LAST_BINARY? Find1stTrue       ( obj l1 l2 [ ob true | false ] ) 
  ITE DROPFALSE TRUE UNROT              ( obj ? l1 l2 )
* {{ ancien nouveau }} 
DEFINE ancien 2GETLAM
DEFINE ancien! 2PUTLAM
DEFINE nouveau 1GETLAM
DEFINE nouveau! 1PUTLAM
  3LAMBIND                              ( obj )
  STURMFLAGEVAL: mevalobj!
  ABND
;

LABEL STURMFLAGEVAL:
::
  'R sturmflag TestSysFlag case EVAL
  sturmflag SetSysFlag 
  EVAL
  sturmflag ClrSysFlag
;

LABEL LAST_BINARY?
* obj -> ?
* true if obj is a symbolic ending with a binary op
::
  DUPTYPESYMB? NOTcase DROPFALSE
  LASTCOMP enumargs SWAPDROP #2=
;

LABEL MAKE2LIST
::                                      ( o1 o2 )
  DUPTYPELIST? ?SKIP ONE{}N             ( o1 l2 )
  DUPLENCOMP                            ( o1 l2 #l2 )
  ROT DUPTYPELIST? ?SKIP ONE{}N         ( l2 #l2 l1 )
  DUP4UNROLL                            ( l1 l2 #l2 l1 )
  LENCOMP #= NOT?DIMERR                 ( l1 l2 )
;

LABEL mevalobj!
::
  CK&DISPATCH1
  list    mevallst!
  TAGGED  mevaltag!
  arry    mevalarry!
  any     mevalrepl!
;

LABEL mevallst!
::
  DUPNULL{}? ?SEMI
  mevallstmeta
  {}N
;

LABEL mevalarry!
::
  DUPTYPEMATRIX? NOT?SEMI
  mevallstmeta []N
;

LABEL mevallstmeta
::
  INNERDUP #1-SWAP
  ZERO_DO 
     get1 mevalobj!
  SWAPLOOP
  #1+ 
;

LABEL mevaltag!
::
  'EvalNoCK: xOBJ>
  SWAP mevalobj!
  SWAP >TAG
;

* ( obj -> obj' remplace les elements de ancien par ceux de nouveau )
LABEL mevalrepl!
::      
  nouveau OVER ancien ListPos           ( obj, l2, pos )
  ROTOVER #0<>                          ( l2, pos, symb, ?pos<>0 )
  casedrop NTHCOMPDROP                  ( l2, pos, symb )
  UNROT2DROP                            ( symb )
  DUPTYPESYMB? NOT?SEMI
  INNERCOMP mevalmeta!
  SYMBN
;

LABEL mevalmeta!
::
  3GETLAM NOTcase mevalmetanotest
  M1st+case mevalbin!
  M1st*case mevalbin!
  M1st-case mevalbin!
  M1st/case mevalbin!
  M1st^case meval^!
  M1stNEGcase mevalmono!
  mevalmetanotest
;

LABEL mevalmetanotest
::
  vxxlflag TestSysFlag NOTcase mevalgen!
  AEQ1stcase xSIGN NOP
  mevalgen!
;

LABEL meval^!
::                                      ( o1, ..., on-1, x^, #n )
  3PICK OBJINT?
  NOTcasedrop mevalgen!
                                        ( o1, ..., on-1, x^, #n, zn-1 )
  4ROLLDROP UNROT
  mevalbin!
;

LABEL mevalbin!
::
  #1-SWAP OBJ>R
  pshzerpsharg mevalmeta!
  psh mevalmeta!
  top& 
  R>OBJ
  SWAP#1+
;


LABEL mevalmono!
::
  #1-SWAP OBJ>R mevalmeta!
  R>OBJ
  SWAP#1+
;

* ( meta -> meta )
LABEL mevalgen!
::
  nouveau 3PICK ancien ListPos          ( meta, nouveau, # )
  DUP#0= case2drop mevalgen2!
  NTHCOMPDROP                           ( meta, newop )
  OVER #1=case
  :: UNROT2DROP SYMBINCOMP ;
  OBJ>R                                 ( meta )
  SYMBN EXPR>                           ( objets, %nbre, oper )
  SWAP COERCEDUP                        ( objets, oper, #, # )
  ZERO_DO (DO)
     get1 mevalobj!
     UNROT
  LOOP
  SWAP 2DROP                            ( objets )
  R>OBJ                                 ( objets oper )
  DUP REP2BY1TRIGOP ListPos             ( objets oper # )
  #0=case EVALSYMBINCOMP                ( -> meta )
                                        ( arg unary_oper )
  SWAP SYMBINCOMP get1 SWAP#1+          ( meta )
;

LABEL EVALSYMBINCOMP
::
  EVAL SYMBINCOMP
;

LABEL mevalgen2!
::
  CKSYMBN
  nouveau OVER ancien ListPos           ( symb, l2, pos )
  DUP#0= case2drop mevalgennot!
  ROTDROP NTHCOMPDROP SYMBINCOMP
;

LABEL mevalgennot!
::                                      ( obj )
  DUPTYPESYMB? NOTcase ONE

* necessaire car OBJ> sur +infini renvoie infini x+ au lieu de infini xNEGNEG
  PLUSINFext OVER EQUALcase SYMBINCOMP

  EXPR>                                 ( objets, %nbre, oper )
  SWAPDUP %0= casedrop ONE
  COERCEDUP                             ( objets, oper, #, # )
  ZERO_DO (DO)
     get1 mevalobj! UNROT
  LOOP
  SWAP                                  ( objets, #nombre, oper )
  DUPEQ: x\.S case2drop MEVAL.s
  DUPEQ:
ASSEMBLE
        IF      f39
RPL
  xint
ASSEMBLE
        ELSE    f39
RPL
  xINT
ASSEMBLE
        ENDIF   f39
RPL

  case2drop MEVALINT
                                        ( objets, #nombre, oper )
  ( * changed IDNTLAM? to DUPTYPECOL? for romptr e.g. psi[n] n=1 subst * )
  DUPTYPECOL? case mevalnidnt!
  symcomp                               ( objets, #nombre, oper )
  2LAMBIND                              ( objets )
  2GETLAM {}N 2GETLAM SWAP              ( # { objets } )
  >R ZEROSWAP                           ( meta # )
  ZERO_DO (DO)
    RSWAP 'R RSWAP                      ( meta obj )
    SYMBINCOMP top&                     ( meta )
  LOOP
  1GETLAM 2GETLAM ABND ' xFCNAPPLY      ( meta oper # fcnapply )
  4ROLL #3+                             ( meta )
;

LABEL mevalnidnt!
::                                      ( objets #nombre oper )
  SWAP #1=case mevalunary
                                        ( objets oper )
  CODE
                A=DAT1  A
                LC(5)   #80000
                ?A<C    A
                GOYES   mevalnock
                LA(5)   =EVAL
                PC=(A)
mevalnock       LA(5)   =EvalNoCK
                PC=(A)
  ENDCODE
  SYMBINCOMP
;

LABEL mevalunary
::                              ( objet oper )
  OBJ>R SYMBINCOMP R>OBJ SWAP#1+
;

**********************************************************************
* Desc:         Execute a change of variables in an integral.
*               The new variable of integration is set to the first
*               idnt found in symb. For example 'int(..,Y)' 'Y=X^2*A'
*               will use X as new integration variable.
* Stack:        ( inf sup func symb --> symb' )
* Note:         if there was actually no change of var in the integral
*               then the new var of integr remains the old one!
**********************************************************************
LABEL MEVAL.s
::                                      ( inf sup func t^2  )

  DUPTYPESYMB?                          ( inf sup func t^2 flag )
  NOTcase :: ' x\.S_ALGEBRAIC FIVE ;
                                        ( inf sup func t^2 )
  ( * first find new integration var * )
  NEWINTVAR                     ( inf sup func[t] t^2 t sqrt[oldvar] oldvar )

  ( * then replace inf and sup by new bounds * )
  2DUP 9ROLL                    ( sup func t^2 t sqrt[old] old sqrt old inf )
  MEVALext 7UNROLL              ( newinf sup func t^2 t sqrt[oldvar] oldvar )
  6ROLL MEVALext                ( newinf func t^2 t newsup )
  4UNROLL                       ( newinf newsup func t^2 t )
  NEWINTFUNC                    ( newinf newsup func* t )
  MEVALMAKEs                    ( meta )
;

LABEL MEVALMAKEs
::                              ( newinf newsup func* t )
  OBJ>R OBJ>R                   ( newinf newsup )
  2SYMBINCOMP top&              ( meta-newinf-newsup )
  R>OBJ R>OBJ ROT #2+           ( meta-newinf-newsup-func-t )
  addt: x\.S_ALGEBRAIC          ( meta )
;

LABEL NEWINTVAR
::                              ( t^2 )
  DUP' TYPEIDNTLAM? Find1stTrue ( t^2 [ t true | false ] )
  NOTcase :: BINT48 ERABLEERROR ;
                                ( t^2 t )
  OVER 'ID210 x-ext SWAP        ( t^2 't^2-Bidon' t )
  ISOL1                         ( t^2 t sqrt[B] )
  'ID210                        ( t^2 t sqrt[B] B )
;


LABEL NEWINTFUNC
::                              ( func t^2 t )
  DUPUNROT DIRECTDERIV          ( func t d[t^2]/dt )
  ROT x*ext SIMP1ext symcomp    ( t func*  )
  SWAP ONESYMBN                 ( func* t )
;

LABEL MEVALINT
::                              ( f[t] t^2 y -> meta_int[f[x],x]|x=y )
  UNROT                         ( y f[t] t^2 )
  IDNTLAM?                      ( y f[t] t^2 flag )
  case xf[t]t2int               ( -> meta )
  NEWINTVAR                     ( y f[t] t^2 t sqrt[B] B )
  6ROLL MEVALext                ( f[t] t^2 t newy )
  4UNROLL NEWINTFUNC            ( newy func* t )
  xf[t]t2int                    ( meta )
;

**********************************************************************
* Desc:         Evaluation of a symbolic
* Stack:        ( symb list1 list2 --> symb' )
* Note:         the list format are 
*               list1={ idnt/lam1.. idnt/lamn }
*               list2={value1..valuen }
*               idnt/lam1..n in list1 are first NOT evaluated
*               then are replaced by value1..valuen
**********************************************************************
NULLNAME CASNUMEVAL
::                                      ( obj list1 list2 )
  ( * swap so that list1 -> lam1 and list2 -> lam2 * )
  OVER 3LAMBIND                         ( obj )
  polarflag TestSysFlag OBJ>R           
  polarflag ClrSysFlag 
  approxflag TestSysFlag OBJ>R
  approxflag SetSysFlag
  fnConstants TestSysFlag OBJ>R
  fnConstants SetSysFlag
  nogcdflag TestSysFlag OBJ>R
  nogcdflag SetSysFlag
  switchflag TestSysFlag OBJ>R
  switchflag SetSysFlag
  symcomp                               ( symb )
  PushVStack ONE KEEP                   ( symb )
  ERRSET
  ::
    INCASCOMPEVAL                       ( meta' )
    replacelam1bylam2
    SYMBN                               ( symb' )
    REPLACEZBYR                         ( symb' with zint replaced by % )
  ;
  ERRTRAP
  ::
   DEPTH NDROPFALSE
   ;        ( res=symb'/false )
  PopVStackAbove                        ( .... symb symb'/false )
  DUPEQ: FALSE ITE_DROP SWAPDROP        ( ... symb' )
  switchflag R>OBJ RestoreFlag
  nogcdflag R>OBJ RestoreFlag
  fnConstants R>OBJ RestoreFlag
  approxflag R>OBJ RestoreFlag
  polarflag R>OBJ RestoreFlag
  ABND
;

LABEL replacelam1bylam2
* meta -> meta
* replace obj of meta that are in the list LAM1 by the object at the
* same position in the list LAM2
::
  SYMBN DUP >R LENCOMP ZEROSWAP         ( # )
  ZERO_DO (DO)
    RSWAP 'R RSWAP                      ( meta obj )
    DUP 1GETLAM ListPos                 ( meta obj # )
    numevalrepl                         ( meta )
  LOOP
;

LABEL numevalrepl
::                                      ( meta obj #pos )
  DUP#0=csedrp SWAP#1+                  ( -> meta )
                                        ( meta obj #pos )
  SWAPDROP                              ( meta #pos )
  2GETLAMSWAP                           ( meta lam2 # )
  NTHELCOMP NcaseSIZEERR                                
                                        ( meta obj )
  SYMBINCOMP top&                       ( meta )
;


**********************************************************************
* Desc:         Evaluation of a symbolic
* Stack:        ( symb --> symb' )
* Algorithm:    Push symb in R, recalls element by element
*               if obj is idnt replace by a-meta
*               if obj is a seco we look for 2 args secondaries, and for int
*               if obj is a bint we have a user function
*               otherwise we have a unary operator
*               we just have to add it to the result meta by SWAP#1+
* Note:         complex are returned in cartesian mode
*               embedded symbolics are protected from evaluation
**********************************************************************
NULLNAME CASCOMPEVAL
::                                      ( symb )
*  ( * first look for a unit object * )*

*  symcomp DUP' TYPEEXT? Find1stTrue    ( symb flag )
*  casedrop COMPEVAL                    ( -> eval symb with 48 algo )

*  ( * no unit object inside * )
                                        ( obj )
  NULL{} DUPDUP 3LAMBIND                ( obj )
  symcomp                               ( symb )
  RunSafeFlagsNoError POLINCASCOMPEVAL  ( meta' )
  METASYMBN                             ( symb' )
  ABND
  DTYPEMATRIX? NOT?SEMI
  COLA BESTMATRIXTYPE			( Convert it into an arry of possible )
;

LABEL POLINCASCOMPEVAL
::
  polarflag TestSysFlag 
  NOTcase INCASCOMPEVAL                 ( -> meta )
  polarflag ClrSysFlag 
  INCASCOMPEVAL
  polarflag SetSysFlag
;

LABEL RestoreFlag
( #flag flag -> )
* restore flag state according to flag
::
  case SetSysFlag ClrSysFlag
;

LABEL INCASCOMPEVAL
::
  ( * insure coherent approx /exact mode * )
  nogcdflag approxflag SysITE SetSysFlag ClrSysFlag

  DUPDUP >R LENCOMP                     ( symb # )
  DUPUNROT NTHCOMPDROP                  ( # last )
  ' SETSIZEERR ROT                              ( last mark # )
  ZERO_DO (DO)
    RSWAP 'R RSWAP                      ( .., obj )
    ( * test if obj is a user RPL program * )
    DUPTYPECOL? ITE INCASCOMPRPL? FALSE ( .., obj ? )
    IT INCASINVALID COMPEVALLP          ( .., meta )
    LOOP                                ( last mark meta )
  get1 EQ: SETSIZEERR NOTcase INTERNALERR
  get1                                  ( meta last )
  OVER #1= casedrop metasimp            ( -> meta )
  ( * no rational normal form if called by CASNUMEVAL * )
  1GETLAM NULLCOMP? NOTcaseDROP         ( -> meta )
  ( * no rational normal form if last operator is not rational * )
  RATIONAL_OP? NOT?SEMI                 ( -> meta )
                                        ( meta )
  metasimp
;

LABEL INCASCOMPRPL?
:: DUPONE NTHCOMPDROP EQ: x<< ;

LABEL INCASINVALID
:: DROP ' ID Invalid_object TOTEMPOB ; 

LABEL COMPEVALLP
::                                      ( meta, obj )
  CK&DISPATCH1
  zint COMP2BY1ZINT
  seco COMP2BY1SECO
  idnt COMPEVALIDNT
  lam  COMPEVALIDNT
  symb COMPEVALSYMB
  any  COMPEVALANY                      ( -> meta )
;

LABEL COMP2BY1ZINT
::
* zint must be converted to real in "numeric" mode, otherwise inf loop
  SYMB? NOTcase Z>RONE
  approxflag TestSysFlag NOTcase ONE
  Z>RONE
;

LABEL Z>RONE
::
  Z>R ONE
;

LABEL COMPEVALSYMB
::                                      ( ... symb )
  EXACT? ?SKIP REPLACEZBYR              ( ... symb )
  DUP LASTCOMP EQ: x=                   ( symb ? )
  NOTcase COMPEVALANY
  SYMBINCOMP                            ( ...meta1..metan )
  ( * mark vxxlflag for CKSYMBS to switch to standard eval loop * )
  vxxlflag ClrSysFlag
  CKSYMBS
;

LABEL COMPEVALANY
::
  DUP' ROMPTR x\9F EQUALcase EVALtometa
  DTYPEFLASHPTR? NOTcase ONE
  COMPEVALFLASH
;

LABEL COMPEVALFLASH
* eval flashptr, (don't know how much args!)
::
  META2SYMB                     ( last mark meta_s flashptr/seco )
  EVAL                          ( last mark symb_s )
  SYMB2META                     ( last mark meta_s )
  ( * test that all meta_s are valid algebraic metas * )
  ( * if not go back to normal evaluation * )
  COLA CKSYMBS
;

LABEL META2SYMB
::
  ( * build symbolics with all meta_s to the mark SETSIZEERR * )
  ONE                                   ( last mark meta_s meta_symb&f )
  [BEGIN]
    DUP#2+PICK EQ: SETSIZEERR
  NOT_WHILE
    BUILDSYMBLP
  [REPEAT]
                                        ( last mark meta_symb&f )
  DROP                                  ( last mark symb1..symbn ob )
;

LABEL BUILDSYMBLP
( meta_s meta_symb&f -> meta_s' meta_symb&f )
::                                      ( meta_s meta_symb&f )
  psh METASYMBN OVER#2+UNROLL #1+
;

LABEL METASYMBN
* meta -> symb/obj
::
  DUP#1= NOTcase SYMBN          ( -> symb )
                                ( obj #1 )
  DROP                          ( obj )
  DUPTYPECOL? case ONESYMBN
  DUPTYPEROMP? case ONESYMBN
  DUPTYPEFLASHPTR? case ONESYMBN
;

LABEL SYMB2META
::
  ( * rebuild meta's from symbs up to the SETSIZEERR mark * )
  ( * clear vxxlflag is one of the symb is not a meta * )
  vxxlflag SetSysFlag
  ZERO
  [BEGIN]                               ( last mark symb_s meta_s #meta_s )
    DUP#2+PICK EQ: SETSIZEERR
  NOT_WHILE
    BUILDMETALP                         ( symb_s' meta_s' #meta_s' )
  [REPEAT]                              ( last mark meta_s #meta_s )
  DROP
;

LABEL BUILDMETALP
( symb_s meta_s -> symb_s' meta_s' )
::
  get1 CKSYMBINCOMP             ( symb_s' meta_s #meta_s meta )
  DUP#1+ pshtop&                ( symb_s' meta_s' #meta_s' )
;

LABEL CKSYMBINCOMP
::
  CK&DISPATCH1
  real ONE
  cmp  ONE
  sym  SYMBINCOMP
  EXT  ONE
  any  :: ONE vxxlflag ClrSysFlag ;
;

**********************************************************************
* Desc:         Evaluation of a symbolic replacing an idnt by a value.
*               (For example evaluation of F(X) for X=1/2)
*               Does not solve for undertermination (use LIMIT for this)
* Stack:        ( symb idnt a --> symb' )
* Algorithm:    Push symb in R, recalls element by element
*               if obj is idnt replace by a-meta
*               if obj is a seco we look for 2 args secondaries, and for int
*               if obj is a bint we have a user function
*               otherwise we have a unary operator
*               we just have to add it to the result meta by SWAP#1+
*               or a data type object that we transform in a meta
* Note:         complex are evaluated in cartesian coordinates
*               embedded symbolics are not protected from evaluation
*               INT is evaluated if keepintflag is cleared
* Added:        If a is +/-/unsigned infinity then does
*               first X->1/X MEVAL then simplify sym (metasimp+LNEXPA+metasimp)
*               and eventually calls REPLACE2BY1 with X and 0
**********************************************************************
NULLNAME REPLACE2BY1
::                                      ( symb X a )
  DUPTYPESYMB? IT SIMP1ext              ( symb X a )
  MAKE2LIST                             ( symb {} {} )
  NULL{} 
DEFINE idnt_name 3GETLAM
DEFINE idnt_value 2GETLAM
DEFINE idnt_name! 3PUTLAM
DEFINE idnt_value! 2PUTLAM
  3LAMBIND                              ( obj )
  polarflag TestSysFlag                 ( obj flag )
  polarflag ClrSysFlag                  ( obj flag )
  SWAP                                  ( flag obj )
  RECURREPLACE                          ( flag obj' )
  ABND SWAP NOT?SEMI 
  polarflag SetSysFlag                  ( symb' )  
;

LABEL RECURREPLACE
* obj -> obj'
::
  CK&DISPATCH1
  sym  SYMREPLACE
  list LISTREPLACE
  arry ARRYREPLACE
  any  NOP
;

LABEL SYMREPLACE
::
  symcomp                               ( flag symb )
  INREPLACE2BY1                         ( flag meta' )
  CKSYMBN                               ( flag symb' )
;

LABEL LISTREPLACE
::
  INNERCOMP metareplace {}N
;

LABEL ARRYREPLACE
::
  ARRAY2MATRIX INNERCOMP metareplace []N
;

LABEL metareplace
::
  DUP#0= ?SEMI
  DUPDUP
  ZERO_DO (DO)
    get1 RECURREPLACE UNROT
  LOOP
  DROP
;

**********************************************************************
* Stack:        ( symb idnt a --> symb' )
* Note:         Like REPLACE2BY1 but prevents evaluation of INT
**********************************************************************
NULLNAME NR_REPLACE
::
  keepintflag TestSysFlag OBJ>R
  keepintflag SetSysFlag                ( F x y )
  REPLACE2BY1                           ( F[y] )
  R>OBJ ?SEMI 
  keepintflag ClrSysFlag
;

LABEL INREPLACE2BY1
::                                      ( symb )
  cascompevalflag TestSysFlag 
  case ALGINREPLACE2BY1
  cascompevalflag SetSysFlag
  ALGINREPLACE2BY1
  cascompevalflag ClrSysFlag
;

LABEL ALGINREPLACE2BY1
::
  DUPDUP >R LENCOMP                     ( symb # )
  DUPUNROT NTHCOMPDROP                  ( # last )
  ' SETSIZEERR ROT                              ( last mark # )
  ZERO_DO (DO)
    RSWAP 'R RSWAP                      ( .., obj )
    REP2BY1LP                           ( .., meta )
    LOOP                                ( last mark meta )
  get1 EQ: SETSIZEERR NOTcase INTERNALERR
                                        ( last meta )
  get1 RATIONAL_OP? NOT?SEMI            ( -> meta )
  idnt_name CARCOMP TYPECOL? ?SEMI      ( no simpl if replacing secondaries )
  metasimp                              ( meta )
;

LABEL RATIONAL_OP?
* obj -> flag
* return true if obj is a rational operator or a zint
::
  DUPTYPEZ? case DROPTRUE
  LISTOPSIMP SWAP Contains? UNROT2DROP
;

LABEL REP2BY1LP
::                                      ( meta, obj )
  CK&DISPATCH1
  seco REP2BY1SECO
  idnt REP2BY1IDNT                      
  lam  REP2BY1IDNT                              
  symb INREPLACE2BY1
  any  COMPEVALANY                      ( was ONE -> meta )
;

LABEL REP2BY1SECO
* added to allow replacement of one seco by a prg acting on meta
::                                      ( ... seco )                    
  DUP idnt_name ListPos                 ( ... seco # )
  DUP#0=csedrp COMP2BY1SECO
                                        ( ... seco # )
  idnt_value ROTDROPSWAP                ( ... {} # )
  NTHCOMPDROP EVAL
;

LABEL COMP2BY1SECO
( * if level 3 is real, cmplx or unit and obj is unary op call it * )
* note: we might add code here testing for binary op and 2 reals for speed
CODE
                C=DAT1  A
                CD1EX                   D1 -> begin seco
                D1=D1+  5
                A=DAT1  A               A = first element
                CD1EX                   D1 -> stack
                LC(5)   =CK1&Dispatch
                ?A=C    A
                GOYES   +
-               LA(5)   =COMP2BY1SECOSLOW
                PC=(A)
+               D1=D1+  10              D1 -> stack level 3
                C=DAT1  A               C -> obj at stack3
                D1=D1-  10
                CD1EX
                A=DAT1  A
                CD1EX
                LC(5)   =DOREAL
                ?A=C    A
                GOYES   ++
                LC(5)   =DOCMP
                ?A=C    A
                GOYES   ++
                LC(5)   =DOEXT
                ?A#C    A
                GOYES   -
++              LA(5)   =COMP2BY1FLOAT
                PC=(A)
ENDCODE

LABEL COMP2BY1FLOAT
:: ROTSWAP EvalNoCK SWAP ;

LABEL COMP2BY1SECOSLOW
::                                      ( .., obj )
  DUP REP2BY1LISTOP ListPos             ( .., obj, # )
  DUP#0=csedrp REP2BY1TRIG?
  REP2BY1EVAL                           ( .., obj, #, list )
  ROTDROPSWAP                           ( .., list, # )
  NTHCOMPDROP EVAL
;

LABEL REP2BY1LISTOP
{
  x+ x- x* x/ x^ xNEG xNEGNEG
  x\.S_ALGEBRAIC
ASSEMBLE
        IF      f39
RPL
  xint
ASSEMBLE
        ELSE    f39
RPL
  xINT
ASSEMBLE
        ENDIF   f39
RPL
  xDER xWHERE xFCNAPPLY xCMDAPPLY xQUOTE
*  xEVAL>
}

LABEL REP2BY1EVAL
{ 
  MetaAdd MetaSub MetaMul MetaDiv meta^meta MetaNeg ( was REPNEG ) NOP
  REP.s REPINT REPDER REPWHERE REPFCNAPPLY REPCMDAPPLY REPQUOTE 
*  REPxEVAL>
}

*LABEL REPxEVAL>
*:: INDEX@ #1+ INDEXSTO RSWAP 'R RSWAP PROTECTEVAL SYMBINCOMP ;

LABEL REP2BY1TRIGOP
{
  xEXP xLN x\v/
  xSIN xCOS xTAN xSINH xCOSH xTANH 
  xASIN xACOS xATAN xASINH xACOSH xATANH
  xABS xARG xSIGN xRE xIM xCONJ
  xNOT
}

LABEL REP2BY1TRIGEVAL
{
  xSYMEXP addtLN addtSQRT
  addtSIN addtCOS addtTAN addtSINH addtCOSH addtTANH
  addtASIN addtACOS addtATAN addtASINH addtACOSH addtATANH
  addtABS addtARG addtSIGN addtRE addtIM addtCONJ
  addtNOT
}
        
LABEL REP2BY1TRIG?
::                                      ( .., op )
  EXACT? NOTcase REP2BY1&
  DUP REP2BY1TRIGOP ListPos             ( .., obj, # )
  DUP#0=csedrp REP2BY1&
  REP2BY1TRIGEVAL                       ( .., obj, #, list )
  ROTDROPSWAP                           ( .., list, # )
  NTHCOMPDROP OBJ>R                     ( meta )
  metasimp R>OBJ EVAL                   ( op[meta] )
;

* non exp/ln/trig operator: find argument numbers of op 
* and glue meta-arguments
* note that we should simplify the meta's
LABEL REP2BY1&
::
  enumargs                              ( ..., op, #k )
  DUP#0=csedrp metasimp1eval

  ( * note: if #=5 then seco is not an operator * )

  DUP #5= casedrop COMPEVALFLASH
  SWAP OBJ>R                            ( meta1 .. metak #k )
  EXACT? MAKE_K_ARG                     ( arg1 ... argk #k )
  DROP R>OBJ                            ( arg1 ... argk operator )
  PROTECTAPPLYNOCK                      ( symb )
  SYMBINCOMP metackundef                ( meta )
;

* Not used anymore since infinity evaluates to +infinity inside evaluation loop
*LABEL REPNEG
* must check for infinity alone, since -infinity in an expression
* is -infinity and not unsigned infinity followed by xNEG
*::                             ( meta )
*  unsignedinf metaEQUAL?       ( meta inf #1 flag )
*  UNROT2DROP case addtNEGonly
*  MetaNeg
*;

*LABEL REPSUM
*::
*  FOUR EXACT? MAKE_K_ARG DROP          ( low high f x )
*  ' xSUM EvalNoCK
*;

LABEL REP.s
::                              ( meta-low meta-high meta-fnc meta-var )
  FOUR EXACT? MAKE_K_ARG DROP                           ( low high f x )
  COLA INT4meta
;

LABEL INT4meta
::
  4PICK 4PICK EQUALcase 4DROPZ0ONE
                                                ( low high f x )
  EXACT? case REP.sEXACT
                                                ( low high f x )
  REP.sTEST
;

LABEL 4DROPZ0ONE
:: 4DROP Z0ONE ;

LABEL REP.sEXACT
:: 
  SIMPSYMBS SYMBINCOMP 
;

LABEL REP.sTEST
::                                              ( low high f x )
  1GETLAM NULLCOMP? case REP.sAPPROX

  ( * for CASNUMEVAL, no integral evaluation * )
                                                ( low high f x )
  SWAPROT 4ROLL                                 ( x f high low )
  SYMBINCOMP get1 SYMBINCOMP top&               ( x f high&low )
  get1 symcomp SWAP#1+                          ( x high&low&f )
  get1 symcomp SWAP#1+                          ( high&low&f&x )
  addt: xINTEGRAL                               ( meta )
;

LABEL REP.sAPPROX
::                                              ( low high f x )
  SWAP REPLACEZBYR                              ( low high x f )
  SWAP DUPTYPESYMB? ?SKIP ONESYMBN              ( low high f 'x' )
  2SWAP                                         ( f 'x' low high )
  ( * if low=minreal or high=maxreal do x=tan[u] change of var * )
  OVER xABSext %MAXREAL EQUALcase REPLINTINF
  DUP xABSext %MAXREAL EQUALcase REPLINTINF
  ' SYMNUMINT PROTECTAPPLY 
  ONE
;

LABEL PROTECTAPPLY
::                                              ( args -> res )
  ERRSET
    EVAL                                        
  ERRTRAP TRAPFCNAPPLY
;

LABEL PROTECTAPPLYNOCK
:: 
  ERRSET
    EvalNoCK                                    ( args -> res )
  ERRTRAP TRAPFCNAPPLY
;

* moved in znext.s at the end of bank 6 as a fptr
*LABEL REPLACEZBYR
*::
*  DUPTYPESYMB? NOTcase IFZTHENZ>R
*  ' IFZTHENZ>R 1LAMBIND LISTSECOext ABND
*;

LABEL IFZTHENZ>R
::
  DUPTYPEZ? case Z>R
;

LABEL REPLINTINF
::                                      ( f x low high )
  fnAngle1 TestSysFlag 5UNROLL		( ?rad f x low high )
  fnAngle1 SetSysFlag
  2SWAP DUPUNROT CARCOMP                ( ?rad low high 'x' f x )
  DUP xTANext                           ( ?rad low high 'x' f x tan[x] )
  MEVALext                              ( ?rad low high 'x' f )
  OVER xTANext xSQext 
  %1 x+ext x*ext REPLACEZBYR SWAP       ( ?rad low high f/[1+x^2] 'x' )
  4ROLL 'EvalNoCK: xATAN 
  4ROLL 'EvalNoCK: xATAN                ( ?rad f/[1+x^2] 'x' atan[low] atan[high] )
  ' SYMNUMINT PROTECTAPPLY ONE		( ?rad result #1 )
  fnAngle1 4ROLL ITE SetSysFlag ClrSysFlag
;

        
LABEL REPINT
::                                      ( meta-f meta-x meta-y )
  THREE EXACT? MAKE_K_ARG                       ( f x y #3 )
  keepintflag TestSysFlag 
  casedrop RETURNINT
  EXACT? NOTcasedrop REPINTAPPROX
  DROP UNROT                            ( y f x )
  CLEANIDLAM                            ( y f x )
  DUPUNROT RISCHext                     ( y x F[x] )
  SWAPROT NR_REPLACE
  SYMBINCOMP                            ( meta )
;

LABEL REPINTAPPROX
::                                      ( f x y ) 
  UNROT %0 4UNROLL                      ( 0 y f x )
  REP.sTEST                             ( meta )
;


LABEL RETURNINT
::                                      ( f x y )
  ROT symcomp ROT symcomp               ( y f x )
  ROT SYMBINCOMP #2+ addt:
ASSEMBLE
        IF      f39
RPL
  xint
ASSEMBLE
        ELSE    f39
RPL
  xINT
ASSEMBLE
        ENDIF   f39
RPL
                                        ( f x y1..yk xINT #k+3 )
;

LABEL REPDER
::                                      ( sym_id_; ONE meta-fcn )
  psh #1= NcaseTYPEERR                  ( meta-fcn symb_id_; )
  SYMBINCOMP #1= NcaseTYPEERR
                                        ( meta-fcn id )
  idntlamcase REPDER1                   
  BINT43 ERABLEERROR
;

LABEL REPDER1
::
  1LAMBIND                              ( meta )
  CLEANMETA                             ( meta )
  INMETADERIV                           ( meta )
  ABND                                  ( meta )
  DUP#0=csedrp Z0ONE
;

LABEL CLEANMETA
::
  DUP#1= NOT?SEMI
  DROP SYMBALGCASCOMPEVAL SYMBINCOMP
;

LABEL REPWHERE
::                                      
  DROP                          ( meta1 ... meta2k+1 #2k+1 )
  EXACT? MAKE_K_ARG SYMBWHERE           ( symb )
  SYMBINCOMP                    ( meta )
;

* The current implementation of where is bugged
* should map back to old SYMWHERE
* to do that change CSR/SYMFCNS.s: uncomment NAMELESS SYMBWHERE
* to NAMELESS jSYMWHERE, change in Header2/list a FakeEntry to SYMWHERE
* and change ?GETWHERE because all WHERE fields have been removed:-(
NULLNAME SYMBWHERE
*:: SYMWHERE ;
::
ASSEMBLE
        IF      f39
RPL
  NOTEQWcase 48SYMBWHERE
ASSEMBLE
        ENDIF   f39
RPL
 #1-
 
 SAFECASEVAL: REPWHERE1
;

NAMELESS SAFECASEVAL:
::
  PushVStack
  'R
  ERRSET
    DoRunSafe
  ERRTRAP :: PopVStack CK0NOLASTWD ERRJMP ;
  2DROP
  DropVStack
;

LABEL REPWHERE1
::
  NULL{}
  DUPROT TWO #/
  SWAP #0= NOTcase SETARGNUMERR
  ZERO_DO                       ( a1, b1, .., an, bn, A, B )
     ROT >HCOMP                 ( a1, b1, .., an, A, B )
     SWAPROT                    ( a1, b1, .., B A an )
     CLEANIDLAM                 ( a1, b1, .., B A an )
     >HCOMP                     ( a1, b1, .., B A )
  SWAPLOOP                      ( symb_1 {} {} )
  ( * replace identifiers by global unassigned names * )
  UNROT DUPLENCOMP ZEROSWAP     ( list_val symb_1 list_orig #0 #n )
  ZERO_DO (DO)                  ( list_val symb_1 list_orig meta )
    %RAN DO>STR $>ID		( list_val symb_1 list_orig meta id )
    SWAP#1+
    LOOP
  {}N                           ( list_val symb_1 list_orig list_idnt )
  DUP 5UNROLL                   ( list_idnt list_val symb_1 list_orig list_idnt )
  ( * added evaluation for examples like DERIV[X^3,X]|[X=2] * )
  SYMBMEVAL SYMBALGCASCOMPEVAL UNROT  ( symb_1 list_idnt list_val )
  SYMBMEVAL                      ( symb )
;

LABEL SYMBMEVAL
::
  fnResults TestSysFlag NOTcase MEVALext
  fnResults ClrSysFlag
  MEVALext
  fnResults SetSysFlag
;

LABEL SYMBALGCASCOMPEVAL
::
  fnResults TestSysFlag NOTcase ALGCASCOMPEVAL
  fnResults ClrSysFlag
  ALGCASCOMPEVAL
  fnResults SetSysFlag
;

LABEL REPQUOTE
::                              ( meta | symb one )
  DUP#1= NOT?SEMI
  DROP SYMBINCOMP
;

LABEL TRAPFCNAPPLY
::                              
  RestoreUStk                   ( last_ob mark metas obj1..objk id )
  ( * convert metas to symb and suppress last_ob and mark * )
  ONE                           ( last_ob objs/metas meta_obj )
  [BEGIN]
    DEPTH OVER #1+ #>   ( [ meta-obj FALSE  ... meta-obj SETSIZEERR ] )
    NOTcase TRAPFCNEXIT ( --> skip direct to error handler )
    get1 DUPEQ: SETSIZEERR              ( last_ob objs/metas meta_obj obj )
  NOT_WHILE
    :: DUPTYPEBINT? NOTcase SWAP#1+ 
                                ( ... meta' meta_obj # )
  ( * check that we have enough objets on the stack * )
       2DUP #+ #4+ DEPTH #< NOTcase :: OVER#2+UNROLL TRAPFCNEXIT ; 
       OVER#2+UNROLL psh METASYMBN
       SWAP#1+ 
    ;
  [REPEAT]
                                ( last_ob meta_obj SETSIZEERR  meta-obj )
  DROP reversym get1 2DROP      ( objs )
  ( * patch for LIMIT * )
  DEPTH #0= ?SKIP :: DUP ' xLIMIT EQUAL IT DROP ;
  ( * set user stack * )
  CK0NOLASTWD
  ( * back to normal error process * )
  ERRJMP
;

LABEL TRAPFCNEXIT
* last_ob objs/metas meta_obj
:: reversym DROP CK0NOLASTWD ERRJMP ;

LABEL REPFCNAPPLY               ( ... meta1 .. metak metafnc #k one )
::
  DROPSWAP                      ( ... meta1 .. metak metafnc' #k # )
  #1= NOTcase INTERNALERR
                                ( ... meta1 .. metak symb #k )
  SWAPDUP DUPTYPESYMB? IT CARCOMP ( ... meta1 .. metak #k symb fnc )
  TYPEIDNT? SWAP OBJ>R          ( ... meta1 .. metak #k ? )
  EXACT? AND MAKE_K_ARG R>OBJ   ( last_ob mark metas obj1..objk #k id )
  ERRSET
   SIMPUSERFCN                  ( ob1..obk id -> symb )
  ERRTRAP TRAPFCNAPPLY
                                ( last_ob mark metas obj )
  ( * check that obj is of symbolic class, if not COMPEVAL * )
  vxxlflag SetSysFlag
  CKSYMBINCOMP                  ( last_ob mark metas )
  COLA CKSYMBS
;

LABEL REPCMDAPPLY               ( meta1 .. metak metafnc #k one )
::
  DROPSWAP                      ( meta1 .. metak metafnc' #k # )
  #1= NOTcase INTERNALERR
                                ( meta1 .. metak id #k )
  OBJ>R META2SYMB               ( last mark symb1..symbn id )
  R>OBJ                         ( last mark symb1..symbn id #k )
  SWAP ONE{}N SWAP 
  ERRSET xCMDAPPLY              ( last mark symb1..symbl/list )
  ERRTRAP TRAPFCNAPPLY
  CKINNERCOMP DROP              ( last mark symb1..symbl )
  SYMB2META                     ( last mark meta1..metal )
  COLA CKSYMBS
;

LABEL CKSYMBS
::
  vxxlflag TestSysFlag ?SEMI    ( -> last mark meta1..metal )
  ( * one object was invalid inside a symbolic * )
        
*  test if CASCOMPEVAL is strict algebraic or not
  cascompevalflag TestSysFlag case NONALGERR
        
  ExitAtLOOP ZERO META2SYMB     ( last mark symb1..symbk #0 )
  DROPRDROP THREE RROLL         ( last mark symb1..symbk )
  ( * standard evaluation loop * )
  LABEL cmdapplylp
  ticR NOTcase ONE              ( must return one arg )
  EVALIFNOTCOL
  GOTO cmdapplylp
;

LABEL EVALIFNOTCOL
* evalute object, check for idnt/lam that evaluate to a user prog
::                              ( obj )
  IDNTLAM? NOTcase PROTECTAPPLY
  DUPSAFE@                      ( idnt obj true | idnt false )
  NOT?SEMI                      ( -> idnt/lam )
                                ( idnt obj )
  DUPTYPECOL? NOTcase SWAPDROP  ( -> obj )
                                ( idnt obj )
  ( * forbid using user prog inside idnt/lam * )
  DUP CARCOMP EQ: x<<           ( idnt obj )
  caseDROP                      ( -> idnt )
                                ( idnt obj )
  SWAPDROP PROTECTAPPLY
;


LABEL MAKE_K_ARG
::                              ( meta1 .. metak #k ?simplify )
  1LAMBIND                      ( meta1 .. metak #k )
  ZEROSWAP
  ZERO_DO (DO)                  ( meta1 .. metak metaarg )
    psh DUPTYPEBINT?            ( meta1 ... metak-1 metaarg metak ? )
    NOTcase INTERNALERR
                                ( meta1 ... metak-1 metaarg metak )
    1GETLAM IT metasimp         ( meta1 ... metak-1 metaarg metak )
    METASYMBN SWAP#1+           ( meta1 ... metak-1 metaarg )
  LOOP                          ( metaarg )
  reversym ABND                 ( metaarg )
;
        
LABEL REP2BY1IDNT
::                              ( idnt )
  DUP idnt_name ListPos         ( idnt # )
  DUP#0=csedrp COMPEVALIDNT
                                ( idnt # )
  idnt_value ROTDROPSWAP        ( {} # ) 
  NTHCOMPDROP SYMBINCOMP        ( meta )
;

LABEL COMPEVALIDNT
::                                      ( idnt )
  1GETLAM OVER Contains?                ( idnt liste1 idnt flag )
  UNROT2DROP                            ( idnt/lam flag )
  case ONE                              ( -> idnt, one )
  DUPSAFE@ NOTcase ONEORUNDEFID         ( -> idnt, one )
                                        ( idnt, obj )
  ( * do not allow user prog inside loop except user defined fcn * )
  DUPTYPECOL? case ::                                   ( idnt col )
                        DUP TWO NTHELCOMP               ( idnt col [el] ? )
                        NOTcase PROTECTEVAL             ( -> )
                                                        ( idnt col el )
                        EQ: xALG-> casedrop ONE         ( -> idnt one )
                        PROTECTEVAL                     ( -> )
                        ; 
  ( * test for other non symbolic objects * )
  SECOMATLIST? case NONALGIDNT          ( -> normal COMPEVAL )
                                        ( idnt, obj )
  DROP 3GETLAM 2GETLAM NULL{} 3LAMBIND
  RCL1IDNT ABND 
  ( * do not allow user prog inside loop * )
  DUPTYPECOL? case :: DUP PROTECTEVAL ;
  ( * added simplification if not called from CASNUMEVAL? * )
  ( * warning: if circular def does not error this will cause infinite loop * )
  1GETLAM NULLCOMP? IT ALGCASCOMPEVAL
  SYMBINCOMP                            ( meta )
;

LABEL NONALGIDNT
::                                      ( ... idnt obj )
  SWAPDROP ONE
  vxxlflag ClrSysFlag
  CKSYMBS                               ( back to normal eval )
;

LABEL SECOMATLIST?
::
  CK&DISPATCH1
  seco TRUE
  list TRUE
  matrix TRUE
  any FALSE
;

LABEL ONEORUNDEFID
::
  SYMB? case ONE 
  SETNONEXTERR
;

LABEL PROTECTEVAL
::                                      ( ... idnt seco )
  PushVStack
  ONE KEEP
  ERRSET 
    EVAL
  ERRTRAP ZEROZERO                      ( so that err will return a list )
  STK>RESULT                            ( res )
  PopVStackAbove                        ( ... idnt obj res )
  UNROT2DROP CKALG? case SYMBINCOMP     ( -> ... meta )
  0LASTOWDOB! NONALGERR
;


**********************************************************************
* Stack:        ( symb  --> symb' )
* Note:         Like CRUNCH but in approx mode
**********************************************************************
NULLNAME CASCRUNCH
::
ASSEMBLE
        IF      f39
RPL
  REPLACEZBYR
ASSEMBLE
        ENDIF   f39
RPL
  APPROXEVAL CRUNCH
;

**********************************************************************
* Stack:        ( symb  --> symb' )
* Note:         Like CASCOMPEVAL but in approx mode
**********************************************************************
NULLNAME APPROXCOMPEVAL
::
  APPROXEVAL ALGCASCOMPEVAL
;
