INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE Series expansion and limit

**********************************************************************
* Algorithm used:
*
* Series expansion, beginning from the most inner variables
* "Variables" depending on the varying one are ordered by their
* decreasing rate at 0 and are normalized to be positive if their sign
* is known. For example 1/ln(x) -> 0^- as x -> 0 hence we use
* INV(-LN(X)) as variable, which tends to 0^+
* The ordering is e.g. EXP(1/X^2) tends to 0 much faster than any
* power of X, X tends to 0 much faster than any power of INV(-LN(X))
* This strategy is always successfull if no algebraic relations
* exists between the "variables" (e.g. X^1/3), like for the Risch algorithm
* Erable tries to handle these variables using TSIMP
* but this is extremely slow and problems occurs with LN expansion
* since we want to keep INV(-LN(X)) type variables :-(
*
* LIMIText	( symb, X, a, direction -> limit )
*
* SYMTAYLOR	( symb id % -> symb )
*		Taylor expansion
*
* LIMSERIES!	( symb, 'X=a' or idnt, [% or hxs] -> 
*		    a, lim, DL-l, reste-l, num-l/deno-l, equiv-l, lvar )
*		a is the point for series expansion, 0 by default
*		lim is the limit (user format)
*		DL-l is the series expansion (internal)
*		reste-l the error term (internal)
*		num-d/deno-l the relative series expansion (internal)
*		equiv-l the equivalent (internal)
*		lvar the list of variables for decoding internal format
*
* LIMIT!	( symb -> 
*		DL-l, reste-l, num-l/deno-l, equiv.-l, lim., lvar, flag )
*		a, X and direction must be bound
*
* LIMSTEP1!	( symb ->  { DL-l, reste-l, num-l/deno-l, equiv.-l }, flag )
*		the list of variables with coresponding series expansion
*		must be bound
*
* LIMLIM!	( #, lvar, equiv-l -> lvar, lim )
*		# is the depth
*
* n{}N		( obj, # -> obj embedded in  #lists )
*
* LIMCMPL!	( reste-1-l, reste-2-l -> reste-l )
*		Compares two relative rests and returns the biggest
*
* LIMEQUFR!	( n/d-l, # -> n/d-l, equiv, % )
*		Extract the equivalent part from n/d-l
*		# is the number of vars depending of X
*
* LIMEQU!	( liste, # ->  liste / equiv-l, equiv-l, { # # # } 
*		Like LIMEQUFR! for a polynomial
*		{ # # # } is the orders of equiv-l with respect to the var
*		depending on X
**********************************************************************

* rationnal operators
*		( DL1, ..., DLn, #n, op -> DL, flag )
* LIM+-!
* LIMNEG!
* LIMRAC!
* LIMPOW!
* LIMINV!
* LIM/!
* LIMSQ!
* LIM*!
* --------------------------------------------------------------------
*
* LIMDIVPC!	( #order, num-l, deno-l -> num-l, deno-l )
*		Ascending power division
* DIVPC!	( reste, n, d, # -> reste, n, d, flag )
*		Used by LIMDIVPC!
* LIMPROF!	( reste-l, num-l, deno-l, #varx, ?num-l -> 
*		  reste-l, num-l, den-l, #prof, #ordre, flag )
* LIM%#!	( num-l, deno-l, { % ... % } -> 
*		  new-num-l, new-deno-l, #depth, { % % % } )
* LIMPROF0!
* LIMPROF1!
* LIMPROF2!
* LIMPROFEND!	( num, deno, #prof -> num, deno )
*
* --------------------------------------------------------------------
* non rational operators
*		( reste , DL-l, num-l/deno-l, equiv, lim, lvar, operator -> 
*		  DL, reste, flag )
* LIMINVLN!
* LIMLN!
* LIMEXP!
* LIMSINCOS!
* LIMATAN!
* LIMASIN!
* LIMSQRT!
* LIMABS!
*
*		invalid operator
* --------------------------------------------------------------------
* LPROF!	( lvar, reste-l, num-l/deno-l ->
*		  #varx, #prof, #ordre, lvar, new reste-l, num-l/deno-l )
* LIM#VARX!	( lvar, lvar -> #varx )
*		Number of var depending on X
* LIMBETA!
* LIMALPHA!
*		Used for DL of (1+x)^alpha
*
* --------------------------------------------------------------------
* Unary Composition of DL
*		( lim, lvar, X-l, reste-l -> lvar, DL, reste-l )
* HORNEXP!	
* HORNLN!
* HORNCOS!
* HORNSIN!
* HORNATAN!
* HORNASIN!
* --------------------------------------------------------------------
* LNOBJ!	( obj -> ln[obj], size )
*		size is equiv[ln(equiv]] if obj is an equivalent
* LIMHORN!	( X, c0, ..., cn, #n -> obj )
*		Used for unary composition of DL

* LIMINFSIGN!
*		Find sign of infinite limit
*
* --------------------------------------------------------------------
* Comparison:
*
* LIMMAX!	( o1, o2 -> o1 or o2 )
*		returns the biggest one
* LIMSORT!	( {} -> {}' )
*		Sort a list of variable depending on X (see beginning of file)
* LIMCOMP!	( o1, o2 -> flag )
*		Comparison of variables that are valide (i.e. EXP() or
*		INV(-LN()) that tend to 0)
* VARCOMP!	( var1, var2 -> flag )
*		TRUE if var1 > var2
* VARCOMP2!	( o1, o2 -> #varx, lvar, equiv of o1/o2-l, # )
* # = 0 if o1/o2 -> finite limit given in equiv of o1/o2-l 
* # = 1 if num/deno negligible 
* # = 2 if deno/num negligible
* VARCOMP3!	 ( #depth, lvar, o1/o2-l -> #, lvar, o-liste, # )
* LIMVALOBJ!	 ( obj, lvar -> symb )
*		 like VAL2ext but at depth #0
* LIMVAL!	 ( obj -> coeff, valuation )
* EQUIV!	 ( {}, { equiv } ) -> equiv, order

* LVARXNX2!	 ( symb -> symb, lvarnx, lvarx )
*		 Like LVARXNX but takes care of valid variables [INV[-LN[]]

* FindCurVar	 ( symb -> symb )
*		 Update current variable using idnt of symb if symb
*		 does not depend on VX
* LIMVAR!	 ( symb -> symb, lvar )
*		 List of sorted var of symb
**********************************************************************

**********************************************************************
* Desc:		Compute limit of symb at X=a
* Stack:	( symb X a direction -> limite )
**********************************************************************
NULLNAME LIMIText
::				( obj X a direction )
  ( * add X to real valued variable list * )
  3PICK ADDTOREAL 		( obj X a direction )

  1LAMBIND			( obj X a )

  ( * first rewrite expr if a=infinity for efficiency * )
  REWRITEIFINF			( obj X a )

  ( * then make a subsitution * )
  3DUP div0by0flag ClrSysFlag	( obj X a obj X a )
  SIMP1ext			( obj X a obj X a )
  NR_REPLACE			( obj X a res )

  ( * test for cases where L'Hopital's rule could resolve undetermined form * )
  div0by0flag TestSysFlag 
  IT LHOPITAL			( obj X a res )

  ( * look if res is undefined or unsigned inf, if not we are done * )
  UNSIGNEDINFOR?		( obj X a res # )
  1GETABND SWAP			( obj X a res direction # )
  #0= casedrop 4UNROLL3DROP	( -> res )
*  :: 	4UNROLL3DROP ( res )
*	POSINFext #0=?SEMI ( -> res )
*	DROP INFINIext ( return unsigned inf instead of any inf type )
*	;	

  ( * bad case: we must do a series expansion to get the limit * )
				( obj, X, a, res, direction )
  4UNROLL DROP 			( obj direction X a )
  EXPR=				( obj, direction, X=a )
  SWAP COMPLEX? 		( obj X=a direction ? )
  NOTcase REALLIMIT		( -> limit )
  ( * in complex mode find real and imaginary part * )
  				( obj X=a direction )
  3PICK xREext 3PICK3PICK	( obj X=a direction real X=a direction )
  REALEVAL REALLIMIT STRIPTAGS 4UNROLL	( re obj X=a direction )
  ROT xIMext UNROT 		( re im X=a direction )
  REALEVAL REALLIMIT STRIPTAGS		( re im )
  'xi x*ext x+ext		( re+i*im )
;

LABEL UNSIGNEDINFOR?
* res -> res #
* return #0 if not ? and unsigned inf ok
::							( res )
  symcomp INFINIext OVER EQUALcase ASKUNSIGNEDINF	( res # )
  POSUNDEFext 						( res # )
;	

LABEL ASKUNSIGNEDINF
::							( res )
ASSEMBLE
	IF	f39
RPL
  InCASApplication? NOT switchflag TestSysFlag OR
ASSEMBLE
	ELSE	f39
RPL
  switchflag TestSysFlag 
ASSEMBLE
	ENDIF	f39
RPL
  case ZERO						( -> res #0 )
							( res )
  % 5000 % 0.3 DOBEEP
  BINT93 GETERABLEMSG AskQuestion			( res flag )
  ITE ONE ZERO
;

* insure exact mode and that stack level 1 is not a floating point
LABEL CKEXACT
::
  EXACT? NOTcase NONEXACTERR
  CK&DISPATCH1
  real NONEXACTERR
  cmp  NONEXACTERR
  EXT  NONEXACTERR
  any  NOP
;

LABEL NONEXACTERR
::
  BINT67 ERABLEERROR
;

LABEL REALLIMIT
::				( obj X=a direction -> limit )
  LIMSERIES!			( a lim DL-l rest-l num/deno-l equiv-l lvar # )
  DROP 5UNROLL 2DROP		( a lim lvar DL-l rest-l )
  CKLIMREST			( a lim lvar )
  ZERO UNROT VAL2ext SWAPDROP	( lim )
;

LABEL CKLIMREST
* lim, lvar, DL-l, rest-l -> lim, lvar
::				( lim lvar DL-l rest-l )
  Z0 EQUALcasedro DROP		( -> lim lvar )
				( lim lvar DL-l rest-l )
  FXNDext CONSTANTPOLY? 	( lim lvar DL-l n d ? )
  NOTcase :: 3DROP LIMORDERWARN ;
				( lim lvar DL-l n d )
  DROPSWAPDROP			( lim lvar n )
  DUPTYPELIST? NOTcasedrop LIMORDERWARN
  DEGREext #0=case LIMORDERWARN
				( lim lvar )
;

LABEL LIMORDERWARN
* lim, lvar -> lim, lvar
::					( lim lvar ) 
  SWAP POSINFext #0<> 			( lvar lim ?inf )
  case SWAP				( -> lim lvar )
					( lvar lim )
  nocareflag TestSysFlag ?SKIP :: DROP ?ext ;
  SWAP					( ? lvar )
  BINT98 GETERABLEMSG 
ASSEMBLE
	IF	f39
RPL
  Ck&DoMsgBox 
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR Ck&DoMsgBox 
ASSEMBLE
	ENDIF	f39
RPL
;

**********************************************************************
* Desc:		Rewrite symb if a tends to infinity so that limit
*		can be computed faster
* Stack:	( symb X a -> symb' X a )
**********************************************************************
NULLNAME REWRITEIFINF
( obj X a -> obj' X a )
::
  DUP SYMBINCOMP 1metainf#		( obj X a meta-a # )
  #0=case NDROP				( -> obj X a )
  NDROP 				( symb X a )

  ( * first change LN containing x * )
  2LAMBIND				( symb )
  LVARext				( symb lvar )
  DUP >R LENCOMP			( symb # )
  ZEROZERO ROT				( symb meta1 meta2 # )
  ZERO_DO (DO)
    RSWAP 'R RSWAP			( symb meta1 meta2 var )
    INFREPLP				( symb meta1 meta2 )
  LOOP					( symb meta1 meta2 )
  {}N OVER#2+UNROLL {}N SWAP MEVALext	( symb' )
  
  ( * now isolate in each fraction the leading power in X * )
  INFEXPAND
  2GETLAM 1GETABND			( symb X a )
;

LABEL INFREPLP
( meta1 meta2 var -> meta1' meta2' )
::					( meta1 meta2 var )
  ( * if var does not contain lam2 exit )
  DUPTYPESYMB? NOTcaseDROP
  2GETLAM Contains? SWAPDROP		( meta1 meta2 var flag )
  NOTcaseDROP				( -> meta1 meta2 )
  					( meta1 meta2 var )
  DUPINCOMP				( meta1 meta2 var meta )
  #1-SWAP numargs #1=			( meta1 meta2 var meta' op ? )
  NOTcasedrop NDROPDROP			( -> meta1 meta2 )
					( meta1 meta2 var meta' op ) 
  OBJ>R infexpand R>OBJ
  DUPEQ: xLN ITE :: DROP METALNEXPA ; SWAP#1+
					( meta1 meta2 var meta' )
  CKSYMBN				( meta1 meta2 var var' )
  2LAMBIND psh 2GETLAM SWAP#1+ 		( meta2 meta1+var )
  psh 1GETABND SWAP#1+			( meta1+var meta2+var )
;

LABEL NDROPDROP
:: NDROP DROP ;

LABEL INFEXPAND
* symb -> symb, rewrite expression with x-equivalent factored
* lam2 contains the X variable
::
  SYMBINCOMP infexpand CKSYMBN
;

LABEL infexpand
* meta -> meta'
* lam2 contains the X variable
::
  dup LISTOPext LVARDext	( meta lvar )
  2GETLAM prepvarlist		( meta lvar )
  DUP 2LAMBIND			( meta )
  VXXL2				( Q )
  ( * now we find the power of X in num and deno of Q * )
  ( * we factor this expression * )
  FXNDext			( num deno )
  REVERT_#			( num deno' #n )
  ROT REVERT_#			( deno' #d num' #n )
  UNROT 4ROLL			( #n #d num' deno' )
  NDXFext			( #n #d Q' )
  ( * change X in 1/X, first occurance of LVAR in LAM2 * )
  2GETLAM DUP CARCOMP SYMINV 	( #n #d Q' lvar 1/x )
  ONE ROT PUTLIST 2PUTLAM	( #n #d Q' )
  ZERO 1PUTLAM			( #n #d Q' )
  VAL1M				( #n #d meta )
  ABND				( #n #d meta )
  get1 rotomo			( meta #d #n )
  2DUP#= case2DROP
  2DUP#< case infexpand<
  				( meta #d #n )
  #- 2GETLAM SYMBINCOMP 	( meta #d-#n meta-x )
  get1 meta^ addtINV MetaMul	( -> meta )
;

LABEL infexpand<
::				( meta #d #n, #d<#n )
  SWAP#- 2GETLAM SYMBINCOMP	( meta #n-#d meta-x )
  get1 meta^ MetaMul		( -> meta )
;

LABEL REVERT_#
* P -> P', #
::
  NOTLISTcase ONE
  DUPLENCOMP SWAPINCOMP			( # meta )
  reversym >POLYTRIM SWAP		( P' # ) 
;

LABEL LHOPITAL
* ( obj X a res -> obj X a res' )
* note that res is undef
* applies L'Hopital's rule to obj at X=a for a faster limit
* does not work inside non rational operator
::
  ( * first insure that a is not infinity * )
					( obj X a res )
  OVER POSINFext SWAPDROP 		( obj X a res # )
  #0<> ?SEMI				( -> obj X a res )
					( obj X a res )
  DROP 3DUP SIMP1ext CKEXACT   		( obj X a obj X a )
  MAKE2LIST NULL{} 3LAMBIND		( obj X a obj )
  SYMBINCOMP lhopital			( obj X a meta-res )
  CKSYMBN ABND
;


LABEL lhopital
* meta -> meta-res
* find limit of meta for variable LAM1 -> LAM2
::
  M1stNEGcase lhopitalCHS
  M1st+case lhopital+
  M1st-case lhopital-
  M1st*case lhopital*
  M1st^case lhopital^
  M1st/case lhopital/
  SYMBN INREPLACE2BY1
; 

LABEL lhopitalCHS
:: pulldrop lhopital MetaNeg ;

LABEL lhopital+
:: lhopitalbin MetaAdd ;

LABEL lhopital-
:: lhopitalbin MetaSub ;

LABEL lhopital*
:: lhopitalbin MetaMul ;

LABEL lhopital^
:: lhopitalbin MetaPow ;

LABEL lhopitalbin
::
  plDRPpZparg lhopital psh lhopital
;

LABEL lhopital/
::
  ( * loop, div0by0flag is used as test to stop the loop * )
  pulldrop larg			( meta-num meta-deno )
  THREE				( maximal number of derivative allowed +1 )
  ZERO_DO (DO)			( meta-num meta-deno )
    dup lhopitalck		( meta-num meta-deno meta-deno[a] )
    rot dup lhopitalck		( meta-deno meta-deno[a] meta-num meta-num[a] )
    rot lhopitallp		( meta-n' meta-d' | meta[a] meta )
  LOOP
  div0by0flag TestSysFlag NOTcase NDROP
  2NDROPundef
;

LABEL lhopitalck
* like lhopital, but avoid infinite recursion if the derivative is 
* a 0/0 form too (might happen e.g. with sqrt[x-a] if x=a)
::
  INDEX@ #0=case lhopital
  SYMBN INREPLACE2BY1
;

LABEL lhopitallp
::					( deno num num[a] deno[a] )
  div0by0flag ClrSysFlag		( deno num num[a] deno[a] )
  ' x? EQUALPOS2META			( deno num num[a] deno[a] ? # )
  SWAPDROP #0<> case 4NDROP2undef	( -> meta-? meta-? )

					( deno num num[a] deno[a] )
  metasimp MetaDiv 
  div0by0flag TestSysFlag 		( meta-deno meta-num meta-[a] flag ) 
  NOTcase lhopitallpok
  					( meta-deno meta-num meta-[a] )
  NDROP 3GETLAM CARCOMP 1LAMBIND 	( meta-deno meta-num )
  lhopitalderiv psh 
  lhopitalderiv ABND			( meta-num' meta-deno' )
  ' xFCNAPPLY EQUALPOS2META		( meta-num' meta-deno fcn # )
  SWAPDROP #0=?SEMI			( -> meta-num' meta-deno' )
  					( meta-num' meta-deno' )
  div0by0flag SetSysFlag ExitAtLOOP
;

LABEL lhopitalderiv
::
  NOSTEPEVAL METADERIV
  DUP#0=csedrp Z0ONE
;

LABEL 4NDROP2undef
:: NDROP NDROP 2NDROPundef dup ExitAtLOOP ;

LABEL lhopitallpok
::
  unrot NDROP ExitAtLOOP 	( -> meta[a] meta )
;

**********************************************************************
* Desc:		Taylor approx. at 0
* Stack:	( symb id %/z -> symb )
**********************************************************************
NULLNAME SYMTAYLOR
NULLNAME SYMPAPRX
::				( symb id %/z )
ASSEMBLE
	IF	f39
RPL
  NOTEQWcase 48SYMPAPRX
ASSEMBLE
	ENDIF	f39
RPL
  EXACTON
  OBJINT? NcaseSIZEERR
				( symb id z )
  CASFLAGEVAL: SYMTAYLOR1
;

LABEL SYMTAYLOR1
::
  COMPLEXOFF LIMSERIES!		( 0 lim DL-l rest-l num/deno-l equiv-l lvar # )
  DROP 5UNROLL SWAPDROP 	( 0 lim lvar DL-l reste-l equiv-l )
  QMul TRUNCDL			( 0 lim lvar truncated_DL-l )
  ZEROSWAP RISCH13		( 0 lim lvar #0 DL-l )
  ROT VAL2ext			( 0 lim DL )
  UNROT2DROP			( DL )  
;

**********************************************************************
* Desc:		Series expansion truncation
* Stack:	( DL-l reste-l -> truncated_DL )
**********************************************************************
NULLNAME TRUNCDL
::
  2DUP EQUAL case2drop Z0	( return 0 if DL=reste )
  Z0 EQUALcasedro NOP		( return DL if reste=0 )
				( DL-l=n2/d2 reste=n1/d1 )
  FXNDext ROT FXNDext		( n1 d1 n2 d2 )
  ( * will truncate n2*d1 w.r.t. n1*d2 and then divide by d1*d2 * )
  3PICKOVER QMul		( n1 d1 n2 d2 d1*d2 )
  ROT 4ROLL QMul		( n1 d2 d1*d2 n2*d1 )
  ROT 4ROLL QMul		( d1*d2 n2*d1 n1*d2 )
  LISTTRUNCDL  			( d1*d2 truncated )
  SWAP RDIVext			( trucated/d1 ) 
;

LABEL LISTTRUNCDL
* Q Q-rem -> Q'
* truncation
::
  NOTLISTcase 2DROPZ0
  LIMVAL!			( Q Q-rem' #valuation )
  ROT DUPCKLEN{}		( Q-rem' #valuation Q #q )
  ( * if #q<=#valuation return Q unchanged * )
  3PICK #> NOTcase UNROT2DROP	( -> Q )
				( Q-rem' #valuation Q )
  CKINNERCOMP			( Q-rem' #valuation qq..q0 #q+1 )
  get1 #1+			( Q-rem' qq..q0 #q+1 #val+1 )
  DUPUNROT #-			( Q-rem' qq..q0 #val+1 #q-val )
  OVER#2+UNROLL pshNDROP	( Q-rem' qval..q0 #val+1 )
  #1- get1 rotomo		( qval+1..qq #val qval Q-rem )
  LISTTRUNCDL			( qval+1..qq #val qval' )
  OVER#2+UNROLL #1+ >POLYTRIM	( Q )
;

**********************************************************************
* Desc:		Series expansion
* Stack:	( expression X=a|X %|zint -> )
*		( a lim DL-l rest-l num-l/deno-l equiv-l lvar # )
* Note:		#=1 for X=a-h or X=-1/h
**********************************************************************
NULLNAME LIMSERIES!
::
  ( %|zint is the asked relative order of expansion )
  ( if % return unidirectional expansion depending on sign of % )
  ( in all cases use of X=a+0 or X=a-0 indicate unidirectional expansion )
  ( * #type= #0=bi, 1=-, 2=+  * )

  ( * first forbid complex and approx mode switch * )
  noswitchflag SetSysFlag 

  ( * parse order and limit type * )
  ::
    CK&DISPATCH1
    real  :: DUP %ABSCOERCE SWAP %0> ITE TWO ONE ;
    zint  :: Z># ZERO ;
  ;
						( symb X=a #ordre #type )
  UNROT ONE #MAX TWENTY #MIN 			( symb #type X=a #ordre )
  LIMSERIESMEVAL				( symb #type #ordre X a # )
  1LAMBIND 5UNROLL STOVX			( a symb #ordre #type )
  ZERO { LAMLIMTYPE LAMLIMORDRE LAMLIMOK } BIND	( a symb )
  NULL$ STOCASINFO				( a symb )
  increaseflag TestSysFlag SWAP SETPLUSAT0	( a ?increaseflag symb )
  VXLVARext DUP STURMFLAGEVAL: SLVARext MEVALext		( a ?increaseflag symb )
  ' x^ ' meta^meta STURMFLAGEVAL: REPLACE2BY1  
  LIMMEVALL1 LIMMEVALL2 REPLACE2BY1		( a ?increaseflag symb )
  LIMIT!	( a ?incrflag DL-l reste-l num-l/deno-l equiv-l lim lvar flag )
  8ROLL ?SKIP CLRPLUSAT0
  			( a DL-l reste-l num-l/deno-l equiv-l lim lvar flag )
  ?SKIP LIMOKERROR	( a DL-l reste-l num-l/deno-l equiv-l lim lvar )
  ABND 1GETABND		( a DL-l reste-l num-l/deno-l equiv-l lim lvar # )
  ROT 7UNROLL 		( a lim DL-l reste-l num-l/deno-l equiv-l lvar # )
  ( * allow again mode switch * )
  noswitchflag ClrSysFlag 
;

LABEL LIMMEVALL1
{	
  xTAN xXROOT 
  xSINH xCOSH xTANH xASINH 
  xACOSH xATANH xACOS
  x\v/ x\.S_ALGEBRAIC
ASSEMBLE
	IF	f39
RPL
  xint
ASSEMBLE
	ELSE	f39
RPL
  xINT
ASSEMBLE
	ENDIF	f39
RPL

}

LABEL LIMMEVALL2
{	
  sin/cos xroot2expln
  sinh2exp cosh2exp tanh2exp asinh2ln 
  acosh2ln atanh2ln pi/2-asin
  addtSQRTonly INVALIDOP 
  INVALIDOP
}

*LABEL sinh2explin
*:: hmeta-meta metasub meta/2 ;
*
*LABEL cosh2explin
*:: hmeta-meta metaadd meta/2 ;
*
*LABEL hmeta-meta
*::
*  dup addtEXPonly psh metaneg addtEXPonly
*;

LABEL LIMOKERROR
:: 
  LAMLIMOK GETERABLEMSG
  DUP STOCASINFO
ASSEMBLE
	IF	f39
RPL
  Ck&DoMsgBox
ASSEMBLE
	ELSE	f39
RPL
  FLASHPTR Ck&DoMsgBox
ASSEMBLE
	ENDIF	f39
RPL
;

LABEL LAMLIMOK
LAM LIMOK

LABEL LAMLIMORDRE
LAM LIMORDRE

LABEL LAMLIMTYPE
LAM LIMTYPE

LABEL LIMSERIESMEVAL
( symb #type X=a #ordre -> symb' #type #ordre X a # )
( * if a is infinite, make change of var X=+/-1/h * )
( * if a is finite, make X=a+-h * )
( * # mod 4 =1 if a-h or -1/h, #>=4 for inf * )
::					( symb #type X=a #ordre )
  SWAP idntlamcase :: Z0 ZERO ;		( -> symb #type #ordre X a #0 )
  EXLR					( symb #type #ordre X a )
  DUP SYMBINCOMP			( symb #type #ordre X a meta )
  ISMETACOMPLEX? case COMPLEXERR	( -> complex not allowed for a )
					( symb #type #ordre X a meta )
  NDROP POSINFext  			( symb #type #ordre X a #inf )
  #0<> case LIMATINF			( -> )
					( symb #type #ordre X a )
  LIMNOTATINF				( symb #newtype #ordre X a )
  4PICK #1= NOTcase LIMMEVAL+		( -> symb #newtype #ordre X a #0 )
					( symb #newtype #ordre X a )
  4ROLLDROP TWO 4UNROLL			( symb #2 #ordre X a )
  LIMMEVAL- 				( symb #2 #ordre X a #1 )
;

LABEL LIMNOTATINF
::				( symb #type #ordre X a )
  DUPTYPESYMB? NOT?SEMI		( -> symb #type #ordre X a )
  DUP EXPR>			( symb #type #ordre X a o1...on %n op )
  SWAP COERCE 			( symb #type #ordre X a o1...on op #n )
  ROTOVER #2=			( ... a o1...on-1 op #n on #n=2? )
  NOTcasedrop NDROP		( -> symb #type #ordre X a )
				( ... a o1...on-1 op #n on )
  Z0 EQUALNOTcase NDROP		( -> symb #type #ordre X a )
				( symb #type #ordre X a o1 op #2 )
  DROPDUP EQ: x+ 		( symb #type #ordre X a o1 op op=x+? )
  casedrop LIMUNIDIR+		( -> )
				( symb #type #ordre X a o1 op op=x+? )
  EQ: x- NOTcaseDROP		( -> symb #type #ordre X a )
				( symb #type #ordre X a o1 )
  SWAPDROP 4ROLLDROP 		( symb #ordre X o1 )
  ONE 4UNROLL 			( -> symb #1 #ordre X o1  )
;

LABEL LIMUNIDIR+
:: 				( symb #type #ordre X a o1 )
  SWAPDROP 4ROLLDROP 		( symb #ordre X o1 )
  TWO 4UNROLL 			( -> symb #2 #ordre X o1 )
;


LABEL LIMMEVAL+
::					( symb #type #ordre X a )
  2DUP OVER 				( symb #type #ordre X a X a X )
  x+ext					( symb #type #ordre X a X a+X )
  7ROLL UNROT MEVALext			( #type ordre X a symb' )
  5UNROLL ZERO				( symb' #type ordre X a #0 )
;

LABEL LIMMEVAL-
::					( symb #type #ordre X a )
  2DUP OVER 				( symb #type #ordre X a X a X )
  x-ext					( symb #type #ordre X a X a-X )
  7ROLL UNROT MEVALext			( #type ordre X a symb' )
  5UNROLL ONE				( symb' #type ordre X a #1 )
;

LABEL LIMATINF
::					( symb #type #ordre X a )
  4ROLLDROP INF>INF# 5ROLL		( #ordre X a # symb )
  4PICK 3PICK #1= ITE Z-1 Z1 		( #ordre X a # symb X +-1 )
  OVER x/ext MEVALext			( #ordre X a # symb )
  5UNROLL TWO 5UNROLL #4+ 		( symb #2 #ordre X a # )
;

LABEL INF>INF#
( '+-infinity' -> '+-infinity' #0|#1|#2 )
::
  DUPLENCOMP #2<>			( symb #len<>2? )
  case ZERO				( -> symb #0 )
  DUPTWO NTHCOMPDROP EQ: xNEG 		( symb ?=- )
  ITE ONE TWO				( -> symb # )
;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMITX!
::	
  Z0 Z1 2DUPSWAP TWO{}POLY DUP4UNROLL
  Z0 LAMLIMTYPE DUP#0= ITE_DROP :: #1= ITE "-" "+" >TAG ;
  RCLVX ONE{}N TRUE 
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMITNOVX!
::			( obj lvarnx )
  VX!			( obj lvarnx )
  OVERSWAP VXXLext	( lim lvarnx dl-l )
  Z0 Z1 		( lim lvarnx dl-l rest num/deno )
  3PICK 6ROLL 		( lvarnx dl-l rest num/deno equiv-l lim )
  6ROLL TRUE		( DL-l reste-l num-l/deno-l equiv.-l lim. lvar flag )
;

**********************************************************************
* Desc:		( var op DL-l rest-l num-l/deno-l equiv-l -> DL rest )
**********************************************************************
NULLNAME LIMERR0!
::	
  ExitAtLOOP 2DROP 2SWAP 2DROP  
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMERR1!
::	
  NDROP 3DROP Z0 Z1
  ?ext DUP NULL{} FALSE		( symb, 0, 0, ?, ?, null{}, false  )
  ( DL-l, reste-l, num-l/deno-l, equiv.-l, lim., lvar, flag )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
* symb -> DL-l, reste-l, num-l/deno-l, equiv.-l, lim., lvar, flag
* lim. = { symf direction }

NULLNAME LIMIT!
::
  { xINV xEXP x\v/ } 
  { addtINV addtEXPonly addtSQRTonly } 
  REPLACE2BY1

  RCLVX EQUALcasedro LIMITX!

  ( 1st step find var depending on X, recursive call on them * )

  LVARXNX2!			( symb lvarnx lvarx )
  DUPNULL{}? casedrop LIMITNOVX!		
				( symb lvarnx lvarx )
  DUPINCOMP DUP #1-SWAP
  ZERO_DO (DO)			( symb lvarnx lvarx var1...varn #n-1 )
     get1 LIMIT1VAR ROT #1+	( symb lvarnx lvarx var2...varn DL1 r1 #n )
  LOOP				( symb lvarnx lvarx DLr1...DLrn #2n-1 )
  LAMLIMOK #0<> case LIMERR1!
				( symb lvarnx lvarx DLr1...DLrn #2n-1 )
  ( * add lvarnx as DL rest * )
  #1+DUP #3+PICK LIMITADDLVARNX	( symb lvarnx lvarx DLr1...DLrn #2n )
  {}N				( symb lvarnx lvarx lDLr )
  4UNROLL			( lDLr symb lvarnx lvarx )
  SWAP &COMP			( lDLr symb lvarxnx )
  ROT tsimpflag TestSysFlag	( symb lvarxnx lDLr ? )
  ?SKIP :: COMPLEXEVAL TSIMP2ext ;		( symb lvarxnx lDLr )
  LVARXNX2!			( symb lvar_orig lDLr lDLvarnx lDLvarx )

  ( * reordering of lDLvarx * )

  LIMSORT!			( symb lvar_orig lDLr lDLvarnx lDLvarx )
* si 1ere var ne tend pas vers 0, pas de dvpt en series possible
  DUP CARCOMP RCLVX CONSTORSINCOS?  ( symb lvar_orig lDLrx DLvarnx DLvarx ? )
  case LIMITOUT0
  DUPLENCOMP OBJ>R		( symb lvar_orig lDLr lDLvarnx lDLvarx )
  SWAP &COMP			( symb lvar_orig lDLr lDLvarxnx )
  VXXLext			( symb lvar_orig lDLvar lDLr-liste )
  R>OBJ 3PICK DUP LIM#VARX! 2LAMBIND ( symb lvar_orig lDLvar lDLr-liste )

* calcul des equivalents des DL
  LIMITEQUIV			( symb lvar_orig lDLvar lDLrnde-list )
  2GETLAM 4UNROLL		( symb # lvar_orig lDLvar lDLrnde-liste )
  1GETABND 5UNROLL	( symb #varxnontrig #varx lvar_orig lDLvar lDLrnde )
  ' NULLLAM FIVE NDUPN DOBIND	( symb )

  ( lam1= lDLrnde, 2= lDLvar, 3=lvar_orig, 4= #varx, 5=#varxnontrig )
* sorte de VXXL su symb: mais on remplace les variables de symb en fonction
* de leur position dans lvar_orig par le DL-liste de la variable en fonction
* de lvar
  LIMSTEP1!			( [symb false | DL r-l n/d eq true] )
  NOTcase LIMITOUT
  				( DL-l reste-l num-l/deno-l equiv-l )
  4GETLAM 2GETLAM 3PICK LIMLIM!
  ABND
  SWAPTRUE
* DL-l, reste-l, num-l/deno-l, equiv.-l, lim., lvar, flag
;

LABEL LIMITADDLVARNX
* meta lvarnx -> meta
::
  DUPNULL{}? caseDROP		( -> meta )
				( meta lvarnx )
  DUPLENCOMP SWAP >R		( meta # )
  ZERO_DO (DO)
    RSWAP 'R RSWAP Z0 ROT #2+	( meta )
  LOOP
;

LABEL LIMITEQUIV
::				( lDLr )
  DUPNULL{}? ?SEMI		( -> )
  ZEROSWAP DUPLENCOMP #2/ 	( meta lDLr #n )
  SWAP >R			( meta #n )
  ZERO_DO (DO)			( meta )
    RSWAP 'R'R RSWAP		( meta DL r )
    OVER			( meta DL rest DL )
    1GETLAM LIMEQUFR!		( meta DL rest DL/equiv equiv { % % % } )
    DROP FOUR			( meta DL r n/d eq #4 )
    top&			( meta )
  LOOP
  {}N
;

LABEL LIMIT1VAR
* var -> DL rest
::					( var )
  idntlamcase Z0			( -> var 0 )			
  ( * var1 is symbolic )
  DUP LASTCOMP enumargs			( var op # )
  #1= NOTcasedrop LIMIT1VARNOTOK	( -> var 0 )
					( var op )
  DROPDUP INNERCOMP			( var o1...on #n )
  #1-SWAP DUP OBJ>R EQ: xINV		( var o1...on-1 #n-1 )

  ( * INV corresponds to INV[-LN[ * )
  IT :: #2- UNROT2DROP ; 

  CKSYMBN R>OBJ SWAP			( var op arg )
  LIMIT! 	( var op DL-l rest-l num-l/deno-l equiv-l lim lvar ok? )
  NOTcase2drop LIMERR0!
		( var op DL-l rest-l num-l/deno-l equiv-l lim lvar )
  7ROLL DUP	( var DL-l rest-l num-l/deno-l equiv-l lim lvar op op )
  LIMIT1VAROP EQLookup 
  ( var DL-l reste-l n/d-l equiv-l lim lvar op [ prg TRUE | obj FALSE ] )
  NOTcasedrop :: 7DROP LIMIT1VARNOTOK ;	( -> var 0 )
  EVAL					( var DL rest_relative ok? )
  ?SKIP ExitAtLOOP
  ROTDROP				( DL rest )
;

LABEL LIMIT1VARNOTOK
:: BINT40 ' LAMLIMOK STO Z0 ExitAtLOOP ;

LABEL LIMIT1VAROP
{ 
  xEXP LIMEXP! xINV LIMINVLN!
  xSIN LIMSINCOS! xCOS LIMSINCOS!
  xATAN LIMATAN! xASIN LIMASIN!
  xLN LIMLN! x\v/ LIMSQRT!
  xABS LIMABS! xFLOOR LIMFLOOR!
}


LABEL LIMITOUT0
:: 				( symb lvar_orig lDLeqre_nx lDLvarnx DLvarx )
  3DROP Z0 Z1 			( symb lvar_orig 0 1 )
  4PICK ?ext 			( symb lvar_orig 0 1 symb ? )
  4ROLL FALSE 			( symb 0 1 symb ? lvar false )
  SIXTYONE ' LAMLIMOK STO
;

LABEL LIMITOUT
:: 					( symb )
  Z0 Z1 3PICK ?ext 2GETLAM ABND FALSE 	( symb 0 1 symb ? lvar false )
;


**********************************************************************
* Desc:		Find DL of symb. LAM1-5 must be bound
* Stack:	( symb ->  DL-l, rest-l, num/deno-l, equiv.-l, TRUE )
*		( symb -> symb, false )
**********************************************************************
NULLNAME LIMSTEP1!
::					( symb )
  DUP 3GETLAM ListPos			( symb # )
  DUP#0=csedrp LIMSTEP2!
					( symb # )
  SWAPDROP FOUR #* #1+DUP #4-		( #*4+1 #*4-3 )
  DO
    1GETLAM INDEX@ NTHCOMPDROP
  LOOP
  TRUE					( DL-l rest-l num/deno-l equiv-l TRUE )
;

**********************************************************************
* Desc:		Find DL of symb. not in LAM3, LAM1-5 must be bound
* Stack:	( symb ->  DL-l, rest-l, num/deno-l, equiv.-l, TRUE )
*		( symb -> symb, false )
**********************************************************************
NULLNAME LIMSTEP2!
::					( symb )
  DUPTYPESYMB? NOTcase LIMSTEP3!	( -> )
					( symb )
  DUPLENCOMP #1=case LIMSTEP3!		( -> )
					( symb )
  DUP EXPR>				( symb arg1...argn %n, op )
  SWAP COERCE				( symb arg1...argn op #n )
  DUP#1= casedrop LIMSTEP4!
  DUP #2= NOTcase :: NDROP DROPFALSE ;
					( symb arg1 arg2 op #2 )
  ( * find DL of args * )
  DROP LIMSTEPBINARY EQLookup	( symb arg1 arg2 [prg TRUE | op FALSE] )
  NOTcase2drop DROPFALSE		( -> symb false )
				( symb arg1 arg2 prg )
  ROT LIMSTEP1!	( symb arg2 prg [ DL1 r1 n/d1 eq1 TRUE | arg1 FALSE ] )
  NOTcase2drop DROPFALSE		( -> symb FALSE : invalid op )
					( symb arg2 prg DL1 r1 n/d1 eq1 )
  6ROLL LIMSTEP1! ( symb prg DL1 r1 n/d1 eq1 [ DL2 r2 n/d2 eq2 tr | arg2 fa ] )
  NOTcasedrop :: 5DROP FALSE ;		( -> symb FALSE )
				( symb prg DL1 r1 n/d1 eq1 DL2 r2 n/d2 eq2 )
  9ROLL EVAL				( -> )
;

LABEL LIMSTEPBINARY
{ x+ LIM+! x- LIM-! x* LIM*! x/ LIM/! x^ LIMPOW! }

**********************************************************************
* Desc:		DL of non-symb or symb of length 1
* Stack:	( symb ->  DL-l, rest-l, num/deno-l, equiv.-l, TRUE )
*		( symb -> symb, false )
**********************************************************************
NULLNAME LIMSTEP3!
::					( % )
  VXXL1ext Z0 Z1 3PICK 
  EXACT? caseTRUE
  3DROP BINT67 ' LAMLIMOK STO FALSE
;	

**********************************************************************
* Desc:		DL of unary op
* Stack:	( symb arg op -> DL-l, rest-l, num/deno-l, equiv.-l, TRUE )
*		( symb arg op -> symb, false )
**********************************************************************
NULLNAME LIMSTEP4!
:: 					( symb arg op )
  LIMSTEPUNARY EQLookup			( symb arg [ prg TRUE | op FALSE ] )
  NOTcasedrop DROPFALSE			( -> symb false : invalid op )
					( symb arg prg )
  SWAP LIMSTEP1!		( symb prg [ DL r n/d eq TRUE | arg FALSE ] )
  NOTcasedrop DROPFALSE			( -> symb false )
					( symb prg DL r n/d eq )
  5ROLL EVAL				( -> )
;

LABEL LIMSTEPUNARY
{  xNEG LIMNEG! xINV LIMINV! x\v/ LIMRAC! xSQ LIMSQ! }

**********************************************************************
* Desc:
* Stack:	( #, lvar, equiv-l -> lvar, lim )
**********************************************************************
NULLNAME LIMLIM!
::
  VARCOMP3!
  FOUR #/			( #, lvar, lim-l, #reste, #quotient )
  5UNROLL
  DUP#0= ITE_DROP
  ::
    { Z0 INFINIext ?ext }
    ROTDROPSWAP NTHCOMPDROP EVAL
  ; 
				( #quotient, #, lvar, lim-l )
  ROT DUP#0= ITE_DROP n{}N	( #quotient, lvar, lim-l )
  OVER LIMVALOBJ!		( #quotient, lvar, lim )
  ::				( #quotient, lvar, lim )
    num#0=case LIMLIM1!
    ROT INFINIext
    3PICK EQUALNOTcase DROP
				( lvar, lim, # )
    TWO EQUALcasedro NOP
    ' xNEGNEG ' xNEG 
    ROT #0=ITE DROP SWAPDROP
    >TCOMP
  ;
				( lvar, lim )
; 

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME n{}N
:: 
  DUP#0=csDROP
  ZERO_DO (DO)
     ONE{}N 
  LOOP 
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMLIM1!
::	
  ROTDUP #2= caseDROP
				( lvar, lim, # )
  #0=ITE "+" "-" >TAG
;

**********************************************************************
* Desc:
* Stack:	( reste-1-l, reste-2-l -> reste-l )
**********************************************************************
NULLNAME LIMCMPL!
::	
  2DUP EQUALcase DROP
  num0case DROP
  SWAP num0case DROP
  SWAP2DUP RDIVext		( reste-1, reste-2, o-l )
  4GETLAM 2GETLAM ROT VARCOMP3!
  4UNROLL3DROP

  ( #, n/d, e, lvar, num-l/deno-l1, equiv-l1, reste-1, reste-2, # )
* #= 0 pour borne, 1 pour 0, 2 pour infini

  FOUR #/ DROP #2= ITE_DROP SWAPDROP
;

**********************************************************************
* Desc:
* Stack:	( n/d, # -> n/d-l, equiv, % )
**********************************************************************
NULLNAME LIMEQUFR!
::
  DUP#0=csedrp :: Z1 SWAP %0 ;
  SWAPFXND 3PICK		( n, d )
  LIMEQU!			( # n deno-l-trimmed deno-equiv-l { # # # } )
  4ROLL 5ROLL LIMEQU!		( den-l den-equiv-l den-# num-l num-equiv-l num-# )
  4ROLL				( num-{ # # # }, deno-{ # # # } -> { % % % } )
  SWAPINCOMP
  get1 INNERCOMP get1		( n1, ..., nk, d1, .., dk, #k, #k )
  ZERO_DO (DO)
     get1 UNCOERCE
     ROT UNCOERCE
     %- OVER #1+UNROLL
  LOOP
  {}N				( deno-l, deno-equiv-l, num-l, num-equiv-l, % )
  SWAP4ROLL RDIVext		( deno-l, num-l, %, equiv )
  ROT 4ROLL RDIVext		( %, equiv, n/d )
  SWAPROT			( n/d, equiv, % )
;

**********************************************************************
* Desc:
* Stack:	( {} # -->  {} / liste-equiv-l liste-equiv-l { # # # } )
**********************************************************************
NULLNAME LIMEQU!
::	
  OVER
  NOTLISTcase LIMEQU0!
				( liste, #, liste )
  NULL{} SWAPROT
  ZERO_DO			( liste, { # # # }, equiv-l )
     LIMVAL!			( liste, { # # # }, equiv-l, # )
     ROTSWAP >TCOMP
  SWAPLOOP
				( liste, { # # # }, equiv-l' )
  OVER LENCOMP ZERO_DO (DO)
     ONE{}POLY
  LOOP
  OVER INNERCOMP
  Z1 SWAP
  ZERO_DO (DO)
     Z0 ROT NDUPN #1+ {}POLY
  LOOP
  RMULText			( liste, { # # # }, equiv-l )
  ROTOVER RDIVext		( { # # # }, equiv-l, liste/liste-equiv-l )
  UNROTSWAP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMEQU0!
::	
  DROP Z1 UNROT			( 1, equiv, # )
  OVER Z0 EQUAL
  ITE MINUSONE ZERO
  SWAP NDUPN {}N 
;

* ====================================================================
*		operateurs rationnels
* ====================================================================

**********************************************************************
* Desc:		Add/Sub 2 DL
* Stack:	( symb DL1 r1 n/d1 eq1 DL2 r2 n/d2 eq2 ?+ -> )
*		( -> symb FALSE | DL r n/d eq TRUE )
**********************************************************************
NULLNAME LIM+-!
::				( symb DL1 r1 n/d1 eq1 DL2 r2 n/d2 eq2 ?+ )
  OBJ>R SWAPDROP QMul		( symb DL1 r1 n/d1 eq1 DL2 r2*eq2 )
  4ROLL DROP 2SWAP QMul		( symb DL1 DL2 r2*eq2 r1*eq1 )
  LIMCMPL!			( symb DL1 DL2 r )
  UNROT R>OBJ ITE QAdd QSub	( symb r DL )

* si DL=0, alors Insufficient order error 
* sauf si le reste est nul, cas qu'il faudrait implementer!!!

  num#0=case LIMERR10!
  ROTDROP			( r DL )
  DUP 5GETLAM LIMEQUFR!		( r_absolu DL nd e { # } )
  DROP 4ROLLOVER QDiv		( DL nd e r_relatif )
  UNROT TRUE			( DL r nd e true )
;

NULLNAME LIMERR10!
::				( symb r DL )
  2DROP FALSE			( symb false )
  INSUFFICIENTORDER
;

LABEL INSUFFICIENTORDER
::
  BINT65 ' LAMLIMOK STO
;

LABEL LIM+!
:: TRUE LIM+-! ;

LABEL LIM-!
:: FALSE LIM+-! ;

**********************************************************************
* Desc:		Neg a DL
* Stack:	( symb DL r n/d eq -> -DL r n/d -eq true )
**********************************************************************
NULLNAME LIMNEG!
::
  5ROLLDROP				( DL r nd e )
  QNeg 4ROLL QNeg 4UNROLL		( -DL r nd -e )
  TRUE					( -> )
;

**********************************************************************
* Desc:		Sqrt of a x-indep DL
* Stack:	( symb DL r n/d eq -> sqrt[eq] 0 1 sxqrt[eq] ) 
**********************************************************************
NULLNAME LIMRAC!
::
  4UNROLL3DROP SWAPDROP			( eq )
  SXSQRext Z0 Z1 3PICK TRUE		( DL r nd e true )
;

**********************************************************************
* Desc:		Invert a DL
* Stack:	( symb DL r n/d eq -> )
*		( -> symb FALSE | DL r n/d eq TRUE )
**********************************************************************
NULLNAME LIMINV!
::
  Z1 5UNROLL Z0 5UNROLL Z1 5UNROLL Z1 5UNROLL 
  LIM/!
;

**********************************************************************
* Desc:		Divide 2 DL
* Stack:	( symb DL1 r1 n/d1 eq1 DL2 r2 n/d2 eq2 -> )
*		( -> symb FALSE | DL r n/d eq TRUE )
**********************************************************************
NULLNAME LIM/!
::					( symb DL1 r1 nd1 e1 DL2 r2 nd2 e2 )
  5ROLL SWAP QDiv			( symb DL1 r1 nd1 DL2 r2 nd2 e1/e2 )
  ROT 6ROLL LIMCMPL!			( symb DL1 nd1 DL2 nd2 e r )
  4ROLLDROP 5ROLL DROP			( symb nd1 nd2 e r )
  2SWAP					( symb e r nd1 nd2 )
  5GETLAM DIVPC!			( symb e r n d flag )
  NOTcase2drop :: INSUFFICIENTORDER 2DROPFALSE ;	( -> symb false )
  5ROLL DROP QDiv			( e r nd )
  ROT2DUP QMul				( r nd e DL )
  4UNROLL TRUE				( DL r nd e true )
;

**********************************************************************
* Desc:		Integral power of a DL
* Stack:	( symb DL1 r1 n/d1 eq1 DL2 r2 n/d2 eq2 -> )
*		( -> symb FALSE | DL r n/d eq TRUE )
**********************************************************************
NULLNAME LIMPOW!
::
  4UNROLL			( symb DL1 r1 n/d1 eq1 eq2 DL2 r2 n/d2 )
  TRIMOBJext Z1 EQUAL		( symb DL1 r1 n/d1 eq1 eq2 DL2 r2 ? )
  NOTcase :: 7DROP FALSE ;	( -> symb false )
				( symb DL1 r1 n/d1 eq1 eq2 DL2 r2 )
  TRIMOBJext Z0 EQUAL		( symb DL1 r1 n/d1 eq1 eq2 DL2 ? )
  NOTcase :: 6DROP FALSE ;	( -> symb false )
				( symb DL1 r1 n/d1 eq1 eq2 DL2 )
  DROP TRIMOBJext		( symb DL1 r1 n/d1 eq1 eq2 )
  DUPTYPEZ? NOTcase :: 5DROP FALSE ;
				( symb DL1 r1 n/d1 eq1 z2 )
  DUP ZABS 7UNROLL 		( |z2| symb DL1 r1 n/d1 eq1 z2 )
  ZIsNeg? ITE LIMINV! TRUE	( z2 symb false | z2 symb DL r n/d eq true )
  NOTcase :: SWAPDROP FALSE ;
  5ROLLDROP			( z DL r n/d eq )
  4ROLL 5PICK RPext		( z r nd e DL )
  2SWAP				( z e DL r nd )
  5PICK RPext 4ROLL 5ROLL RPext	( DL r nd e )
  TRUE				( DL r nd e true )
;

**********************************************************************
* Desc:		Square of a DL
* Stack:	( symb DL r n/d eq -> DL^2 r n/d^2 eq^2 true )
**********************************************************************
NULLNAME LIMSQ!
::	
  4PICK 4PICK 4PICK 4PICK LIM*!
;

**********************************************************************
* Desc:		Multiplies 2 DL
* Stack:	( symb DL1 r1 n/d1 eq1 DL2 r2 n/d2 eq2 -> )
*		( -> symb FALSE | DL r n/d eq TRUE )
**********************************************************************
NULLNAME LIM*!
::				( symb DL2 r2 nd2 e2 DL1 r1 nd1 e1 )
  5ROLL QMul 7UNROLL		( symb e DL2 r2 nd2 DL1 r1 nd1 )
  4ROLL QMul 6UNROLL		( symb nd e DL2 r2 DL1 r1 )
  ROT LIMCMPL! 5UNROLL		( symb r nd e DL2 DL1 )
  QMul 				( symb r nd e DL )
  4UNROLL 5ROLLDROP TRUE	( DL r nd e true )
;

**********************************************************************
* Desc:
* Stack:	( #ordre, num-l, deno-l -> num-l, deno-l )
**********************************************************************
NULLNAME LIMDIVPC!
::
  CKINNERCOMP
  reversym {}POLY
  3PICK RRDMext				( #ordre, num-l, deno-l )
  SWAP CKINNERCOMP reversym {}POLY	( #ordre, deno-l, num-l )
  3PICK #2* RRDMext			( #ordre, deno-l, num-l )
  SWAP PSEUDODIV DROP
  CKINNERCOMP reversym >POLYTRIM	( #ordre, denolcm, quotient-l*denolcm )
  ROTDROPSWAP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME DIVPC!
::				( reste, nd1, nd2, # )

* division en puissance croissantes: reste, n, d, # -> reste, n, d, flag

  UNROT RDIVext FXNDext 
  ROTOVER TYPELIST? NOTcasedrop TRUE

				( reste-l,  num-l, deno-l, # )

* determine la profondeur courante et l'ordre pour le prochain DL

  FALSE LIMPROF!		( reste-l, num-l, deno-l, #prof, #ordre, flag )
  NOTcase2drop FALSE
  2SWAP				( reste-l, #prof, #ordre, num-l, deno-l )
  LIMDIVPC!			( reste-l, #prof, num-l, deno-l )
  ROT				( reste-l, num-l, deno-l, #prof )
  LIMPROFEND!
  TRUE
;

**********************************************************************
* Desc:
* Stack:	( num, deno, #prof -> num, deno )
**********************************************************************
NULLNAME LIMPROFEND!
::
  DUP#0=csDROP
  ZERO_DO (DO)
     SWAP ONE{}POLY SWAP ONE{}POLY
  LOOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMPROF!

* reste-l, num-l, deno-l, #varx, ?num-l -> 
* reste-l, num-l, den-l, #prof, #ordre, flag
* determine la profondeur de la variable par rapport a laquelle on peut faire
* un developpement et l'ordre dudit developpement
* flag est FALSE s'il n'est plus possible de faire de developpement, sinon 
* la profondeur est l'ordre de la 1ere variable par rapport a laquelle
* le reste ET num-l ont une valuation non nulle si ?num-l est TRUE
* ou le reste seul a une valuation non nulle si ?num-l est FALSE
* si #profondeur[num-l]>#profondeur[reste] ou si #valuation[reste]>#limordre
* alors reste devient { ... { 1 0 ... 0 }... } avec #limordre 0 
* et max[#profondeur] { }
* sinon reste est inchange

::				( rest-l num-l deno-l #varx ?num-l )
  4ROLL num0case LIMPROF0!	( rest deno # ?num-l num )
  4ROLL 5PICK 5PICK		( rest # ?num-l num deno rest # )
  LIMEQUFR! UNROT2DROP		( rest-l # ?n-l n-l deno-l { % % % } )
  LIM%#!			( rest-l # ?n-l n-l deno-l #prof { %%% } )
  CARCOMP DUP %0>		( rest-l # ?n-l n-l deno-l #prof % ? )
  NOTcasedrop ::		( rest-l # ?n-l n-l deno-l #prof )
		4ROLLDROP 4ROLL FALSE ( rest-l n-l deno-l #prof #ordre false )
		;
  COERCE		( rest-l # ?num-l num-l deno-l #prof-r #ordre-r )
  4ROLL 6ROLL 4PICK #-	( rest-l ?num-l deno-l #prof-r #ordre-r num-l #varx )
  6ROLL ITE
  ::			( rest-l deno-l #prof-r #ordre-r num-l #varx )
    OVERSWAP LIMEQUFR!
    UNROT2DROP		( rest-l deno-l #prof-r #ordre-r num-l { % % % } )
    5ROLL SWAP
    LIM%#! DROP		( rest-l #prof-r #ordre-r num-l deno-l #prof-n )
  ;
  ::			( rest-l deno-l #prof-r #ordre-r num-l #varx )
    DROP 4ROLL ZERO	( rest-l #prof-r #ordre-r num-l deno-l #prof-n )
  ;

* si prof-n=0: si #ordre-r > #limordre -> LIMPROF2!, sinon on ne change rien

  DUP#0=csedrp
  ::			( rest-l #prof-r #ordre-r num-l deno-l )
    2SWAP		( rest-l num-l deno-l #prof #ordre )
    LAMLIMORDRE #1+
    2DUP#> 		( rest-l num-l deno-l #prof #ordre #limordre ?> )
    NOTcase DROPTRUE
			( r-l num-l den-l #prof #ordre #limor )
    SWAPDROP#1- LIMPROF2!
  ;
			( rest-l #prof-r #ordre-r num-l deno-l #prof-n )

* si prof-n<>0, alors prof=prof-r+prof-n et -> LIMPROF2!

  4ROLLDROP 4ROLL #+ LAMLIMORDRE LIMPROF2!
;

**********************************************************************
* Desc:
* Stack:	( num-l, deno-l, { % ... % }
*		  --> new-num-l, new-deno-l, #profondeur, { % % % } )
**********************************************************************

* determine la profondeur: le 1er element non nul de { % % % }
* et reajuste en consequence la profondeur de num-l et deno-l

NULLNAME LIM%#!
::
  ZEROSWAP			( num-l, deno-l, #prof, { %%% } )
  DUP' NONZERO Find1stTrue	( num-l deno-l #prof { %%% } [obj true|false] )
  NOT?SEMI
  DROP
  BEGIN				( num-l, deno-l, #prof, { %%% } )
    DUP CARCOMP %0=
  WHILE
    CDRCOMP SWAP#1+SWAP 	( num-l, deno-l, #prof, { %%% } )
    LIMPROF1! LIMPROF1! 2SWAP
  REPEAT
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMPROF0!
::				( reste, deno, #, ?num-l, 0 )
  4UNROLL 2DROP
  ZERO LAMLIMORDRE TRUE		( reste-l, num-l, den-l, #prof, #ordre, flag )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMPROF1!
::
  4ROLL DUPTYPELIST? NOT?SEMI
  DUPLENCOMP NTHCOMPDROP 
; 

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMPROF2!
::				( reste-l, num-l, deno-l, #prof, #limor )
  5ROLL DROP			( num-l, deno-l, #prof, #limor )
  Z1Z0 3PICK #1+ NDUPN
  #1+ {}POLY			( num-l, deno-l, #prof, #limor, { 1 0 ... 0 } )
  3PICK n{}N
  5UNROLL TRUE
;

* ====================================================================
*		operateurs ln, exp, sin, cos, etc.
* ====================================================================

* reste , DL-l, num-l/deno-l, equiv, lim, lvar, operateur -> DL, reste, flag
* pour les operateurs courants ln, exp, ... ( attention le reste est relatif )

* Note pour les DL non asymptotiques
* si x_0 = 0, f(x(1+r))= [\sigma _{k=k_0} ^{k_0+n} f[k](0) x^k ](1+r)
* ou n est tel que x^n =O(r) [ si r=0, on prend n=LIMORDRE er r=x^n ]
* si x_0 est non nul et f(x_0) non nul, r reste constant avec le choix
* de n ci-dessus
* si f(x_0) est nul, il faut changer r en r/e ou e est l'equivalent de x
* et on se ramene au 1er cas avec f[k](x_0) au lieu de f[k](0)
* en pratique, on a les cas particuliers suivants:
* ln: x_0=1, k_0=1
* sin: x_0 = n \pi (n \neq 0), k_0=1 (diviser r par e)
* sin: x_0 = 0, k_0=1 
* cos: x_0 = (n+1/2)*pi, k_0=1 (diviser r par e)
* atan: x_0 = 0, k_0=1
* Pour les DL asymptotiques

**********************************************************************
* Desc:		operateur INV[-LN]
* Stack:
**********************************************************************
NULLNAME LIMINVLN!
::
  LIMLN!
  Z1 4ROLL RDIVext UNROT		( ln, reste, ?ok )
;

**********************************************************************
* Desc:		operateur LN
* Stack:
**********************************************************************
NULLNAME LIMLN!
::
  DROPSWAP		( DL-l rest-l num-l/deno-l equiv-l lvar lim )
  6ROLL DROP		( rest-l num-l/deno-l equiv-l lvar lim )

* tester si la limite est negative -> erreur

  ZSIGNECK SWAPDROP		( rest-l num-l/deno-l equiv-l lvar flag )
  NOTcase
  :: BINT64 ' LAMLIMOK STO 4DROP ?ext %0 FALSE ;

			( rest-l num-l/deno-l equiv-l lvar )
  ROT Z1 RSUBext	( rest-l equiv-l lvar num-l/deno-l )
  4ROLLSWAP		( equiv-l lvar rest-l  num-l/deno-l )
  TRUE LPROF!		( equiv-l #varx #prof #ordre lvar new rest-l num-l/deno-l )
  LAMLIMOK #0<> 
  case
  :: 7DROP ?ext %0 FALSE ;	( ln rest ok? )

			( equiv-l #varx #prof #ordre lvar new rest-l num-l/deno-l )
  SWAP3PICK LIMVALOBJ!
  UNROT			( equiv-l #varx #prof #ordre new rest lvar num-l/deno-l )
  7ROLL

* a faire: si equiv-l=1 alors chercher la valuation de num-l/deno-l 
* pour augmenter limordre

  UNROT

  ( #varx, #prof, #ordre, reste relatif, equiv-l, lvar, num-l/deno-l )

* developpement de Taylor du ln applique a X 

  ( X -> X-X^2/2... )

  5ROLL DUP1LAMBIND HORNLN!	( X, c0, ..., cn )
  1GETABND #1+ LIMHORN!		( #varx #prof rest equiv-l lvar l-ln[1+X] )

* ajustement de la profondeur

  FXNDext 6ROLL LIMPROFEND! NDXFext

* fin du dvpt de Taylor

			( #varx, reste, equiv-l, lvar, l-ln[1+X] )
  ROT ZEROSWAP 4PICK	( #varx, reste, lvar, l-ln[1+X], equiv-l, lvar )
  VAL2ext		( #varx, reste, lvar, l-ln[1+X], equiv )
  LNOBJ!		( #varx, reste, lvar, ln[1+X]-l, ln[equiv], taille )
  DUP Z0 EQUAL ITE
  ::
    DROPOVER		( #varx, reste, lvar, ln[1+X]-l, ln[equiv], ln[1+X]-l )
    6ROLL LIMEQUFR!	( reste, lvar, ln[1+X]-l, ln[equiv], o, taille-l, o )
    DROPSWAPDROP ZEROSWAP
    5PICK VAL2ext
  ;
  ::
    6ROLL DROP
  ;

  ( reste, lvar, ln[1+X], ln[equiv], taille ou X^[valuation] )

  5ROLL SWAP x/ext	( lvar, ln[1+X], ln[equiv], nouveau reste )

* simplification du nouveau reste

  SIMP1!		( lvar, ln[1+X], ln[equiv], nouveau reste )
  4UNROLL		( reste, lvar, ln[1+X], ln[equiv] )
  UNROTSWAP		( reste, ln[equiv], ln[1+X], lvar )

* methode de Horner

  SWAPFXND			( lvar, num-l, deno-l )
  CKCARCOMP 
  OVER TYPELIST? ITE
    >HPOLY
    :: SWAP TWO{}POLY ; 
  ONESWAP
  3PICK CDRCOMP
  VAL2ext			( lvar, { deno num-l } )

  INNERCOMP
  #1- reversym
  DUP #3+ ROLL CARCOMP
  1LAMBIND			( deno, a0, .., an, #n+1)
  ::
    DUP#1= casedrop ABND
				( a0, ..., an, #n+1 )
    ONE_DO (DO)
       1GETLAM 
       x*ext x+ext
    LOOP
    ABND
  ;
				( deno, a0, .., an, #n+1)
  SWAP RDIVext			( reste, ln[equiv], ln[1+X] )

* fin de la methode de Horner

  x+ext SWAPTRUE		( ln, reste, ok? )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMEXP!
::

* DL-l, reste-l , num-l/deno-l, equiv-l, lim, lvar, operateur 
* -> DL, reste, flag

  LIMDL! caseTRUE
  LIMDLINF!			( ?-inf { a#...a0 } deno-l lim lvar X reste-l )
  LAMLIMOK #0<> case
  :: 7DROP %0 %0 FALSE ;
  HORNEXP!			( ?-inf { a# ... a0 } deno-l lvar DL reste-l )
  6UNROLL 2SWAP
  5PICK ?SKIP RNEGext
  RDIVext			( reste-l  ?-inf lvar DL partie->infini-liste )
  3PICK LIMVALOBJ! xEXPext	( reste-l  ?-inf lvar DL partie->infini )
  4ROLL ITE x*ext NDXFext	( reste-l lvar exp )
  UNROT LIMVALOBJ!		( DL reste )
  TRUE
;
 
**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMSINCOS!
::

* DL-l, reste-l , num-l/deno-l, equiv-l, lim, lvar, operateur 
* -> DL, reste, flag

  LIMDL! caseTRUE		( DL-l rest-l num-l/den-l equiv-l lim lvar op )
  DUP1LAMBIND LIMDLINF!
  1GETABND			( ?-inf {a#..a0} den-l lim lvar X, rest-l op )
  8ROLL DROP
  7ROLL 7ROLL RDIVext		( lim lvar X rest-l op partie_infinie-list )
  5PICK LIMVALOBJ!		( lim lvar X rest-l op partie_infinie )
  6ROLL x+ext 5UNROLL
  EQ: xCOS
  ITE HORNCOS! HORNSIN!		( lvar, DL, r-l )
  ROT LIMVALOBJ! TRUE		( DL, reste, flag )
;
 
**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMATAN!
::				( DL-l rest-l num-l/den-l equiv-l lim lvar op )
  LIMDL! caseTRUE		( DL-l rest-l num-l/den-l equiv-l lim lvar op )

  ( x -> 1/x )
  DROPSWAP LIMINFSIGN!		( DL-l rest-l num-l/den-l equiv-l lvar ?-+ )
  pisur2 SWAP IT RNEGext	( DL-l rest-l num-l/den-l equiv-l lvar +-pi/2 )
  6ROLL DROP			( rest-l  num-l/den-l equiv-l lvar +-pi/2 )
  5ROLL 5ROLL			( equiv-l lvar +/-pi/2 rest-l num-l/den-l )
  FXNDext SWAP			( equiv-l lvar +/-pi/2 rest-l den-l num-l )

  5PICK LVARXNXext
  4UNROLL3DROP LENCOMP DIVPC!	( equiv-l lvar +-pi/2 rest-l num-l den-l flag )
  NOTcase
  :: 3DROP UNROT2DROP FALSE 	( DL rest flag )
     BINT56 ' LAMLIMOK STO
  ;
				( equi-l lvar +-pi/2 rest-l num-l/den-l )
  RDIVext Z0 5ROLL ROT		( equi-l +-pi/2 rest-l lim=0 lvar num-l/den-l )
  6ROLL RDIVext 4ROLL HORNATAN!	( +-pi/2 lvar DL rest-l )
  ROT LIMVALOBJ!
  UNROT x-ext SWAPTRUE		( DL rest flag )
;
	
**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMASIN!
::			( DL-l rest-l  num-l/deno-l equiv-l lim lvar op )
  LIMDL! caseTRUE	( DL-l rest-l  num-l/deno-l equiv-l lim lvar op )
  5DROP FALSE 
  BINT57 ' LAMLIMOK STO
* erreur: asin non defini en +/- infini
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMSQRT!
::				( DL-l rest-l num-l/den-l equiv-l lim lvar op )
  DROPSWAPDROP
  5ROLL DROP			( rest-l num-l/den-l equiv-l lvar )
  2SWAP Z1 RSUBext
  onehalf TOTEMPOB
  LIMBETA!
  SWAPROT LIMVALOBJ! TRUE	( DL rest true )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMFLOOR!
::

* DL-l, reste-l, num-l/deno-l, equiv-l, lim, lvar, operateur -> 
* DL, reste, TRUE
   				( Dl-l rest num/deno equiv lim lvar op )
  3PICK TYPEZINT? 
  NOTcase2drop LIMFLOORNOTZ
				( Dl-l rest num/deno equiv lim lvar op )
* check if limit is reached from the right or left
  DROP OBJ>R			( Dl-l rest num/deno equiv lim )
  5UNROLL SWAP 4UNROLL3DROP	( lim num/deno )
  OVER QSub			( lim num/deno-lim )
  FXNDext 			( lim n d )
  DUPTYPELIST? IT LASTCOMP	( lim n obj )
  SWAP LIMVAL!			( lim obj coeff #valuation )
  TWO #/ DROP			( lim obj coeff #valuation_mod_#2 )
  #1= LAMLIMTYPE #0= 		( lim obj coeff ? ? )
  ANDcase SIGNEERROR		( -> )
				( lim obj coeff )
  QMul				( lim obj*coeff-list )
  R>OBJ CDRCOMP SWAP R2SYM	( lim obj*coeff )
  ZSIGNECK SWAPDROP		( lim ?>0 )
  ?SKIP :: Z1 QSub ;	( lim )
  Z0 TRUE			( lim-1 rest=0 true )
;

LABEL LIMFLOORNOTZ
::
  SYMFLOOR			( Dl-l rest num/deno equiv lim )
  DUPTYPEREAL? IT R>Z
  5UNROLL 4DROP Z0 TRUE		( lim_as_DL rest=0 true )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMABS!
::

* DL-l, reste-l, num-l/deno-l, equiv-l, lim, lvar, operateur -> 
* DL, reste, TRUE

  DROPSWAP ZSIGNECK	( DL-l rest-l num-l/den-l equiv-l lvar lim flag )

* attention: si la limite est nulle non signee

  SWAP num0=case
  :: 5DROP BINT66 ' LAMLIMOK STO FALSE ;
  DROP 2SWAP 2DROP 2SWAP		( lvar, flag, DL-l, reste-l )
  ROT ?SKIP
  :: RNEGext SWAP RNEGext SWAP ;	( lvar, DL-l, reste-l )
  SWAP3PICK LIMVALOBJ!			( lvar, reste-l, DL )
  SWAPROT LIMVALOBJ! TRUE
;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LPROF!
::

* lvar rest-l num-l/den-l flag -> 
* #varx #prof #ord lvar new rest-l num-l/den-l 

  1LAMBIND
  FXNDext			( lvar rest-l  num-l den-l )

* detecte la 1ere var ne contenant pas x

  4PICK DUP LIM#VARX!		( lvar rest-l  num-l den-l #varx )
  DUP 6UNROLL 1GETABND
  LIMPROF!			( #varx lvar rest-l num-l den-l #prof #ord flag )
  ?SKIP
  :: BINT62 ' LAMLIMOK STO ;
  6UNROLL 6UNROLL NDXFext	( #varx #prof #ord lvar new-rest-l num-l/den-l )
;

**********************************************************************
* Desc:		First var of lvar that does not depend of X or is trig
* Stack:	( lvar, lvar -> #varx )
**********************************************************************
NULLNAME LIM#VARX!
::	
  RCLVX ' CONSTORSINCOS? POSCOMP	( lvar # )
  DUP#0=csedrp LENCOMP			( -> #lvar )
  DUP#1= case INTERNALERR
  SWAPDROP#1-				( -> #-1 )
;

LABEL CONSTORSINCOS?
( obj X -> flag )
* flag is true if obj is cst or sin/cos function
::						( obj X )
  OVER TYPESYMB? NOTcase EQUALNOT		( -> ? )
						( symb X )
  OVER LASTCOMP { xSIN xCOS } ListPos		( symb X # )
  #0= NOTcase2drop TRUE
						( symb X )
  EQUALPOSCOMP #0=
;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMBETA!
LABEL limbeta!
::				( equiv-l lvar reste-l num-l/deno-l alpha )
  LIMALPHA!			( equiv-l alpha lvar new_reste-l DL-l )
  3PICK LIMVALOBJ!		( equiv-l alpha lvar new_reste-l DL )
  5ROLL 4PICK 			( alpha lvar new_reste-l DL equiv-l lvar )
  LIMVALOBJ!			( alpha lvar new_reste-l DL equiv )	
  5ROLL 			( lvar new_reste-l DL equiv alpha )
*  ZEROSWAP NULL{} VAL2ext ( lvar, new reste-l, DL, equiv, alpha )
  EXPAND^			( lvar, new reste-l, DL, equiv^alpha )

* probleme avec sqrt[x] ou abs[x] -> variable non valide et limite = ?
				( lvar, new reste-l, DL, equiv^alpha )
  { x^ xABS x\v/ }
  { meta^meta limabs seriessqrt }
  REPLACE2BY1		( lvar, new reste-l, DL, equiv^alpha )
  x*ext			( lvar rest-l equiv^alpha*DL[1+num-l/deno-l-1]^alpha )
;

LABEL limabs
:: 						( meta )
  dup CKSYMBN 
  LIMIT! 			( meta dl rest num/den eq lim lvar flag )
  2DROP 5UNROLL 4DROP 				( meta lim )
  ZSIGNECK 					( meta lim ? )
  caseDROP					( -> meta ) 
  DROP metaneg					( -> -meta ) 
;


**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMALPHA!
LABEL limalpha!
::				( lvar rest-l num-l/den-l-1=X alpha )
  4UNROLL			( alpha  lvar rest-l num-l/den-l-1=X )
  Z0 EQUALcasedro Z1		( -> alpha lvar rest-l 1 )
  TRUE LPROF!			( alpha #varx #prof #ordre lvar new-rest-l X )
  Z1 8PICK 6ROLL
  DUP1LAMBIND	( alpha #varx #prof #ordre lvar new-rest-l X 1 mult #ordre )
  #1+_ONE_DO (DO)		( ck mult )
     2DUP RMULText		( ck mult ck*mult )
     INDEX@ #>Z RDIVext		( ck mult ck+1=ck*mult/k )
     SWAP Z1 RSUBext
  LOOP
  DROP
  1GETABND #1+ LIMHORN!		( alpha #varx #prof  lvar new-rest-l DL )
  FXNDext 5ROLL LIMPROFEND!
  RDIVext 4ROLL DROP		( alpha lvar new-rest-l DL-l )
;

**********************************************************************
* Desc:
* Stack:	( lim, lvar, X-l, reste-l -> lvar, DL, reste-l )
**********************************************************************
NULLNAME HORNEXP!
LABEL hornexp!
::					( lim lvar X rest-l ) 
  4UNROLLROT
  DUPTYPETAG? IT STRIPTAGS
  UNROT					( rest-l lim lvar X )
  num0case :: DROPSWAP xEXPext ROT ;
					( rest-l lim lvar X )
  4ROLLSWAP				( lim lvar rest-l num-l/den-l )
  TRUE LPROF!		( lim #varx #prof #ordre lvar new rest-l num-l/den-l )
  Z1 DUP
  6ROLL DUP1LAMBIND
  #1+_ONE_DO		( c0 ... ck k! )
      INDEX@ #>Z RMULText
      Z1 OVER NDXFext
  SWAPLOOP
  DROP			( lim #varx #prof lvar rest-l X c0 ... ck ) 
  1GETABND #1+ LIMHORN!	( lim #varx #prof lvar  rest-l f[x_0+X]-f[x_0]-l )
  FXNDext
  5ROLL LIMPROFEND!
  NDXFext		( lim #varx lvar rest-l  f[x_0+X]-f[x_0]-l )
  4ROLLDROP		( lim lvar  rest-l f[x_0+X]-f[x_0]-l )
  3PICK LIMVALOBJ!	( lim lvar rest-l f[x_0+X]-f[x_0] )
  4ROLL xEXPext x*ext	( lvar rest-l f[x_0+X]-f[x_0] )
  SWAP			( lvar DL rest-l )
;

**********************************************************************
* Desc:
* Stack:	( lim, lvar, X-l, reste-l -> lvar, DL, reste-l )
**********************************************************************
NULLNAME HORNCOS!
::				( lim lvar X rest-l )
  SWAP				( lim lvar rest-l X )
  num0case
  :: DROPROT xCOSext SWAP ;

* si cos[x_0]=0, rajouter 1 a l'ordre du DL a effectuer

  4PICK xCOSext DUP 5UNROLL	( lim cos[lim] lvar rest-l X cos[lim] )
  Z0 EQUAL			( lim cos[lim] lvar rest-l X ?=0 )
  LIMSC0!			( lim cos[lim] r-l lvar X sin-l cos-l ) 
  2OVER ROT 8ROLL		( lim r-l lvar X sin-l lvar X cos-l cos[lim] )
  LIMSC1!			( lim r-l lvar X sin-l cos[lim]*cos[X] )
  UNROT 4PICK UNROT
  7ROLL xSINext		( r-l lvar sin[lim]*cos[X] lvar X sin-l sin[lim] )
  LIMSC1! x-ext			( r-l lvar cos[lim]*cos[X]-sin[lim]*sin[X] )
  ABND ROT			( lvar DL r-l )
;

**********************************************************************
* Desc:		SIN of a DL with finite limit
* Stack:	( lim, lvar, X-l, reste-l -> lvar, DL, reste-l )
**********************************************************************
NULLNAME HORNSIN!
::				( lim lvar X rest-l )
  SWAP				( lim lvar rest-l X )
  num0case :: DROPROT xSINext SWAP ;	( -> lvar sin[limit] rest-l )

* si sin[x_0]=0, rajouter 1 a l'ordre du DL a effectuer
				( lim lvar rest-l X )
  4PICK xSINext DUP 5UNROLL	( lim sin[lim] lvar rest-l X sin[lim] )
  Z0 EQUAL			( lim sin[lim] lvar rest-l X ?=0 )
  LIMSC0!			( lim sin[lim] r-l lvar X sin-l cos-l ) 
  2OVER ROT 8ROLL		( lim r-l lvar X sin-l lvar X cos-l sin[lim] )
  LIMSC1!			( lim r-l lvar X sin-l sin[lim]*cos[X] )
  UNROT 4PICK UNROT
  7ROLL xCOSext		( r-l lvar sin[lim]*cos[X] lvar X sin-l cos[lim] )
  LIMSC1! x+ext			( r-l lvar sin[lim]*cos[X]+cos[lim]*sin[X] )
  ABND ROT			( lvar DL r-l )
;

**********************************************************************
* Desc:		Find DL of sin and cos at the right order
* Stack:	( lvar r-l X ? -> r-l lvar X sin-l cos-l )
* Note:		? is true if one must add 1 to the order
*		Binds #prof to lam1
**********************************************************************
NULLNAME LIMSC0!
::				( lvar rest-l X ? )
  4UNROLL			( ? lvar rest-l X )
  TRUE LPROF!			( ? #varx #prof #ordre lvar new-rest-l X )
  6ROLL DROP ROTSWAP		( ? #prof #ordre r-l lvar X )
  Z1Z0 2DUPSWAP DUP		( ? #prof #ordre r-l lvar X 1 0 0 1 1 )
  9ROLL				( ? #prof r-l lvar X 1 0 0 1 1 #ordre )
  ELEVEN ROLL			( #prof r-l lvar X 1 0 0 1 1 #ordre ? )
  IT #1+ DUP1LAMBIND		( #prof r-l lvar X 1 0; 0 1 1! # )
  DUPTWO #<case
  :: 2DROP TWO{}POLY UNROT TWO{}POLY 6ROLL 1PUTLAM ;
  ONE_DO (DO)				( cos sin k! )
     INDEX@ #1+ #>Z RMULText		( cos sin +-k! )
     INDEX@ TWO #/ DROP
     #0= DUPUNROT ?SKIP RNEGext		( cos sin flag +-k! )
     Z1 OVER x/ext			( cos sin flag +-k! +-1/k! )
     3PICK ITE
        UNROT
        :: INDEX@ #4+ UNROLL ;
     Z0 ROT ITE
        :: INDEX@ #3+ UNROLL ;
        SWAP
     LOOP
     DROP 1GETLAM #1+ {}POLY
   1GETABND #1+ DUPUNROT
   #2+UNROLL {}POLY		( #prof r-l lvar X sin-l cos-l )
   6ROLL 1LAMBIND		( r-l lvar X sin-l cos-l )
;

**********************************************************************
* Desc:		Utility for SIN and COS
* Stack:	lvar X-l P-l coeff -> coeff*P[X]
* Note:		lam1 contains #prof
**********************************************************************
NULLNAME LIMSC1!
::					( lvar X-l P-l coeff )
  num0case 4UNROLL3DROP			( -> 0 )
  4UNROLL				( coeff lvar X-l P-l )
  CKINNERCOMP LIMHORN!			( coeff lvar P[X]-l )
  FXNDext 1GETLAM LIMPROFEND! NDXFext	( coeff lvar P[X]-l )
  SWAP LIMVALOBJ! x*ext			( coeff*P[X] )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME HORNATAN!
::			( lim lvar X-l rest-l )
  LIMATAS!
  7PICK 6PICK VXXLext
  SWAPDROP		( lim #prof lvar new-rest-l X #ord num lim-l )
  DUPDUP RMULText
  Z1 RADDext
  SWAP Z2 RMULText	( lim #prof lvar new-rest-l X #ord num 1+l^2 2l )
  Z1Z0 6ROLL		( lim #prof lvar new-rest-l X num 1+l^2 2l 1 0 #ord )
  TWO 2DUP#<
  IT 2DROP00 #-		( lim #prof lvar new-rest-l X num 1+l^2 2l 1 0 #ord-2 )
  NDUPN #3+ {}POLY		( lim #prof lvar rest-l X num 1+[X+lim]^2 )
  PSEUDODIV DROP
  CKINNERCOMP reversym {}N	( lim #prof lvar rest-l X dl-d dl-n )
  INTEGRext SWAPNDXF RISCH13	( lim #prof lvar rest-l X dl )
  SWAPFXND 6ROLL LIMPROFEND!	( lim lvar rest-l dl X-n X-d )
  RDIVext
  SWAPINCOMP reversym LIMHORN!	( lim lvar rest-l dl )
  3PICK LIMVALOBJ!
  4ROLL xATANext x+ext SWAP	( lvar DL rest-l )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMATAS!
::
  SWAP				( lim lvar rest-l X )
  num0case
  :: DROPROT xATANext SWAP ;
  TRUE LPROF!			( lim #varx #prof #ordre lvar new-rest-l X )

* si lim=0 rajouter 1 a l'ordre du DL a effectuer

  6ROLL DROP			( lim #prof #ordre lvar new-rest-l X )
  4ROLL 6PICK
  Z0 EQUAL ?SKIP #1-		( lim #varx #prof lvar new-rest-l X #ordre )
  Z1Z0
  3PICK #2* NDUPN #1+ {}POLY	( lim #prof lvar new-rest-l X #ordre num )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME HORNASIN!
::				( lim lvar X-l rest-l )
  ROTSWAP			( lim X-l lvar rest-l )
  4PICK 3PICK VXXLext SWAPDROP	( lim X-l lvar rest-l lim-l )
  DUPDUP RMULText Z1
  EQUALcase HORNASIN1!
				( lim X-l lvar rest-l lim-l )
  DUPDUP RMULText Z1 RSUBext	( lim X-l lvar rest-l l l^2-1 )
  DUP RNEGext 5UNROLL		( lim X-l 1-l^2 lvar rest-l l l^2-1 )
  SWAP Z2 RMULText		( ... lvar rest-l l^2-1 2l )
  OVER RDIVext ONE>POLY		( ... lvar rest-l l^2-1 -2l/[1-l^2] )
  Z1 ROT RDIVext ONE>POLY	( ... lvar rest-l -2l/[1-l^2] -1/[1-l^2] )
  INTERNALX RMULText RADDext
  INTERNALX RMULText		( ... lvar rest-l num-l/den-l-1=-[2lx+x^2]/[1-l^2] )
  minusonehalf TOTEMPOB		( ... lvar rest-l num-l/den-l-1 alpha )
  LIMALPHA!			( ... alpha lvar rest-l DL[ 1+ num-l/den-l-1 ]^alpha-l )
  FXNDext CARCOMPext
  SWAP INTEGRext
  SWAP' RDIVext LOP1ext		( lim X-l 1-l^2 alpha lvar rest-l DL[ 1+ num-l/den-l-1 ]^alpha-l )
  6ROLL SWAP CKINNERCOMP reversym
  LIMHORN!			( lim 1-l^2 alpha lvar rest-l DL-l )
  3PICK LIMVALOBJ!
  5ROLL 4PICK LIMVALOBJ!	( lim lvar rest-l DL-l 1-l^2 )
  LNOBJ! DROP
  5ROLL x*ext xEXPext
  ZEROSWAP NULL{}
  VAL2ext x*ext			( lim lvar rest-l DL-l )
  4ROLL xASINext
  HSECO2RCext x+ext SWAP	( lvar DL rest-l )
; ( end hornasin! )

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME HORNASIN1!
::
  ( lim-l=+/-1 )		( lim X-l lvar rest-l lim-l )
  Z0 TWO{}POLY Z2 NDXFext	( lim X-l lvar rest-l num/den-1 )
  minusonehalf TOTEMPOB		( lim X-l lvar rest-l num-l/den-l-1=X alpha )
  LIMALPHA!
		( lim X-l alpha lvar rest-l DL[ 1+ num-l/den-l-1 ]^alpha-l )

* version speciale de INTEGRext: toutes les puissances sont decalees de -1/2

  FXNDext 1LAMBIND CKINNERCOMP #1+DUP
  ONE_DO (DO)
     SWAPINDEX@ #>Z
     minusonehalf TOTEMPOB
     RADDext
     1GETLAM RMULText
     RDIVext
     OVER UNROLL
  LOOP
  ABND Z0 SWAP {}POLY

		( lim X-l alpha lvar rest-l DL[ 1+ num-l/den-l-1 ]^alpha-l )

  5PICK SWAPINCOMP reversym LIMHORN!	( lim X-l alpha lvar rest-l DL-l )
  3PICK LIMVALOBJ!			( lim X-l alpha lvar rest-l DL )

* multiplier par ( X/equ[X] )^alpha-l
* puis selon lim par (-/+ 2 equ[X])^alpha

  6UNROLL 4ROLL 3PICK DUP LIM#VARX!	( DL lim alpha lvar rest-l X-l #varx )
  LIMEQUFR! DROP		( DL lim alpha lvar rest-l X-l/equiv equiv-l )
  6ROLL
  ZIsNeg? DUP
  ITE pisur-2 pisur2
			( DL alpha lvar rest-l X-l/equiv equiv-l sign +-pi/2 )
  8UNROLL
  ITE Z2 Z-2
  RMULText		( +-pi/2 DL alpha lvar rest-l X-l/equiv equiv-l*+/-2 )
  5UNROLL Z1 RSUBext
  4ROLL			( +-pi/2 DL l*2*equiv-l lvar rest-l X-l/equiv-1 alpha )
  LIMBETA!		( +-pi/2 DL lvar rest-l equiv^alpha*DL[1+num-l/den-l-1 ]^alpha )
  4ROLL x*ext 4ROLL
  x+ext SWAP		( lvar DL rest-l )
;	

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME HORNLN!
::	
  Z0 Z-1 ROT
  #1+_ONE_DO			( c0, ..., ck, +- 1 )
     RNEGext
     DUPINDEX@ #>Z NDXFext
  SWAPLOOP
  DROP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LNOBJ!
LABEL lnobj!
::
  ( obj -> ln[obj], taille: taille est aussi equiv[ln(equiv]] si obj = equiv )

  SIMP1!

  '
  ::					( LAM4 op=x*)
					( symb, o1,..., on-1, #n )
    #1- pZpargSWAPUn			( symb, meta1, meta2 )
    CKSYMBN
    OVER#2+UNROL CKSYMBN		( symb, symb1, symb2 )
    ROTDROP				( symb1, symb2 )
    DUPTYPEREAL? IT
    :: DUP %0< NOT?SEMI
       RNEGext SWAP RNEGext
    ; 
    2GETEVAL				( symb1, log2, taille2 )
    ROT 2GETEVAL			( log2, taille2, log1, taille1 )
    4ROLLROT x+ext			( taille2, taille1, log )
    UNROT LIMMAX!
  ;

  '
  ::					( LAM3: op=x^ )
					( symb, o1,..., on-1, #n )
    #1- pZpargSWAPUn			( meta, meta-exposant)
    #1= NOTcase :: BINT39 ERABLEERROR ;
    OVER#2+UNROL CKSYMBN
    ROTDROP 2GETEVAL			( exposant, log, equiv )
    UNROTOVER x*ext UNROT x*ext		( log, taille )
  ;

  '
  ::					( LAM2: obj -> logarithme, taille  )
    idntlamcase :: LN2ext DUP ;
    DUPTYPESYMB? NOTcase
    :: CKLN HSECO2RCext num#0=case Z0 Z1 ;

    DUPINCOMP SWAP			( symb, o1, ..., on-1, #n, on )
    { x* 4GETLAM x^ 3GETLAM x/ 1GETLAM }
    EQLookup case :: EVAL EVAL ;
					( symb, o1, ..., on-1, #n, on )
    DROP#1- NDROP			( symb ) 
    LN2ext DUP
  ;

  '
  ::					( LAM1: op=x/ )
					( symb, o1,..., on-1, #n )
    #1- pZpargSWAPUn			( symb, meta1, meta2 )
    CKSYMBN
    OVER#2+UNROL CKSYMBN		( symb, symb1, symb2 )
    ROTDROP				( symb1, symb2 ) 
    DUPTYPEREAL? IT
    :: DUP %0< NOT?SEMI
       RNEGext SWAP RNEGext SWAP
    ; 
    2GETEVAL				( symb1, log2, taille2 )
    ROT 2GETEVAL			( log2, taille2, log1, taille1 )
    4ROLLROT x-ext			( taille2, taille1, log )
    UNROT LIMMAX!			( log, taille )
  ;
  4NULLLAM{} BIND
  2GETEVAL
  ABND
;

**********************************************************************
* Desc:		Horner evaluation for limits when X is a fraction
* Stack:	( a0, .., an+1, #n+1, lcm, X -> P[X] )
* Note:		lam1=D^k, lam2=D, lam3=N, lam4=lcm
**********************************************************************
NULLNAME NEWLIMHORN
::				( a0...an #n+1 lcm X )
  3PICK #1= casedrop :: SWAPDROP QDiv ;	( -> an/lcm )
				( a0...an #n+1 lcm X )
  FXNDext Z1 4NULLLAM{} BIND	( a0...an #n+1 )
  SWAP 4GETLAM QMul		( a0...an-1 #n+1 cn )
  OVER ONE_DO (DO)		( a0...an-1 #n+1 res )
    1GETLAM 2GETLAM QMul 1PUTLAM ( a0...an-1 #n+1 res )
    3GETLAM QMul		( a0...an-1 #n+1 res*N )
    OVER #1- LRDM!		( a0...an-1 #n+1 res*N )
    ROT 4GETLAM QMul 		( a0...an-2 #n+1 res*N an-1*lcm )
    1GETLAM QMul QAdd		( a0...an-2 #n+1 res*N+an-1*lcm*D^1 )
  LOOP				( #n+1 res )
  SWAPDROP 			( res )
  4GETLAM 1GETABND QMul QDiv	( res/D^n*lcm )
;

LABEL LIM*LCM
::				( a0, ..., an, #n+1, lcm )
  SWAPDUP #3+			( a0, ..., an, lcm, #n+1, #n+4 )
  THREE DO			( a0, ..., an, lcm, #n+1 )
     INDEX@ ROLL		
     3PICK QMul UNROT
  LOOP				( cn, ..., c0, deno, #n+1 )
;
			
**********************************************************************
* Desc:		Compute DL: c0+c1*X+...+cn*X^n
* Stack:	X, c0, ..., cn, #n+1 -> DL
**********************************************************************
NULLNAME LIMHORN!
::				( X, c0, ..., cn, #n+1 )
  DUP#1= casedrop SWAPDROP
  dup METADENOLCM		( X, a0, ..., an, #n+1, lcm )
  rotomo			( a0, ..., an, #n+1, lcm, X )
  ( * if X = INTERNALX it's much easier than the general case! * )
  INTERNALX EQUALcasedro LIMHORNX
				( a0, ..., an, #n+1, lcm, X )
  DUPTYPESYMB? case NEWLIMHORN	( -> DL )
				( a0, ..., an, #n+1, lcm, X )
  4ROLL 3PICK RMULText		( X, a0, ..., an-1, #n+1, lcm, X, an*lcm )
  SWAP4ROLL			( X, a0, ..., an-1, lcm, res, X, #n+1 )
  ONE_DO			( a0, ..., an-1, lcm, res, X )
     DUPUNROT RMULText		( a0, ..., an-1, lcm, X, res*X )
     ISTOP@ #1- LRDM!		( a0, ..., an-1, lcm, X, res*X )
     4ROLL 4PICK
     RMULText RADDext		( a0, ..., an-2, lcm, X, res*X )
  SWAPLOOP
				( lcm, res, X )
  DROPSWAP RDIVext		( res/lcm )
;

LABEL LIMHORNX
* warning: lcm should be an integer (never use seco)
::				( a0...an #n+1 lcm )
  LIM*LCM			( cn...c0 lcm #n+1 )
  SWAP OBJ>R >POLYTRIM 		( P )
  R>OBJ NDXFext			( P/lcm )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LRDM!
::	
  OVER TYPELIST? NOTcaseDROP
  #1+ OVER LENCOMP 2DUP#< 

  ( liste, entier+1, longueur de liste , ?entier+1<longueur de liste)

  NOTcase2DROP

  ( si liste trop longue: on supprime a gauche )

  DUPROT #-+1 SWAP SUBCOMP TRIMOBJext
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMDL!

* DL reguliers de la forme f(x_0+x) avec x_0 fini 
* si x_0 non nul et f(x_0)=0, r -> r/e

::

* DL-l, reste-l, num-l/deno-l, equiv-l, lim, lvar, operateur -> 
* DL, reste, TRUE
* ou DL-l, reste-l, num-l/deno-l, equiv-l, lim, lvar, operateur, FALSE
  3PICK
  ::
    DUPTYPESYMB? NOTcasedrop TRUE
    CARCOMP INFINIext CARCOMP EQUALNOT
* a rajouter un eventuel test d'egalite avec ?
  ;

  NOTcaseFALSE

  ( DL-l, reste-l, num-l/deno-l, equiv-l, lim, lvar, operateur )

  7UNROLL 4UNROLL

  ( operateur, DL-l, reste-l, lvar, num-l/deno-l, equiv-l, lim )

  DUPTYPETAG? IT STRIPTAGS DUP 5UNROLL 

  ( operateur, DL-l, reste-l, lim, lvar, num-l/deno-l, equiv-l, lim )

  ::
    num#0=case :: 3DROP 4ROLL ;

* test si f(x_0)=0, dans ce cas r -> r/e (car f'(x_0) non nul pour sin, cos )

    { xCOS xSIN } NINE PICK ' EQ POSCOMP

    ( operateur, DL-l, reste-l, lim, lvar, num-l/deno-l, equiv-l, lim, # )

    DUP#0= ITE_DROP
    ::
      OVER pisur2 x/ext SIMP1ext
      ( operateur, DL-l, r-l, lim, lvar, n/d-l, e-l, lim, #, 2*l/pi )
      dZINTNcase 2DROP
      Z2 ZMod SWAP #1= ITE Z1 Z0 EQUAL
      ( operateur, DL-l, r-l, lim, lvar, n/d-l, e-l, lim, ? 2*l/pi mod 2=% )
      NOT?SEMI
      6ROLL 4PICK RDIVext 6UNROLL
    ;
    DROPSWAPDROP 
    ( op, DL-l, reste-l, lim, lvar, equiv-l )
    5ROLL SWAP RSUBext
  ;
			  ( operateur, reste-l, lim, lvar, X )
  4ROLL 5ROLL		( lim, lvar, X-l, reste-l, operateur )

* liste de liste a 2 elements: operateur a appliquer a la limite
* et prog: lim, lvar,X-l, reste-l -> lvar, DL, reste-l

  { xEXP HORNEXP! xCOS HORNCOS! xSIN HORNSIN! xATAN HORNATAN! xASIN HORNASIN! }
  EQLookup NOTcase :: BINT40 ERABLEERROR ; 
  EVAL
					( lvar, DL, reste-l  )
  ROT LIMVALOBJ!			( DL, reste )
  TRUE
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMDLINF!
::

* DL-l, rest-l , num-l/den-l, equiv-l, lim=+/-inf, lvar, operateur
* -> ?-inf, { a# ... a0 }, den-l, lim, lvar, X, rest-l

  ROTSWAP
  EQ: xEXP
  IT LIMINFSIGN!
  2SWAP 2DROP			( DL-l rest-l lvar ?-inf )
  SWAPROT DUP4UNROLL		( DL-l ?-inf lvar rest-l )
  5ROLL				( ?-inf rest-l lvar rest-l num-l/den-l )
  FALSE LPROF!	( ?-inf r-l #varx #prof #ord lvar new-rest-l num-l/den-l )
  8ROLL
  ::
    num0case ROTDROPSWAP
    DROP
  ;
			( ?-inf #varx #prof #ord lvar rest-l num-l/den-l  )
  6ROLL 6PICK
  #- 6UNROLL		( ?-inf #var-#prof #prof #ord lvar rest-l num-l/den-l )

  DUP FXNDext SWAPDROP 7PICK
  ( ?-inf #varx-#prof #prof #ord lvar rest-l num-l/den-l den-l #var-#prof )
  LIMEQU!
  ( ?-inf, #varx-#prof, #prof, #ord, lvar, rest-l, num-l/den-l, den-l trimmed, den-equiv-l, { # # # }  )

  ROTDROP CARCOMP 2DUP Z1Z0 ROT NDUPN #1+ {}POLY RDIVext 
  ( ?-inf, #varx-#prof, #prof, #ord, lvar, rest-l, num-l/den-l, den-equiv-l, #valuation, coeff  )
  CARCOMPext
  ( ?-inf, #varx-#prof, #prof, #ord, lvar, rest-l, num-l/den-l, den-equiv-l, #valuation, coeff  )

* reajuste le rest

  5ROLL 4ROLL 8PICK LIMPROFEND!
  RDIVext
  ( ?-inf, #varx-#prof, #prof, #ord, lvar, num-l/den-l, #val, coeff, rest-l  )
  6ROLL DROP
  ( ?-inf, #varx-#prof, #prof, lvar, num-l/den-l, #val, coeff, rest-l )
  4ROLL FXNDext
  ( ?-inf, #varx-#prof, #prof, lvar, #val, coeff, rest-l, num-l, den-l  )
  4ROLL 5ROLL 4ROLL
  ( ?-inf, #varx-#prof, #prof, lvar, rest-l, den-l, coeff-l, #val, num-l )
  CKINNERCOMP get1
  ( ..., den-l, coeff-l, an...a0, #n+1, #val )
  DUP#1+ #2+PICK
  ( ...,  den-l, coeff-l, an...a0, #n+1, #val, a#val )
  3PICK #4+ ROLL
  ( ...,  den-l, an...a0, #n+1, #val, a#val, coeff-l )

* teste si a#val/coeff-l -> inf: sinon, a#val va dans la partie bornee

  RDIVext 3PICK #6+ PICK ( lvar ) 4PICK #8+ PICK ( #prof )
  #1+ MINUSONE SUBCOMP CDRCOMP ( lvar tronquee )
  4PICK #9+ PICK #1- ( #varx-#prof-1 )
  SWAPROT
  ( ..., den-l, an...a0, #n+1, #val, #, lvar, coeff-l/a#val )
  VARCOMP3!
  (  ..., den-l, an...a0, #n+1, #val, #, lvar, o-liste, #  )
  4UNROLL3DROP FOUR #/ DROP #2=
  (  ...,  den-l, an...a0, #n+1, #val, ?a#val trop grand )
  SWAPDUP 1LAMBIND OVER IT #1+
  (  ...,  den-l, an...a0, #n+1, ?a#val trop grand, # )
  SWAP3PICK #3+ UNROLL
  ( ..., den-l, ?a#val trop grand, an...a0, #n+1, # )
  DUPUNROT 
  :: 2DUP#> 
     NOTcase :: ( ..., den-l, ?a#val trop grand, an...a0, #, #n+1, # ) 
	      DROPSWAPDROP {}POLY Z0 
	     ;
     #-
     (  ..., den-l, ?a#val trop grand, an...a0, #, #n+1-# )
     OVER#2+UNROL >POLYTRIM OVER#2+UNROL >POLYTRIM
  ;
  (  ..., den-l, ?a#val, {a#...a0}, { an ... } )
  ROT IT :: Z0 OVER TYPELIST? ITE >TPOLY TWO{}POLY ; 
  (  ..., den-l, {a#...a0}, { an ... } )
  ROT DUPDUP
  ( ?-inf, #varx-#prof, #prof, lvar, rest-l, {a#...a0}, { an ... }, den-l, den, den )
  ::
    DUPTYPELIST? NOTcasedrop ABND
    LENCOMP 1GETABND #- ONESWAP SUBCOMP
  ;
  ( ?-inf, #varx-#prof, #prof, lvar, rest-l, {a#...a0}, { an ... }, den-l, den-stripped )
  ROTSWAP RDIVext
  ( ?-inf, #varx-#prof, #prof, lvar, rest-l, {a#...a0}, den-l, partie bornee )
  7ROLL 6ROLL 3PICK
  ( ?-inf, #prof, rest-l, {a#...a0}, den-l, bornee, #, lvar, partie bornee )
* #, lvar, equiv-l -> #, lvar, lim-l, #

  VARCOMP3!
  ( ?-inf, #prof, rest-l, {a#...a0}, den-l, bornee, #, lvar, lim-l, # )
  FOUR #/ DROP #1= IT DROPZ0
  ( ?-inf, #prof, rest-l, {a#...a0}, den-l, bornee, #, lvar, lim-l )
  ROT :: DUP#0=csDROP ZERO_DO (DO) ONE>POLY LOOP ;
  ( ?-inf, #prof, rest-l, {a#...a0}, den-l, bornee, lvar, lim-l )
  ROTOVER RSUBext
  ( ?-inf, #prof, rest-l, {a#...a0}, den-l, lvar, lim-l, borne-lim )
  ROT 6UNROLL
  ( ?-inf, #prof, lvar, rest-l, {a#...a0}, den-l, lim-l, borne-lim )
  FIVE
  ZERO_DO (DO)
     5ROLL FXNDext 8PICK LIMPROFEND!
     RDIVext
  LOOP
  ( ?-inf, #prof, lvar, rest-l, {a#...a0}, den-l, lim-l, X=borne-lim )
  7ROLL DROP
  ( ?-inf, lvar, rest-l, {a#...a0}, den-l, lim-l, X=borne-lim )
  ZERO ROT 7PICK VAL2ext
  ( ?-inf, lvar, rest-l, {a#...a0}, den-l, X, lim )
  6ROLL ROT 6ROLL
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMINFSIGN!
::					( symb )
  INNER#1= 				( o1 true | o1 o2 false )
  casedrop :: BINT58 ERABLEERROR ;	( -> unsigned inf error )
  					( inf xNEGNEG ou xNEG )
  SWAPDROP EQ: xNEG			( ?=-inf )
;

* ====================================================================
*		Routines de comparaison
* ====================================================================

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMMAX!
::
  2DUP EQUALcase DROP
  2DUP LIMCOMP!
  ITE_DROP SWAPDROP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMCOMP!

* comparaison d'objets: o1, o2 -> flag
* Attention: les variables doivent etre valides 
* i.e. inv[ln] et exp tendant vers 0 

::
  SWAP num#0=case 2DROPFALSE
  SWAP num#0=case 2DROPTRUE
  VARCOMP2!
  4UNROLL3DROP FOUR #/ DROP #0=
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME VARCOMP2!

* o1, o2 -> #varx, lvar, equiv de o1/o2-l, #
* # = 0 si o1/o2 -> limite finie donnee par equiv de o1/o2-l 
* # = 1 si le numerateur est negligeable devant den
* # = 2 si le den est negligeable devant le num

::
  x/ext LVARXNX2!		( o1/o2, lvarnx, lvarx )
  LIMSORT!			( o1/o2, lvarnx, lvarx )
  DUPLENCOMP 4UNROLL
  SWAP &COMP VXXLext
  VARCOMP3!
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMSORT!
::
  ' VARCOMP! SortList
;

**********************************************************************
* Desc:
* Stack:	( var1, var2 -> flag )
**********************************************************************
NULLNAME VARCOMP!
::	
  2DUP EQUAL case2drop TRUE
  2DUP VAR%
  SWAP VAR%			( var1, var2, #var2, #var1 )
  2DUP#= NOTcase
  :: #> UNROT2DROP ; 

  ( type 0, 1 et 4: pas de comparaison possible )

  #0=         case2drop DROPTRUE
  DUP FOUR #= case2drop DROPTRUE
  DUP#1=      case2drop DROPTRUE

  ( type 2: enlever les ln et rappeler limcomp )

  #2= case VARCOMPLN!

  ( type 3: prendre les ln des 2 membres, chercher la lim de la diff )

  SWAP CKLN SWAP CKLN x-ext LIMIT!

  ( DL-l, rest-l, DL/equiv-l, equiv-l., lim., lvar, flag )	

  NOTcase :: 6DROP BINT59 ' LAMLIMOK STO TRUE ;
  DROP 5UNROLL 4DROP

  ZSIGNECK SWAPDROP
;
  
**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME VARCOMPLN!
::				( var1=inv[ln], var2=inv[ln] )
  DUPLENCOMP #2- ONESWAP SUBCOMP
  SWAP DUPLENCOMP #2- ONESWAP SUBCOMP
  SWAP LIMCOMP!
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME VARCOMP3!
::				( #profondeur, lvar, o1/o2-l )
  ZERO 1LAMBIND

* 1GETLAM contient ZERO pour + borne, ONE pour 0+ , TWO pour +infini
* FOUR pour - borne, FIVE pour 0-, SIX pour -infini
* EIGHT pour borne (signe indetermine), NINE pour 0, TEN pour infini
* THREE pour undef

  3PICK VARCOMP3IN		( #prof lvar o-liste )
  VARCOMP3!#
				( #prof lvar o-liste # )
  ABND
;

LABEL VARCOMP3!#
( #prof lvar obj -> #prof lvar obj # )
::
  1GETLAM SEVEN #>case 1GETLAM		( -> #prof lvar obj # )
					( #prof lvar obj )
  DUP3PICK				( #prof lvar obj obj lvar )
  DUPLENCOMP 6PICK			( #prof lvar obj obj lvar #lvar #prof )
  #1+SWAP SUBCOMP SWAP R2SYM		( #prof lvar obj user-obj )
  ZSIGNE 				( #prof lvar obj z )
  Z1 EQUALcasedro 1GETLAM		( -> obj # )
					( obj z )
  Z0 EQUALcase LAM1UNSIGNED		( -> obj [#lam1 or #8] )
					( obj )
  LAM1NEGSIGNED 			( -> obj # )
;


LABEL VARCOMP3IN
::					( #profondeur lvar o1/o2-l #prof )
  DUP#0=csedrp VARCOMP31!
  #1+_ONE_DO (DO)			( # lvar o-liste )
     FXNDext LIMVAL! ROT		( # lvar den-ak den-#k num )
     LIMVAL!				( # lvar den-ak den-#k num-ak num-#k )
     SWAP4ROLL NDXFext 			( # lvar den-#k num-#k obj-liste )
     ( * added test for constant object to exit the loop * )
     DUP DEPTHext #0= IT ExitAtLOOP
     UNROT				( # lvar obj-liste den-#k num-#k )
     4PICK INDEX@ NTHCOMPDROP		( # lvar obj-liste den-#k num-#k var )
     VAR%IDNT4				( # lvar o-liste den-#k num-#k type )

       ( type #4=idnt, #3=exp, #2=ln, #0 borne, #1 undef )

     DUP4UNROLL		( # lvar o-liste type den-#k num-#k type )
     VARCOMPINLP	( # lvar objet type )
     DROP		
  LOOP			( # lvar objet )
;

LABEL VARCOMPINLP
::				( # lvar o-liste type den-#k num-#k type )
  ONE #>case VARCOMP32!
				( # lvar o-liste type den-#k num-#k )
  ( type: #0 borne, #1 undef )
  2DROP ONE
  EQUALcasedro LIMERR6!		( # lvar o-liste type )

  ( la variable est bornee: si -> 0, ca continue a -> 0 )

  1GETLAM THREE #AND
  #1=case :: NINE 1PUTLAM ;

  ( #0: borne par 1, regarder si on peut faire qqchose? )

  BINT61 ' LAMLIMOK STO
  THREE 1PUTLAM
  ExitAtLOOP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME VARCOMP31!
::	
  DUP FXNDext 
  num0case      :: 2DROP TWO 1PUTLAM ;
  DROP num0case :: DROPONE 1PUTLAM ;
  DROP
;

**********************************************************************
* Desc:
* Stack:	( var -> 0: compter les zeros finaux lvar, o-liste )
**********************************************************************
NULLNAME VARCOMP32!
::				( lvar, o-liste, type, den-#k, num-#k )
  2DUP#= case2DROP
  VARCOMP33!			( lvar, o-liste, type, den-#k, num-#k )
  1GETLAM THREE #AND
  #0= NOTcase2DROP
  #> ITE TWO ONE
  1GETLAM #+ 1PUTLAM
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME VARCOMP33!
::	
  ( ajuste le signe )
  1GETLAM SEVEN #> ?SEMI
  2DUP #- TWO #/ DROP #0=?SEMI
  3PICK

  ( #4 regarder la direction, #3: 0+, #2 0+ )
  ( num-#, den-#, type -> num-#, den-# )

  THREE EQUALcasedro NOP
  TWO EQUALcasedro NOP
  FOUR #= NOT?SEMI
  LAMLIMTYPE			( 0: bi, 1: -; 2: + )
  DUP #2= caseDROP
  #0=ITE LAM1UNSIGNED LAM1NEGSIGNED
  1PUTLAM
;

LABEL LAM1UNSIGNED
:: 1GETLAM THREE #AND EIGHT #+ ; 

LABEL LAM1NEGSIGNED
:: 1GETLAM FOUR 2DUP#> ITE #- #+ ;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME LIMERR6!
::	
  ( undef: sans espoir )
  DUP
  BINT60 ' LAMLIMOK STO
  THREE 1PUTLAM
  ExitAtLOOP
;

**********************************************************************
* Desc:
* Stack:	( obj, lvar -> symb )
**********************************************************************
NULLNAME LIMVALOBJ!
::
  ZERO UNROT VAL2ext
;

**********************************************************************
* Desc:
* Stack:	( obj -> coeff, valuation )
**********************************************************************
NULLNAME LIMVAL!
::	
  num#0=case :: LAMLIMORDRE #1+ ;
  NOTLISTcase ZERO
  INNERCOMP
  ZERO_DO (DO)				( on-1, ..., ok )
    LIMVALLP
  LOOP
;

LABEL LIMVALLP
  ::
    num#0=case DROP
    ISTOP-INDEX #1- DUPUNROT
    #2+UNROLL NDROP 
    INDEX@				( ok, #k )
    ExitAtLOOP
  ;


**********************************************************************
* Desc:
* Stack:	( liste, lequiv -> equiv, ordre )
**********************************************************************
NULLNAME EQUIV!
::
  Z1 OVER LENCOMP  
  ZERO_DO (DO)			( liste, lequiv, equiv )
     OVERINDEX@ NTHCOMPDROP
     4ROLL LIMVAL!		( lequiv, equiv, equiv-k, an, # )
     INDEX@ DUP#0= IT DUP1LAMBIND
     SWAP 5UNROLL x^ext x*ext
  LOOP
  UNROT2DROP 1GETABND
;

**********************************************************************
* Desc:
* Stack:	( obj -> obj, lvarx, lvarnx  )
**********************************************************************
NULLNAME LVARXNX2!
::
  DUP LISTOPext
  DUPLENCOMP ONESWAP #5- SUBCOMP
  LVARXNX2ext ROTDROP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME SIMP1!
::
  LVARXNX2! SWAP &COMP VXXLext ZEROSWAP ROT VAL2ext 
;

**********************************************************************
* Desc:		set a new current var if needed
* Stack:	( symb -> symb )		
**********************************************************************
NULLNAME FindCurVar
::
  DUP LIDNText DUPNULL{}? case :: BINT41 ERABLEERROR ;
						( symb lindt )
  DUP RCLVX EQUALPOSCOMP			( symb lidnt # )
  #0= NOTcaseDROP				( -> symb )
						( symb lidnt )
  CARCOMP STOVX					( symb )
  ( symb, rest )
;

**********************************************************************
* Desc:
* Stack:	( symb -> symb, lvar )		( ??? )
**********************************************************************
NULLNAME LIMVAR!
::
  LVARXNX2!			( { symb, rest }, lvarnx, lvarx )
  LIMSORT! SWAP &COMP		( { symb, rest }, lvar )
;

**********************************************************************
* Desc:		Return type of variable
* Stack:	obj -> #
* Note:		4=idnt, 3=exp, 2=inv[ln], 1=undef, 0=sin/cos
**********************************************************************
LABEL VAR%IDNT4
:: 
  IDNTLAM? casedrop FOUR
  VAR%
;

**********************************************************************
* Desc:		Return type of variable
* Stack:	obj -> #
* Note:		like VAR%IDNT4 but return 3 for idnt
**********************************************************************
NULLNAME VAR%
CODE
var%		C=DAT1	A
		CD1EX
		A=DAT1	A
		CD1EX
		LC(5)	=DOIDNT
		?A#C	A
		GOYES	var%1
var%2		LA(5)	=THREE
		DAT1=A	A
varloop		GOVLNG	=Loop
var%1		LC(5)	=DOLAM
		?A=C	A
		GOYES	var%2
		A=PC
		LC(5)	(var%symb!)-(*)
limbranch	A=A+C	A
		PC=(A)
ENDCODE

* renvoie 3 pour exp, 2 pour inv[ln], 1 undef, 0 pour sin/cos
LABEL var%symb!
::
  DUPTYPESYMB? NOTcase INTERNALERR
  DUPLENCOMP NTHCOMPDROP
  DUPEQ: xEXP casedrop THREE
  DUPEQ: xINV casedrop TWO
  DUPEQ: xSIN casedrop ZERO
  EQ:    xCOS ITE ZERO ONE
;
**********************************************************************
