INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE global.s
INCLUDE keydefs.h

	STITLE  Factorization of integers/polynomials with multiplicity

**********************************************************************
* EXPORT:
*
* MSQFF		( obj -> meta )
*		Full square-free factorization of obj as a meta
*		factor1 mult1 ... factor_n multiplicity_n, #2n
*
* MZSQFF	( Z -> meta )
*		returns the full factorization of an integer as a meta
*
* MLISTSQFF	( P -> meta )
*		Full square-free factorization of a polynomial with recursive
*		call on the gcd of all coefficients
*
* SQFFext	( Q -> list )
*		generic square-free factorization
*
* SUMSQRext	( Z -> Z, C ) 
*		returns a gauss integer C s.t. |C|^2=Z
*		Z must be 2 or s.t. Z=1 mod 4 (if Z != 1 mod 4 -> SIZEERR)
*		Z shoud be prime to insure existence of a solution
*
* CSQFFext	( C -> list ) 
*		returns the factorization of a Gauus integer 
*		as a list of the form 
*		{ factor1 mult1 ... factor_n multiplicity_n }
**********************************************************************

**********************************************************************
* Desc:		Generic square-free factorization.
* Stack:	( Q --> { F1 mult1 .. Fn multn } )
* Note:		il peut y rester des facteurs 1 ou de degre 0
**********************************************************************
NULLNAME SQFFext
::
  DUPTYPECOL? case SECOSQFFext
  MSQFF {}N
;

**********************************************************************
* Desc:		Full square-free factorization of object. The result
*		is given as a Meta object.
* Stack:	( Q --> F1 mult1 .. Fn multn #2n )
**********************************************************************
NULLNAME MSQFF
::
  nogcdflag TestSysFlag case %1TWO
  CK&DISPATCH1
  zint	MZSQFF
  seco	MSECOSQFF
  list	MLISTSQFF
  any	%1TWO
;

* --------------------------------------------------------------------
* Desc:		Square free factorization of unknown (?) object
* Stack:	( ob --> ob %1 #2 )
* Note:		See MSQFF
* --------------------------------------------------------------------
NULLNAME %1TWO
::
  %1 TWO
;

**********************************************************************
* Desc:		Full factorization of an integer.
* Stack:	( Z --> Z1 mult1 .. Zn multn #2n )
**********************************************************************
NULLNAME MZSQFF
::
  num#0=case %1TWO
  DupZIsNeg? DUPUNROT IT RNEGext
  NFactor				( neg? lfac )
  ZEROSWAP DUPDUP >R			( flag meta lfac lfac )
  CARCOMP %0				( flag meta lfac curfac multiplic )
  ROT LENCOMP				( flag meta curfac multiplic #k )
  ZERO_DO (DO)				( flag meta curfac multiplic )
	RSWAP 'R RSWAP			( flag meta curfac multiplic newfac )
	3PICKOVER EQUAL			( flag meta curfac multiplic newfac ? )
	MZSQFF1				( flag meta curfac multiplic )
  LOOP
					( flag meta curflag multiplic )
  ROT #2+ get1 NOT?SEMI
  Z-1 %1 ROT #2+
;

* --------------------------------------------------------------------
* Desc:		Adds integer factor to factor list. If the factor
*		is the same as the last time, only the multiplicity
*		is increased.
* Stack:	( Meta curfac %n newfac TRUE  --> Meta curfac %n+1 )
*		( Meta curfac %n newfac FALSE --> Meta' newfac %1
* --------------------------------------------------------------------
NULLNAME MZSQFF1
::
  casedrop %1+
  4ROLL #2+ SWAP %1
;

**********************************************************************
* Desc:		Factorization of an extension.
* Stack:	( ob --> Meta )
**********************************************************************
NULLNAME MSECOSQFF
::
  SECOSQFFext INNERCOMP
;

**********************************************************************
* Desc:		Full square-free factorization of a polynomial
*		with a recursive call on the GCD of all coefficients.
* Stack:	( P --> Meta )
**********************************************************************
NULLNAME MLISTSQFF
::
  nogcdflag TestSysFlag case %1TWO

  METASQFFext

  ( appel recursif de SQFF )
						( S1..Se-1 %e-1 %e e Te Re )
  CKCARCOMP MSQFF				( S1..Se-1 e-1 e e Te meta )

  ( reajustement des profondeurs et retrait des 1 )

  DUP #1-SWAP #2/
  ZERO_DO (DO)					( f1 m1 ... fk mk #2k-1 )
     get1 ONE>POLY
     SWAP get1 SWAP
     3PICK Z1 EQUAL IT :: UNROT2DROP #2- ;
  LOOP
  #1+ {}POLY

  ( fin d'appel recursif )
						( S1..Se-1 %e-1 %e e Te {} )
  2SWAP ROTSWAP					( S1..Se-1 %e-1 Te %e {} e )
  #2* DUPUNROT #2+UNROLL			( {} S1..Se-1 %e-1 Te %e 2*e )
  get1 INNERCOMP top&				( meta )
;

**********************************************************************
* Desc:		Square-free factorization
* Stack:	( P-list -> S1 %1 ..Se-1 %e-1 %e e Te Re )
**********************************************************************
NULLNAME METASQFFext

* rajouter 	
*	2SWAP ROT	
*	( S1..Se-1 %e-1 Te %e e Re )
*	%1 ROT#1+ #2*
* pour obtenir le meta au complet

::
  ONESWAP					( #1 P )
  NOTLISTcase :: %1 UNROT Z1 ;			( -> %1 #1 P 1 )
  DUP DERext					( #1 P P' )
  SIMP3ext					( #1 R1 P/R1 P'/R1 )
  DROPSWAP					( #1 T1 R1 )

* T1= Pi_{i} q_i, R1= Pi_{i>=2} q_i^{i-1}

  BEGIN					( S1..Si-1 %i-1 i Ti Ri )
     3PICK UNCOERCE 4UNROLL
     ?ATTNQUIT
     DUPCKLEN{} #1<>			( S1..Si-1 %i-1 %i i Ti Ri flag )
  WHILE
     ROT#1+UNROT			( S1..Si-1 %i-1 %i i+1 Ti Ri )
     SIMP3ext

* Tk=Pi_{i>=k} q_k, Sk=q_k, R_k=Pi_{i>k} q_i^{i-k}

					( S1..Si-1 %i-1 %i i+1 Ti+1 Si Ri+1 )
    SWAP 

* si on teste 1, probleme car e n'indique plus la taille du meta
*	Z1 EQUALcasedro 4ROLLDROP

    5UNROLL			( S1, .., Si-1, %i-1, Si, %i, i+1, Ti+1, Ri+1 )
  REPEAT			
				( S1..Se-1 %e-1 %e e Te Re )
;

**********************************************************************
* Desc:		factorization of irrquads and Gauss integers
* Stack:	:: x<< a b c x>> -> { fact1 mult1 ... factn multn }
**********************************************************************
NULLNAME SECOSQFFext
:: 
  nogcdflag TestSysFlag case :: %1 TWO{}N ;
  DUPLENCOMP #2= case CSQFFext
  DUP QXNDext					( q, a, b, c )
  2SWAP Z0					( b, c, q, a, 0 )
  EQUALcasedro SECOSQFFa=0
						( b, c, q, a )
  Z0 5ROLL 5ROLL NDXQext GCDext			( q, d )
  DUP %1 2SWAP RDIVext %1 FOUR {}N
;

LABEL SECOSQFFa=0
::
  DROPSWAP					( c b )
  ( * extract gcd of real and imaginary part of b * )
  DUPDUP RREext SWAP RIMext GCDext		( c b b' )
  DUPUNROT QDiv					( c b' b/b' )
  MSQFF get1 rotomo 				( meta b' c )
  SIMP3ext					( meta d b'/d c'/d )
  SXSQRext UNROT %1 UNROT			( meta sqrt_c'/d %1 d b'/d )
  ( * if d=1, simply sqff b'/d * )
  OVER Z1 EQUALcase 
  :: 						( meta sqrt_c'/d %1 d b'/d )
     SWAPDROP					( meta sqrt_c'/d %1 b'/d )
     4ROLL #2+ SWAP				( meta b'/d )
     MSQFF top& {}N				( {} )
     ;
						( meta sqrt_c'/d %1 d b'/d )
  ( * compute actual multiplicity of d by dividing b'/d by d * )
  %3
  [BEGIN]					( ... d b'/d % )
    OVER 4PICK RDIVext DUPTYPESYMB?		( ... d b" % b"/d ? )
  NOT_WHILE
    :: ROTDROPSWAP %2 %+ ;
  [REPEAT]
						( ... d b'/d % b"/d )
  DROPROT SXSQRext  			( meta sqrt_c' %1 b'/d % sqrt_d )
  SWAPROT 				( meta sqrt_c' %1 sqrt_d % b'/d )
  FOUR SWAP MSQFF 			( meta meta b'/d )
  top& top& {}N				( {} )
;


**********************************************************************
* Desc:		Factorization of Gauss integers
* Stack:	( C --> { factor1 mult1 ... factn multn } )
* Note:		This is not the complete factorization of C over
*		Gauss integers since the GCD of the real part and
*		imaginary part of c is factored only over R.
**********************************************************************
NULLNAME CSQFFext
::
  DUP RCONJext OVER RMULText SQFFext		( A, FACT )
  NULL{} 
  OVER LENCOMP #2/
  #1+_ONE_DO (DO)				 ( A, FACT, RES )
     OVERINDEX@ #2* #1-  NTHCOMPDROP		 ( A, FACT, RES, F )
     ::
       num#1=case CSQFF1
       Z2 EQUALcasedro
       ::
         Zi+1 CSQFF1
       ;
       DUP Z4 ZMod
       Z1 EQUALcase
       ::
         SUMSQRext SWAPDROP			( A, FACT, RES, C+i*D )
         4PICKOVER CMODext SWAPDROP
         Z0 EQUALNOTcase
         ::
           RCONJext CSQFF1
         ;
						( A, FACT, RES, C+i*D )
         4PICKOVER RCONJext CMODext SWAPDROP
         Z0 EQUALNOTcase CSQFF1
						( A, FACT, RES, C+i*D )
         DUP RCONJext
         4PICK INDEX@ #2* NTHCOMPDROP %2 %/ 
         SWAPOVER FOUR {}N &COMP
       ;
       >TCOMP
       OVERINDEX@ #2* NTHCOMPDROP
       %2 %/ >TCOMP
     ;
						( A, FACT, RES )
  LOOP
  SWAPDROP					( a, fact )
  DUPINCOMP #2/ Z0 SWAP
  ZERO_DO (DO)				( fact1 num1 ... factk numk sigma )
     UNROTSWAP
     DUPTYPEZ? ITE Z>R SECO2CMPext
     xLNext x*ext x+ext CASCRUNCH
  LOOP
					( z lst sigmalnlist )
  ROT SECO2CMPext
  xLNext x-ext CASCRUNCH		( lst ln[z]-sigma_ln[list] )
  C>Im% %PI %/ %2 %* %1 RNDXY		( lst Im[ln[z]-sigma_ln[list]]/[pi/2] )
  num0=case DROP
  %CHS >HCOMP Zi >HCOMP			( faclist )
;

LABEL CSQFF1
:: >TCOMP OVERINDEX@ #2* NTHCOMPDROP >TCOMP ; 

**********************************************************************
* Desc:		Returns a Gauss integer C so that |C|^2=Z.
*		Z must be 2 or so that Z=1 mod 4.
*		If Z<>1 mod 4, "Z is not 1 mod 4" error.
*		Z should be prime to ensure the existence of a solution.
* Stack:	( Z --> Z C )
**********************************************************************
NULLNAME SUMSQRext
::
  DupZIsNeg? case Z<0ERR
  num#1=case Zi
  Z2 EQUALcasedro
    :: Z2 Zi+1 ;

* decomposition of p=1 [4] as a^2+b^2
* step 1: research sqrt[-1] mod p
							( p -> p, i )
  DUP Z4 ZDIVext
  Z1 EQUALNOTcase :: BINT29 ERABLEERROR ;
							( p, [p-1]/4 )
  MINUSONE TWO
  DO
     2DUP INDEX@ #>Z			( p, [p-1]/4, p, [p-1]/4, a )
     SWAPROT ModPow			( p, [p-1]/4, a^[p-1]/4 )
     DUPDUP RMULText
     4PICK QMODSYMext			( p, [p-1]/4, a^[p-1]/4, a^[p-1]/2 )
     ::
       Z1 EQUALcasedro DROP
       Z-1 EQUALcase ExitAtLOOP
       BINT30 ERABLEERROR		( p n'est pas premier )
     ;
  LOOP
  SWAPDROP

* step 2: reduction du reseau engendre par (1,i=sqrt(-1)) et (0,p)

					( p, i )
  Z1 IRXCext				( p , [1,sqrt[-1]] )
  Z0 3PICK RIXCext			( p, u0, v0 )
  [BEGIN]

     ( calcul des normes de un et vn et test s'il faut echanger un et vn )

     2DUP CNORMext SWAP CNORMext	( p, un, vn, |vn|, |un| )
     2DUP ZNLT?				( p, un, vn, |vn|, |un|, ?|vn|<|un| )
     ITE_DROP :: SWAPDROP ROTSWAP ;	( p, un, vn, |vn| )

     ( calcule m, si m est nul c'est fini )

     UNROT 2DUP				( p, |vn|, un, vn, un, vn )
     CXRIext ROT CXRIext ROT
     RMULText UNROT RMULText
     RADDext				( p |vn| un vn un.vn )
     DUP 5PICK QMODSYMext		( p |vn| un vn un.vn, un.vn mod |vn| )
     RSUBext 4ROLL ZQUOText DUP		( p un vn m m )
     Z0 EQUAL
  NOT_WHILE
     ::
       OVER RMULText ROTSWAP RSUBext
     ;
  [REPEAT]
  DROP					( p, un, vn )
  SWAPDROP INCOMPDROP
  ZAbs SWAP ZAbs SWAP
  RIXCext
;
**********************************************************************



