INCLUDE cas2romid.h
INCLUDE erextdec.h
INCLUDE cas2ext.h
INCLUDE ../CAS/keydefs.h
INCLUDE ../CAS/global.s

	STITLE Ordinary differential equations (1st order)


**********************************************************************
* EXPORT:
*
* DESOLVE	( list symb -> symb1 symb2 )
*		( symb symb -> symb1 symb2 )
*		list contains the equation followed by initial bound cond
*
* ID0		( idnt, idnt_0 )
*		Adds 0 to the name of the idnt, recalls it if in search path
*		Example if idnt='X' and X0 contains 5 returns 'X', 5
*
* ODE_INT	( symb, x -> int[x0,x,symb, x] )
*		Evaluate the integral with RISCHext
*
* LINSOLV	( b a -> y )
*		Solves the linear diff equation y'=ay+b
*
* LDESOLVext	( symb2, symb1 -> symb )
*		Solves a linear ode with constant coeff
*		symb2 is the second member of the diff equ
*		symb1 is the charact. polynomial (using the current var)
*		Example y''+4y=sin(x)
*		symb2='SIN(X)' and symb1='X^2+4'
*
* LDEGENE	( symb1 -> symb )
*		Returns the general solution of a lin ode with const coeff
*
* LDEPART	( symb2, symb1 -> symb1, symb )
*		Finds a particular solution of a lin ode (by Lap/inv Lap)
*
* LDSSOLVext	( vector, matrix -> vector )
*		Solves a 1st order linear differential system with cst coeff
*		matrix is the matrix of the system
*		vector is the vector of the 2nd members
*
* ODETYPESTO	( type -> )
*		Stores the equation type in the ODETYPE var
*
* ODE_SEPAR	( P -> P, FALSE )
*		( P, P-y, P-x, TRUE )
*		Tries to separate symb as a product of a function of y and x
*		(y is the out variable, x the second one)
**********************************************************************

**********************************************************************
* Desc:		Solve ordinary differential equation
* Stack:	( list symb1 --> list_sols )
*		( symb symb1 --> list_sols )
* Note:		for some ode's returned level2 is not symb1
**********************************************************************
NULLNAME DESOLVE
::
						( obj fnc )
  ( * First check that symb1 is a user function * )
  DUP F2ARG&NAME? NcaseTYPEERR			
						( obj fnc var fnc_name )
  SWAP IDNTLAM? NcaseTYPEERR			( obj fnc fnc_name var )

  ( * stores var as current var * )
  DUP STOVX

  ( * create temp env * )
  RCLVX NULL{} ZEROZERO		( obj fnc fnc_name var vx bound # ode# )
  
  ' NULLLAM EIGHT NDUPN DOBIND 			( )

DEFINE ode_eq 8GETLAM 
DEFINE ode_eq! 8PUTLAM
DEFINE fnc_applied 7GETLAM ( the function applied to the var )
DEFINE fnc_name 6GETLAM ( the function name to solve for )
DEFINE ode_var 5GETLAM ( the var of this ode )
DEFINE ode_savevx  4GETLAM ( copy of cur var )
DEFINE ode_unknown 4GETLAM ( use same lam as ode_savevx )
DEFINE ode_unknown! 4PUTLAM ( for the list of constants )
DEFINE boundary 3GETLAM
DEFINE boundary! 3PUTLAM ( boundary conditions )
DEFINE stackdepth 2GETLAM ( error trapping )
DEFINE stackdepth! 2PUTLAM
DEFINE ode# 1GETLAM ( counter for constant number )
DEFINE ode#! 1PUTLAM

  ( * Checks that obj2 is either a sym or a list of symb * )
  ( * and binds boundary conditions and ode_eq * )
  ode_eq CKREPLACE=BY- 				( eqdiff )
  HYP2EXPext 					( eqdiff )


  LVARext					( eq lvar ) 

  fnc_name ID>$					( eq lvar fnc$ )

  ( * check that all derivative of fnc are order 1 or 2 * )
  SWAPINCOMP get1			( eq var1 .. varn #n fnc$ )
  ZERO SWAPROT 				( eq var1 .. varn #order fnc$ #n )
  ZERO_DO (DO)				( eq var1 .. varn #order fnc$ )
    ROT USERFCN? 			( eq var1..varn-1 #order fnc$ varn ? )
    ITE TESTDEPEND DROP			( eq var1..varn-1 #order fnc$ )
  LOOP					( eq #order fnc$ )
  SWAP					( eq fnc$ #order )
	
  ( * solve ode without boundaries conditions * )
	
  DESOLVEGENE				( liste_sols )
  ( * restore VX variable * )
  ode_savevx STOVX

  ( * now look for boundary conditions * )
  DUPNULL{}? case ABND
  boundary NULLCOMP? case ABND
  					( liste_sols )
  DUP >R LENCOMP ZEROSWAP		( #0 #n )
  ZERO_DO (DO)				( meta-sols )
    RSWAP 'R RSWAP			( meta-sols sol )
    SOLVFORBOUND			( meta-sols meta )
    top&
  LOOP
  {}N
; 

LABEL DESOLVEGENE
* dispatch ode solving by order
( eq fnc$ #order -> sol )
::
  CSTLINEAR? 				( eq fnc$ #order #0 | 2nd carac #1 )
  #1=case DESOLVCSTLIN			( -> solution )
  DUP#0=csedrp ZEROORDER
  DUP#1= casedrop FIRSTORDER
					( eq fnc$ #order )
  #2= NOTcase :: BINT47 ERABLEERROR ;

  ( * second order equation * )
					( eq fnc$ )
  SWAP fnc_name Contains? SWAPDROP	( fnc$ eq flag )
  NOTcase INCOMPLETE2Y			( -> )
  BINT74 ERABLEERROR
;

LABEL DESOLVCSTLIN
:: SWAP QNeg SWAP LDECSOLV fnc_applied SWAP SYM= ONE{}N ;

LABEL CKREPLACE=BY-
::
  CK&DISPATCH1
  matrix BINDBOUND
  symb REPLACE=BY-
  sym  NOP
;

LABEL BINDBOUND
* matrix -> symb, bind boundary conditions to LAM1 (boundary!)
::
  DUPTYPEMATRIX? NcaseSIZEERR
  DUP CARCOMP 			( matrix 1st )
  DUPTYPEMATRIX? caseSIZEERR	( must be a vector )
				( matrix 1st )
  CKREPLACE=BY-			( matrix symb )
  SWAP CDRCOMP boundary!	( symb )
;

LABEL REPLACE=BY-
* symb -> symb
:: 
  DUPINCOMP
  OVER 
*ASSEMBLE
*	IF	f39
*RPL
*  DUPEQ: x== InCASApplication? NOT AND SWAP EQ: x= OR
*ASSEMBLE
*	ELSE	f39
*RPL
  EQ: x= 
*ASSEMBLE
*	ENDIF	f39
*RPL
  NOTcase NDROP 
  SWAPDROP ' x- SWAP SYMBN SWAPDROP
;

LABEL TESTDEPEND
* #order fnc$ var -> #order fnc$
* test that var depends from fnc or its derivative, if not drops var
* else modifies #order of diff equation
* SETSIZEERR if arg is not EQUAL to ode_var
::
  F2ARG&NAME? NOTcaseDROP
				( #order fnc$ arg name )
  NAME2#NAME$			( #order fnc$ arg # name$ )
  4PICK EQUAL			( #order fnc$ arg # flag )
  NOTcase2DROP			( -> #order fnc$ )
  SWAP ode_var EQUAL		( #order fnc$ # flag )
  NcaseSIZEERR
  ROT #MAX SWAP			( -> #order fnc$ )
;

LABEL NAME2#NAME$
* returns the fnc name without derivatives prefix and the # of derivatives
* name -> # name$
::					( name )
  ZEROSWAP ID>$				( # name$ )
  BEGIN
    DUPONE TWO SUB$ d1$ EQUAL		( # name$ flag )
  WHILE
    SWAP#1+SWAP THREE OVERLEN$ SUB$	
  REPEAT
					( # name$ )
;

LABEL F2ARG&NAME?
* tests if obj is a user fcn of 1 arg
( obj -> arg idnt_name true | obj false )
::
  DUPTYPESYMB? NOTcaseFALSE	( -> obj false )
				( symb )
  DUPINCOMP			( symb a1..ak symb_id_name_; #arg xFCN #k+3 )
  OVER EQ: xFCNAPPLY		( symb a1..ak symb_id_name_; #arg xFCN #k+3 ? )
  NOTcase NDROPFALSE		( -> symb false )

  3PICK #1=			( symb a1..ak symb_id_name_; #arg xFCN #k+3 ? )
  NOTcase NDROPFALSE		( -> symb false )
  				( symb a1..ak symb_id_name_; #arg xFCN #k+3 )
  #3- UNROT2DROP SWAP OBJ>R 	( symb a1..ak #k ) 
  CKSYMBN R>OBJ CARCOMP 	( symb arg id_name )
  DUPTYPEIDNT? 			( symb arg id_name flag )
  NOTcase2drop FALSE		( -> symb false )
				( symb arg id_name )
  ROTDROP TRUE			( -> arg id_name )
;

LABEL 'IDt
:: ' ID tt TOTEMPOB ;

LABEL CSTLINEAR?
* eq fnc$ #order -> 2nd char #1 | eq fnc$ #order #0
* tests if equation is linear with cst coeff
* if true returns 2nd member and char. equation, 
* binds order to lam1 and "Linear with cst coeff" to ODETYPE
* otherwise returns false
* Algorithm: take derivative of eq wrt "d1...d1" fnc$[x] for #0 to #order "d1"
* If result is cst [by taking second derivative] then add coeff to char
* poly, otherwise stops and returns false
*
* improvements: should return the coeff in a list if non cst linear ode
* for Euler-type equation testing
::				( eq fcn$ #order )
  DUP#1= caseFALSE		( -> never try for 1st order eqn )
  3DUP				( eq fnc$ # eq fnc$ # )
  NULL{} TRUE		( eq fnc$ #order 2nd cur_fnc$ #ord char_poly try? )
  ROT #1+ ZERO_DO (DO)		( ... 2nd cur_fnc$ char_poly try? )
    CSTLINEARLP
  LOOP				( ... 2nd cur_fnc$ char_poly try? )
  NOTcase2drop DROPZERO		( -> eq fcn$ #order false )
  				( eq fcn$ #order 2nd cur-fcn$ char_poly )
  ( * look if char_poly is indep of dependent var * )
  ( * if not we have a non cst linear, if true it is cst linear * )
  LVARext ode_var Contains? 	( eq fcn$ #order 2nd cur-fcn$ char_p var x ? )
  UNROT2DROP			( eq fcn$ #order 2nd cur-fcn$ char_p ? )
  case2drop DROPZERO		( -> eq fcn$ #order false )
				( eq fcn$ #order 2nd cur-fcn$ char_p_list )
  ode_var ONE{}N SWAP R2SYM	( eq fcn$ #order 2nd cur-fcn$ char_p_sym )
  SWAPDROP ROT ode#!		( eq fcn$ 2nd char_poly )
  2SWAP 2DROP 			( 2nd char_poly )
  "Linear w/ cst coeff" 
  ODETYPESTO ONE		( 2nd char-poly TRUE )
;

LABEL CSTLINEARLP
::				( ... 2nd cur_fnc$ char_poly try? )
  NOTcaseFALSE			( -> immediate reloop )
				( ... 2nd cur_fnc$ char_poly )
  UNROTDUP			( ... char_poly 2nd cur_fcn$ cur_fcn$ )
  d1$ SWAP&$ 4UNROLL		( ... new_fcn$ char_poly 2nd cur_fcn$ )
  $>ID ode_var 			( ... new_fcn$ char_poly 2nd cur_fcn var )
  SWAP QUOTEAPPLY		( ... new_fcn$ char_poly 2nd cur_fcn[var] )
  'IDt MEVALext			( ... new_fcn$ char_poly 2nd[t] )
  DUP 'IDt DIRECTDERIV		( ... new_fcn$ char_poly 2nd[t] 2nd[t]' )
  SIMP1ext 'IDt Contains?	( ... new_fcn$ char_poly 2nd[t] 2nd[t]' t ? )
  case2drop FALSE		( -> ... new_fcn$ char_poly 2nd[t] FALSE )
				( ... new_fcn$ char_poly 2nd[t] 2nd[t]' t )
  4ROLL 3PICK  			( ... new_fcn$ 2nd[t] 2nd[t]' t char_pol coef )
  >HCOMP			( ... new_fcn$ 2nd[t] 2nd[t]' t new_char_pol )
  4UNROLL x*ext x-ext 		( ... new_fcn$ new_char_pol new_2nd )
  SIMP1ext UNROT TRUE		( ... new_2nd new_fcn$ new_char_pol TRUE )
;


LABEL ZEROORDER
* solve 0th-order diff eq (this is not an ode BTW)
::					( eq fnc$ )
  DROP fnc_applied			( eq var )
  ( simply calls internal solve )
  ISOLALL				( var liste_sols )
  VAR=LIST				( -> list of solutions )
;

LABEL QUOTEAPPLY
::					( var idnt_name )
  ONE SYMBN				( var symb )
  ONE ' xFCNAPPLY			( var symb #1 xFCN )
  FOUR SYMBN
;


*--------------------------------------------------------------------
* First order equations
*--------------------------------------------------------------------

LABEL FIRSTORDER
::					( eq fnc$ )
  ( * first try to isolate d1fnc in eq * )
  d1$ SWAP&$				( eq d1y$ )
  $>ID ode_var SWAP QUOTEAPPLY		( eq d1y[x] )
  ISOLALL				( d1y[x] list_sols )
  DUPNULL{}? case ISOLERR
					( d1y[x] list_sols )
  ( * now solve each equation * )
  INNERCOMP get1 SWAPDUP		( eq1..eqn d1y #n #n )
  ZERO_DO (DO)				( eq1..eqn d1y #n )
    get1 3PICK 				( eq2..eqn d1y #n eq1 d1y[x] )
    2DUP 				( eq2..eqn d1y #n eq1 d1y[x] eq1 d1y )
    DEPTH stackdepth!
    ERRSET
      FIRSTSOLVE 
    ERRTRAP TRAP1STORDER
					( eq2..eqn d1y #n eq1 d1y[x] sol1 )
    5UNROLL 2DROP			( eq2..eqn sol1 d1y #n )
  LOOP
  SWAPDROP {}N				( -> list of solutions )
;

LABEL TRAP1STORDER
:: 
  ERROR@ # 203 #<>case ERRJMP 
  DEPTH #3+ stackdepth #- NDROP
					( eq2..eqn d1y #n eq1 d1y[x] )
  2DUPSWAP SYM=				( eq2..eqn d1y #n eq1 d1y[x] d1y=eq )
  boundary DUP NULLCOMP? 		( eq2..eqn d1y #n eq1 d1y d1y=eq {} ? )
  ITE_DROP SWAP>HCOMP			( eq2..eqn d1y #n eq1 d1y[x] d1y=eq )
  fnc_applied ' xDESOLVE ONE SYMBN
  TWO ' xFCNAPPLY FIVE SYMBN		( eq2..eqn d1y #n eq1 d1y[x] sol )
;

LABEL d1$
"d1"

LABEL FIRSTSOLVE
( eq d1y[x] -> solution )
::
  ( * first look if y[x] is there, if not just take the antiderivative * )
   					( eq d1y[x] )
  DROP fnc_name Contains? SWAPDROP	( eq flag )
  NOTcase INCOMPLETEY

					( eq )
  ( * look for a linear equation * )
  fnc_applied				( eq y[x] )
  2DUP 'IDt MEVALext 			( eq y[x] eq[t] )
  DUP 'IDt DIRECTDERIV			( eq y[x] eq[t] eq'[t] )
  SIMP1ext 'IDt Contains? SWAPDROP 	( eq y[x] eq[t] a flag )
  NOTcase FIRSTLINEAR

					( eq y[x] eq[t] a )
  ( * look for an autonomous equation * )
  DROP ode_var Contains?		( eq y[x] eq[t] x flag )
  ( * #pos is the position of x in lvar; if #0 then ode is autonomous * )
  NOTcase INCOMPLETEX

  					( eq y[x] eq[t] x )
  ( * now look for separate var * )
  2DUP DIRECTDERIV				( eq y[x] eq[t] x eq'[t] )
  3PICKSWAP x/ext SIMP1ext		( eq y[x] eq[t] x eq[t]/eq'[t] )
  'IDt Contains?			( eq y[x] eq[t] x eq/eq' t ? )
  NOTcasedrop SEPARABLE
					( eq y[x] eq[t] x eq/eq' t )
  2DROP					( eq y[x] eq[t] x )
  SWAP FXNDext				( eq y[x] x M N )

  ( * integrating factor? test if [D_yM+D_xN]/N =f[x] or over M=g[t] * )
  2DUP 5PICK DIRECTDERIV			( eq y[x] x M N M D_xN )
  SWAP 'IDt DIRECTDERIV			( eq y[x] x M N D_xN D_tM )
  x+ext DUP				( eq y[x] x M N D_xN+D_tM D_xN+D_tM )
  3PICK x/ext SIMP1ext			( eq y[x] x M N D_xN+D_tM D_xN+D_tM/N )
  'IDt Contains? NOTcasedrop EXACT1F_x	( -> )

  2DROP					( eq y[x] x M N D_xN+D_tM )
  3PICK x/ext SIMP1ext			( eq y[x] x M N D_xN+D_tM/M )
  4PICK Contains? NOTcasedrop EXACT1F_y	( -> )
  
  ( * now look for an homogeneous equ. * )
  4DROP					( eq y[x] x )  
  2DUP TWO{}N				( eq y[x] x { y[x] x } )
  UNROTSWAP 'IDt x*ext			( eq { y[x] x } x t*y[x] )
  SWAP 'IDt x*ext			( eq { y[x] x } t*y[x] t*x )
  TWO{}N				( eq {} { t*y[x] t*var } )
  3PICK UNROT MEVALext			( eq t_eq )

  OVER x-ext TEXPAext SIMP1ext		( eq t_eq-eq )
  j#0=case7 HOMOGENE
 					( eq )

  ( * look for Bernouilli equation * )
  BINT52 ERABLEERROR
;

LABEL FIRSTLINEAR
::
  ( * linear equation, eq=a*y[x]+b * )
					( eq y[x] eq[t] a )
  DUPUNROT 'IDt x*ext			( eq y[x] a eq[t] a*t )
  x-ext SIMP1ext			( eq y[x] a b )
  SWAP LINSOLV				( eq y[x] solution )
  ROTDROP SYM=				( y[x]=... )
  "1st order linear" ODETYPESTO
;

LABEL INCOMPLETEY
::					( eq )
  fnc_applied SWAP 			( y[x] eq )
  ode_var ODE_INT SYM=			( -> y[x]=int[eq,x] )
;

LABEL INCOMPLETE2Y
::					( fnc$ eq )
  OVER d1$ SWAP&$			( fnc$ eq d1y$ )
  DUP $>ID ode_var SWAP QUOTEAPPLY	( fnc$ eq d1y$ d1y[x] )
  SWAP d1$ SWAP&$ 			( fnc$ eq d1y[x] d1d1y$ )
  $>ID ode_var SWAP QUOTEAPPLY  	( fnc$ eq d1y d1d1y )
  TWO{}N fnc_applied OVER CARCOMP TWO{}N ( fnc$ eq { d1y d1d1y } { y d1y } )
  MEVALext				( fnc$ eq )
  SWAP FIRSTORDER			( {} )
  ( * now isolate y[x] and integrate each fnc wrt x )
  DUPNULL{}? ?SEMI
  DUP >R LENCOMP			( # )
  ZEROSWAP				( #0 # )
  ZERO_DO (DO)				( meta )
    RSWAP 'R RSWAP			( meta sol )
    fnc_applied ISOLALL SWAPDROP	( meta {sols} )
    INCOMPLETE2YINT top&		( meta )
  LOOP
  {}N
;

LABEL INCOMPLETE2YINT
::					( { sols } )
  DUPNULL{}? ?SEMI
  DUP >R LENCOMP ZEROSWAP		( meta # )
  ZERO_DO (DO)
    fnc_applied RSWAP 'R RSWAP		( meta y fcn )
    ode_var ODE_INT SYM= SWAP#1+	( meta )
  LOOP					( meta )
;

LABEL INCOMPLETEX
* solves equation y'=f[y]
* equation is independant of x
::					( eq y[x] eq[t] x )
  2SWAP 2DROP				( eq[t] x )
  SWAP					( x eq[t] )
  'IDt fnc_name MEVALext		( x f[y] )
  Z1 SWAPNDXF				( x 1/f[y] )
  fnc_name ODE_INT			( x F[y] )
  fnc_name fnc_applied MEVALext		( x F[y[x]] )
  SYM=					( x=F[y[x]] )
  "Autonomous 1st order" ODETYPESTO
;

LABEL EXACT1F_x
* a function of x is an integrating factor
::					( eq y[x] x M N D_xN+D_tM v[x] )
  SWAPDROP				( eq y[x] x M N v[x] )
  QNeg 4PICK NOCAREEVAL7 RISCHext	( eq y[x] x M N int[-v[x]] )
  xEXPext				( eq y[x] x M N u[x] )
  SWAP 'IDt RISCHext 			( eq y[x] x M u[x] int[N,t] )
  OVER x*ext SIMP1ext 			( eq y[x] x M u[x] u*int[N,t] )
  DUP 5PICK DIRECTDERIV			( eq y[x] x M u F D_x[F] )
  2SWAP x*ext x+ext SIMP1ext		( eq y[x] x F D_xF+u*M )
  ROT ODE_INT  				( eq y[x] F int[D_xF+u*M] )
  SWAP 'IDt 4ROLL REPLACE2BY1		( eq F G )
  ROTDROPSWAP SYM=			( res )
;

LABEL EXACT1F_y
* a function of y[x]=t is an integrating factor
::					( eq y[x] x M N v[t] )
  QNeg 'IDt NOCAREEVAL7 RISCHext	( eq y[x] x M N int[-v[t]] )
  xEXPext				( eq y[x] x M N u[t] )
  ROT 4ROLL RISCHext			( eq y[x] N u[t] int[M,x] )
  OVER x*ext SIMP1ext			( eq y[x] N u[t] u[t]*int[M,x] )
  DUP 'IDt DIRECTDERIV			( eq y[x] N u F D_tF )
  2SWAP x*ext x+ext SIMP1ext		( eq y[x] F uN+D_tF )
  'IDt ODE_INT				( eq y[x] F int[uN+D_tF] )
  SYM= 'IDt ROT REPLACE2BY1		( eq res )
  SWAPDROP				( res )
;

LABEL HOMOGENE
* les solutions singulieres ne sont pas renvoyees
* attention, si l'equation est y'=y/x, il n'y a que des solutions singulieres
::					( eq )
   fnc_applied ode_var TWO{}N		( eq { y[x] x } )
   'IDt Z1 TWO{}N MEVALext		( f[t;1] )
   'IDt x-ext Z1 SWAPNDXF		( 1/[f[t,1]-t] )
   'IDt NOCAREEVAL7 RISCHext 
   SIMP1ext EXPEXPA			( F[t] )
   SYMBINCOMP 				( meta-F[t] )
   odeconst				( meta-F[t] C )
   SYMBOL Z1 xi ID t x* x+ ; INNERCOMP	( meta-F[t] C meta-1+i*t )
   #1+ addt*				( meta-F[t] meta-C*[1+i*t] )
   pshtop& addt* SYMBN			( C*[1+i*t]*F[t] )
   'IDt ode_var QUOTEAPPLY		( C*[1+i*t]*F[t] x[t] )
   'IDt fnc_name QUOTEAPPLY 		( C*F[t]*[1+i*t] x[t] y[t] )
   SWAPINCOMP get1			( C*F[t]*[1+i*t] meta-x[t] y[t] )
   'xi SWAPINCOMP #1+ addt* 		( C*F[t]*[1+i*t] meta-x[t] meta-i*y[] )
   top& addt+ SYMBN			( C*F[t]*[1+i*t] x[t]+i*y[t] )
   SWAP SYM=				( param_solution )
   "Homogeneous" ODETYPESTO
;

LABEL SEPARABLE
( separable variables )
::				( eq y[x] eq[t] x eq/eq' )
  ( * we might get the two factors by square-free factorization? * )
  ( * here we will integrate w.r.t. x to get the x factor * )
  FXNDext SWAPNDXF 		( eq y[x] eq[t] x eq'/eq )
  SWAP NOCAREEVAL7 RISCHext	( eq y[x] eq{t] ln-x-part )
  SIMP1ext EXPEXPA SWAPDROP	( eq y[x] x-part )
  ROTOVER x/ext			( y[x] x-part y-part )
  3PICK 'IDt MEVALext		( y[x] x-part y-part[t] )
  Z1 SWAP x/ext SIMP1ext 	( y[x] x-part 1/y-part[t] )
  'IDt RISCHext			( y[x] x-part int[y][t] )
  'IDt 4ROLL REPLACE2BY1	( x-part int[y] )
  SWAP ode_var ODE_INT		( int[y] int[x] )
  SYM=				( -> solution )
  "Separable" ODETYPESTO
;

LABEL ODE_VARX
( list -> list list_x )
* return list of var with x dependecy
::
  DUPDUP >R LENCOMP ZEROSWAP	( list meta_x )
  ZERO_DO (DO)
    RSWAP 'R RSWAP		( list meta-x var )
    ode_var Contains?		( list meta-x var x flag )
    SWAPDROP ITE SWAP#1+ DROP	( list meta-x )
  LOOP
  {}N
;

LABEL BERNOUILLI?
( Test for BERNOUILLI odetype, we know that #d!=0 and #N+1>2 ) 
:: 					( lvar deno_d #d_order num )
  INNERCOMP get1			( lvar deno_d nN..n0 #N+1 #d )
  BERNOUILLI1				( lvar deno_d nN n0 zN )
  4UNROLL  				( lvar zN deno_d nN n0 )
  4PICK QMul 3PICK QDiv			( lvar zN deno_d nN n0*zN/deno )
  SWAP					( lvar zN deno_d n0*zN/deno nN )
  4PICK QMul ROT QDiv 			( lvar zN nN*zN/deno_d n0*zN/deno_d )
  5ROLL CDRCOMP				( zN nN*zN/deno_d n0*zN/deno_d lvar' )
  DUPUNROT R2SYM			( zN nN*zN/deno_d lvar' n0*zN/deno_d )
  SWAP R2SYM SWAP			( zN nN*zN/deno_d n0*zN/deno_d )
  LINSOLV				( zN solution )
  fnc_applied ROT x^ext SWAP SYM=	( solution )
  "Bernouilli" ODETYPESTO
;

LABEL BERNOUILLI1
:: 					( nN..n0 #N+1 #d )
  DUP#0=csedrp BERNOUILLI0
					( nN..n0 #N+1 #d )
  OVER #-+1 #0= NcaseSIZEERR
					( nN...n0 #N+1 )
  #1-SWAP OVER#2+UNROL			( n0 nN..n1 #N )
  DUP#1= case SETSIZEERR
  DUP #1-SWAP 				( n0 nN nN-1..n1 #N-1 #N )
  OVER#2+UNROL				( n0 nN #N nN-1..n1 #N-1 )
  ZERO_DO (DO)
    TRIMOBJext				( n0 nN #N nN-1..nk )
    Z0 EQUAL NcaseSIZEERR
  LOOP					( n0 nN #N )
  ROTSWAP #>Z				( nN n0 zN )
;


LABEL BERNOUILLI0
:: 					( nN..n0 #N+1 )
  #1-SWAP TRIMOBJext			( nN..n1 #N n0 )
  Z0 EQUAL NcaseSIZEERR			( nN...n1 #N )
  SWAPOVER #1+UNROLL			( n1 nN...n2 #N ) 
  :: 
     DUP #2= case ROTSWAP		( -> n2 n1 #N )
					( n1 nN...n2 #N )
     DUP #1-SWAP OVER#2+UNROL		( n1 #N nN nN-1..n2 #N-1 )
     ONE_DO (DO)
     TRIMOBJext
       Z0 EQUAL NcaseSIZEERR
     LOOP
					( n1 #N nN )
     UNROT				( nN n1 #N )
  ;
  #1- #>Z QNeg				( nN n1 z1-N )
;

*--------------------------------------------------------------------
* Boundary conditions
*--------------------------------------------------------------------

LABEL SOLVFORBOUND
* equation -> meta
* [tries] to determine constant coefficients Ck using boundary equations
* LAM boundary
::
  DUP DEPTH stackdepth!		( eq eq )
  ERRSET SOLVBOUND		( eq meta )
  ERRTRAP BOUNDTRAP
  get1 DROP			( -> meta )
;

LABEL SOLVBOUND
* eq -> meta
* find value of constant coeff, SETSIZEERR if SOLVBOUND can not solve
::
  ( * construct system of equations from boundary * )
  ( * we accept only polynomial equations * )
  boundary INNERCOMP get1		( meta-cond eq )
  SWAPDUP
  ZERO_DO (DO)				( c1..cn eq #n )
    get1 3PICK				( c2..cn eq #n c1 eq )
    APPLYBOUND UNROT			( c2..cn C1 eq #n )
  LOOP
					( C1..Cn eq #n )
  DUPUNROT #2+UNROLL {}N		( eq system )
  ( * fetch number unknowns from ode# * )

  ode# 
  ZERO_DO (DO) INDEX@ odeconst# LOOP 
  ode# {}N				( eq {system} {unknown} )
  
  ( * solves system of equations * )

  BEZOUTMSOLV				( eq {unknown} liste_sols )
  SWAP ode_unknown!			( eq liste_sols )

  ( * replace solutions in eq * )

  INNERCOMP DUP#0=case SWAPDROP		( -> return null-meta )
  get1 SWAPDUP				( meta-sols' eq #n #n )
  ZERO_DO (DO)				( sol1..soln eq #n )
    get1 3PICK				( sol2..soln eq #n sol1 eq )
    ode_unknown ROT 			( sol2..soln eq #n eq unknown sol )
    BOUNDMEVAL				( sol2..soln eq #n eq1 )
    UNROT
  LOOP					( sol1..soln eq #n )
  SWAPDROP				( -> meta )
;

LABEL APPLYBOUND
* c1 eq -> C1
* applies condition 1 to eq returning a condition on the unknowns
* condition accepted: fnc_name[value]=value or d1fnc_name[value]=value
::
  SWAP
  DUPTYPESYMB? NcaseTYPEERR 		( must be an equation )
  					( eq c1 )
  INNERCOMP #1-SWAP			( eq meta-c1 op )
  EQ: x= NcaseTYPEERR
  larg CKSYMBN 
  OVER#2+UNROLL SYMBN			( eq right left )
  ( * must be ode_var[] user fcn * )
  F2ARG&NAME? NcaseTYPEERR				
					( eq value arg name )
  NAME2#NAME$ $>ID			( eq value arg #deriv name )
  fnc_name EQUAL NcaseTYPEERR
					( eq value arg #deriv )
  ( * now look if solution eq is parametric * )
  4PICK 'IDt Contains?			( eq value arg #deriv eq 't' flag )
  UNROT2DROP case HOMOBOUND
					( eq value arg #deriv )
  DUP#0=csedrp BOUND0

  ( * if y can not be uniquely isolated in eq, invalid boundary cond * )
  4ROLL fnc_applied ISOLALL		( value arg #deriv fnc_applied {} )
  DUPLENCOMP #1= NcaseSIZEERR
  CARCOMP				( value arg #deriv fnc_applied sol )
  SWAPDROPSWAP				( value arg sol #deriv )
  ZERO_DO (DO) ode_var DIRECTDERIV LOOP	( value arg d1..d1sol )
  ode_var ROT NR_REPLACE		( value symb )
  SIMP1ext SWAP SYM=			( condition )
;

LABEL BOUND0
* if #deriv=0 replace y[x] by value and x by arg
* eq value arg -> symb
::
  TWO{}N fnc_applied ode_var TWO{}N 	( eq { value arg } { y[x] x } )
  SWAP MEVALext				( condition )
  SIMP1ext				( condition )
;

LABEL HOMOBOUND
* boundary condition for a parametric solution x[t]+i*y[t]=f[t]
* since only homogeneous equations give such solution at this time
* y[t]=t*x[t], eq=C*[1+i*t]*F[t]
::					( eq value arg #deriv )
  4ROLL INNERCOMP plDRPpZparg		( value arg #deriv meta1 meta2 )
  NDROP pulldrop larg			( value arg #deriv meta-C* meta-F )
  SYMBN OVER#2+UNROLL			( value arg #deriv F[t] meta-C*[1+it] )
  pulldrop larg NDROP SYMBN		( value arg #deriv F[t] C )
  5UNROLL SWAP 				( C value arg F[t] #deriv )
  DUP#0=csedrp HOMOBOUND0
  #1= NcaseSIZEERR
					( C value arg F[t] )
  ( * dy/dx=[F+tF']/F=1+t*F'/F, then x=CF * )
  DUP 'IDt DIRECTDERIV			( C value arg F[t] F'[t] )
  OVER x/ext 'IDt x*ext Z1 x+ext	( C value arg F[t] 1+tF'/F )
  4ROLL SYM= 'IDt ISOLALL		( C arg F[t] t { value } )
  DUPLENCOMP #1= NcaseSIZEERR
  CARCOMP NR_REPLACE			( C arg F )
  QDiv SIMP1ext				( C arg/F )
  SYM=					( condition )
;

LABEL HOMOBOUND0
* boundary condition of order 0 for an homogeneous equation
::					( C y_0 x_0 F[t] )
  ( * y_0/x_0=t, x=CF[t] * )
  UNROT DUPUNROT QDiv			( C F[t] x_0 t_0 )
  ROT 'IDt ROT				( C x_0 F[t] t t_0 )
  NR_REPLACE				( C x_0 F[t_0] )
  QDiv SIMP1ext				( C x_0/F[t_0] )
  SYM=					( condition )
;

LABEL BOUNDMEVAL
* special replacement, does not make 'C=?' replacements
* this is necessary because BEZOUTMSOLV returns ? if every value of an
* unknown is allowed
* eq {} {} -> eq'
::					( eq unknown sols )
  SWAP MATINNERCOMP get1 SWAPDROP	( eq u1..un sols )
  MATINNERCOMP				( eq u1..un s1..sn #n )
  DUPDUP ZERO_DO (DO)			( eq u1..un s1..sn #n #n )
    BOUNDMEVALLP
  LOOP					( eq u1..un s1..sn #n #n )
  OVER#2+UNROLL {}N 			( eq u1..un sols )
  OBJ>R {}N R>OBJ			( eq unknown sols )
  MEVALext				( eq' )
;

LABEL MATINNERCOMP
::
  DUPTYPEMATRIX? 
  ITE INNERCOMP CKINNERCOMP
;

LABEL BOUNDMEVALLP
::  
  get1 ?ext 			( eq u1..un s2..sn #n #n s1 undef )
  OVER 				( eq u1..un s2..sn #n #n s1 flag )
  EQUALNOTcase UNROT		( -> eq u1..un s2..sn s1 #n #n )
				( eq u1..un s2..sn #n #n s1 )
  2DROP				( eq u1..un s2..sn #n )
  #1-DUP			( eq u1..un s2..sn #n-1 #n-1 )
  DUPDUP INDEX@#- #+ 		( eq u1..un s2..sn #n-1 #n-1 2n-2-k )
  #3+ROLL DROP			( eq u2..un s2..sn #n-1 #n-1 )
;


LABEL BOUNDTRAP
::
  ERROR@ # 203 #<>case ERRJMP 
  SHALT DEPTH #3+ stackdepth #- NDROP
					( eq )
  boundary OVER >HCOMP			( eq {eq} )
  fnc_applied ' xDESOLVE ONE SYMBN	( eq {eq} symb )
  TWO ' xFCNAPPLY FIVE SYMBN		( eq sol )
  ONE					( eq meta )
;

*--------------------------------------------------------------------
* Linear ode solvers
*--------------------------------------------------------------------

* ====================================================================
* Desc:		Integration with addition of a constant
* Stack:	( symb idnt -> symb )
* ====================================================================
NULLNAME ODE_INT
::				( symb, vx )
  RISCHext odeconst x+ext
;

LABEL odeconst
::
  1GETLAM DUP#1+ 1PUTLAM
  odeconst#
;

LABEL odeconst#
::
*ASSEMBLE
*	IF	f39
*RPL
*  UNCOERCE
*ASSEMBLE
*	ELSE	f39
*RPL
  #>Z 
*ASSEMBLE
*	ENDIF	f39
*RPL
  DO>STR CHR_C >H$ CHR_c >H$  $>ID
;


* ====================================================================
* Desc:		Solve y'=ay+b
* Stack:	( b a -> y )
* ====================================================================
NULLNAME LINSOLV
::
  SLOPPY? UNROT
  nocareflag SetSysFlag			( ? b a )
  ode_var RISCHext SIMP1ext		( ? b A )
  EXPEXPA				( ? b EXP[A] )
  SWAPOVER NDXFext			( ? EXP[A] b/EXP[A] )
  ode_var RISCHext			( ? EXP[A] int[b/EXP[A]] )
  odeconst x+ext SWAP x*ext		( ? y )
  nocareflag ROT ITE_DROP ClrSysFlag	( y )
;

* ====================================================================
* Desc:		linear diff equ with const coeff
* Stack:	( 2nd_member char_eq -> solution )
* ====================================================================
NULLNAME LDECSOLV
::
  ( 2nd membre, eq. caracteristique -> solution )
  SWAP LAPext OVER

  ( eq carac, laplace[2nd membre], eq. carac )
  VX>LVARext VXXLext FXNDext SWAP

  ( eq carac, laplace[2nd membre], lvar, deno, p-liste )
  DUPTYPELIST? NcaseSIZEERR
  DUPLENCOMP DUP#1= caseSIZEERR

  ( eq carac, laplace[2nd membre], lvar, deno, p-liste, #ordre+1 )
  Z0 SWAP#1-
  ZERO_DO (DO)
      ( eq carac, laplace[2nd membre], lvar, deno, p-liste, res )
      SWAPINCOMP pulldrop {}N
      ( eq carac, laplace[2nd membre], lvar, deno, res, p-liste )
      DUPUNROT
      ( eq carac, laplace[2nd membre], lvar, deno, p-liste, res, p-liste )
      INDEX@ 
*ASSEMBLE
*	IF	f39
*RPL
*	UNCOERCE
*ASSEMBLE
*	ELSE	f39
*RPL
	#>Z 
*ASSEMBLE
*	ENDIF	f39
*RPL
  DO>STR CHR_C >H$ CHR_c >H$  $>ID CASSYM* CASSYM+
  LOOP

  ( eq carac, laplace[2nd membre], lvar, deno, p-liste, res )
  SWAPDROPSWAP CASSYM/

  ( eq carac, laplace[2nd membre], lvar, res/deno )
  R2SYM

  ( eq carac, laplace[2nd membre], partie solut generale )
  x+ext SWAP CASSYM/ ILAPext
;

* ====================================================================
* Desc:		
* Stack:	( eq. carac -> sol generale )
* ====================================================================
NULLNAME LDEGENE
::
  ZERO COMPLEXEVAL7 RFACText	( sol, lvar, obj interne, meta- rac mult )
  DUP#0=csedrp SETSIZEERR
  %0 1LAMBIND 
  Z0 SWAP #2/
  ZERO_DO (DO)				( rn, mn, ..., r1, m1, gen )
     ROT Z0 TWO >POLYTRIM xEXPext	( rn, mn, ..., m1, gen, exp[r1*x] )
     ROTDUP %0> NcaseSIZEERR		( rn, mn, ..., gen, r1, m1 )
     COERCE ZEROSWAP
     ZERO_DO (DO)			( .., meta )
        1GETLAM %1+ DUP 1PUTLAM
ASSEMBLE
	IF	f39
RPL
ASSEMBLE
	ELSE	f39
RPL
        R>Z 
ASSEMBLE
	ENDIF	f39
RPL
	DO>STR CHR_C >H$ CHR_c >H$  $>ID SWAP#1+
     LOOP
     {}N
     SWAP CASSYM* x+ext
  LOOP

  ( sol part, lvar, obj interne, sol gene )
  ABND
  ZEROSWAP ROTDROP ROT VAL2ext
;

* ====================================================================
* Desc:
* Stack:	( 2nd membre, eq carac -> eq. carac, sol part )
* ====================================================================
NULLNAME LDEPART
::
  SWAP LAPext SWAP		( laplace[2nd membre], eq. carac )
  DUPUNROT CASSYM/ ILAPext
;

* ====================================================================
* Desc:
* Stack:	
* ====================================================================
NULLNAME LDSSOLVext
::
  ( 2nd membre, matrice -> solution )

  DTYPENDO? NcaseTYPEERR
  SWAP' FLASHPTR LAPext TOTEMPOB rpnQOBJext SWAP
  DUPLENCOMP 3PICK LENCOMP #= NOT?DIMERR

					( lap[2nd membre], matrice )
  TWO{}N DUP LISTOPext LVARDext		( {}, lvar )
  VX! VXXLext INCOMPDROP		( lvar, lapl[2nd membre], matrice )
  MADJ 2SWAP 2DROP			( lvar, v, M[A,p], pcar )
  SWAPINCOMP DUPDUP			( lvar, v, pcar, A1, ..., An, #n, #l )
  2LAMBIND

  ( lam1= nbre de lignes des matrices, lam2= nombre de matrices )

  ZERO_DO (DO)
     INDEX@ 2PUTLAM
     1GETLAM ZERO_DO (DO)
        1GETLAM ZERO_DO (DO)		( lvar v pcar An..A1 l1..lj a1..ak )
           1GETLAM JINDEX@ #+
           2GETLAM #+PICK
           2GETLAM #1+ NTHCOMPDROP
           JINDEX@ #1+ NTHCOMPDROP
        LOOP
        1GETLAM >POLYTRIM
     LOOP
     1GETLAM []N
  LOOP
  1GETLAM []N

  1GETABND DUPUNROT #2+UNROLL NDROP	( lvar, v, pcar, [p-A]^-1 )
  { ONE ZERO TWO } TOTEMPOB FROMLISText	( v, pcar, [p-A]^-1 )
  ROT DUPLENCOMP			( pcar, [p-A]^-1, v, # )
  DUP1LAMBIND
  #1+_ONE_DO (DO)
      INDEX@ 
*ASSEMBLE
*	IF	f39
*RPL
*	UNCOERCE
*ASSEMBLE
*	ELSE	f39
*RPL
	#>Z
*ASSEMBLE
*	ENDIF	f39
*RPL
      DO>STR CHR_V >H$ CHR_c >H$  $>ID
  LOOP
  1GETABND []N				( pcar, [p-A]^-1, v, v0 )
  VADD MVMULT				( pcar, v' )
  SWAPONE 				( v' pcar #1 )
  DEPTHEVAL:7 
  :: EvalNoCKx/ symcomp ILAPext ;	( v" )
;

* ====================================================================
* Desc:		Store ode type in variable ODETYPE
* Stack:	( type -> )
* ====================================================================
NULLNAME ODETYPESTO
::
  ' ID ODETYPE STO
;

* ====================================================================
* Desc:
* Stack:	( deno -> deno, [ deno-y, deno-x, TRUE | FALSE ] )
* ====================================================================
NULLNAME ODE_SEPAR
::
  NOTLISTcase :: Z1 OVER TRUE ;

  LPGCDext				( deno deno-x )
  2DUP LOPDext TRIMext			( deno deno-x deno-y )
  DUP FALSE
  BEGIN					( deno deno-x deno-y deno-y' bad? )
     DUP ITE TRUE
       :: SWAP DUPNULL{}? ROTSWAP ;
				( deno deno-x deno-y deno-y' bad? arreter? )
     NOT			( deno deno-x deno-y deno-y' bad? cont? )
  WHILE
     DROPDUP CDRCOMP
     SWAP CARCOMP  TYPELIST?
  REPEAT
				( deno, deno-x, deno-y, deno-y', bad? )
  case2drop DROPFALSE
  DROPSWAP TRUE
;

