**********************************************************************
* This file defines various useful assembly language utilities.
* This file defines various useful real math utilities.
**********************************************************************

INCLUDE ../CAS/global.s

**********************************************************************
* Various long precision constants
**********************************************************************

LABEL %%MAXREAL
%% 9.99999999999999E49999

LABEL %%-MAXREAL
%% -9.99999999999999E49999

* Mathematical constants

LABEL %%LNPI
%% 1.14472988584940017414		( ln[pi]         )

LABEL %%PI/2
%% 1.57079632679489661923		( pi/2           )

LABEL %%PI/4
%% 7.85398163397448309616E-1		( pi/4           )

LABEL %%SQRT2
%% 1.41421356237309504880		( sqrt[2]        )

LABEL %%SQRT3
%% 1.732050807568877293527		( sqrt[3]        )

LABEL %%INVSQRT2
%% 7.07106781186547524401E-1		( sqrt[2]/2      )

LABEL %%SQRT2/PI
%% 7.9788456080286535587989E-1		( sqrt[2/pi]     )

LABEL %%LN2
%% 6.93147180559945309417E-1		( ln[2]          )

LABEL %%LN4
%% 1.3862943611198906188E0		( ln[4]          )

LABEL %%INVLN2
%% 1.4426950408889634073599		( 1/ln[2]        )

LABEL %%LN2/2
%% 3.46573590279972654709E-1		( ln[2]/2        )

LABEL %%3PI/4
%% 2.35619449019234492885		( 3*pi/4         )

LABEL %%2/PI
%% 6.36619772367581343075535E-1     	( 2/pi           )

LABEL %%INVSQRTPI
%% 5.64189583547756286948E-1		( 1/sqrt[pi]     )

LABEL %%SQRT2PI
%% 2.50662827463100050242E0		( sqrt[2*pi]     )

LABEL %%LNSQRT2PI
%% 0.91893853320467274178		( ln[sqrt[2*pi]] )

LABEL %%EUL
%% 0.57721566490153286061		( Euler constant )

* Machine precision parameters

LABEL %%MAXLOG
%% 115129.254649702			( ln[maxnum] )

LABEL %%MACHEP
%% 1E-14				( 1+MACHEP > 1 )

**********************************************************************


**********************************************************************
* Desc:		Convert complex to extended complex
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL !C%>C%%
:: C%>%% %%>C%% ;

**********************************************************************
* Desc:		Complex inverse
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL C%%1/
:: %%1 SWAP C%%R/C ;

**********************************************************************
* Desc:		Complex square root
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL C%%SQRTRAD
::
  DUP C%%0= ?SEMI
  C%%>%% MAG
  3PICK %%ABS %%+ %%2 %%/ %%SQRT
  DUPDUP %%+ ROTSWAP %%/ ROT %%0>= case %%>C%%
  DUP %%ABS UNROT %%0>= ?SKIP %%CHS %%>C%%
;

**********************************************************************
* Desc:		Complex exponential
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL C%%EXP
::
  C%%>%%
  SWAP %%EXP OVER %%COSRAD OVER %%*ROT %%SINRAD ROT %%*
  %%>C%%
;

**********************************************************************
* Desc:		Complex logarithm
* Stack:	( C%% --> C%%' )
**********************************************************************

ASSEMBLE
*=DOREC%LNZ#0	EQU #105FA
*=C%%LNZ=0	EQU #103C0
*=%%CMP*CMP	EQU #0FEE2
RPL

LABEL C%%LN
::
  DUP C%%0= case C%%LNZ=0
  C%%>%% DOREC%LNZ#0 %%ANGLERAD %%>C%%
;

LABEL C%%LNZ=0
::
  handleDVZ
  DROP %%-MAXREAL %%0 %%>C%%
;

**********************************************************************
* Desc:		Complex argument
* Stack:	( C%% --> %%' )
**********************************************************************

LABEL C%%ARG
::
  C%%>%% %%ANGLERAD
;

**********************************************************************
* Desc:		Complex sine
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL C%%SIN
::
  C%%>%%
  2DUP %%COSH SWAP %%SINRAD %%*UNROT %%SINH SWAP %%COSRAD %%*
  %%>C%%
;

**********************************************************************
* Desc:		Complex cosine
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL C%%COS
::
  C%%>%%
  2DUP %%COSH SWAP %%COSRAD %%*UNROT %%SINH SWAP %%SINRAD %%* %%CHS
  %%>C%%
;

**********************************************************************
* Desc:		Complex tangent
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL C%%TAN
::
  C%%>%%
  DUP %%SINH DUPDUP %%* 4PICK %%COSRAD DUPDUP %%*ROT %%+
  5ROLL %%SINRAD ROT %%* OVER %%/
  4ROLL %%COSH 4ROLL %%*ROT %%/
  %%>C%%
;

**********************************************************************
* Desc:		Complex power
* Stack:	( C%% --> C%%' )
**********************************************************************

LABEL C%%R^C
::
  SWAP %%0 %%>C%% SWAP C%%C^C
;

LABEL C%%C^R
::
  %%0 %%>C%% C%%C^C
;

LABEL C%%C^C
::
  OVER C%%>%% %%0= SWAP %%0= ANDcase DOC%%W^ZW=0
  C%%>%%
  DUP %%0<> case
  :: ROT C%%>%% DOREC%LNZ#0 %%ANGLERAD %%CMP*CMP DOC%%EXP ;
  DROPSWAP C%%>%%
  DUP %%0<> ITE
  :: DOREC%LNZ#0 %%ANGLEDEG UNROTOVER %%*UNROT %%*SWAP
     %%EXP OVER %%COSDEG OVER %%* ROT %%SINDEG ROT
  ;
  :: DROP 2DUP %%ABS SWAP %%^ UNROT %%0 %%ANGLEDEG %%* DUP
     %%COSDEG SWAP %%SINDEG UNROTOVER %%*UNROT
  ;
  %%* %%>C%%
;

LABEL DOC%%W^ZW=0
::
  SWAPDROP C%%>%%
  OVER %%0> case2drop C%%0
  %%0= NOTcase
  :: DROP SETIVLERR ;
  DUP %%0= casedrop C%%1
  %%MAXREAL %%0 %%>C%%
  handleDVZ
;

LABEL DOC%%EXP
::
  SWAP %%EXP OVER %%COSRAD OVER %%*ROT %%SINRAD ROT %%*
  %%>C%%
;

**********************************************************************
* Desc:		Extended real equality test
* Stack:	( %%1 %%2 --> flag )
**********************************************************************

LABEL %%=
CODE
	P=	2		equality test
	GOSBVL	=POP2%%
	GOSBVL	=TST15
	P=	0
	SETHEX
	GOVLNG	=GPPushT/FLp
ENDCODE

**********************************************************************


ASSEMBLE

**********************************************************************
*		Closely packed reals
**********************************************************************

*	X1 = R1[S] R1[14:0] R4[4:0]
*	X2 = R2[S] R2[14:0] R4[9:5]
*	X3 = R3[S] R3[14:0] R4[14:10]
*
* Leaving R0 free for example for EXPF and LN15F
* Also R4[S] is free for sign saves etc

=STABX1		ABEX	W
		ABEX	S
		R1=A	W
		ABEX	W
		ABEX	S
		R4=A.F	A
		RTN

=STCDX1		CDEX	W
		CDEX	S
		R1=C	W
		CDEX	W
		CDEX	S
		R4=C.F	A
		RTN

=STABX2		ABEX	W
		ABEX	S
		R2=A	W
		ABEX	W
		ABEX	S
		CR4EX	W
		CSRC
		CSRC
		CSRC
		CSRC
		CSRC
		C=A	A
		CSLC
		CSLC
		CSLC
		CSLC
		CSLC
		CR4EX	W
		RTN

=STCDX2		CDEX	W
		CDEX	S
		R2=C	W
		CDEX	W
		CDEX	S
		AR4EX	W
		ASRC
		ASRC
		ASRC
		ASRC
		ASRC
		A=C	A
		ASLC
		ASLC
		ASLC
		ASLC
		ASLC
		AR4EX	W
		RTN

=STABX3		ABEX	W
		ABEX	S
		R3=A	W
		ABEX	W
		ABEX	S
		CR4EX	W
		CSLC
		CSLC
		CSLC
		CSLC
		CSLC
		CSLC
		C=A	A
		CSRC
		CSRC
		CSRC
		CSRC
		CSRC
		CSRC
		CR4EX	W
		RTN		

=RCABX1		A=R1	W
		B=A	W
		A=B	S
		B=0	S
		A=R4.F	A
		RTN

=RCABX2		A=R4	W
		ASR	W
		ASR	W
		ASR	W
		ASR	W
		ASR	W
		B=A	W
		A=R2	W
		ABEX	W
		ABEX	S
		RTN

=RCABX3		A=R4	W
		ASLC
		ASLC
		ASLC
		ASLC
		ASLC
		ASLC
		B=A	W
		A=R3	W
		ABEX	W
		A=B	S
		B=0	S
		RTN

=RCCDX1		C=R1	W
		D=C	W
		D=0	S
		C=R4.F	A
		RTN

=RCCDX2		C=R4	W
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		D=C	W
		C=R2	W
		CDEX	W
		CDEX	S
		RTN

=RCCDX3		C=R4	W
		CSLC
		CSLC
		CSLC
		CSLC
		CSLC
		CSLC
		D=C	W
		C=R3	W
		CDEX	W
		C=D	S
		D=0	S
		RTN

**********************************************************************
*		 Miscellaneous arithmetic
**********************************************************************

**********************************************************************
* Name:		ADDF1
* Input:	X
* Output:	X=X+1
* Alters:	A B C D D0
* Stack:	1	(call to Y=1)
**********************************************************************

=ADDF1		GOSUB	=Y=1
		GOVLNG	=ADDF

**********************************************************************
* Name:		SUBF1
* Input:	X
* Output:	X=X-1
* Alters:	A B C D D0
* Stack:	1	(call to Y=1)
**********************************************************************

=SUBF1		GOSUB	=Y=1
		C=-C-1	S
		GOVLNG	=ADDF

**********************************************************************
* Name:		SUBF
* Input:	X Y
* Output:	X=X-Y
* Alters:	A B C D D0
* Stack:	0
**********************************************************************

=SUBF		C=-C-1	S
		GOVLNG	=ADDF

**********************************************************************
* Name:		PUTCD1
**********************************************************************

=PUTCD1		DAT1=C	A
		D1=D1+	5
		CDEX	W
		C=D	S
		DAT1=C	W
		D1=D1+	16
		C=0	S
		CDEX	W
		RTN

**********************************************************************
* Name:		SWAPDIVF
* Input:	X Y
* Output:	X=Y/X
**********************************************************************

=SWAPDIVF	GOSBVL	=XYEX
		GOVLNG	=DIVF

**********************************************************************
* Name:		SQF
* Input:	X
* Output:	X=X^2
**********************************************************************

=SQF		C=B	W
		D=C	W
		C=A	W
		GOVLNG	=MULTF

**********************************************************************
* Name:		INVFMULPI
* Input:	X
* Output:	X=PI/X
**********************************************************************

=INVFMULPI	GOSUB	=Y=PI
		GOSBVL	=XYEX
		GOVLNG	=DIVF

**********************************************************************
* Name:		ATOREAL
* Input:	A[A] = signed BCD integer
* Output:	X = float
**********************************************************************

=ATOREAL	B=0	W
		B=A	A
		A=A+A	A		CS if A[A] is negative
		A=0	W
		GONC	+
		B=-B	A
		A=-A-1	S
+		A=A-1	A
-		A=A+1	A
		BSRC
		?B#0	A
		GOYES	-
		BSR	W
		RTN

**********************************************************************
* Name:		CTOREAL
* Input:	C[A] = signed BCD integer
* Output:	Y = float
**********************************************************************

=CTOREAL	D=0	W
		D=C	A
		C=C+C	A		CS if C[A] is negative
		C=0	W
		GONC	+
		D=-D	A
		C=-C-1	S
+		C=C-1	A
-		C=C+1	A
		DSRC
		?D#0	A
		GOYES	-
		DSR	W
		RTN

**********************************************************************
* Name:		Y=1
* Input:
* Output:	Y = 1
**********************************************************************

=Y=1		P=	1
=Y=P		C=0	W
		D=0	W
		C=P	14
		P=	14
		D=C	P
		P=	0
		RTN

**********************************************************************
* Name:		X=1
* Input:
* Output:	X = 1
**********************************************************************

=X=1		P=	1
=X=P		A=0	W
		B=0	W
		CBEX	W
		C=P	14
		CBEX	W
		P=	0
		RTN

**********************************************************************
* Name:		Y=PI
* Input:
* Output:	Y = PI
**********************************************************************

=Y=PI		D0=(5)	=%%PI
		D0=D0+	5
		GOVLNG	=GETCD0

**********************************************************************
* Name:		aSINRAD
* Input:	X
* Output:	X = SIN(X) in radian mode  (R0/R1=abs(1/tan(x)))
* Alters:	A,B,C,D R0,R1 D0 P SB XM
* Note:		ST11 is set if tan(x) is negative
**********************************************************************

=aSINRAD	ST=1	9		sRAD=1
		ST=0	4		sGRAD=0
		GOVLNG	=aSIN15F

**********************************************************************
* Name:		aCOSRAD
* Input:	X
* Output:	X = COS(X) in radian mode  (R0/R1=abs(tan(x)))
* Alters:	A,B,C,D R0,R1 D0 P SB XM
* Note:		ST11 is set if tan(x) is negative
**********************************************************************

=aCOSRAD	ST=1	9		sRAD=1
		ST=0	4		sGRAD=0
		GOVLNG	=aCOS15F

**********************************************************************
* Name:		aCOSRAD
* Input:	X
* Output:	X = COS(X) in radian mode
* Alters:	A,B,C,D R0,R1 D0 P SB XM
**********************************************************************

=aTANRAD	ST=1	9		sRAD=1
		ST=0	4		sGRAD=0
		GOVLNG	=aTAN15F

**********************************************************************
* Name:		aEXPBIG
* Input:	X
* Output:	X = EXP(X)
* Notes:	The built-in EXP is limits the results below
*		exponent 20000. This allows me to use it up
*		to MAXREAL precision.
**********************************************************************

=aEXPBIG	P=	0
		LCHEX	0460517018598809	LIM = 46051.7018598809
		D=C	W
		C=0	W
		LCHEX	4

		GOSUB	=TSTLT		CS if Y < LIM
		GOC	++		If so, use regular EXP

		C=-C-1	S
		GOSBVL	=ADDF		Y = Y - LIM
		GOSUB	++		exp(Y-LIM)

* And multiply by exp(LIM) = 9.999999999863196 40170999888838 E 19999

		P=	0
		LCHEX	0999999999863196
		D=C	W
		C=0	W
		LCHEX	19999
		GOVLNG	=MULTF

++		GOVLNG	=aEXP15F


**********************************************************************
* Name:		TSTEQ
* Input:	X Y
* Output:	X Y  CS if X==Y
**********************************************************************

=TSTEQ		P=	2
		GOVLNG	=TST15

**********************************************************************
* Name:		TSTLT
* Input:	X Y
* Output:	X Y  CS if X<Y
**********************************************************************

=TSTLT		P=	1
		GOVLNG	=TST15

**********************************************************************
* Name:		TSTGT
* Input:	X Y
* Output:	X Y  CS if X>Y
**********************************************************************

=TSTGT		P=	4
		GOVLNG	=TST15

**********************************************************************
* Name:		TSTGE
* Input:	X Y
* Output:	X Y  CS if X>=Y
**********************************************************************

=TSTGE		P=	6
		GOVLNG	=TST15

**********************************************************************
* Name:		TSTEQ1
* Input:	X
* Output:	CS if X=1
**********************************************************************

=TSTEQ1		P=	14
		A=B	P
		B=0	S
		?B=0	W
		GOYES	+		CS if mantissa is 1
		B=A	P
		P=	0
		RTNNC			CC: mantissa was not 1
		?A#0	A
		GOYES	+		exponent not zero - not 1
		?A=0	S
		RTNYES			CS: sign is correct too - is 1
+		RTNCC			CC: is not 1

**********************************************************************
* Name:		MULTPI
* Input:	X
* Output:	X = X*PI
* Alters:	A,B,C,D
**********************************************************************

=MULTPI		GOSUB	=Y=PI
		GOVLNG	=MULTF

**********************************************************************
* Name:		TSTEVEN
* Input:	X
* Output:	X	CS if FLOOR(MOD(X,2))==0
* Alters:	C P=0
* Notes:	The fractional part is ignored so that
*		2.1 is even, 2.9 is even, 3.123 is odd etc
**********************************************************************

=TSTEVEN	P=	0
		?B=0	W
		RTNYES			CS: zero is even
		C=A	A
		C=C+C	A
		RTNC			CS: numbers < 1 are even
		LCHEX	00015
		?A>=C	A
		RTNYES			CS: numbers > 10^15 are even

		C=A	A		convert exponent to hex
		CSR	A		exponent/10
		?C=0	A
		GOYES	+
		LCHEX	A
		SETHEX
+		C=C+A	A
		SETDEC
		P=C	0		shift mantissa
		C=B	W
-		CSL	W
		P=P-1
		GONC	-
		P=C	15
		CPEX	0		C[0] = nibble	(P=0 due to CSL)
		?CBIT=0	0
		RTNYES			CS: is even
		RTNCC			CC: is odd

**********************************************************************
* Name:		INFRES
* Input:	-
* Output:	X = MAXREAL
* Alters:	A B D0
* Errors:	If infinite results disallowed, generates error instead
**********************************************************************

=INFRES		P=	0
		D0=(5)	(=SystemFlags)+(OverfActNib)	Test if overflow
		A=DAT0	B				should error or
		?ABIT=0	(OverfActBit)			just return maxreal
		GOYES	infresdef

		SETHEX
		LCHEX	00303	ERROR(OVERFLOW)
		GOVLNG	=GPErrjmpC

infresdef	D0=(5)	(=SystemFlags)+(OverflowNib)
		A=DAT0	B
		ABIT=1	(OverflowBit)
		DAT0=A	B

		SETDEC
		B=0	W
		B=B-1	W
		B=0	S
		A=0	W
		LAHEX	49999
		P=	3		Indicate infinite result
		RTNSXM

**********************************************************************

* %%LNSQRT2PI = ln(sqrt(2*pi)) = 0.91893853320467274178

=Y=LNSQRT2PI	P=	0
		LCHEX	0918938533204673
		GOTO	+

* %%INVSQRTPI = inv(sqrt(pi)) = 5.6418958354775628695E-1

=Y=INVSQRTPI	P=	0
		LCHEX	0564189583547756
+		D=C	W
		C=0	W
		C=C-1	A
		RTN

* %%LNPI = ln(pi) = 1.14472988584940017414 
=Y=LNPI		P=	0
		LCHEX	0114472988584940
		D=C	W
		C=0	W
		RTN

* %%PI/2 = 1.57079632679489661923

=Y=PI/2		P=	0
		LCHEX	0157079632679490
		D=C	W
		C=0	W
		RTN

* %%EUL = 0.57721566490153286061

=Y=EUL		P=	0
		LCHEX	0577215664901533
		D=C	W
		C=0	W
		C=C-1	A
		RTN
RPL


**********************************************************************
