INCLUDE cas2romid.h
INCLUDE erextdec.h
INCLUDE cas2ext.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE Sign of a polynomial using Sturm sequences
	
* Interface
* FLAGSTURM STURM:	P -> []		compute the Sturm sequence
* sturm:		Q -> meta	internal version
* MAPSTURM:		[P] a -> [P[a]]	evaluate array of poly at a
* SIGNCHANGE#:		[] -> #		find # of sign change in an array
* ZERO#IN[AB]		P a b -> #	find # of real roots of poly P in [a,b]
*
* Notes:	does not use the subresultant algorithm
*		Warning: all polynomials are assumed to be square-free

******************************************************************
* Desc.:	compute the Sturm sequence of polynomials from P
* Stack:	P -> list of [] of P | cst coeff
* Note:		works on symb-poly, protect flags
******************************************************************
NULLNAME FLAGSTURM
::
   					( P )
  ' :: VX>LVARext FLASHPTR STURM ;
  TOTEMPOB CASFLAGEVAL		( meta )
  {}N
;
	
******************************************************************
* Desc.:	compute the Sturm sequence of polynomials from P
* Stack:	P lvar -> meta of []_of_P  %mult
* Note:		works on symb-poly
******************************************************************
NULLNAME STURM
::
  					( P lvar )
  ( * keep num only * )
  VXXLext FXNDext			( lvar N D )
  ( * replacing D by D^-1 does not change sign * )
  ROT ZERO 2LAMBIND			( N D )
  RFACTSTEP3 %-1 METAMM2		( N meta )
  get1 RFACTSTEP3 top&			( meta )
  DUP#0=csedrp :: ABND Z1 %1 TWO ;
					( meta )
  DUP #2- SWAP #2/
  ZERO_DO (DO)				( f1 m1 f2..fn mn #2n-2 )
    get1 rotomo				( f2..fn mn #2n-2 m1 f1 )
    ZEROSWAP sturm			( f2..fn mn #2n-2 m1 # meta )
    METAVAL2 []N			( f2..fn mn #2n-2 m1 []_of_polys )
    SWAPROT 				( f2..fn mn []_of_polys m1 #2n-2 )
  LOOP					( f1 m1 meta )
  #2+ ABND 
;

******************************************************************
* Desc.:	compute the Sturm sequence of polynomials from P
* Stack:	P -> meta-P
* Note:		works on list-poly
******************************************************************
NULLNAME sturm
::				( P )
  EXACT? NcaseTYPEERR
  DUPTYPELIST? NOTcase ONE
				( P )
  DUP PDer QNeg	TWO		( P -P' #2 )
  [BEGIN]
    OVER Z0 EQUAL		( meta ? )
  NOT_WHILE
    sturmlp
  [REPEAT]
  ( * remove 0 polynomial * )
  pulldrop			( meta )
;

LABEL sturmlp
::
    3PICK 3PICK			( P1..Pk #k Pk-1 Pk )
    NEWDIVext SWAPDROP		( P1..Pk #k rem )
    FXNDext ZSIGNECK SWAPDROP	( P1..Pk #k remN remDsign )
    IT QNeg			( P1..Pk #k +/-remN )
    SWAP#1+			( meta )
;

******************************************************************
* Desc.:	compute the Sturm sequence of polynomials from P at a
* Stack:	[]_of_P, var, a -> []_of_P[a]
* Note:		work with symb P
******************************************************************
NULLNAME MAPSTURM
::				( [] var a )
  SWAP 1LAMBIND			( [] a )
  SWAPDUP 			( a [] [] )
  LENCOMP			( a [] # )
  SWAP >R			( a # )
  ZERO UNROT			( meta a # )
  ZERO_DO (DO)			( meta a )
    RSWAP 'R RSWAP		( meta a P )
    OVER STURMHORN		( meta a val )
    UNROT SWAP#1+SWAP		( meta a )
  LOOP
  DROP ABND []N			( [] )
;

LABEL STURMHORN
::					( P r )
  TWO dup LISTOPext LVARDext		( P r #2 {} )
  1GETLAM prepvarlist			( P r #2 {} )
  DUP1LAMBIND OVER#2+UNROLL		( {} P r #2 )
  METALISTVXXL DROP ABND CKCARCOMP	( lvar P-list r )
  HORNext				( lvar P_div_X-r r P[r] )
  UNROT2DROP ONE>POLY			( lvar P[r] )
  R2SYM					( P[r] )
;


******************************************************************
* Desc.:	find # of sign changes in a vector
* Stack:	[] -> #
* Note:		elements of [] are symbolic
******************************************************************
NULLNAME SIGNCHANGE#
::
					( [] )
  2DMATRIX? caseSIZEERR
  >R					( )
  ( * get first non-zero element * )
  BEGIN
    RSWAP 'R RSWAP DUP NULLCOMP?	( a0 false | ::; true )
    MAPSTURM1				( z false | true )
  UNTIL					( z )
  num#0=case7 DROPZERO
					( z )
  ZERO					( z # )
  [BEGIN]
    RSWAP 'R RSWAP DUP NULLCOMP?	( z # ak false | z # ::; true )
  NOT_WHILE				( z # ak )
    MAPSTURM2
  [REPEAT]
					( z # ::; )
  DROPSWAPDROP
;

LABEL MAPSTURM1
* ::; true -> z0 true
* z0 false -> false
* !=0 false -> +-1 true
::					( ::; true | a0 false )
  casedrop Z0TRUE			( -> 0 true )
  num#0=case7 DROPFALSE			( -> false )
  ZSIGNECKZ TRUE			( -> a0 true )
;

LABEL ZSIGNECKZ
::
  ZSIGNECK SWAPDROP
  case Z1 Z-1
;

LABEL MAPSTURM2
* z # ak -> newz #
* if ak=0 does nothing
* else take sign, if same as z does nothing, otherwise -> newz and #++
::				( z # ak )
  num#0=case7 DROP		( -> z # )
				( z # ak )
  ZSIGNECKZ 			( z # zk )
  ROTOVER EQUAL			( # zk z=zk? )
  case SWAP SWAP#1+		( -> zk # )
;

LABEL Z0TRUE
::
  Z0 TRUE
;

******************************************************************
* Desc.:	find # of real roots & poles of a rat frac in an interval
* Stack:	P a b -> cst_coeff zint
* Note:		P is a symb polynomial wrt the current var, a and b are symb
******************************************************************
NULLNAME STURMAB
::
  RCLVX
  ' FLASHPTR ZERO#INAB TOTEMPOB CASFLAGEVAL
  #>Z
;

******************************************************************
* Desc.:	find # of real roots & poles of a rat frac in an interval
* Stack:	P a b var -> cst_coeff #
* Note:		P is a symb polynomial wrt the current var, a and b are symb
******************************************************************
NULLNAME ZERO#INAB
::
  ( * lam5=a, lam4=b, lam3=var, lam2=cst_coeff, lam1=#signchange * )
  Z1 ZERO ' NULLLAM FIVE NDUPN DOBIND 	( P )
  DUP LISTOPext LVARDext 3GETLAM prepvarlist STURM	( meta_of_[]_ % )
  #2/
  ZERO_DO (DO)
    ZERO#INABLP
  LOOP
  2GETLAM 1GETABND
;

LABEL ZERO#INABLP
* [] % ->
* adjust #sign, else adjust cst_coeff
::					( [] mult )
  %ABSCOERCE SWAPDUP 			( mult [] [] )
  3GETLAM 5GETLAM MAPSTURM 		( mult [] [] )
  ( * adjust sign with P[a] * )
  3PICK TWO #/ DROP #0=			( mult [] [] ? )
  ?SKIP ZERO#INABLPLAM2
  SIGNCHANGE#				( mult [] # )
  SWAP 3GETLAM 4GETLAM MAPSTURM 	( mult # [] )
  SIGNCHANGE#				( mult # # )
  2DUP#< ITE SWAP#- #-			( mult # )
  #* 1GETLAM #+ 1PUTLAM
;

LABEL ZERO#INABLPLAM2
:: DUP CARCOMP 2GETLAM QMul 2PUTLAM ;

******************************************************************
* Desc.:	large sign of a fraction
* Stack:	P lvar a b -> 1/0/-1
* Note:		
******************************************************************
NULLNAME STURMLSIGN
::
  
  ( * like ZERO#INAB but throw even multiplicity * )
  ( * lam5=a, lam4=b, lam3=var, lam2=cst_coeff, lam1=#signchange * )
  3PICK CARCOMP 			( P lvar a b var )
  Z1 ZERO ' NULLLAM FIVE NDUPN DOBIND 	( P lvar )

  ( * first look if P[a] or P[b] =0 * )
  VXXLext				( lvar P )
  5GETLAM ROT VXXLext			( P lvar a )
  4GETLAM ROT VXXLext			( P a lvar b )
  4ROLL					( a lvar b P )
  FXNDext 				( a lvar b n d )
  3PICK FALSE STURMREDUCE 		( a lvar b n d' )
  5PICK TRUE STURMREDUCE		( a lvar b n d" )
  SWAPROT FALSE STURMREDUCE		( a lvar d" n' )
  4ROLL TRUE STURMREDUCE		( lvar d" n" )
  SWAP SIMPNDXFext			( lvar P' )
  ZEROSWAP 3PICK VAL2ext		( lvar P )
  SWAP STURM				( meta_of_[]_or_cst_coeff | % )
  #2/
  ZERO_DO (DO)
    DUP %2 %MOD %0= ITE 2DROP ZERO#INABLP
    ( * if lam1!=0 then exit * )
    1GETLAM #0= ?SKIP :: ISTOP-INDEX #1- #2* NDROP ExitAtLOOP ;
  LOOP
  2GETLAM 1GETABND #0= NOTcasedrop Z0 ZSIGNE
;

******************************************************************
* Desc.:	Divide P by a-X (false) or X-a (true) as long as P[a]=0
* Stack:	P a ? -> P'
* Note:		P and a must be in internal form
******************************************************************
NULLNAME STURMREDUCE
::		
  1LAMBIND				( P a )
  [BEGIN]
    OVERSWAP HORNext			( P P_div_X-a a P[a] )
    Z0 EQUAL				( P P_div_X-a a ? )
  [WHILE]
    ROTDROP?QNeg			( P_div_X-a a )
  [REPEAT]
					( P P_div_X-a a )
  2DROP					( P )
  ABND
;

LABEL ROTDROP?QNeg
::					( P P_div_X-a a )
  ROTDROP 1GETLAM ?SEMI 		( -> P_div_X-a a )
  SWAP QNeg SWAP			( -> -P_div_X-a a )
;

******************************************************************
* Desc.:	strict sign of a polynomial
* Stack:	P lvar a b -> 1/0/-1
* Note:		
******************************************************************
NULLNAME STURMSSIGN
::
  ( * like ZERO#INAB but throw negative even multiplicity * )
  ( * lam5=a, lam4=b, lam3=var, lam2=cst_coeff, lam1=#signchange * )
  3PICK CARCOMP 			( P lvar a b var )
  Z1 ZERO ' NULLLAM FIVE NDUPN DOBIND 	( P lvar )
  STURM					( meta_of_[]_or_cst_coeff | % )
  #2/
  ZERO_DO (DO)
    DUP %0< OVER %2 %MOD %0= AND ITE 2DROP ZERO#INABLP
    ( * if lam1!=0 then exit * )
    1GETLAM #0= ?SKIP :: ISTOP-INDEX #1- #2* NDROP ExitAtLOOP ;
  LOOP
  2GETLAM 1GETABND #0= NOTcasedrop Z0 ZSIGNE
;


******************************************************************
* Desc.:	split an assume hypothesis in < and > parts
* Stack:	symb -> sym< sym>
* Note:		
******************************************************************
NULLNAME SPLITHYPOTHESIS
::
  DUP EXPR> SWAP %2 %= NcaseSIZEERR
					( sym sym1 sym2 op )
  DUPEQ: xAND casedrop ROTDROP		( -> sym1 sym2 )
  					( sym sym1 sym2 op )
  UNROT2DROP				( sym op )
  DUPEQ: x< casedrop Z1
  EQ: x<= case Z1
  Z1 SWAP
;

******************************************************************
* Desc.:	find assumption interval on var
* Stack:	sym -> low high
* Note:		
******************************************************************
NULLNAME VARINTERVAL
::
  idntlamcase IDLAMINTERVAL
  ( * I do not take care of the arg currently ! * )
  LASTCOMP					( op )
  { xSIN VARSIN xCOS VARSIN xATANH VARSIN
    xABS VARABS x\v/ VARABS xEXP VARABS xCOSH VARABS xACOSH VARABS
    xATAN VARATAN xASIN VARATAN xACOS VARATAN
    xe VARe x\pi VARpi
  } EQLookup					( seco TRUE | op FALSE )
  NOTcasedrop MINUSPLUSINFext
  EVAL
;

LABEL VARSIN
:: Z-1 Z1 ;

LABEL VARABS
:: Z0 PLUSINFext ;

LABEL VARATAN
:: pisur-2 pisur2 ;

LABEL VARe
:: Z2 Z3 ;

LABEL VARpi
:: Z3 Z4 ;

******************************************************************
* Desc.:	find assumption interval on idnt/lam
* Stack:	id/lam -> low high
* Note:		
******************************************************************
NULLNAME IDLAMINTERVAL
::
  ISIDREAL? 				( id obj TRUE | id FALSE )
  NOTcasedrop MINUSPLUSINFext
					( id obj )
  SWAPDROP DUPTYPESYMB? NOTcasedrop MINUSPLUSINFext
					( symb )
  SPLITHYPOTHESIS			( x< x> )
  IDLAM>VALUE SWAP IDLAM<VALUE		( low high )
;

LABEL MINUSPLUSINFext
:: MINUSINFext PLUSINFext ;

LABEL IDLAM>VALUE
:: 
  DUPTYPESYMB? NOTcasedrop MINUSINFext
  IDLAMVALUE
;

LABEL IDLAM<VALUE
:: 
  DUPTYPESYMB? NOTcasedrop PLUSINFext
  IDLAMVALUE
;

LABEL IDLAMVALUE
::
  EXPR> DROP %2 %= NcaseSIZEERR
  SWAPDROP				( -> value )
;

******************************************************************
* Desc.:	Translate sign tab to OR symbolic
* Stack:	list flag -> symb
* Note:		flag is true for strict and false for large
******************************************************************
NULLNAME SIGNTABTOSYMB
::
  ITE :: ' x< ' x> ; :: ' x<= ' x>= ;	( list op< op> )
  RCLVX 3LAMBIND			( list )
  ( * lam3=op<, lam2=op>, lam1=var * )
  ZEROSWAP DUPLENCOMP #2/		( meta-res list #signs )
  SWAP >R				( meta-res #signs )
  'R SWAP				( meta-res inf #signs )
  ZERO_DO (DO)				( meta-res inf )
    RSWAP 'R'R RSWAP			( meta-res inf sign sup )
    ( * if sign=- then sup -> inf * )
    SWAP EQ: x+ 			( meta-res inf sup ? )
    ITE SIGNTABTOSYMB+ SWAPDROP 
  LOOP					( meta-res last_sup )
  DROP DUP#0= NOTcase SYMBN
  ( * no solution * )
  DROP ' ID Impossible TOTEMPOB
;

LABEL SIGNTABTOSYMB+
* meta-res inf sup -> meta-res' sup
::
  DUP OBJ>R				( meta inf sup )
  SWAPDUP MINUSINFext EQUAL		( meta sup inf ?=-inf )
  ITE SIGNTABTOSYMB- SIGNTABTOSYMBNOT- 	( meta meta' )
  DUP #2+PICK #0= 			( meta meta' ? )
  ITE top& :: top& addt: xOR ;		( meta' )
  R>OBJ					( meta' sup )
;

LABEL SIGNTABTOSYMB-
::					( meta sup -inf )
  DROPDUP PLUSINFext EQUAL		( meta sup ?=+inf )
  casedrop :: 1GETLAM ONE ;
  1GETLAM SWAP 3GETLAM THREE
;

LABEL SIGNTABTOSYMBNOT-
::					( meta sup inf )
  1GETLAM SWAP 2GETLAM THREE		( meta sup x>inf-meta )
  get1 DUP PLUSINFext EQUAL		( meta x>inf-meta sup ?=inf )
  caseDROP				( -> meta x>inf-meta )
					( meta x>inf-meta sup )
  1GETLAM SWAP 3GETLAM THREE		( meta x>inf-meta x<sup-meta )
  top& addt: xAND			( meta meta )
;