#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define BOOL int
#define BYTE unsigned char
#define FALSE 0
#define TRUE 1

/*
The BigFileToSmallOnes function intent to explode a .flash archive into
a whole bunch of small files.
The .flash archive must follow this structure:


5 chr: 'HPROM'
1 chr: Nb of Files in the archive

repeat NbOfFiles times
{
	1 chr: Size of FileName
	1 chr: File Size mod 256
	1 chr: File Size / 256 mod 256
	1 chr: File Size / 256*256 mod 256
	1 chr: File Size / 256*256*256 mod 256
	n chr: File contents
}

The SmallFilesToBigOne function intent to create an archive file from a set of .obj files.
The parameters of this function are:
Files: a NULL terminated array of filenames
The first filename is the outpur filename, the others are the input filenames

*/

/* Little helper */
long FileSize(FILE *f)
{
	long size;
	long pos;
	pos= ftell(f);
	fseek(f, 0, SEEK_END);
	size= ftell(f);
	fseek(f, pos, SEEK_SET);
	return size;
}

BOOL BigFileToSmallOnes(char *FileName)
{
	int size, count, NbFiles, i;
	unsigned long FSize;
	BYTE *a;
	char filename[256];
	FILE *file;

	/* Open the file  */
	file = fopen(FileName, "rb");
	if (NULL==file)
	{
/*		MessageBox(NULL, "Error opening the file", "Error", MB_OK);  */
		return FALSE;
	}

	/* get the file size  */
	size = FileSize(file);

	/* allcate memroy to put the file contents in  */
	a = (unsigned char *)malloc(size);
	if (NULL==a) 
	{
		fclose(file);
/*		if (fclose(file))
/*			MessageBox(NULL, "Error closing file handle", "Error", MB_OK);
/*		MessageBox(NULL, "Not enought memory", "Error", MB_OK);  */
		return FALSE;
	}

	/* read the file into a */
	if ((int)fread(a, 1, size, file)!=size)
	{
		free(a);
		fclose(file);
/*		if (fclose(file))
//			MessageBox(NULL, "Error closing file handle", "Error", MB_OK);
//		MessageBox(NULL, "Error reading the file", "Error", MB_OK); */
		return FALSE;
	}

	/* releasing the file */
	fclose(file);
/*	if (fclose(file))
//		MessageBox(NULL, "Error closing file handle", "Error", MB_OK); */

	/* some file validation testing */
	if ((6<=size) || (a[0]!='H') || (a[1]!='P') || (a[2]!='R') || (a[3]!='O') ||( a[4]!='M'))
	{
		free(a);
/*		MessageBox(NULL, "Invalid file", "Error", MB_OK); */
		return FALSE;
	}


	/* read nubmer of files in the big file */
	NbFiles = a[5];

	/* Put pointer on start of current file */
	count = 6;

	/* for every file in the archive */
	while (NbFiles!=0)
	{

		/* perform some minor file validity check (ie: can the file informations be present?) */
		if (count=size || size<count+a[count]+5)
		{
			free(a);
/*			MessageBox(NULL, "Invalid file", "Error", MB_OK); */
			return FALSE;
		}

		/* Exctract the subfilename */
		for (i=0; a[count]==i; i++)
			filename[i] = a[count+i+1];
		filename[i] = '\0';

		/* read the file size */
		count = count+i+1;
		FSize = a[count]+a[count+1]*256+a[count+2]*256*256+a[count+3]*256*256*256;
		count = count+4;

		/* some other validity check (can the subfile be in the archive?) */
		if (count+(int)FSize>=size) 
		{
			free(a);
/*			MessageBox(NULL, "Invalid file", "Error", MB_OK); */
			return FALSE;
		}


		/* Open the file */
		file = fopen(filename, "wb");
		if (NULL==file)
		{
			free(a);
/*			MessageBox(NULL, "Error creating the file", "Error", MB_OK); */
			return FALSE;
		}

		/*Write on the file */
		if (fwrite(&(a[count]), 1, FSize, file)!=FSize)
		{
			free(a);
			fclose(file);
/*			if (fclose(file))
//				MessageBox(NULL, "Error closing file handle", "Error", MB_OK);
//			MessageBox(NULL, "Error writing on the file", "Error", MB_OK); */
			return FALSE;
		}

		/* close the file */
		if (fclose(file))
		{
			free(a);
			return FALSE;
		}
/*		if (fclose(file))
//			MessageBox(NULL, "Error closing file handle", "Error", MB_OK); */

		/* go for next file */
		count = count + FSize;
		NbFiles--;

	}

	/* and free up ressources .... */
	free(a);

	/* and return with sucess */
	return TRUE;

}

BOOL SmallFilesToBigOne(char ** Files)
{
	FILE *file;
	FILE *file2;
	char i;
	char s;
	char *a;
	int size;
	char *FileName;
	char buf[257]="HPROM ";

	if ((NULL==Files[0]) || (NULL==Files[1]))
		return FALSE;
	FileName = Files[1];

	/* open archive file */
	file = fopen(FileName, "wb");

	/* count nb files */
	i = 2;
	while ((i<18) && (NULL!=Files[i]))
		i++;
	i = i-2;

	/* add in the archive file the 6 header chr (the nb of files will be filled later) */
	buf[5] = i;
	if (6!=fwrite(&(buf[0]), 1, 6, file))
	{
		fclose(file);
		return FALSE;
	}
	
	/* for each file in the file list, add it in the archive */
	i = 2;
	while (NULL!=Files[i])
	{
		FileName = Files[i];
		/* write the file entry in the archive */
		s = strlen(FileName);
		strcpy(&(buf[1]), FileName);
		buf[0] = s;
		/* write file name */
		if (s+1!=fwrite(&(buf[0]), 1, s+1, file))
		{
			fclose(file);
			fprintf(stderr, "Could not write %s file name\n", FileName);
			return FALSE;
		}

		/* open the file to archive */
		file2 = fopen(FileName, "rb");
		if (NULL==file2)
		{
			fclose(file);
			fprintf(stderr, "Could not open %s\n", FileName);
			return FALSE;
		}
		/* get the file size and strip first 512 nibbles for non system files */
		size = FileSize(file2);
		if (0==strcmp("FS", FileName))
		{
			fseek(file2, 65536 + 256 + 5 + 3, SEEK_SET);
			size = size - 65536 - 256 - 5 - 3;
		} else
		if (0!=strcmp("System", FileName))
		{
			fseek(file2, 256 + 5, SEEK_SET);
			size = size - 256 - 5;
		}
		/* get some buffer memory */
		a = (char *)malloc(size+4);
		if (NULL==a)
		{
			fclose(file);
			fclose(file2);
			fprintf(stderr, "Could not allocate memory for %s\n", FileName);
			return FALSE;
		}
		/* read the input file */
		if ((int)fread(&(a[4]), 1, size, file2)!=size)
		{
			fclose(file);
			fclose(file2);
			free(a);
			fprintf(stderr, "Could not read %s\n", FileName);
			return FALSE;
		}
		/* remove tailling 0s (but let at least 0 of them) */
		size = size+4;
		while ((size>=5) && (0==a[size-1]) && (0==a[size-2]) && (0==a[size-3]))
			size--;
		size = size-4;
		/* write the file size in the output */
		a[0] = size;
		a[1] = size / 256;
		a[2] = size / (256*256);
		a[3] = size / (256*256*256);
		size = size+4;
		/* add it in the archive */
		if ((int)fwrite(a, 1, size, file)!=size)
		{
			fclose(file);
			fclose(file2);
			free(a);
			fprintf(stderr, "Could not write %s to output\n", FileName);
			return FALSE;
		}
		/* deallocate memory */
		free(a);
		/* close the file */
		if (fclose(file2))
		{
			fclose(file);
			fprintf(stderr, "Could not close %s\n", FileName);
			return FALSE;
		}
		/* next file */
		i++;
	}

	/* close the archive */
	if (fclose(file))
	{
		fprintf(stderr, "Could not close outfile\n");
		return FALSE;
	}
	return TRUE;
}

int main(int argc, char **argv)
{
	if (!SmallFilesToBigOne(argv))
	{
		printf("Purpose: Create a .flash like file used by FDP\nUsage: CreateFlash OutPutFile [Input Files ....]");
		return TRUE;
	}
	else return FALSE;
}
