#include <stdio.h>

int ChipsetSize= 508;
#define RomSize 1024*128*16*2

int error()
{
  printf("Usage: RomName(emu48) RomName(bin/manufacturing) [Files to use to create ROM] [-B Bootsector file] [-UserFile] [More_UserFiles]\n");
  printf("User files can not exceed 128KB\n");
  return 1;
}

int StringToFile(char *data, char *name, int size)
{
	int i= 0;
	FILE *f= fopen(name, "wb");
	if (f==NULL) { fprintf(stderr, "Error, can not open %s for writing\n", name); return 0; }
	while (i<size)
	  i+= fwrite(data+i, 1, size-i, f);
	fclose(f);
	return 1;
}

long FileSize(FILE *f)
{
	long size;
	long pos;
	pos= ftell(f);
	fseek(f, 0, SEEK_END);
	size= ftell(f);
	fseek(f, pos, SEEK_SET);
	return size;
}

int FileToString(char **data, char *name, char *text)
{
	int i= 0, size;
	FILE *f= fopen(name, "rb");
//	printf("%s %s\n", text, name);
	if (f==NULL) { fprintf(stderr, "Error, can not open %s for reading\n", name); return -1; }
	size= FileSize(f);
	*data= (char*)malloc(size*2);
	while (i<size)
	  i+= fread((*data)+i, 1, size-i, f);
	fclose(f);
  if (((*data)[0]=='H') && ((*data)[1]=='P') && ((*data)[2]=='H') && ((*data)[3]=='P') && ((*data)[4]=='4'))
  {
    memcpy(*data, (*data)+8, size-8);
    size-= 8;
  }
  for (i=size-1; i>=0; i--)
  {
  	(*data)[i*2+1]= ((*data)[i]>>4) & 15;
  	(*data)[i*2]= (*data)[i] & 15;
  }
	return size*2;
}

int HeaderPos[]= {0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 1, };

int main(int argc, char **argv)
{
  int ParamPos, i, j, bpos;
  static char b[RomSize];

  ParamPos = 3;
  bpos= 0;

  if (argc<2)
    return error();
  else 
  {
    // prepare flash (ensure all bank are user by default)
    memset(b, 15, RomSize);
    for(i=0; i<17; i++)
    {
    	char *a= b+((128*1024*HeaderPos[i])+0x200);
    	*a++= 0; *a++= 15;
    }

    // reading the rom files
    while ((ParamPos<argc) && ((argv[ParamPos][0]=='-' && argv[ParamPos][1]=='B') || (argv[ParamPos][0]!='-')))
    {
     	char *a;
      if (argv[ParamPos][0]=='-' && argv[ParamPos][1]=='B')
      { 
      	int size;
        if (ParamPos+1>argc) { fprintf(stderr, "Error, no filename after -B\n"); return error(); }
        // loading the file in memory
        size= FileToString(&a, argv[ParamPos+1], "Load Boot File"); if (size==-1) return error();
        // writing the file in ROM
        for (i=0x20000; i<(0x40000<size?0x40000:size); i++) b[i]= a[i];
        ParamPos+= 2;
        free(a);
      } else {
      	int size;
        // opening the file in memory
        size= FileToString(&a, argv[ParamPos], "Load File");  if (size==-1) return error();
        // writing the file in ROM
        for (i=0; i<size; i++) b[bpos++]= a[i];
        // 128 Kb padding
        bpos= bpos-size+128*2*1024;
        ParamPos++;
        free(a);
      }
    }

    if (ParamPos<argc)
    {
    	bpos+= 512;
    	b[bpos++]= 0; b[bpos++]= 15; b[bpos++]= 0; b[bpos++]= 0; b[bpos++]= 0; b[bpos++]= 0; b[bpos++]= 0; b[bpos++]= 0; b[bpos++]= 0; b[bpos++]= 0;
	    // reading the user files (no more than 128k)
	    while (ParamPos<argc) 
	    {
	      // opening the file in memory
	      int size;
	      char *a;
        if (argv[ParamPos][0]=='-') 
          size= FileToString(&a, &(argv[ParamPos][1]), "Load additional File");
        else
          size= FileToString(&a, argv[ParamPos], "Load additional File");
        if (size==-1) return error();
	      // writing the file in ROM
	      b[bpos++]= 3; b[bpos++]= 5;
	      for (i=0; i<6; i++)
	        b[bpos++]= (((size+7) & 0xFFFFFE) >> (4*i))&15;
        for (i=0; i<size; i++) b[bpos++]= a[i];
	      if (size&1==1)
	        bpos++;
	      free(a);
  	    ParamPos++;
	    }
	  }

    // overrite bank number and erase numbers
    for(i=0; i<17; i++)
    {
    	char *a= b+((128*1024*HeaderPos[i])+0x200+2);
    	*a++= HeaderPos[i]>>1; *a++= 0; *a++= 0; *a++= 0; *a++= 0; *a++= 0; *a++= 0; *a++= 0; 
    }

    if (!StringToFile(b, argv[1], RomSize)) return error();
    b[(128*1024*2*15)+0x202+1-2]= 15;
   	for (i=0; i<RomSize/2; i++)
   	  b[i]= (b[i*2]<<4)+b[i*2+1];
    if (!StringToFile(b, argv[2], RomSize/2)) return error();
    return 0;
  }
}
