#include <stdio.h>
#include <unistd.h>

extern char *optarg;
extern int optind;

char *name_output = NULL;
char *name_input = NULL;
char *listfptr = NULL;
char *id = "2";
char *libs_name = NULL;
int mode = 0;

#define LISTF	1
#define HEADER	2
#define FEXT	4
#define PORT0	8
#define LIBS    16

void get_options( int argc, char **argv )
{
  char	i;

  while( ( i = getopt( argc, argv, "f:i:lhxpb:" ) ) != (char)( -1 ) )
    switch( i ) {
    case 'f':
      name_output= optarg;
      break;
    case 'i':
      id = optarg;
      break;
    case 'l':
      mode |= LISTF;
      break;
    case 'x':
      mode |= FEXT;
      break;
    case 'h':
      mode |= HEADER;
      break;
    case 'p':
      mode |= PORT0;
      break;
    case 'b':		/* Added by cyrille to allow adding libs in a rom module */
      libs_name= optarg;
      mode |= LIBS;
      break;
    default:
      fprintf(stderr,"Usage: [-f file] [-l] [-i id] [-h] [-b file]\n"
	      "-f file\t: Output file (default is stdout)\n"
	      "-h\t: Output Flash Module Header\n"
	      "-i id\t: Flash ROM ID (default=2)\n"
	      "-l\t: Create list of flash pointer reference\n"
	      "-p\t: Create list of flash pointer indirect reference\n"
	      "-x\t: Create list of FEXTERNAL reference\n"
	      "-b file\t: lib definition file (included before the CON(3) #FFF)\n"
	      "\n");
      exit( -1 );
      break;
    }
}

int main(int argc, char **argv)
{
  FILE *fr, *output = stdout, *input = stdin;
  char buf[256],*p;
  int i;

  get_options(argc, argv);

  switch( argc - optind ) {
  case 1: /* progr arg1 */
    if ((input = fopen(argv[optind],"r")) == NULL) {
      perror(argv[optind]);
      exit(-1);
    }
  case 0:
    break;
  default:
    fprintf(stderr,"Bad argument number/type\n");
    exit(-1);
  }
  if (name_output) {
    if ((output = fopen(name_output,"w")) == NULL) {
      perror(name_output);
      exit(-1);
    }
  }
  if (mode & HEADER) {
    fprintf(output,"\nASSEMBLE\n"
	    "	ABS	#40200\n"
	    "	CON(2)	=FRBankRom	ROM bank ID\n"
	    "	CON(2)	#FF		Physical Bank Number\n"
	    "	CON(6)	#FFFFFF		Number of erases\n"
	    "=CrcStart%s\n"
	    "	CON(5)	(=CrcVger%s)-(CrcStart%s)-4		Offset to Crc\n"
	    "	CON(2)	#00		Object revision\n"
	    "	CON(5)	#%s		Module ID\n"
	    "	CON(7)	#0		Initialization program (RTNSXM)\n"
	    "	CON(5)	NbCmd%s		Number of Commands\n\n",id,id,id,id,id);
  }
  if (!(mode & LISTF))
    fprintf(output,"\tCON(5)\t=EVAL\n=^EvalInModule%s\tEQU\t#%s+4096*0\n",id,id);

  i = 1;
  while (fgets(buf,sizeof(buf),input)) {
    p = (char *)strchr(buf,'\n');
    if (p)
      *p = '\0'; /* remove the \n */
    p = (char *)strchr(buf,'\r');
    if (p)
      *p = '\0'; /* remove the \r */
    p = buf;
    /* Ignore any starting spaces */
    while (*p && (*p==' '))
      p++;
    /* Ignore Line starting with * */
    if ((*p == '*') || !strlen(p))
      continue;
    if (mode & LISTF) /* produce a list for PORT0 */
      fprintf(output,"NAMELESS %s\nASSEMBLE\n\tNIBHEX\t1\nRPL\n",p);
    else
      if (mode & FEXT) /* produce a list of FEXTERNAL reference */
	fprintf(output,"FEXTERNAL %s\n",p);
      else
	if (mode & PORT0) /* produce a list for Vger1/Header2/header.s */
	  fprintf(output,"\tCON(5)\t=j%s\n=^%s\tEQU\t#%s+4096*%d\n",
		  p,p,id,i++);
	else
	  fprintf(output,"\tCON(5)\t=%s\n=^%s\tEQU\t#%s+4096*%d\n",
		  p,p,id,i++);
  }
  if (mode & LISTF)
    fprintf(output,"ASSEMBLE\n\tNIBHEX\t1111\tPadding for valid address\n");
  if (input != stdin)
    fclose(input);

  if (mode & HEADER) {
    fprintf(output,"NbCmd%s\tEQU\t%d\n\n",id,i);
    if (mode & LIBS)
    { /* added by cyrille to allow libs in rom modules */
      if ((input = fopen(libs_name,"r")) == NULL) 	/* open the libs file */
      {
        perror(libs_name);
        exit(-1);
      }
      while (fgets(buf, sizeof(buf), input))		/* copy the lib files */
      {
        fprintf(output, buf);
      }
      if (input != stdin)				/* close the lib file */
        fclose(input);
      /* end added by cyrille to allow libs in rom modules */
    }
    fprintf(output,"	CON(3)	#FFF		No library attached\n\n");
  }

  return(0);
}
